/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _HomgPoint1D_hh_
#define _HomgPoint1D_hh_
#include "bias_config.h" 

#include <Base/Debug/Error.hh>
#include <Base/Math/Vector2.hh>



// the type of each element:
#define HOMGPOINT1D_TYPE double
// values whose abs is < epsilon are handled as zero
#define HOMGPOINT1D_TYPE_EPS 1E-12

namespace BIAS {

  /** @class HomgPoint1D
      @brief class HomgPoint1D describes a point with 2 degrees of freedom in 
      projective coordinates. 
    
      manual loop unrolling is used if possible. 
      The Vector is in row-major order (3 rows, 1 column)
      the indizes begin with zero (to size-1)
    
      constructors and assignment operators are never inherited, so they are 
      wrapped (or reimplemented) here. 
      @author Jan Woetzel
      @status untested  (03/04/2002)
  **/

  class BIASGeometryBase_EXPORT HomgPoint1D:public Vector2 < HOMGPOINT1D_TYPE > {

  public:

    /** destructor
        @author Jan Woetzel
        @status untested (02/28/2002)
    **/
    ~HomgPoint1D() {}
    ;


    /**	default constructor
        @author Jan Woetzel
        @status untested (02/28/2002)
    **/
    inline HomgPoint1D():Vector2 < HOMGPOINT1D_TYPE > () {}
    ;

    /**	copy constructor
        @author Jan Woetzel
        @status untested (03/05/2002)
    **/
    inline HomgPoint1D(const HomgPoint1D & p):Vector2 < HOMGPOINT1D_TYPE >
    (p) {}
    ;


    /** assignment with a constant value for all elements
        @author Jan Woetzel
        @status untested (03/01/2002) **/
    explicit inline HomgPoint1D(const HOMGPOINT1D_TYPE & scalar):Vector2 <
      HOMGPOINT1D_TYPE > (scalar) {}
    ;

    /**	assignment with an array of values
        which is copied into this ones class members
        @author Jan Woetzel
        @status untested (02/28/2002) **/
    explicit inline HomgPoint1D(const HOMGPOINT1D_TYPE * pv):Vector2 <
      HOMGPOINT1D_TYPE > (pv) {}
    ;

    /**	constructor with element assignment
        @author Jan Woetzel
        @status untested (03/01/2002) **/
    explicit inline HomgPoint1D(char *s):Vector2 < HOMGPOINT1D_TYPE > (s) {}
    ;

    /** assignment operator
        set the elements of this HomgPoint to the element of vec
        @author Jan Woetzel
        @status untested (02/25/2002)
    **/
    inline HomgPoint1D & operator =
    (const Vector2 < HOMGPOINT1D_TYPE > &vec) {
      copy(vec.GetData());	// copy the elements of vec to this HomgPoint coordinates
      return *this;
    };

    /** assignment operator
        set the vector elementwise to scalar value
        @author Jan Woetzel
        @status untested (02/28/2002)
    **/
    inline HomgPoint1D & operator = (const HOMGPOINT1D_TYPE & scalar) {
      Set(scalar);
      return *this;
    };


    /** homogenize class data member elements to W==1 by divison by W
        @author Jan Woetzel
        @status untested (02/28/2002)
    **/
    inline void Homogenize() {
      // w = data_[3]  is not in [-eps..eps]  (means non-zero)
      BIASASSERT(!((GetW() > -1 * HOMGPOINT1D_TYPE_EPS)
                   && (GetW() < HOMGPOINT1D_TYPE_EPS)));
      data_[0] = data_[0] / GetW();
      data_[1] = data_[1] / GetW();	// should be 1
    }
    ;


    /**	return the euclidian coordinates
        Only possible for w !=0 because w==0 describes points of infinity who are'nt not in euclidian space (but may be interpreted as directions)
        @author Jan Woetzel
        @status untested (03/04/2002)
    **/
    inline HOMGPOINT1D_TYPE GetEuclidian() {
      BIASASSERT(!((GetW() > -1 * HOMGPOINT1D_TYPE_EPS) && (GetW() < HOMGPOINT1D_TYPE_EPS)));	// test if w is zero (in eps env.)
      return (GetX() / GetW());
    };


    ////////////////////////////////////////////////////////////////////////
    ///////////////// Comparison operators /////////////////////////////////
    ////////////////////////////////////////////////////////////////////////


    /** Comparison operator 'equal'
        @author Ingo Thomsen
        @status tested
    **/
    inline bool operator== (HomgPoint1D& arg) {
      Homogenize();
      arg.Homogenize();

      // If BIAS_BOUNDS_CHECk is active, Homogenize() does not allow an w component as zero
      // and it is enough to compare the non w components
      // In the other w component may be zero (and indicating a direction neither than a point)
#ifdef BIAS_BOUNDS_CHECK
      return (data_[0] == arg.data_[0]);
#else
      return (data_[0] == arg.data_[0]) &&  (data_[1] == arg.data_[1]);
#endif
    }

    /** Comparison operator 'not equal'
        @author Ingo Thomsen
        @status tested
    **/
    inline bool operator!= (HomgPoint1D& arg) {
      return !(*this == arg);
    }



  }
  ;				// class HomgPoint1D ------------------------------------------
} // namespace BIAS
#endif
// /////////////    do not add code after this line !    /////////////
