/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _BIAS_IMAGE_CONVERSION_H_
#define _BIAS_IMAGE_CONVERSION_H_

#include "bias_config.h"

#include "Image.hh"

#ifdef BIAS_HAVE_OPENCV
#  include <cv.h>
#endif


#ifdef BIAS_HAVE_DV
#  include <dv.h>
#endif


#define D_IMCONV_IPL  0x00000001

namespace BIAS {  
    /** @class ImageConvert
    @ingroup g_image_io
    @brief class with static functions 
    for converting all kinds of color formats and storage types

    As all fucntions are static they have to be called by
    eg. ImageConvert::Convert() and not with an object.
    @author Daniel Grest, Nov. 2002 */
    class BIASImageBase_EXPORT ImageConvert : Debug {
        friend class ImageBase;
    public:
        /** main general function, calls desired specialized functions,
        always initializes the destImage(if not source==dest)
        @autor Daniel Grest, Nov. 2002*/
      static int Convert(BIAS::ImageBase &source, BIAS::ImageBase &dest, 
            enum BIAS::ImageBase::EColorModel targetColorModel,
            bool bPlanar=false);

        /** ask Felix or Friso
        @author Felix Woelk or Friso Evers
        */
        static int ConvertST(const BIAS::ImageBase &source, BIAS::ImageBase &dest, 
            ImageBase::EStorageType targetST);


        /** In place conversion to gray image. No further memory is allocated,
        because only the the first part of the existing image is used.
        @author Felix Woelk **/
        template <class StorageType>
          BIASImageBase_EXPORT 
            static int IP_ToGrey(Image<StorageType> &img);

        /** Create a grey converted copy of source image in this. 
        Release() and Init() are called if necessary.
        @author Felix Woelk **/
        template <class StorageType>
          BIASImageBase_EXPORT 
            static int ToGrey(const Image<StorageType>& source, 
            Image<StorageType>& dest);

        /** Create a RGB converted copy of source image in this. 
        Release() and Init() are called if necessary.
        @author Felix Woelk **/
        template <class StorageType>
          BIASImageBase_EXPORT 
            static int ToRGB(Image<StorageType>& source, Image<StorageType>& dest);

        /** create a HSV converted copy of source image in this
        Release() and Init() are called if necessary.
        HSV space is implemented as a cylinder not as a cone!
        only for float and unsigned char
        @author Felix Woelk  partially tested **/
        template <class StorageType>
          BIASImageBase_EXPORT 
            static int ToHSV(Image<StorageType>& source, Image<StorageType>& dest);

        /** create a HSL converted copy of source image in this
        Release() and Init() are called if necessary.
        only for float and unsigned char
        @author Daniel Grest, April 2003
        @status tested for unsigned char interleaved
        */
        template <class StorageType>
          BIASImageBase_EXPORT 
            static int ToHSL(Image<StorageType>& source, Image<StorageType>& dest);

        /** create a hsL converted copy of source image in this
        Release() and Init() are called if necessary.
        only for unsigned char
        similar to HSL but in euclidean coordinates for the CNCC
        @author Daniel Grest, June 2003
        @status tested for unsigned char interleaved
        */
        template <class StorageType>
          BIASImageBase_EXPORT 
            static int TohsL(Image<StorageType>& source, Image<StorageType>& dest);

        /** Takes the source image, and build an interleaved version 
        @author Ingo Thomsen
        @date 11.06.2002
        @status tested
        @todo make it work for yuv **/
        template <class StorageType>
          BIASImageBase_EXPORT 
            static int ToInterleaved(Image<StorageType>& source,
            Image<StorageType>& dest);

        /** Takes the three channels an compose one interleaved RGB image */
        template <class StorageType>
          BIASImageBase_EXPORT 
            static int ToInterleavedRGB(Image<StorageType>& red, 
            Image<StorageType>& green,
            Image<StorageType>& blue,
            Image<StorageType>& dest);

        /** Converts the existing image to an planar version, 
        @author Ingo Thomsen
        @date 03.04.2002
        @status none
        case CM_YUYV422 not yet tested **/
        template <class StorageType>
          BIASImageBase_EXPORT 
            static int ToPlanar(Image<StorageType>& source,
            Image<StorageType>& dest);

        /** extracts color planes to argument images without changing source.
        So far only implemented for CM_YUYV422
        @author Felix Woelk
        @status alpha (18/04/2002) */
        template <class StorageType>
          BIASImageBase_EXPORT 
            static int FromInterleaved(Image<StorageType>& source,
            Image<StorageType>& Image1, 
            Image<StorageType>& Image2,
            Image<StorageType>& Image3);

        /////////////////////////////////////////////////////////////////
        /// foreign image formats, ipl, vil
        /// so far untested
        /////////////////////////////////////////////////////////////////

#ifdef BIAS_HAVE_DV
        /** Can specify speed/quality trade off
        0 : monochrome , DC coeffs only
        1 : monochrome , AC 1 coeffs
        2 : color ,  DC coeffs only
        3 : color , AC 1 coeffs
        4 : color, AC 2 coeffs (slowest but best quality)
        */
        template <class StorageType>
            static int DVToRGB_(const Image<StorageType>& source,
            Image<StorageType>& dest,unsigned char quality=4);
#endif


#ifdef BIAS_HAVE_OPENCV
        /** Returns a new, separate IplImage for use with OpenCV created rom teh source BIAS Image.
        The dest has to be a call by reference "*&" because the pointer itself is changed due to new (JW).
        The colorChannel order is swapped because OpenCv used BGR - which may be inefficient.
        See WrapBias2Ipl class, too (jw).
        @return newly created IplImage
        @todo   ROIs are not yet kept, when converted to IplImages 
        @author Ingo Thomsen 04.05.2002, Jan Woetzel 2005
        */
        template <class StorageType>
            static int BIAS2ipl(const Image<StorageType>& source, IplImage* &dest);

        /** Takes the image information and data of an existing IplImage
        and builds a Image. The existing data is discarded. 

        @author Ingo Thomsen
        @date 04.17.2002
        @return -1 in case of failure
        @param image The IplImage to be transfered
        @status untested

        @todo Make it work for colored images
        @todo Make it work for ROIs **/
        template <class StorageType>
            static int Ipl2BIAS(IplImage* source, Image<StorageType>& dest);
#endif // BIAS_HAVE_OPENCV


    protected:  
        template <class StorageType>
            static int IP_RGBToGrey_(Image<StorageType> &img);

        template <class StorageType>
            static int IP_YUV422ToGrey_(Image<StorageType> &img);

        template <class StorageType>
            static int BGRToRGB_(Image<StorageType>& source, Image<StorageType>& dest);

        template <class StorageType>
            static int YUV422ToRGB_(Image<StorageType>& source, 
            Image<StorageType>& dest);


        template <class StorageType>
            static int YUV420PToRGB_(Image<StorageType>& source, 
            Image<StorageType>&  dest ) ;

        template <class StorageType>
            static int YUV420PToGrey_(const Image<StorageType>& source, 
            Image<StorageType>&  dest ) ;
        template <class StorageType>
            static int GreyToRGB_(Image<StorageType>& source,
            Image<StorageType>& dest);


        template <class StorageType>
            static int RGBToGrey_(const Image<StorageType>& source,
            Image<StorageType>& dest);

        template <class StorageType>
            static int YUV422ToGrey_(const Image<StorageType> &source,
            Image<StorageType>& dest);

    public:
        /** added border calculation, woelk
        @authour evers, woelk*/
        template <class StorageType>
            static int BayerToRGB_(Image<StorageType>& source, 
            Image<StorageType>&  dest );
    protected:
        /** this is the right way, green interpoaltion is mostly done
        out of two values
        no visible difference to fast calculation as used above
        @author woelk 03 2003 */
        template <class StorageType>
            static int BayerToRGBSlow_(Image<StorageType>& source, 
            Image<StorageType>&  dest );
        /** interpolates, better than above
        @author woelk 03 2003 */
        template <class StorageType>
            static int BayerToGrey_(const Image<StorageType>& source,
            Image<StorageType>& dest);



    protected:
        /** just copies the L channel 
        @author Daniel Grest */
        template <class StorageType>
            static int HSLToGrey_(const Image<StorageType>& source,
            Image<StorageType>& dest);



        // helper function for ToPlanar(Image<StorageType>& source)
        template <class StorageType>
            static int ToPlanarYUYV422_(Image<StorageType>& source,
            Image<StorageType>& dest);

        template <class StorageType>
            static int ToPlanarRGB_(Image<StorageType>& source,
            Image<StorageType>& dest);

        // helper function for ToInterleaved(Image<StorageType>& source)
        template <class StorageType>  
            static int ToInterleavedYUYV422_(Image<StorageType>& source,
            Image<StorageType>& dest);
        template <class StorageType>
            static int ToInterleavedRGB_(Image<StorageType>& source,
            Image<StorageType>& dest);

        template <class StorageType>
            static int FromInterleavedYUYV422_(Image<StorageType>& source,
            Image<StorageType>& Y, 
            Image<StorageType>& U,
            Image<StorageType>& V);

        template <class StorageType>
            static int FromInterleavedRGB_(Image<StorageType>& source,
            Image<StorageType>& R, 
            Image<StorageType>& G,
            Image<StorageType>& B);

        /// so far only implemented for StorageType=float
        //template <class StorageType>
        template <class StorageType>
            static int RGBToHSV_(Image<StorageType>& source,
            Image<StorageType>& dest);


        template <class StorageType>
            static int RGBToHSL_(Image<StorageType>& source,
            Image<StorageType>& dest);


        template <class StorageType>
            static int BGRToHSL_(Image<StorageType>& source,
            Image<StorageType>& dest);

        template <class StorageType>
            static int BGRAToHSL_(Image<StorageType>& source,
            Image<StorageType>& dest);


        template <class StorageType>
          BIASImageBase_EXPORT 
            static int RGBTohsL_(Image<StorageType>& source,
            Image<StorageType>& dest);


        template <class StorageType>
          BIASImageBase_EXPORT 
            static int RGBAToRGB_(Image<StorageType>& source,
            Image<StorageType>& dest);

        template <class StorageType>
          BIASImageBase_EXPORT 
            static int BGRAToRGB_(Image<StorageType>& source,
            Image<StorageType>& dest);

        //in WIN32 forward declarations for specialized function templates are needed
#ifdef _WIN32
        template <> 
          //BIASImageBase_EXPORT 
          static int RGBToHSV_(Image<float>& source,
            Image<float>& dest);

        template <> 
          //BIASImageBase_EXPORT 
          static int RGBToHSV_(Image<unsigned char>& source,
            Image<unsigned char>& dest);

        template <> 
          //BIASImageBase_EXPORT 
          static int RGBToHSL_(Image<float>& source,
            Image<float>& dest);

        template <> 
          //BIASImageBase_EXPORT 
          static int RGBToHSL_(Image<unsigned char>& source,
            Image<unsigned char>& dest);

        template <> 
          //BIASImageBase_EXPORT 
          static int BGRToHSL_(Image<unsigned char>& source,
            Image<unsigned char>& dest);

        template <> 
          //BIASImageBase_EXPORT 
          static int RGBTohsL_(Image<int>& source,
            Image<int>& dest);

        template <> 
          //BIASImageBase_EXPORT 
          static int RGBTohsL_(Image<unsigned char>& source,
            Image<unsigned char>& dest);

#endif
        //...


        static int FloatToUC_(const ImageBase &source, ImageBase &dest);



#ifdef BIAS_HAVE_DV
        static dv_decoder_t *DVDecoder_;
#endif

    };

} // namespace


#endif
