/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "ImageConvert.hh"

using namespace BIAS;
using namespace std;



#ifdef BIAS_HAVE_OPENCV
template <class StorageType>
int ImageConvert::BIAS2ipl(const Image<StorageType>& source, IplImage*  &dest) 
{
    //#ifdef USE_MY_OPENCV
    //BIASDOUT(D_IMCONV_IPL, "using own CreateIplImage");
    //  cerr << "using own CreateIplImage\n";
    //dest = CreateIplImage(source.GetWidth(), source.GetHeight(), 
    //                      sizeof(StorageType) * 8 , source.GetChannelCount());
    //#else
    //BIASDOUT(D_IMCONV_IPL, "using official CreateIplImage");
    //cerr << "using official CreateIplImage\n";
    dest = cvCreateImage(cvSize(source.GetWidth(), source.GetHeight()), 
        sizeof(StorageType) * 8, source.GetChannelCount());
    //#endif

    StorageType* newImageData = (StorageType*) dest->imageDataOrigin;

    //
    // IplImages in OpenCV are always:
    //   - GREY 
    //   - RGB  (with byte Order BGR due to Windows DIB format) 
    //   - RGBA (with byte Order BGRA)
    //
    // As an limitation of opencv compared to IPL, color images are ALWAYS 
    //   interleaved!

    //cerr << "source.GetColorModel() :"<<source.GetColorModel()<<endl;
    switch (source.GetColorModel()) {

  case ImageBase::CM_Grey: 
      //cerr <<"grey\n";
      memcpy(newImageData, source.GetImageData(), source.GetSize());
      break;

  case ImageBase::CM_BGR : 
      //cerr <<"BGR\n";
      if (source.IsPlanar()) {
          //cerr <<"Planar\n";
          // planar BGR image
          for(int channel = 0; channel < 3; channel++) {
              register StorageType* src  = source.GetImageData() + 
                  channel*source.GetPixelCount();
              register StorageType* act = newImageData + channel;
              register StorageType* stop = src + source.GetPixelCount();
              for(;src < stop; src++) {
                  *act = *src;
                  act += 3;
              }
          }
      } else {
          //cerr <<"Interleaved\n";
          // interleaved BGR image - the same format as IplImage itself
          memcpy((void*)newImageData, (void*)source.GetImageData(), source.GetSize());
      };
      break;

  case ImageBase::CM_RGB : 
      //cerr <<"RGB\n";
      if (source.IsPlanar()) {
          //cerr <<"Planar\n";
          // planar RGB image
          for(int channel = 0; channel < 3; channel++) {
              register StorageType* src  = source.GetImageData() + (2 - channel) * 
                  source.GetPixelCount(); // RGB instead of BGR order!
              register StorageType* act = newImageData + channel;
              register StorageType* stop = src + source.GetPixelCount();
              for(;src < stop; src++) {
                  *act = *src;
                  act += 3;
              }
          }
      } else { 
          //cerr <<"Interleaved\n";
          // interleaved RGB image
          register StorageType* src  = source.GetImageData();
          register StorageType* stop = src + 3 * source.GetPixelCount();
          register StorageType* act = newImageData;
          while (src < stop) {
              // swap RGB because OpenCV uses BGR as native channel order
              // only the color conversion routines need this, I fixed highgui to handle channelSeq correctly (jw)
              act[0]=src[2];
              act[1]=src[1];
              act[2]=src[0];
              act+=3; 
              src+=3;
          }
      }    
      break;

  default: 
      BIASERR("Only Grey, BGR and RGB images can be converted to IplImages yet."
          <<" Aborting!");
      return -1;
    }

    return 0;
}

template <class StorageType>
int ImageConvert::Ipl2BIAS(IplImage* source, Image<StorageType>& dest) 
{
    // Does the color depth match?
    if (8*sizeof(StorageType) != source->depth) {
        BIASERR("Cannot convert an IplImage into an Image of different color "
            <<"depth!");
        return -1;
    }

    if (source->width <=0 || source->height <= 0) {
        BIASERR("IplImage to be converted to Image has no size!");
        return -1;
    }
    if ( !dest.IsEmpty() )
        dest.Release();     

    // What kind of image do we have?
    switch (source->nChannels) {
  case 1: // gray image
      // Release the old image data and allocate new memory
      dest.Init(source->width, source->height, 1);
      memcpy(dest.GetImageData(), source->imageDataOrigin, dest.GetSize());
      break;
  case 3:
      if (strcmp(source->colorModel, "RGB")==0 && 
          strcmp(source->channelSeq, "BGR")==0 &&
          source->dataOrder==IPL_DATA_ORDER_PIXEL &&
          (source->depth>>3)==sizeof(StorageType) ){
              dest.Init(source->width, source->height, 3);
              register unsigned char* src = (unsigned char*) source->imageDataOrigin;
              register unsigned char* stop = src + 3 * source->width * source->height;
              register StorageType* dst = dest.GetImageData();
              while (src < stop) {
                  dst[0]=src[2];
                  dst[1]=src[1];
                  dst[2]=src[0];
                  dst+=3; src+=3;
              }
          } else {
              BIASERR("Ipl2BIAS not implemented for this color model or inavlid storage type");
              return -1;
          }
          break;
  default:
      BIASERR("Only gray IplImages can be converted to Images yet!");
      return -1;      
    }
    return 0;
}
#endif // BIAS_HAVE_OPENCV



/*
template <class StorageType>
int Image<StorageType>::CopyInNoInit(const IplImage& Image)
{
#ifdef BIAS_DEBUG
if (Image.imageSize != GetSize() )
BIASERR("Image<StorageType>::CopyIn() Image has different size from IplImage_");
#endif  
memcpy(GetImageData(), Image.imageData, Image.imageSize); 

if (Image.nChannels == 1)
ColorModel_ = CM_Grey;
else if (strcmp("BGR", Image.colorModel)==0)
ColorModel_ = CM_BGR;
else if(strcmp("RGB", Image.colorModel)==0)
ColorModel_ = CM_RGB;
else
BIASERR("ColorModel_ not determined");
return 0;
}*/


