/* 
   This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003, 2004    (see file CONTACTS for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel


   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.

   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ROI.hh"

#include <Base/Debug/Error.hh>
#include "ImageBase.hh"

#include <algorithm>

using namespace BIAS;
using namespace std;

#define ROI_IDENTIFIER "MIP-ROI"
#define ROI_IDENTIFIER_LENGTH 7


//////////////////////////////////////////////////////////////////////
// operator<< 
/////////////////////////////////////////////////////////////////////

ostream& BIAS::operator<<(ostream& os, const Position& pos)
{
  os << "["<<pos.x<<", "<<pos.y<<"]";
  return os;
}

ostream& BIAS::operator<<(ostream& os, const ROI& roi)
{
  os << "ROI ("<<roi.UpperLeftX_<<", "<<roi.UpperLeftY_<<")  <-->  ("
     <<roi.LowerRightX_<<", "<<roi.LowerRightY_<<")";
  os << "\tMaskValid = "<<boolalpha<<roi.MaskValid_;
  os << "\tVectorValid = "<<boolalpha<<roi.VectorValid_;
  if (roi.VectorValid_){
    vector<Position>::const_iterator it;
    for (it=roi.Vector_.begin(); it!=roi.Vector_.end(); it++)
      os << " "<<*it;
  }
  return os;
}


//////////////////////////////////////////////////////////////////////
// class
/////////////////////////////////////////////////////////////////////

const int ROI::Version_ =  ROI_VERSION_NUMBER;


ROI::~ROI()
{
  Release();
}


ROI::ROI()
  : MaskData_(NULL),
    MaskDataArray_(NULL)
{
  VectorValid_=false;
  MaskWidth_=MaskHeight_=0;
  UnsetROI();
}


void ROI::Release()
{
  if (MaskData_ != NULL) { 
    delete [] MaskData_; 
    MaskData_=NULL; 
  }
  if (MaskDataArray_ != NULL ) {
    delete [] MaskDataArray_;  //jw
    MaskDataArray_=NULL; 
  }
  MaskWidth_=0;
  MaskHeight_=0;
  Vector_.clear();
  MaskValid_=false;
  VectorValid_=false;
  UpperLeftX_=UpperLeftY_=0; 
  LowerRightX_=MaskWidth_; 
  LowerRightY_=MaskHeight_;
}

void ROI::UnsetROI() 
{
  UpperLeftX_=UpperLeftY_=0; 
  LowerRightX_=MaskWidth_; 
  LowerRightY_=MaskHeight_;
  MaskValid_=VectorValid_=false;
}

void ROI::Resize(unsigned w, unsigned h)
{

  if (MaskWidth_!=w || MaskHeight_!=h){
    Release();

    BIASASSERT(MaskData_ == NULL);
    BIASASSERT(MaskDataArray_ == NULL);
    MaskData_=new MASK_TYPE[w*h];
    MaskDataArray_=new MASK_TYPE*[h];

    for (unsigned i=0; i<h; i++)
      MaskDataArray_[i] = &(MaskData_[i*w]);

    MaskWidth_  = w;
    MaskHeight_ = h;
  }
}

int ROI::SetCorners(unsigned UpperLeftX, unsigned UpperLeftY, 
                    unsigned LowerRightX, unsigned LowerRightY)
{
  int res=0;
#ifdef BIAS_DEBUG
  if (UpperLeftX < 0 || UpperLeftX > MaskWidth_) {
    BIASERR( "UpperLeftX "<<UpperLeftX<<" is out of image range "
	     <<MaskWidth_);
    //BIASASSERT(false);
    return -1;
  }
  if (UpperLeftY < 0 || UpperLeftY > MaskHeight_) {
    BIASERR("UpperLeftY "<<UpperLeftY<<" is out of image range "
            <<MaskHeight_);
    //BIASASSERT(false);
    return -2;
  }
  if (LowerRightX < 0 || LowerRightX > MaskWidth_ ) {
    BIASERR("LowerRightX "<<LowerRightX<<" is out of image range "
            <<MaskWidth_);
    //BIASASSERT(false);
    return -3;
  }
  if (LowerRightY < 0 || LowerRightY > MaskHeight_ ) {
    BIASERR("LowerRightY "<<LowerRightY<<" is out of image range "
            <<MaskHeight_);
    //BIASASSERT(false);
    return -4;
  }

  if (LowerRightX < UpperLeftX) {
    unsigned int temp = UpperLeftX;
    UpperLeftX = LowerRightX;
    LowerRightX = temp;
    BIASERR("WARNING: You swapped the x coordinates of the upper left and the"
            <<" lower right corner. It's been re-swapped!");
    //BIASASSERT(false);
    abort();
  }

  if (LowerRightY < UpperLeftY) {
    unsigned int temp = UpperLeftY;
    UpperLeftY = LowerRightY;
    LowerRightY = temp;
    BIASERR("WARNING: You swapped the y coordinates of the upper left and the"
            <<" lower right corner. It's been re-swapped!");
    //BIASASSERT(false);
    abort();
  }
#endif  

  UpperLeftX_  = UpperLeftX;
  LowerRightX_ = LowerRightX;
  UpperLeftY_  = UpperLeftY;
  LowerRightY_ = LowerRightY;

  return res;
}



int ROI::GetMaskImage(ImageBase& im) const
{
  BIASASSERT(im.GetStorageType()==ImageBase::ST_invalid || 
	     im.GetStorageType()==ImageBase::ST_unsignedchar);
  if (im.GetWidth()!=MaskWidth_ || im.GetHeight()!=MaskHeight_ ||
      im.GetChannelCount()!=1)
    im.Release();
  if (im.IsEmpty())
    im.Init(MaskWidth_, MaskHeight_, 1, ImageBase::ST_unsignedchar);

  if (MaskValid_){
    memcpy(im.GetImageData(), MaskData_, im.GetPixelCount()*sizeof(MASK_TYPE));
    /*
      bool *p, *end;
      unsigned char *ucp=(unsigned char)im.GetImageData();
      p=Mask_; 
      end=Mask_+(MaskWidth_*MaskHeight_);

      while (p<end){
      *ucp=(*p):((unsigned char)255):((unsigned char)0);
      ucp++;
      p++;
      }
    */
  }
  return (MaskValid_)?(0):(-1);
}

int ROI::SetMaskImage(const ImageBase& im)
{
  int res=0;
  if (im.GetWidth()==MaskWidth_ && im.GetHeight()==MaskHeight_){
    MaskValid_=true;
    VectorValid_=false;
    if (im.GetStorageType()==ImageBase::ST_unsignedchar){
      memcpy(MaskData_, im.GetImageData(), 
	     im.GetPixelCount()*sizeof(MASK_TYPE));
    } else {
      BIASERR("unfinished");
      res=-1;
    }
  } else {
    BIASERR("wrong image size "<<im.GetWidth()<<"x"<<im.GetHeight()
            <<"    expected: "<<MaskWidth_<<"x"<<MaskHeight_<<endl);
    res = -2;
  } 
  return res;
}


void ROI::SetVector(std::vector<Position>& pos)
{
  VectorValid_=true;
  MaskValid_=false;
  Vector_=pos;
  sort(Vector_.begin(), Vector_.end());
}

bool ROI::GetVector(std::vector<Position>& vec) const
{
  vec = Vector_;
  return VectorValid_; 
}

ROI& ROI::operator=(const ROI& roi)
{
  Resize(roi.GetMaskWidth(), roi.GetMaskHeight());
  MaskValid_ = roi.MaskValid_;
  VectorValid_ = roi.VectorValid_;
  UpperLeftX_ = roi.UpperLeftX_;
  UpperLeftY_ = roi.UpperLeftY_; 
  LowerRightX_ = roi.LowerRightX_;
  LowerRightY_ = roi.LowerRightY_; 

  memcpy(MaskData_, roi.MaskData_, MaskWidth_*MaskHeight_*sizeof(MASK_TYPE));
  Vector_ = roi.Vector_;

  // TODO initialize MaskDataArray_ (JW)
  // HACK to avoid access to uninitialized values (JW)
  //  BIASERR("not correctly implemented. Please fix MaskDataArray_ (JW)")
  if (this->MaskDataArray_ != NULL) {
	  delete [] this->MaskDataArray_;
	  this->MaskDataArray_=NULL;
  };
  //MaskDataArray_=new MASK_TYPE*[h];
  //for (unsigned i=0; i<h; i++) MaskDataArray_[i] = &(MaskData_[i*w]);


  return *this;
}

bool ROI::SameSize(const ROI& roi) const
{
  return ((LowerRightX_-UpperLeftX_)-(roi.LowerRightX_-roi.UpperLeftX_))==0 &&
    ((LowerRightY_-UpperLeftY_)-(roi.LowerRightY_-roi.UpperLeftY_))==0;
}

int ROI::WriteBinary(std::ostream& os) const
{
  if (!os)
    return -1;
    
  // 'header'
  os.write(ROI_IDENTIFIER, ROI_IDENTIFIER_LENGTH); 
  os.write(reinterpret_cast<const char*>(&Version_), sizeof(int));
  // rectangular roi
  os.write(reinterpret_cast<const char*>(&UpperLeftX_), sizeof(unsigned));
  os.write(reinterpret_cast<const char*>(&UpperLeftY_), sizeof(unsigned));
  os.write(reinterpret_cast<const char*>(&LowerRightX_), sizeof(unsigned));
  os.write(reinterpret_cast<const char*>(&LowerRightY_), sizeof(unsigned));
  // size of mask image
  os.write(reinterpret_cast<const char*>(&MaskWidth_), sizeof(unsigned));
  os.write(reinterpret_cast<const char*>(&MaskHeight_), sizeof(unsigned));
  // size of vector
  unsigned length = Vector_.size();
  os.write(reinterpret_cast<const char*>(&length), sizeof(unsigned));
  // validity flags
  os.write(reinterpret_cast<const char*>(&MaskValid_), sizeof(bool));
  os.write(reinterpret_cast<const char*>(&VectorValid_), sizeof(bool));
  BIASCDOUT(D_ROI_IO,  "wrote MaskValid "<<boolalpha<<MaskValid_
	    << "  VectorValid "<<boolalpha<<VectorValid_<<endl);
  // mask image
  if (MaskValid_)
    {	
      BIASASSERT(MaskData_ !=NULL );
	
      BIASCDOUT(D_ROI_IO,  "mdp: "<< (void*)MaskData_ << endl);
      BIASCDOUT(D_ROI_IO,  "md: " << boolalpha
		<<MaskData_[MaskWidth_*MaskHeight_-1] << endl);
      BIASCDOUT(D_ROI_IO,  "mda: " <<boolalpha
		<<MaskDataArray_[MaskHeight_-1][MaskWidth_-1]<<endl);
      // TODO: uninitialized values. 
      os.write(reinterpret_cast<const char*>(MaskData_), 
	       MaskWidth_*MaskHeight_*sizeof(MASK_TYPE));
    };
  // position vector
  if (VectorValid_)
    {
      vector<Position>::const_iterator it;
      for (it=Vector_.begin(); it!=Vector_.end(); it++){
	BIASCDOUT(D_ROI_IO,  "writing " << *it << endl);
	os.write(reinterpret_cast<const char*>(&(it->x)), sizeof(unsigned));
	os.write(reinterpret_cast<const char*>(&(it->y)), sizeof(unsigned));
      }
    }

  return (os.good())?0:-2;
}


int ROI::ReadBinary(std::istream& is) 
{
  //SetDebugLevel(D_ROI_IO);
  int res = 0;
  char identifier[ROI_IDENTIFIER_LENGTH];
  int Version;
  unsigned length;
  vector<Position>::iterator it;

  // 'header' 
  is.read(identifier, ROI_IDENTIFIER_LENGTH);
  if (strncmp(identifier, ROI_IDENTIFIER, ROI_IDENTIFIER_LENGTH)!=0){
    BIASERR("invalid ROI identifier "<<identifier<<endl);
    return -1;
  }
  is.read(reinterpret_cast<char*>(&Version), sizeof(int));

  switch (Version){
  case 100:
    // rectangular roi
    is.read(reinterpret_cast<char*>(&UpperLeftX_), sizeof(unsigned));
    is.read(reinterpret_cast<char*>(&UpperLeftY_), sizeof(unsigned));
    is.read(reinterpret_cast<char*>(&LowerRightX_), sizeof(unsigned));
    is.read(reinterpret_cast<char*>(&LowerRightY_), sizeof(unsigned));
    // size of mask image
    is.read(reinterpret_cast<char*>(&MaskWidth_), sizeof(unsigned));
    is.read(reinterpret_cast<char*>(&MaskHeight_), sizeof(unsigned));
    Resize(MaskWidth_, MaskHeight_);
    // size of vector
    is.read(reinterpret_cast<char*>(&length), sizeof(unsigned));
    Vector_.resize(length);
    // validity flags
    is.read(reinterpret_cast<char*>(&MaskValid_), sizeof(bool));
    is.read(reinterpret_cast<char*>(&VectorValid_), sizeof(bool));
    BIASCDOUT(D_ROI_IO,  "read MaskValid "<<boolalpha<<MaskValid_
	      << "  VectorValid "<<boolalpha<<VectorValid_<<endl);
    // mask image
    if (MaskValid_){
      BIASCDOUT(D_ROI_IO, "reading mask image of size "
		<<MaskWidth_<<"x"<<MaskHeight_<<endl);
      is.read(reinterpret_cast<char*>(MaskData_), 
	      MaskWidth_*MaskHeight_*sizeof(MASK_TYPE));
    }
    // position vector
    if (VectorValid_){
      BIASCDOUT(D_ROI_IO, "reading vector of length "<<Vector_.size()<<endl);
      for (it=Vector_.begin(); it!=Vector_.end(); it++){
	is.read(reinterpret_cast<char*>(&(it->x)), sizeof(unsigned));
	is.read(reinterpret_cast<char*>(&(it->y)), sizeof(unsigned));
	BIASCDOUT(D_ROI_IO,  "read "<<*it<<endl);
      }
    }
    res = (is.good())?0:-3;
    break;
  default:
    BIASERR("unknown ROI version");
    res = -2;
    break;
  }
  return res;
}

int ROI::Mask2Vector() 
{
  register unsigned y=0, x=0;
  Vector_.clear();
  Vector_.reserve(MaskWidth_*MaskHeight_);
  for (y=0; y<MaskHeight_; y++){
    for (x=0; x<MaskWidth_; x++){
      if (GetMaskDataArray()[y][x])
	Vector_.push_back(Position(x, y));
    }
  }

  if (MaskValid())
    SetVectorValidFlag(true);
  return (MaskValid_)?0:-1;
}

int ROI::Vector2Mask() 
{
  memset(GetMaskData(), 0, GetMaskWidth()*GetMaskHeight()*sizeof(MASK_TYPE));

  vector<Position>::iterator it;

  for (it=Vector_.begin(); it!=Vector_.end(); it++){
#ifdef BIAS_DEBUG
    if (!(it->y >= 0 && it->x >= 0 && 
	  it->y < GetMaskHeight() && it->x < GetMaskWidth() ))
      BIASERR((*it)<<"  "<<MaskWidth_<<"x"<<MaskHeight_<<endl);
    BIASASSERT(it->y >= 0 && it->x >= 0 && 
	       it->y < GetMaskHeight() && it->x < GetMaskWidth() );
#endif
    GetMaskDataArray()[it->y][it->x]=UCHAR_MAX;
  }

  if (VectorValid())
    SetMaskValidFlag(true);
  return (VectorValid_)?0:-1;
}

int ROI::Vector2Corners() 
{
  if (VectorValid_)
    SetCorners(Vector_.begin()->x, Vector_.begin()->y,
	       Vector_.rbegin()->x, Vector_.rbegin()->y);
  return (VectorValid_)?0:-1;
}

int ROI::Mask2Corners() 
{
  BIASERR("unfinished");
  return (MaskValid_)?0:-1;
}

void ROI::Corners2Mask() 
{
  BIASERR("unfinished");
  SetMaskValidFlag(true);
}

void ROI::Corners2Vector() 
{
  BIASERR("unfinished");
  SetVectorValidFlag(true);
}

