/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Matrix.hh"

//additional math constants are not used by default in WIN32 - so set usage
#ifdef _WIN32
#define _USE_MATH_DEFINES
#endif

#include "math.h"
#include <algorithm>
#include <fstream> 

using namespace BIAS;
using namespace std;

template<class T> 
Matrix<T>::~Matrix() 
{}



template <class T>
void Matrix<T>::Fill(const T & scalar)
{
  TNT::Matrix<T>::operator=(scalar);
}


template <class T>
void Matrix<T>::normalizeColsToOne(const int row_to_use) 
{
  if (row_to_use >= this->num_rows()) {
    // Error, index for row is expected to be in [0...num_rows-1]
    BIASERR("dim of matrix is (" << this->num_rows() << " x " << this->num_cols() <<
           ").");
    BIASERR("cannot normalize cols by elements in row " << row_to_use <<
           ". aborting. ");

  } else {

    // value of row index is OK for normalization
    double divisor;
    for (int col = 0; col < this->num_cols(); col++) {
      // get the divisor to normalize with
      divisor = (*this)[row_to_use][col];
      if (fabs(divisor) > DOUBLE_EPSILON) {

        // normalize one column (= all rows of one column)
        for (int row = 0; row < this->num_cols(); row++) {
          // normalize element by dividing it by the row_to_use element of 
          //the same column
          (*this)[row][col] = ((*this)[row][col]) / (T) divisor;
        };
      } else {
        // could not divide by (nearly) zero
        BIASERR("could not divide by " << divisor <<
               " because it is too close to zero. DOUBLE_EPSILON= "
               << DOUBLE_EPSILON);
        BIASERR("skipping normalization of this column.");
      };
    };
  };
}




template <class T>
void Matrix<T>::normalizeRowsToOne(const int col_to_use) 
{
  if (col_to_use >= this->num_cols()) {
    // Error, index for row is expected to be in [0...num_rows-1]
    BIASERR("dim of matrix is (" << this->num_rows() << "," << this->num_cols() <<
           ").");
    BIASERR("cannot normalize rows by elements in col " << col_to_use 
           << ". aborting. ");
  } else {
    // value of col index is OK for normalization
    double divisor;
    for (int row = 0; row < this->num_rows(); row++) {
      // get the divisor to normalize with
      divisor = (*this)[row][col_to_use];
      if (fabs(divisor) > DOUBLE_EPSILON) {

        // normalize one row (= all cols of one row)
        for (int col = 0; col < this->num_rows(); col++) {
          // normalize element by dividing it by the col_to_use 
          //element of the same column
          (*this)[row][col] = ((*this)[row][col]) / (T) divisor;
        };
      } else {
        // could not divide by (nearly) zero
        BIASERR("could not divide by " << divisor <<
               " because it is too close to zero. DOUBLE_EPSILON= "
               << DOUBLE_EPSILON);
        BIASERR("skipping normalization of this column.");
      };
    };
  };
}


template <class T>
void Matrix<T>::NormalizeRows()
{
  for (int i = 0; i< this->num_rows() ; i++){
    double scale = GetRow(i).norm_L2();
    for (int j = 0; j < this->num_cols(); j++)
      (*this)[i][j]=(T)( (*this)[i][j]/scale);
  }
}


template <class T>
void Matrix<T>::NormalizeCols()
{
  for (int j = 0; j< this->num_cols() ; j++){
    double scale = GetCol(j).norm_L2();
    for (int i = 0; i < this->num_rows(); i++)
      (*this)[i][j]=(T) ((*this)[i][j]/scale);
  }
}


template<class T>
void Matrix<T>::ScaleRow(int NoRow, T scale)
{
  for (int j = 0; j < this->num_cols(); j++)
    //((*this)[NoRow][j]) *= scale;  
    ((*this)[NoRow][j]) = ((*this)[NoRow][j]) * scale;  
}


template<class T>  
void Matrix<T>::ScaleCol(int NoCol,T scale)
{
  for (int i = 0; i < this->num_rows(); i++)
    //((*this)[i][NoCol]) *= scale;
    (*this)[i][NoCol] =  (*this)[i][NoCol] * scale;
}


template<class T>
Vector<T> Matrix<T>::GetRow(const int& row) const
{
  BIASASSERT(row >= 0 && row < this->num_rows());
  // create a new Vector which will be returned
  // ? faster: copy elements during constructor
  const unsigned int numcols = this->num_cols();
  Vector<T> vec(numcols);

  // copy the elements into the vector:
  for (unsigned int col = 0; col < numcols; col++) {
    vec[col] = (*this)[row][col];
  };
  
  return vec;
}



template <class T>
Vector<T> Matrix<T>::GetCol(const int& col) const
{
  BIASASSERT(col >= 0 && col < this->num_cols());
  const unsigned int numrows = this->num_rows();
  Vector<T> vec(numrows);
  // copy the elements into the vector:
  for (unsigned int row = 0; row < numrows; row++) {
    vec[row] = (*this)[row][col];
  };
  return vec;
}

template<class T>
void Matrix<T>::SetRow(const int row, const Vector<T>& data) 
{
  BIASASSERT(row >= 0 && row < this->num_rows());
  const  int numcols = (int)this->num_cols();
  BIASASSERT(data.size() == numcols);
  for (register int col=0; col<numcols; col++)
    (*this)[row][col] = data[col];
}

template<class T>
void Matrix<T>::SetCol(const int col, const Vector<T>& data) 
{
  BIASASSERT(col >= 0 && col < this->num_cols());
  const  int numrows = (int)this->num_rows();
  BIASASSERT(data.size() == numrows);
  for (register int row=0; row<numrows; row++)
    (*this)[row][col] = data[row];
}

template<class T>
void Matrix <T>::SetIdentity() 
{
  SetZero();
  for (int i = 0; i < this->num_cols() && i < this->num_rows(); i++) {
    //cout << "actual: " << i << endl;
    (*this)[i][i] = 1;
  }
}


/*
template<class T>
void Matrix<T>::set_diagonal_elements(const Vector<T>& argVec, 
                                      const bool zero_others) 
{
  // Check the dimensions
  BIASASSERT( argVec.size() == min(num_cols(), num_rows()) );
  // Are the otther elements to be set to zero?
  if (zero_others)
    SetZero();
  for(int i = 0; i < argVec.size(); i++)
    (*this)[i][i] = argVec[i];
}
*/

template<class T>
T Matrix<T>::minVal() const 
{
  // initialize result value with last element
  register T* dataP = GetDataLast();
  T result = *dataP;

  for(; dataP >= GetData(); dataP--)
    if (*dataP < result)
      result = *dataP;

  return result;
}

template<class T>
T Matrix<T>::Normalize()
{
  T greatest = 0;
  for (int i=0; i < this->num_rows(); i++)
    for (int j=0; j < this->num_cols(); j++)
      if ( ((*this)[i][j])*((*this)[i][j]) > greatest*greatest ) 
        greatest = (*this)[i][j];
  (*this) /= greatest;
  return greatest;
}

template<class T>
T Matrix<T>::maxVal() const 
{
  // initialize result value with last element
  register T* dataP = GetDataLast();
  register T result = *dataP;

  for(; dataP >= GetData(); dataP--)
    if (*dataP > result)
      result = *dataP;

  return result;
}

template<class T>
T Matrix<T>::meanVal() const 
{
  // initialize result
  T result = 0;
  for(register T* dataP = GetDataLast(); dataP >= GetData(); dataP--)
      result = result + T(*dataP);
    //result += T(*dataP);

  result = result/T(num_elements());
  return result;
}

template<class T>
bool Matrix<T>::Load(const std::string & filename){
  std::ifstream fs( filename.c_str() );
  if (!fs) {
    return false; // error, could not open file. 
  } else {
    // read in 
    fs>>(*this);
    fs.close();
  };
  return true;
}


template<class T>
bool Matrix<T>::Save(const std::string & filename) const {
  std::ofstream fs( filename.c_str() );
  if (!fs) {
    return false; // error, could not open file. 
  } else {
    // write out to disk
    fs<<(*this);
    fs.close();
  };
  return true;
}


template<class T>
bool Matrix<T>::BinRead(const char * filename) const
{
  ifstream filein;
  //filein.open(filename, ios::in);
  filein.open(filename, ios::in|ios::binary);
  if ( !filein.is_open() )
    {
      cerr << "Error opening file '" << filename << "'" << endl;
      return false;
    }

  // read and verify dimension
  int m,n;
  filein.read( (char*) &m, sizeof(int) );
  filein.read( (char*) &n, sizeof(int) );

  if ( m!=this->m_ || n!=this->n_ )
    {
      BIASERR("Matrix has wrong dimension! This one is "
              << this->m_ << "x" << this->n_ << " but file is "
              << m << "x" << n << "." << endl);
      return false;
    }

  // read values
  for(int i=0; i<this->m_; i++)
    for(int j=0; j<this->n_; j++)
      filein.read( (char*) &((*this)[i][j]), sizeof(T) );

  filein.close();
  return true;
}


template<class T>
bool Matrix<T>::BinWrite(const char * filename) const
{
  ofstream fileout;
  //fileout.open(filename, ios::out);
  fileout.open(filename, ios::out|ios::binary);
  if ( !fileout.is_open() )
    {
      cerr << "Error opening file '" << filename << "'" << endl;
      return false;
    }

  // store the dimension
  fileout.write( (char*) &this->m_, sizeof(int) );
  fileout.write( (char*) &this->n_, sizeof(int) );

  // store values
  for(int i=0; i<this->m_; i++)
    for(int j=0; j<this->n_; j++)
      fileout.write( (char*) &((*this)[i][j]), sizeof(T) );

  fileout.close();
  return true;
}


template<>
bool Matrix<double>::IsZero( double eps /*=0.0*/) const
{
  bool result=true;
  for (int i=0; i<this->num_rows(); i++)
    for (int j=0; j<this->num_cols(); j++)
      if ( fabs((*this)[i][j]) > eps )
        result=false;
  return result;
}


template<class T>
bool Matrix<T>::IsZero( double eps /*=0.0*/) const
{
  if ( eps != 0 )
    {
      BIASERR("Epsilon environment is not implemented for this type!");
    }
  bool result=true;
  for (int i=0; i<this->num_rows(); i++)
    for (int j=0; j<this->num_cols(); j++)
      if ( (*this)[i][j] != 0 )
        result=false;
  return result;
}


template<>
bool Matrix<double>::IsIdentity( double eps /*=0.0*/ ) const
{
  bool result=true;
  for (int i=0; i<this->num_rows(); i++)
    for (int j=0; j<this->num_cols(); j++)
      {
        if ( i==j && fabs((*this)[i][j]-1.0) > eps )
          result=false;
        else if ( i!=j && fabs((*this)[i][j]) > eps )
          result=false;
      }
  return result;
}

template<class T>
bool Matrix<T>::IsIdentity( double eps /*=0.0*/ ) const
{
  if ( eps != 0 )
    {
      BIASERR("Epsilon environment is not implemented for this type!");
    }
  bool result=true;
  for (int i=0; i<this->num_rows(); i++)
    for (int j=0; j<this->num_cols(); j++)
      {
        if ( i==j && (*this)[i][j]!=1 )
          result=false;
        else if ( i!=j && (*this)[i][j] != 0 )
          result=false;
      }
  return result;
}

