/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _Matrix_hh_
#define _Matrix_hh_
#include "bias_config.h" 

#define _USE_MATH_DEFINES
#include <math.h>

#include <Base/Debug/Error.hh>

#include "tnt/tnt.h"
#include "tnt/cmat.h"
#include "tnt/fmat.h"
#include "Operators.hh"


namespace BIAS {

#ifndef DOUBLE_EPSILON
#define DOUBLE_EPSILON 1E-12
#endif

  template<class T> class Vector;

  /** @enum MatrixInitType
      @brief can be passed to matrix constructors to init the matrix with
      the most often used values */
  typedef enum {MatrixZero, MatrixIdentity} MatrixInitType;

  /**    @class Matrix
         @ingroup g_math
         @brief  matrix class with arbitrary size, indexing is row major. <br>
         Access the data by M[row][col].
         
         class Matrix is the matrix class that should be used by common BIAS 
         algorithms and data structures with arbitrary size <br>

         Access the data simply by M[row][col]. <br>
         
         It is derived from TNT::Matrix to inherit the basic operations and 
         algorithm.
         special implementations should be done here and NOT in TNT:Matrix 
         because the true 'base' Matrix class "TNT::Matrix"
         should be interchangeable and aside from this maintained by NIST.
    
         changed from MIP to BIAS (j1. 11/05/2002)
         operators are no longer class members, they are defined after the
         class, woelk

         @author Jan Woetzel, woelk 
  */
  template <class T=double>
  class BIASMathBase_EXPORT Matrix : public TNT::Matrix<T> 
  {
  public:
    virtual ~Matrix();

    Matrix() : TNT::Matrix<T>() {};

    

    /** @author Jan Woetzel
     **/
    Matrix(int rows, int cols) :TNT::Matrix<T>(rows, cols) {};

    /** assignment with a constant values for all elements (=set)
        @author Jan Woetzel **/
    Matrix(int rows, int cols, const T value) 
      : TNT::Matrix<T>(rows, cols, value) {};

    /** init with standard form
        @author koeser **/
    Matrix(int rows, int cols, const MatrixInitType& i) 
      : TNT::Matrix<T>(rows, cols, T(0)) {
      switch (i) {
      case MatrixZero:return;
      case MatrixIdentity:{ 
        for (int r=0; r<rows; r++) {
          if (r==cols) return;
          (*this)[r][r] = T(1);
        }
        return;
      }
      default: BIASERR("undefined MatrixInitType"); 
        abort();
      }
    };


    /** @author Jan Woetzel **/
    Matrix(int rows, int cols, const char *s) 
      : TNT::Matrix<T>(rows, cols, s){};

    /** @author Jan Woetzel **/
    Matrix(const Matrix<T> &A) : TNT::Matrix<T>(A) {};

    /** @author Jan Woetzel **/
    Matrix(const TNT::Matrix<T> &A) : TNT::Matrix<T>(A) {};

    /** transpose function, storing data destination matrix
        @author JMF   **/
    inline Matrix<T> Transpose() const;

    /////////////////////// Get Functions ////////////////////////////////
    /** @name Get Functions
        @{
    */
    /** get the pointer to the data array of the matrix 
        (for faster direct memeory access)
        
        the order of the data is linewise, which means an elemnt 
        sequence of e.g.
        [0] = M(0,0)
        [1] = M(0,1)
        [2] = M(0,2)
        [3] = M(1,0) // next line 
        [4] = M(1,1) ... 
        @author Jan Woetzel
        @return a pointer to the data array (beginning with GetData[0])
        @status alpha (02/27/2002)
        -added const (jw) 03/12/2002 **/
    inline  T *GetData() const 
    { return this->v_; };

    /** returns zero based arry for data access */
    inline T **GetDataArray() const
    { return this->row_; }

    /** returns 1 based array to data access */
    inline T **GetDataArray1() const
    { return this->rowm1_; };

    inline unsigned int GetRows() const { return this->m_; }

    inline unsigned int GetCols() const { return this->n_; }

    /** Get a pointer to the last data element
        @author Ingo Thomsen
        @date 04/11/2002 **/
    inline T* GetDataLast() const 
    { return GetData() + num_elements() - 1; };

    /** @brief return a copy of row "row" of this matrix, zero based counting
        @author Jan Woetzel **/
    Vector<T> GetRow(const int& row) const;

    /** @brief return a copy of column "col", zero based counting
        @author Jan Woetzel **/
    Vector<T> GetCol(const int& col) const;

    /** @brief set a row of matrix from vector
        @author woelk 08/2004 */
    void SetRow(const int row, const Vector<T>& data);

    /** @brief set a col of matrix from vector
        @author woelk 08/2004 */
    void SetCol(const int row, const Vector<T>& data);

     /** Returns the total number of elements
        @author Ingo Thomsen
        @date 04/11/2002
        @status tested    **/
    inline int num_elements() const 
    { return this->num_cols() * this->num_rows(); };

    /** Returns the minimum value of the matrix elements
        @author Ingo Thomsen
        @date 04/11/2002
        @status tested    **/
    T minVal() const;

    /** Returns the maximum value of the matrix elements
        @author Ingo Thomsen
        @date 04/11/2002
        @status tested    **/
    T maxVal() const;

    /** Returns the mean value of the matrix elements
        @author Ingo Thomsen
        @date 04/11/2002
        @status tested    **/
    T meanVal() const;

    /** Return the L1 norm: |a| + |b| + |c| + ...
        @author Ingo Thomsen
        @date 04/11/2002
        @status untested    **/
    inline T NormL1() const;
    inline T norm_L1() const
    { return NormL1(); }

    /** Return the L2 norm: a^2 + b^2 + c^2 + ...
        @author woelk 07/2004 */
    inline T NormL2() const;
    
    /// return the biggest entry
    inline T Max() const;
    /// return the smalles entry
    inline T Min() const;
    /// return biggest and smallest entry
    inline void MaxMin(T& max, T&min) const;

    /** @brief Return Frobenius norm = sqrt(trace(A^t * A))
        @author koeser 02/2004 */
    inline T NormFrobenius() const;


    //@}

    //////////////////////////// Arithmetic ////////////////////
    /** @name Arithmetic
        @{
    */
    /** \brief assignment operators
        calling corresponding operator from base class if appropriate 
        @author Jan Woetzel
        @status alpha (02/25/2002) */
    inline Matrix<T>& operator=(const TNT::Matrix<T> &mat);
    inline Matrix<T>& operator=(const Matrix<T> &mat);

    /** in place addition function
        @author Ingo Thomsen
        @status tested **/
    inline void AddIP(const T& scalar) 
    { Add(scalar, *this); };

    /** Adds arg to this 
        @author grest,2004
    */
    inline void AddIP(const Matrix<T> &arg);

    /** Subtracts arg from this <br>
        this -= arg 
        @author grest, 2004
    */
    inline void SubIP(const Matrix<T> &arg);

    /** addition function, storing data destination matrix
        @author Ingo Thomsen
        @status tested **/
    inline void Add(const T& scalar,  Matrix<T> &dest) const;

    /** in place subtraction function
        @author Ingo Thomsen
        @status tested **/
    inline void SubIP(const T& scalar) 
    { Sub(scalar, *this); };

    /** substraction function, storing data destination matrix
        @author Ingo Thomsen
        @status tested    **/
    inline void Sub(const T& scalar,  Matrix<T> &dest) const;

    /** in place multiplication function
        @author Ingo Thomsen
        @status tested **/
    inline void MultiplyIP(const T& scalar) 
    { Multiply(scalar, *this); };

    /** multiplication function, storing data destination matrix
        @author Ingo Thomsen
        @status tested    **/
    inline void Multiply(const T& scalar,  Matrix<T> &dest) const;

    /** matrix multiplication, result is not allocated
        @author Felix Woelk */
    inline void Mult(const Matrix<T>& arg, Matrix<T>& result) const;

    /** in Place matrix multiplication
        this is equal to M = M * arg, but faster
        @author Daniel Grest */
    inline void Mult(const Matrix<T>& arg);
  
    /** in Place matrix multiplication
        this is equal to M = arg*M, but faster
        @author Daniel Grest */
    inline void MultLeft(const Matrix<T>& arg);

    /** matrix vector multiplication, result is not allocated
        @author Felix Woelk */
    inline void Mult(const Vector<T>& arg, Vector<T>& result) const;

    /** vector matrix multiplication result=arg*this.
        @author Marcel Lilienthal */
    inline void MultLeft(const Vector<T>& arg, Vector<T>& result) const;

    /** in place division function
        @author Ingo Thomsen
        @status tested    **/
    inline void DivideIP(const T& scalar) 
    { Divide(scalar, *this); };

    /** division function, storing data destination matrix
        @author Ingo Thomsen
        @status tested    **/
    inline void Divide(const T& scalar,  Matrix<T> &dest) const;
  

    /** @brief compute square system matrix dest = A^T * A
	@param dest holds result of this->Transpose * this

        If you want to solve A * x = b, where A has more rows than columns,
        a common technique is to solve x = (A^T * A)^-1 * A^T * b.
        This function provides a fast way to compute A^T*A from A.
        @author grest/koeser */
    inline void GetSystemMatrix(Matrix<T> &dest) const;

    //@}
    ///////////////  Set Functions //////////////////////////////////////////
    /** @name Set Functions
        @{
    */
    /** Sets all values to zero
        @author Ingo Thomsen
        @date 04/11/2002
        @status tested    **/
    inline void SetZero();
    inline void set_zero()
    { return SetZero(); };
    

    /** @brief Converts matrix to identity matrix

    Even if the matrix is not quadratic, the diagonal values
    starting with the upper left element are set to 1. Example:

    1 0 0       1 0 0 0
    0 1 0   or  0 1 0 0 
    0 0 1       0 0 1 0
    0 0 0

    @author Ingo Thomsen       
    **/
    void SetIdentity();
    inline void set_identity()
    { SetIdentity(); };
    
    /** Takes the elements of a Vector and put them as diagonal elements into
        a matrix. The size of this Vector must be exactly 
        min( matrix.width, matrix.height). Optionally all other elements may 
        be set to zero
        @author Ingo Thomsen
        @param zero_others Must be true, if all non-diagonal elements shall be
        set to zero    **/
    //void set_diagonal_elements(const Vector<T>& argV, const bool zero_others);
   
    void Fill(const T& scalar);

    // @}
    
    //////////////////// Scaling and Normalization //////////////////////////
    /** @name  Scaling and Normalization 
        @{
    */
    /** divides each column of the matrix through the element norm_row_index.
        index runs [0..num_rows-1]
        for example:
        2 3  
        4 5  
        normalized with row '1' will be
        2/4  3/5
        1    1     **/
    void normalizeColsToOne(const int norm_row_index);

    /** divides each row of the matrix through the element norm_col_index.
        index runs [0..num_cols-1]
        for example:
        2 3  
        4 5  
        normalized with row '1' will be
        2/3  1
        4/5  1      **/
    void normalizeRowsToOne(const int norm_col_index);

    /** Normalizes each row to L2 norm one. Attention each row is normaized
        by its own scale! */
    void NormalizeRows();

    /** Normalizes each coloumn to L_2 norm one. Attention each column is 
        normaized by its own scale!   */
    void NormalizeCols();

    /** Normalizes the matrix by the entry with the biggest absolute value
        by dividing all elements with this one
        @return value of the element the matrix is normalized with
        @author mdunda 09 2003 */
    T Normalize();
  
    /** Scales row NoRow with scale.      */
    void ScaleRow(int NoRow,T scale);

    /** Scales column NoCol with scale.     */
    void ScaleCol(int NoCol,T scale);
    // @}

    /////////////////// Input / Output ////////////////////////////////////
    /** @name Input / Output
        @{
    */
    /** method to load directly from a given filename. 
        internally using stream operator 
        @author Jan Woetzel 05/2003 
        @return false in case of error, true in case of success
    */
    bool Load(const std::string & filename); 
    
    /** method to save directly to a given filename. 
        internally using stream operator
        @author Jan Woetzel 05/2003 
        @return false in case of error, true in case of success
    */
    bool Save(const std::string & filename) const;

    /** This method reads a matrix from a given file in binary format.
        The file is not human read-/editable but provides full precision.
        A dimension check is performed.
        @return true on success
        @author mdunda 04 2004 */
    bool BinRead(const char * filename) const;

    /** This method writes a matrix to a given file in binary format.
        The file is not human read-/editable but provides full precision.
        @return true on success
        @author mdunda 04 2004 */
    bool BinWrite(const char * filename) const;

    // @}

    /** Checks if the matrix is a null matrix.
        @author mdunda 12 2003
        @return true if all elements are zero */
    bool IsZero( double eps=0.0 ) const;

    /** Checks if the matrix a an identity. I.e. all elements with index
        i==j are equal 1 and all others are zero.
        @author mdunda 12 2003
        @return true if matrix is identity */
    bool IsIdentity( double eps=0.0 ) const;

  }; // class

  //////////////////////////////////////////////////////////////////
  ///  operators
  ///////////////////////////////////////////////////////////////////

  /**  @relates Matrix
       addition operator
       @author Ingo Thomsen **/
  template<class T>
  inline Matrix<T>& operator+=(Matrix<T>& mat, const T& scalar)
  { mat.AddIP(scalar); return mat; }

  /**  @relates Matrix
       addition in place operator
       @author grest **/
  template<class T>
  inline Matrix<T>& operator+=(Matrix<T>& mat, const Matrix<T>& arg)
  { mat.AddIP(arg); return mat; }

  /**  @relates Matrix
       subtracts arg from mat, mat = mat - arg
       @author grest **/
  template<class T>
  inline Matrix<T>& operator-=(Matrix<T>& mat, const Matrix<T>& arg)
  { mat.SubIP(arg); return mat; }

  /** @relates Matrix
      substraction operator
      @author Ingo Thomsen */
  template<class T>
  inline Matrix<T>& operator-=(Matrix<T>& mat, const T& scalar)
  { mat.AddIP(scalar); return mat; }

  /** @relates Matrix
      multiplication operator
      @author Ingo Thomsen */
  template<class T>
  inline Matrix<T>& operator*=(Matrix<T>& mat, const T& scalar)
  { mat.MultiplyIP(scalar); return mat; }

  
  
  
  /** @relates Matrix
      operator for equal
      @author Ingo Thomsen (07/02/2002)
      @status tested  */
  template<class T>
  inline bool operator==(const Matrix<T>& argl, const Matrix<T>& argr)
  {
    BIASASSERT( argl.size() == argr.size() );
    register T* argP = argl.GetData();
    register T* matP = argr.GetData();
    for (; matP < argr.GetDataLast() ; matP++, argP++)
      if (*matP != *argP)
        return false;
    return true;
  }

  /** @relates Matrix
      operator for not equal
      @author Ingo Thomsen (07/02/2002) **/
  template<class T>
  inline bool operator!=(const Matrix<T>& argl, const Matrix<T>& argr) 
  { return !(argl==argr); }

  ////////////////////////////////////////////////////////////////////
  /// implementation
  ///////////////////////////////////////////////////////////////////

  template<class T>
  inline void Matrix<T>::Add(const T& scalar,  Matrix<T> &dest) const 
  {
    if (this->num_rows() != dest.num_rows() || 
        this->num_cols() != dest.num_cols())
      dest.newsize(this->num_rows(), this->num_cols());
    register T* destP = dest.GetData();
    register T* matP = GetData();
    for(;matP <= GetDataLast(); matP++, destP++) {
      *destP = *matP + scalar;
    }
  }
  
  template<class T>
  inline void Matrix<T>::Sub(const T& scalar,  Matrix<T> &dest) const 
  {
    if (this->num_rows() != dest.num_rows() || 
        this->num_cols() != dest.num_cols())
      dest.newsize(this->num_rows(), this->num_cols());
    register T* destP = dest.GetData();
    register T* matP = GetData();
    for(;matP <= GetDataLast(); matP++, destP++) {
      *destP = *matP - scalar;
    }
  }

  template<class T>
  inline void Matrix<T>::Multiply(const T& scalar,  Matrix<T> &dest) const 
  {
    if (this->num_rows() != dest.num_rows() || this->num_cols() != dest.num_cols())
      dest.newsize(this->num_rows(), this->num_cols());
    register T* destP = dest.GetData();
    register T* matP = GetData();
    for(;matP <= GetDataLast(); matP++, destP++) {
      *destP = *matP * scalar;
    }
  }

  template<class T>
  inline void Matrix<T>::Divide(const T& scalar,  Matrix<T> &dest) const 
  {
    if (this->num_rows() != dest.num_rows() || this->num_cols() != dest.num_cols())
      dest.newsize(this->num_rows(), this->num_cols());
    if (scalar == 0) {
      BIASERR("Division by Zero");
      //BIASASSERT(false);
      abort();
    } else {
      register T* destP = dest.GetData();
      register T* matP = GetData();
      for(;matP <= GetDataLast(); matP++, destP++) {
        *destP = *matP / scalar;
      }
    }
  }

  template<class T>
  inline Matrix<T> Matrix<T>::Transpose() const 
  {
    Matrix<T> Out(this->num_cols(),this->num_rows());
    for (int i=0; i<this->num_rows(); i++)
      for (int j=0; j<this->num_cols(); j++)
        Out[j][i]= (*this)[i][j];
    return Out;  
  }

  template<class T>
  inline void Matrix<T>::SetZero() 
  {
    for(register T* dataP = GetData(); dataP <= GetDataLast(); dataP++)
      *dataP = (T)(0);
  }

  template<class T>
  inline T Matrix<T>::NormL1() const 
  {
    T result = 0;
    for(register T* dataP = GetDataLast(); dataP >= GetData(); dataP--) {
      result += *dataP > 0 ? *dataP : *dataP * -1;
    }
    return result;
  }

  template<class T>
  inline T Matrix<T>::NormL2() const 
  {
    T result = 0;
    for (register T* dataP = GetDataLast(); dataP >= GetData(); dataP--) {
      result += *dataP * *dataP;
    }
    return (T) sqrt((double)result);
  }

  template<class T>
  inline T Matrix<T>::Max() const 
  {
    T result = *GetDataLast();
    for(register T* dataP = GetDataLast(); dataP >= GetData(); dataP--) {
      if (*dataP>result) result = *dataP;
    }
    return result;
  }

  template<class T>
  inline T Matrix<T>::Min() const 
  {
    T result = *GetDataLast();
    for(register T* dataP = GetDataLast(); dataP >= GetData(); dataP--) {
      if (*dataP<result) result = *dataP;
    }
    return result;
  }


  template<class T>
  inline void Matrix<T>::MaxMin(T& max, T& min) const 
  {
    max = min = *GetDataLast();
    for(register T* dataP = GetDataLast(); dataP >= GetData(); dataP--) {
      if (*dataP<min) min = *dataP;
      else if (*dataP>max) max = *dataP;
    }
  }



  template<class T>
  inline T Matrix<T>::NormFrobenius() const {
    register T result = 0;
    // compute only diagonal elements of A^T*A for trace
    register int x, i;
    for (x=0; x<this->num_rows(); x++) {
      if (x<this->num_cols()) {
        for (i=0; i<this->num_cols(); i++) {
          result += (*this)[x][i]*(*this)[x][i];
        }
      }
    }
    // frobenius norm is square root of that trace
    return (T) sqrt((double)result);
  }

  template<class T>
  inline void Matrix<T>::Mult(const Matrix<T>& arg, Matrix<T>& result) const
  {
#ifdef BIAS_DEBUG
    if (this->num_cols() != arg.num_rows() ||
        result.num_cols() != arg.num_cols() ||
        result.num_rows() != this->num_rows()){
      BIASERR("invalid matrix sizes "<<this->num_rows()<<"x"<< this->num_cols()
              <<" * "<<arg.num_rows()<<"x"<<arg.num_cols()
              <<"  =  "<<result.num_rows()<<"x"<<result.num_cols());
      //BIASASSERT(false);
      exit(-1);
    }
#endif    
    if (result.num_cols() != arg.num_cols() ||
        result.num_rows() != this->num_rows()){
      result.newsize(this->num_rows(), arg.num_cols());
    }
    result.SetZero();
    register int x, y, i;
    for (x=0; x<result.num_cols(); x++)
      for (y=0; y<result.num_rows(); y++)
        for (i=0; i<this->num_cols(); i++)
          result[y][x]+=(*this)[y][i]*arg[i][x];
  }

  template<class T>
  inline void Matrix<T>::Mult(const Matrix<T>& arg)  {
    Matrix<T> result(this->num_rows(),arg.num_cols());
    Mult(arg,result);
    (*this)=result;
  }
  
  template<class T>
  inline void Matrix<T>::MultLeft(const Matrix<T>& arg)  {
    Matrix<T> result(arg.num_rows(),this->num_cols());
    arg.Mult(*this,result);
    (*this)=result;
  }
  
  template<class T>
  inline void Matrix<T>::Mult(const Vector<T>& arg, Vector<T>& result) const {
#ifdef BIAS_DEBUG
    if (this->num_cols() != arg.size() ||
        result.size() != this->num_rows()) {
      BIASERR("invalid matrix/vector sizes "<<this->num_rows()<<"x"<< this->num_cols()
              <<" * "<<arg.size()<<"x1"
              <<"  =  "<<result.size()<<"x1");
      //BIASASSERT(false);
      exit(-1);
    }
#endif    
    if (result.size() != this->num_rows()) {
      result.newsize(this->num_rows());
    }
    result.Fill((T)0);
    register int y, i;
    for (y=0; y<result.size(); y++)
      for (i=0; i<this->num_cols(); i++)
        result[y]+=(*this)[y][i]*arg[i];
  }

  template<class T>
  inline void Matrix<T>::MultLeft(const Vector<T>& arg, 
                                  Vector<T>& result) const {
#ifdef BIAS_DEBUG
    if (this->num_rows() != arg.size() ||
        result.size() != this->num_cols()) {
      BIASERR("invalid matrix/vector sizes "
              <<arg.size()<<"x1"<<" * "
              <<this->num_rows()<<"x"<< this->num_cols()
              <<"  =  "<<result.size()<<"x1");
      exit(-1);
    }
#endif    
    if (result.size() != this->num_cols()) {
      result.newsize(this->num_cols());
    }
    result.SetZero();
    register int y, i;
    for (y=0; y<this->num_cols(); y++)
      for (i=0; i<this->num_rows(); i++)
        result[y]+=(*this)[i][y]*arg[i];

  }

  template<class T>
  inline void Matrix<T>::AddIP(const Matrix<T> &arg)
  {
#ifdef BIAS_DEBUG
    if (this->num_cols() != arg.num_cols() ||
        this->num_rows() != arg.num_rows()) {
      BIASERR("invalid matrix sizes "<<this->num_rows()<<"x"<< this->num_cols()
              <<" + "<<arg.num_rows()<<"x"<<arg.num_cols());
      //BIASASSERT(false);
      exit(-1);
    }
#endif    
    T* thisData=GetData();
    T* argData=arg.GetData();
    register unsigned int size=this->num_rows()*this->num_cols();
    for (unsigned int i=0;i<size;i++) {
      *thisData = *thisData + T(*argData);
      thisData++; argData++;
    }
  }

  template<class T>
  inline void Matrix<T>::SubIP(const Matrix<T> &arg)
  {
#ifdef BIAS_DEBUG
    if (this->num_cols() != arg.num_cols() ||
        this->num_rows() != arg.num_rows()) {
      BIASERR("invalid matrix sizes "<<this->num_rows()<<"x"<< this->num_cols()
              <<" - "<<arg.num_rows()<<"x"<<arg.num_cols());
      //BIASASSERT(false);
      exit(-1);
    }
#endif    
    T* thisData=GetData();
    T* argData=arg.GetData();
    register unsigned int size=this->num_rows()*this->num_cols();
    for (unsigned int i=0;i<size;i++) {
      *thisData = *thisData  - T(*argData);
      thisData++; argData++;
    }
  }

  template <class T>
  inline void Matrix<T>::GetSystemMatrix(Matrix<T> &dest) const {
    const unsigned int colsJ = this->num_cols(), rowsJ = this->num_rows(); 
    // resize result to square matrix   
    dest.newsize(colsJ, colsJ);
    
    // get output array, account for ++prefix by -1
    T *pDestData = dest.GetData()-1;
    for (unsigned int row = 0; row<colsJ; row++) {
      for (unsigned int col = 0; col<colsJ; col++) {
        // init matrix field [row][col] with 0
	*(++pDestData) = 0;
	for (unsigned int k = 0; k<rowsJ; k++) {
          // avoid unneccessary pointer calculations
          // dest[row][col] += J[k][row] * J[k][col];
	  const T* pJ_k_ = this->row_[k];
	  *pDestData += pJ_k_[row] * pJ_k_[col];
	}
      }
    }
  }
  
  template <class T>
  inline Matrix<T> &Matrix<T>::operator=(const TNT::Matrix<T> &mat) 
  {
    TNT::Matrix<T>::operator=(mat);	// call the operator of the base class
    return *this;
  }
  
  template <class T>
  inline Matrix<T> &Matrix<T>::operator=(const Matrix<T> &mat) 
  {
    TNT::Matrix<T>::operator=(mat);	// call the operator of the base class
    return *this;
  }
  
} // namespace BIAS
#endif // _Matrix_hh_
