/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Base/Common/BIASpragma.hh>

#include "Matrix3x3.hh"

#include <vector>
#include <sstream>
#include <fstream>


using namespace BIAS;
using namespace std;


template<class T>
Matrix3x3<T>::Matrix3x3(const Matrix<T>& mat) : Matrix<T>(mat)
{
  if ((mat.num_rows() != 3) || (mat.num_cols() != 3)) {
    BIASERR("wrong matrix size "<<mat.num_rows()<<"x"
           << mat.num_cols());
  }
}

template<class T>
Matrix3x3<T>::Matrix3x3(const TNT::Matrix<T>& mat) : Matrix<T>(mat)
{
  if ((mat.num_rows() != 3) || (mat.num_cols() != 3)) {
    BIASERR("wrong matrix size "<<mat.num_rows()<<"x"
           << mat.num_cols());
  }
}

template<class T>
Matrix3x3<T>::~Matrix3x3() 
{}


template<class T> 
Matrix3x3<T> Matrix3x3<T>::Transpose() const
{
    Matrix3x3<T> ret;
    ret = *this;

    T *retd = ret.GetData();
    T *thisd = this->GetData();
    retd[1] = thisd[3];
    retd[3] = thisd[1];
    retd[2] = thisd[6];
    retd[6] = thisd[2];
    retd[5] = thisd[7];
    retd[7] = thisd[5];
    return ret;
}



template <class T> 
Matrix3x3<T>& Matrix3x3<T>::newsize(int rows, int cols) 
{
  if ((rows != 3) || (cols != 3)) {
    // do nothing
    // print error message if the new size should not be a 3x4 matrix
    BIASERR("The size of a Matrix3x3<T> has to be 3 rows, 3 cols. "
           <<"newsize can't create a " 
           << rows << " x " << cols << " Matrix3x3 !");
#if DEBUG
    exit(-1);
#endif
    return *this;
  } else {
    BIASASSERT( rows==3 ); // 3x3 matrix
    BIASASSERT( cols==3 );
    // call base class function 
    // (usually unnecessary except for first construction)
    Matrix<T>::newsize(3, 3);
    return *this;
  }
}

template<class T>
void Matrix3x3<T>::SetFromVector(const TNT::Vector<T>& vec) 
{
  BIASASSERT( vec.size() == 9 );
  // fill the matrix rowwise with the first 9 elements of the vector
  // example:
  // 1 2 3 4 5 6 7 8 9
  //  -->
  // 1 2 3
  // 4 5 6
  // 7 8 9
  for (int i=0; i<9; i++) {
    this->GetData()[i] = vec[i];
  }
}

//JW 07/2003
template<class T>
void Matrix3x3<T>::SetFromColumnVectors(const BIAS::Vector3<T> & v0,
                                        const BIAS::Vector3<T> & v1,
                                        const BIAS::Vector3<T> & v2 )
{
  // will be loop-unrolled ..
  // for all elements of a column:
  for (unsigned int row=0; row<3; row++){
    (*this)[row][0]=v0[row];
    (*this)[row][1]=v1[row];
    (*this)[row][2]=v2[row];
  };
}

//JW 07/2003
template<class T>
void Matrix3x3<T>::SetFromRowVectors(const BIAS::Vector3<T> & v0,
                                     const BIAS::Vector3<T> & v1,
                                     const BIAS::Vector3<T> & v2 )
{
  // will be loop-unrolled ..
  // for all elements of a row:
  for (unsigned int col=0; col<3; col++){
    (*this)[0][col]=v0[col];
    (*this)[1][col]=v1[col];
    (*this)[2][col]=v2[col];
  };
}

// JW
template<class T>
void Matrix3x3<T>::GetRow(const unsigned int row, Vector3<T>& v) const 
{
  BIASASSERT(row<3);// this has only 3 rows[0,1,2]
  T* p=this->GetData()+3*row;
  v[0]=*p++;
  v[1]=*p++;
  v[2]=*p;
  //  BIAS::Vector3<T> v( &(*this)[row][0] );
  //return v;
}

// JW
template<class T>
BIAS::Vector3<T> Matrix3x3<T>::GetColumn(const unsigned int col) const {
  BIASASSERT(col<3);// this has only 3 cols[0,1,2]
  BIAS::Vector3<T> v( (*this)[0][col], (*this)[1][col], (*this)[2][col] );
  return v;
}

template<class T>
T Matrix3x3<T>::GetDeterminant() const
{
  double t4 = (*this)[0][0]*(*this)[1][1];
  double t6 = (*this)[0][0]*(*this)[1][2];
  double t8 = (*this)[0][1]*(*this)[1][0];
  double t10 = (*this)[0][2]*(*this)[1][0];
  double t12 = (*this)[0][1]*(*this)[2][0];
  double t14 = (*this)[0][2]*(*this)[2][0];
  return T((t4*(*this)[2][2]-t6*(*this)[2][1]-t8*(*this)[2][2]+
          t10*(*this)[2][1]+t12*(*this)[1][2]-t14*(*this)[1][1]));
}

template<class T>
Matrix3x3<T> Matrix3x3<T>::Invert() const
{
    /// JW: for what reason should this be deprecated???
    //BIASERR("warning Matrix3x3<T>::Invert() is deprecated");


    Matrix3x3<T> i;
#ifdef BIAS_DEBUG
    int res =
#endif
        Invert(i);

#ifdef BIAS_DEBUG
    if (res!=0) BIASERR("unable to invert matrix "<<*this);
#endif
    BIASASSERT(res==0);

    return i;
}



template<class T>
int Matrix3x3<T>::Invert(Matrix3x3<T>& x) const
{
  int res=0;
  BIASASSERT(x.GetData()!=this->GetData());
  // following lines have been produced by Maple  
  const double t4 = (*this)[0][0]*(*this)[1][1];
  const double t6 = (*this)[0][0]*(*this)[1][2];
  const double t8 = (*this)[0][1]*(*this)[1][0];
  const double t10 = (*this)[0][2]*(*this)[1][0];
  const double t12 = (*this)[0][1]*(*this)[2][0];
  const double t14 = (*this)[0][2]*(*this)[2][0];
  
  const double det=(t4*(*this)[2][2]-t6*(*this)[2][1]-t8*(*this)[2][2]+t10
                    *(*this)[2][1]+t12*(*this)[1][2]-t14*(*this)[1][1]);
  // matrices with zero determinante ar not invertible
  if (det==0.0)
    return -1;

  const double t17 = 1.0/det;

  x[0][0] = T(((*this)[1][1]*(*this)[2][2]-(*this)[1][2]*(*this)[2][1])*t17);
  x[0][1] = T(-((*this)[0][1]*(*this)[2][2]-(*this)[0][2]*(*this)[2][1])*t17);
  x[0][2] = T(-(-(*this)[0][1]*(*this)[1][2]+(*this)[0][2]*(*this)[1][1])*t17);
  x[1][0] = T(-((*this)[1][0]*(*this)[2][2]-(*this)[1][2]*(*this)[2][0])*t17);
  x[1][1] = T(((*this)[0][0]*(*this)[2][2]-t14)*t17);
  x[1][2] = T(-(t6-t10)*t17);
  x[2][0] = T(((*this)[1][0]*(*this)[2][1]-(*this)[1][1]*(*this)[2][0])*t17);
  x[2][1] = T(-((*this)[0][0]*(*this)[2][1]-t12)*t17);
  x[2][2] = T((t4-t8)*t17);

  return res;
}

template<class T>
int Matrix3x3<T>::InvertIP()
{
  int res=0;
  // following lines have been produced by Maple  
  const double t4 = (*this)[0][0]*(*this)[1][1];
  const double t6 = (*this)[0][0]*(*this)[1][2];
  const double t8 = (*this)[0][1]*(*this)[1][0];
  const double t10 = (*this)[0][2]*(*this)[1][0];
  const double t12 = (*this)[0][1]*(*this)[2][0];
  const double t14 = (*this)[0][2]*(*this)[2][0];
  
  const double det=(t4*(*this)[2][2]-t6*(*this)[2][1]-t8*(*this)[2][2]+t10
                    *(*this)[2][1]+t12*(*this)[1][2]-t14*(*this)[1][1]);
  // matrices with zero determinante ar not invertible
  if (det==0.0) return -1;

  const double t17 = 1.0/det;

#ifdef BIAS_DEBUG
  Matrix3x3<T> tm=(*this);
#endif

  T tmp[9];

  tmp[0] = T(((*this)[1][1]*(*this)[2][2]-(*this)[1][2]*(*this)[2][1])*t17);
  tmp[1] = T(-((*this)[0][1]*(*this)[2][2]-(*this)[0][2]*(*this)[2][1])*t17);
  tmp[2] = T(-(-(*this)[0][1]*(*this)[1][2]+(*this)[0][2]*(*this)[1][1])*t17);
  tmp[3] = T(-((*this)[1][0]*(*this)[2][2]-(*this)[1][2]*(*this)[2][0])*t17);
  tmp[4] = T(((*this)[0][0]*(*this)[2][2]-t14)*t17);
  tmp[5] = T(-(t6-t10)*t17);
  tmp[6] = T(((*this)[1][0]*(*this)[2][1]-(*this)[1][1]*(*this)[2][0])*t17);
  tmp[7] = T(-((*this)[0][0]*(*this)[2][1]-t12)*t17);
  tmp[8] = T((t4-t8)*t17);

  memcpy((void*)(this->GetData()), (void*)tmp, 9*sizeof(T));

#ifdef BIAS_DEBUG
  Matrix3x3<T> tm2;
  tm.Invert(tm2);
  int tres=memcmp(this->GetData(), tm2.GetData(), 9*sizeof(T));
  if (tres!=0) cout << (*this)<<"\n"<<tm2<<endl;
  BIASASSERT(tres==0);
#endif

  return res;
}



// JW 03/2005
template <class T>
int Matrix3x3<T>::LoadBogRC(const std::string & filename)
{
    this->SetZero();// reset this
    ifstream fs( filename.c_str() );
    if (!fs) {
        BIASERR("could not open file "<<filename);
        return -1; // error, could not open file.
    };
    // parse and read in 
    const unsigned int nMax=1024;
    char caLine[nMax];
    // std string for find
    string strLine;
    // helper for string to number /atof)
    istringstream iss;
    unsigned int pos=0, line=0;
    vector<string> pattern;
    pattern.clear();
    pattern.push_back("RC11 ="); // 0
    pattern.push_back("RC12 ="); // 1
    pattern.push_back("RC13 ="); // 2

    pattern.push_back("RC21 ="); // 3
    pattern.push_back("RC22 ="); // 4
    pattern.push_back("RC23 ="); // 5

    pattern.push_back("RC31 ="); // 6
    pattern.push_back("RC32 ="); // 7
    pattern.push_back("RC33 ="); // 8

    vector< bool > found;
    found.resize( pattern.size(), false);
    BIASASSERT(found.size()>8);
    while (!fs.eof() && (fs) 
        && (
        !found[0] 
        || !found[1] 
        || !found[2] 
        || !found[3] 
        || !found[4] 
        || !found[5] 
        || !found[6] 
        || !found[7] 
        || !found[8]) 
            ) 
        {
            line++;
            fs.getline(&caLine[0], nMax);
            strLine = caLine;
            for (unsigned int i=0; i<pattern.size(); i++){
                pos = strLine.find(pattern[i]);
                if (pos != string::npos){
                    found[i]=true;
                    iss.clear();
                    iss.str(  // after pattern: start, length
                        strLine.substr(pos+pattern[i].size(),  strLine.size()-pos-pattern[i].size()) 
                        );
                    BIASASSERT(i<9); // dim
                    iss>> this->GetData()[i];
                };
            };
        };
        // check if we found all patterns    
        BIASASSERT(found.size()==pattern.size());
        bool foundAll=true;
        for (unsigned int j=0; j<found.size(); j++) foundAll=foundAll&&found[j];
        if (!foundAll) return -99; // missed something
        return 0; // OK.
}



// JW 03/2005
template <class T>
int Matrix3x3<T>::LoadBogK(const std::string & filename)
{
    SetIdentity();// reset this
    ifstream fs( filename.c_str() );
    if (!fs) {
        BIASERR("could not open file "<<filename);
        return -1; // error, could not open file.
    };
    // parse and read in 
    const unsigned int nMax=1024;
    char caLine[nMax];
    // std string for find
    string strLine;
    // helper for string to number /atof)
    istringstream iss;
    unsigned int pos=0, line=0;
    vector<string> pattern;
    pattern.clear();

    pattern.push_back("FC_U ="); // 0
    pattern.push_back("SKEW ="); // 1
    pattern.push_back("CC_U ="); // 2
    // unused place holder: 
    pattern.push_back("UNUSED_UNUSED-x82S6jd_UNUSED "); // 3
    pattern.push_back("FC_V ="); // 4
    pattern.push_back("CC_V ="); // 5

    vector< bool > found;
    found.resize( pattern.size(), false);
    BIASASSERT(found.size()>5);
    while (!fs.eof() && (fs) 
        && (
        !found[0] 
        || !found[1] 
        || !found[2] 
        //|| !found[3] //unused
        || !found[4] 
        || !found[5] ) 
            ) 
        {
            line++;
            fs.getline(&caLine[0], nMax);
            strLine = caLine;
            for (unsigned int i=0; i<pattern.size(); i++){
                if (i!=3){
                    pos = strLine.find(pattern[i]);
                    if (pos != string::npos){
                        found[i]=true;
                        iss.clear();
                        iss.str(  // after pattern: start, length
                            strLine.substr(pos+pattern[i].size(),  strLine.size()-pos-pattern[i].size()) 
                            );
                        BIASASSERT(i<9); // dim
                        iss>> this->GetData()[i];
                    }
                }
            };
        };
        // check if we found all patterns    
        BIASASSERT(found.size()==pattern.size());
        bool foundAll=true;
        for (unsigned int j=0; j<found.size(); j++) {
            if (j!=3) foundAll=foundAll&&found[j];
        }
        if (!foundAll) return -99; // missed something
        return 0; // OK.
}

