/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "CornerDetectorHarris.hh"

using namespace std;
using namespace BIAS;


//////////////////////////////////////////////////////////////////////////
//                 implementation
//////////////////////////////////////////////////////////////////////////



template <class StorageType, class CalculationType>
CornerDetectorHarris<StorageType, CalculationType>::
CornerDetectorHarris(): 
CornerDetectorGradient<StorageType,CalculationType>()
{
  this->ParamsAllocated_ = true;
  _Scale = new double;
  *_Scale = 0.04;
}

template <class StorageType, class CalculationType>
CornerDetectorHarris<StorageType, CalculationType>::
CornerDetectorHarris(Param& para, const std::string& postfix): 
CornerDetectorGradient<StorageType,CalculationType>(para, postfix)
{
  this->ParamsAllocated_ = false;
 int grp=para.GetFreeGroupID();
  _AddParam(para, grp, "-CornerDetectorHarris"+postfix);
  if (para.IsUsedGroupID(grp)) para.SetGroupName(grp,"CornerDetectorHarris");
}

template <class StorageType, class CalculationType>
CornerDetectorHarris<StorageType,CalculationType>::
CornerDetectorHarris(Param& para, const int grp, const std::string& postfix)
  : CornerDetectorGradient<StorageType, 
                           CalculationType>(para, grp, postfix)
{
  this->ParamsAllocated_ = false;
  _AddParam(para, grp, "-CornerDetectorHarris"+postfix);
}

template <class StorageType, class CalculationType>
CornerDetectorHarris<StorageType,CalculationType>::
~CornerDetectorHarris() 
{
  if (this->ParamsAllocated_)
    delete _Scale;
}

template <class StorageType, class CalculationType> int
CornerDetectorHarris<StorageType, CalculationType>::
_ComputeCornerness(Image<CalculationType>& Cornerness)
{
  // preserver ROI
  int tlx, tly, brx, bry;
  this->_sgxx.GetROI()->GetCorners(tlx, tly, brx, bry);
  Cornerness.GetROI()->SetCorners(tlx, tly, brx, bry);

  const unsigned int width=Cornerness.GetWidth();
  const unsigned int height=Cornerness.GetHeight();

  // work pointers to summed multiplied derivatives
  CalculationType *psgxx=this->_sgxx.GetImageData(), 
    *psgyy=this->_sgyy.GetImageData(), 
    *psgxy=this->_sgxy.GetImageData();
  CalculationType *ptr, *end, tmp;

  // init min und max with impossible values
  //_max = FLT_MIN;
  //_min = FLT_MAX;
  float scale = float ( *_Scale );
  // init all pointers to first usable position (skip upper two lines)
  /// \todo use ROI instead of border=2
  unsigned offs = 2*width+2;
  unsigned int x=2,y=2;
  ptr = Cornerness.GetImageData()+offs;
  psgxx += offs;
  psgxy += offs;
  psgyy += offs;
  end   = Cornerness.GetImageData()+width*(height-2)-2;
  const double me=(double)*this->_MinCornerness;
  Feat fp; 
  this->_FeatList.clear(); 
  this->_FeatList.reserve(end-ptr);
  // run over all pixels, dont care about borders
  while (ptr<end) {
    const CalculationType xx = (CalculationType) *psgxx++;
    const CalculationType xy = (CalculationType) *psgxy++;
    const CalculationType yy = (CalculationType) *psgyy++;

    // cornerness = det - scale*trace*trace
    tmp = (CalculationType) (xx*yy-xy*xy - scale * (xx+yy) * (xx+yy));

    // save highest/lowest values for quality analysis later
    //if (tmp>_max) _max = tmp;
    //else if (tmp<_min) _min=tmp;
    if (tmp>me){
        fp.x=x;
        fp.y=y;
        fp.val=(tmp>(double)INT_MAX)?(INT_MAX):((int)tmp);
        this->_FeatList.push_back(fp);
        //cout <<"Found a point above min cornerness: "<<endl;
      }
    *ptr++ = tmp;
    if (++x>=width){
      y++;
      x=0;
    }
  }
  return 0;
}

template <class StorageType, class CalculationType>
void CornerDetectorHarris<StorageType, CalculationType>::
_AddParam(Param& para, const int grp, const std::string& postfix) 
{
  if (para.CheckParam("ScaleTrace"+postfix)){ 
    _Scale = para.GetParamDouble("ScaleTrace"+postfix);
  } else { 
    _Scale = 
      para.AddParamDouble("ScaleTrace"+postfix,
                          "Weight of trace in Harris min. "
                          "eigenvalue approximation", 
                          0.04, 0.0, DBL_MAX, 0, grp);
  }
  para.SetEnhancedFlag("ScaleTrace"+postfix, true);
}

//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////

template class CornerDetectorHarris<unsigned char, float>; 
template class CornerDetectorHarris<float, float>; 

// fill in instances as required
#ifdef BUILD_IMAGE_INT
template class CornerDetectorHarris<unsigned char, int>; 
#endif
#ifdef BUILD_IMAGE_CHAR
template class CornerDetectorHarris<unsigned char, char>; 
#endif
#ifdef BUILD_IMAGE_SHORT
#endif
#ifdef BUILD_IMAGE_USHORT
template class CornerDetectorHarris<unsigned short, float>; 
#ifdef BUILD_IMAGE_INT
template class CornerDetectorHarris<unsigned short, int>; 
#endif
#endif
#ifdef BUILD_IMAGE_UINT
#endif
#ifdef BUILD_IMAGE_DOUBLE
#endif
