/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "EMatrix.hh"
#include <limits.h>

using namespace std;
using namespace BIAS;

int EMatrix::InitFromF(const FMatrixBase &F,  KMatrix &K1, 
                       KMatrix &K2) {
  FMatrix Fnorm;
  
  // normalize 
  double normalization = F.NormFrobenius();
  if (fabs(normalization)>DBL_EPSILON)
    Fnorm = (F / normalization );
  else return -1;
  
  Matrix3x3<double> tmp(K2.Transpose()*Fnorm*K1);
  
  normalization = tmp.NormFrobenius();
  if (fabs(normalization)>DBL_EPSILON)
    Fnorm = tmp / normalization;
  else return -1;

  // compose nearest essential matrix
  SVD Fnorm_svd(Fnorm);
  Vector<double> S(Fnorm_svd.GetS());
  tmp.SetZero();
  tmp[1][1] = tmp[0][0] = 0.5*(S[0]+S[1]);
  tmp[2][2] = 0.0;
  (*this) = Matrix3x3<double>( Fnorm_svd.GetU()*tmp*Fnorm_svd.GetVT() );
  return 0;
}



int EMatrix::GetRotationTranslation(RMatrix &R, Vector3<double> &C,
                                    std::vector<HomgPoint2D> &inlier1,
                                    std::vector<HomgPoint2D> &inlier2)
{
  //AddDebugLevel(D_EMATRIX_ROT_TRANS);
  BIASCDOUT(D_EMATRIX_ROT_TRANS, "####################################"<<endl);
  // at first compute possible candidates for baseline and rotation 
  // details see Hartley Zisserman pp. 239...
  SVD svd(*this);
  Matrix<double> U = svd.GetU();
  Vector3<double> e(U.GetCol(2));
  BIASCDOUT(D_EMATRIX_ROT_TRANS,"epipol: "<<e<<endl);
#ifdef BIAS_DEBUG
  HomgPoint2D te1, te2;
  this->GetEpipolesHomogenized(te1, te2);
  BIASCDOUT(D_EMATRIX_ROT_TRANS, "te1: "<<te1<<"\tte2: "<<te2<<endl);
#endif
  Matrix3x3<double> W; W.SetZero();
  W[0][1] = -1.0; W[1][0] = 1.0; W[2][2] = 1.0;
  
  std::vector<RMatrix > RotCand;
  // first candidate
  R = svd.GetU()*W*svd.GetVT();
  R = R.Transpose();
  RotCand.push_back(R.det()*R); 
  // second candidate
  R = svd.GetU()*W.Transpose()*svd.GetVT(); 
  R=R.Transpose();
  RotCand.push_back(R.det()*R);

  // now evaluating the candidates to find the best one
  PMatrix P0, P1; P0.SetIdentity();
  BIASCDOUT(D_EMATRIX_ROT_TRANS, "P0 is "<<P0<<endl);
  KMatrix K; K.SetIdentity();
  unsigned int inlier_count, best_inlier_count=0;
  Triangulation triangulator;
  HomgPoint3D p3d;
  RMatrix RCurrent;
  BIAS::Vector3<double> CCurrent;
  for (unsigned int i = 0; i < RotCand.size(); i++){
    for (double base_sign = -1.0; base_sign <= 2.0 ; base_sign += 2.0){
      // since we want to set +/-e as the last column of P
      // (which is -R^T*C)
      // we have to compute a CCurrent that produces this row
      // therefore we undo R^T by multiplying by R
      CCurrent = Vector3<double>(base_sign*RotCand[i]*e);
      RCurrent = RotCand[i];
      // set P1
      P1.Compose(K, RCurrent, CCurrent);
      BIASCDOUT(D_EMATRIX_ROT_TRANS, "-----------------------------"<<endl);
      BIASCDOUT(D_EMATRIX_ROT_TRANS, "P1 is "<<P1<<endl);
      inlier_count = 0;
      for (unsigned int n =0; n < inlier1.size(); n++){
        // now run triangulation
        int tri_res;
        if ((tri_res=triangulator.Triangulate(P0, P1, inlier1[n], 
                                              inlier2[n],p3d))==0){
          inlier_count++;   
        }
#ifdef BIAS_DEBUG
        if (n<10){
          BIASCDOUT(D_EMATRIX_ROT_TRANS, inlier1[n]<<"\t"<< inlier2[n]
                    <<"\t"<<p3d<<"\t"<<tri_res<<endl);
        }
#endif
      }
      if (inlier_count > best_inlier_count){
        BIASCDOUT(D_EMATRIX_ROT_TRANS, "Set best inlier count " 
                  << inlier_count <<endl << "R = " << RCurrent 
                  << ", C = " << CCurrent  << endl);
        best_inlier_count = inlier_count;
        // save the same R and C we used for the test
        R = RCurrent;
        C = CCurrent;
      } else {
        BIASCDOUT(D_EMATRIX_ROT_TRANS, "inlier count " << inlier_count <<endl
                  << "R = " << RCurrent << ", C = " << CCurrent  << endl);
      }
    } // loop over baselines
  } // loop over RotCand
  
 
  if (best_inlier_count == 0) {
    BIASERR("No inlier found during triangulation for essential candidates!");
    return -1;
  }
  return 0;
}
