/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "EpipolarLine.hh"

#include <Base/Debug/Debug.hh>
#include <Base/Image/ImageDraw.hh>

#include <cmath>
#include <cfloat>

#define MY_DBL_PRECISION 1e-14

#include <Base/Common/BIASpragma.hh>


using namespace BIAS;
using namespace std;


ostream& operator<<(ostream& os, const enum TEdge edge)
{
  switch(edge){
  case TopEdge:     os << "top";    break;
  case RightEdge:    os << "right";    break;
  case BottomEdge:    os << "bottom";    break;
  case LeftEdge:    os << "left";    break;
  case InvalidEdge:    os << "invalid";    break;
  }
  return os;
}


EpipolarLine::EpipolarLine() : Debug()
{
  HomgLine2D::HomgLine2D();
  Width_ = Height_ = 0;
  Angle_ = DeltaR_ = RMinLine_ = RMaxLine_ = 0.0;
  // set all HomgPoints2D to (0,0,1)
  CloseIntersection_.Set (0.0, 0.0);
  FarIntersection_ = Current_ = TransformedStart_ = TransformedEnd_ 
    = TransformedCurrent_ = Epipole_ = PInfinity_ = CloseIntersection_;
  dx_ = dy_ = 0.0;
}

EpipolarLine::EpipolarLine(const EpipolarLine& line)
  :HomgLine2D(line)
{
  Epipole_ = line.GetCopyOfEpipole();
  EpipolePosition_ = line.GetEpipolePosition();
  Type_ = line.GetLineType();
  Width_ = line.GetImageWidth();
  Height_ = line.GetImageHeight();
  Angle_ = line.GetAngle();
  FarIntersection_ = line.GetCopyOfFarIntersectionWithImageBorder();
  CloseIntersection_ = line.GetCopyOfCloseIntersectionWithImageBorder();
  FarEdge_ = line.GetFarEdge();
  CloseEdge_ = line.GetCloseEdge();
  RMinLine_ = line.GetRMin();
  RMaxLine_ = line.GetRMax();
  dx_ = line.GetDx();
  dy_ = line.GetDy();
}


EpipolarLine& EpipolarLine::operator=(const EpipolarLine& line)
{
  (*this).HomgLine2D::operator=(line);
  Epipole_ = line.GetCopyOfEpipole();
  EpipolePosition_ = line.GetEpipolePosition();
  Type_ = line.GetLineType();
  Width_ = line.GetImageWidth();
  Height_ = line.GetImageHeight();
  Angle_ = line.GetAngle();
  FarIntersection_ = line.GetCopyOfFarIntersectionWithImageBorder();
  CloseIntersection_ = line.GetCopyOfCloseIntersectionWithImageBorder();
  FarEdge_ = line.GetFarEdge();
  CloseEdge_ = line.GetCloseEdge();
  RMinLine_ = line.GetRMin();
  RMaxLine_ = line.GetRMax();
  dx_ = line.GetDx();
  dy_ = line.GetDy();
  return *this;
}

EpipolarLine::EpipolarLine(const Vector3<double>& vec) 
  :HomgLine2D(vec)
{}

EpipolarLine::~EpipolarLine () 
{}

void
EpipolarLine::GetNextLine(bool clockwise, EpipolarLine& nextline,
                          double dist)
{
  double dx, dy;
  HomgPoint2D newpoint;

  Homogenize();
  //    cerr <<"slope: "<<slope<<"\tinv. slope: "<<bool2string(invslope)
  //         <<"\tdr: "<<GetDeltaR)<<"\ttype: "<<GetLineType()
  //         <<"\tangle: "<<GetAngle()<<"\tclockwise: "<<bool2string(clockwise)
  //         <<endl;
  
  dx = dist*fabs(data_[0]);
  dy = dist*fabs(data_[1]);

  if (data_[0] * data_[1] < 0)
    dx *= -1;

  if (fabs(GetAngle())>M_PI_2){
    dx *= -1;
    dy *= -1;
  }

  if (!clockwise){
    dx *= -1;
    dy *= -1;
  }

  newpoint = GetFarIntersectionWithImageBorder();

  // maybe newpoint.Homogenize() ?
  newpoint[0] += dx;
  newpoint[1] += dy;

  nextline.Init(GetImageWidth(), GetImageHeight(), GetEpipole(), 
                newpoint, GetEpipolePosition());

  BIASDOUT(D_EL_NEXTLINE,"dx: "<<dx<<"\tdy: "<<dy<<"\tnewpoint: ("
          <<newpoint[0]<<", "<<newpoint[1]<<")   far int.: ("
          <<nextline.GetFarIntersectionWithImageBorder()[0]<<", "
          <<nextline.GetFarIntersectionWithImageBorder()[1]<<")  angle: "
          <<nextline.GetAngle());
}


int 
EpipolarLine::Recalc(unsigned int width, unsigned int height, 
                     HomgPoint2D& epipole, enum HomgPoint2D::EPosition epos)
{
  Width_ = width;
  Height_ = height;
  Epipole_ = epipole;
  EpipolePosition_ = epos;
  return Recalc();
}


int
EpipolarLine::Recalc()
{
  Homogenize();
  // argument point is only used as dummy, to
  // verify if (argument-)point sits on image border
  HomgPoint2D point(-1.0, -1.0);
  if (!DetermineIntersectionsAndEpipoleDistances_(point)) return -1;
  Angle_ = atan2(FarIntersection_[1]-Epipole_[1], 
		 FarIntersection_[0]-Epipole_[0]);
  Type_ = Sym_.DetermineLineType(Angle_);
  Homogenize();
  DeltaR_ = (fabs(data_[0]) > fabs(data_[1]))?(fabs(data_[0]/data_[1])):
    (fabs(data_[1]/data_[0]));
  return 0;
}

void 
EpipolarLine::Init(unsigned int width, unsigned int height, 
                   HomgPoint2D& epipole, HomgPoint2D& otherpoint, 
                   enum HomgPoint2D::EPosition epos)
{
  PInfinity_ = otherpoint;
  PInfinityPosition_ = otherpoint.DeterminePosition(width, height);
  epipole.CrossProduct(otherpoint, *this);
  Homogenize();
  Epipole_ = epipole;
  Width_ = width;
  Height_ = height;
  Homogenize();
  DeltaR_ = (fabs(data_[0]) > fabs(data_[1]))?(fabs(data_[0]/data_[1])):
    (fabs(data_[1]/data_[0]));
  
  DetermineIntersectionsAndEpipoleDistances_(otherpoint);
  Angle_ = atan2(FarIntersection_[1]-Epipole_[1], 
		 FarIntersection_[0]-Epipole_[0]);
  Type_ = Sym_.DetermineLineType(Angle_);
  EpipolePosition_ = epos;
}

void 
EpipolarLine::Init(unsigned int width, unsigned int height, 
                   HomgPoint2D& epipole, HomgPoint2D& pinf)
{
  Init(width, height, epipole, pinf, 
       epipole.DeterminePosition(width, height));
}


//  template <class PixelType> double
//  EpipolarLine::ScanLine(Image<PixelType>& Source, HomgPoint2D& epipole, 
//  		       double ROffset, unsigned int& DestLength, 
//  		       PixelType *Destination, bool forward)
double
EpipolarLine::ScanLine(Image<unsigned char>& Source, HomgPoint2D& epipole, 
                       enum HomgPoint2D::EPosition epos, double RMinImage, 
                       double RMaxImage, bool forward, 
                       unsigned int& DestLength, unsigned char *Destination)
{
  DestLength = 0;
  Homogenize();
  Epipole_ = epipole;
  Width_ = Source.GetWidth();
  Height_ = Source.GetHeight();
  Homogenize();
  DeltaR_ = (fabs(data_[0]) > fabs(data_[1]))?(fabs(data_[0]/data_[1])):
    (fabs(data_[1]/data_[0]));
  if (!DetermineIntersectionsAndEpipoleDistances_(epipole))
    return -1.0;
  Angle_ = atan2(FarIntersection_[1]-Epipole_[1], 
		 FarIntersection_[0]-Epipole_[0]);
  EpipolePosition_ = epos;

  Type_ = Sym_.DetermineLineType(Angle_);

  CalcTransformed_();

  BIASDOUT(D_EL_TRANSF,"TransformedLine ("<<TransformedStart_[0]<<", "
	  <<TransformedStart_[1]<<")  ->  ("<<TransformedEnd_[0]<<", "
	  <<TransformedEnd_[1]<<")");

  ROffset_ = (unsigned int)ceil(RMinLine_) - (unsigned int)ceil(RMinImage);

  BIASDOUT(D_EL_OFFSET,"RMinLine_: "<< RMinLine_ <<"\tRMaxLine_: "<< RMaxLine_ 
          <<"\tRMinImage:"<<RMinImage<<"\tRMaxImage:"<<RMaxImage
          <<"\tROffset_: "<< ROffset_);

  return ScanLine_(Source, ROffset_, DestLength, Destination, forward);
}
 

// input : Source, ROffset
// output : DestLength, Destination
// assumes that EpipolarLine is fully initialized
//  template <class PixelType> double
//  EpipolarLine::ScanLine_(Image<PixelType>& Source, double ROffset,
//  			unsigned int& DestLength, PixelType *Destination,
//  			bool forward) 
double
EpipolarLine::ScanLine_(Image<unsigned char>& Source, unsigned int ROffset,
                        unsigned int& DestLength, unsigned char *Destination,
                        bool forward) 
{ 
  unsigned int counter= 0;
  int counterincr = 0;
  unsigned int DestImageWidth = 
    (unsigned int)floor(sqrt(pow(Source.GetWidth(),2.0) + 
                             pow(Source.GetHeight(),2.0)))+ 1;

  if (Destination == NULL){
    Destination = new unsigned char[DestImageWidth];
    ROffset = 0;
  }  

  DestLength = ROffset;

  if(forward){
    counter = DestLength;
    counterincr = 1;
  } else {
    counter = DestImageWidth - DestLength;
    counterincr = -1;
  }
  while ((TransformedCurrent_[0]<=TransformedEnd_[0])&& 
         (TransformedCurrent_[1]<=TransformedEnd_[1])){
    // bilinear interpolation
    //Current_=InverseTransform_(TransformedCurrent_);
    Sym_.InverseTransform(Type_, TransformedCurrent_, Current_);
    BIASDOUT(D_EL_ROUND,"x: "<<Current_[0]<<" y: "<<Current_[1]
            <<"\tdest: "<<(int)Destination[DestLength]);
    Destination[counter] = (unsigned char)
      Source.BilinearInterpolationGrey(Current_[0], Current_[1]);
      
    TransformedCurrent_[0]+=dx_;
    TransformedCurrent_[1]+=dy_;
    DestLength++;
    counter += counterincr;
  }
  BIASDOUT(D_EL_OFFSET,"DestLength: " << DestLength);
  CalcTransformed_();
  Length_ = DestLength - ROffset_;
  return ROffset - RMinLine_;
}

void
EpipolarLine::Interpolate_(enum TLineType& type, unsigned char **ida, int& x,
                           int& y, double& g, unsigned char& result)
{
  result = 0;
  double gm = 1.0 - g;
  switch(type){
  case SE_E:
    BIASDOUT(D_EL_BRESROUND,x << ", " << y << " : "<< gm << "\t" 
	    << x << ", " << y+1 << " : "<<g);
    result = (unsigned char)((double)ida[y][x]*gm + 
			     (double)ida[y+1][x]*g);
    break;
  case S_SE:
    BIASDOUT(D_EL_BRESROUND,y << ", " << y << " : "<< gm << "\t" 
	    << y << ", " << x+1 << " : "<<g);
    result = (unsigned char)((double)ida[x][y]*gm + 
			     (double)ida[x+1][y]*g);
    //        Original[0] = Transformed[1];
    //        Original[1] = Transformed[0];
    break;
  case SW_S:
    BIASDOUT(D_EL_BRESROUND,y << ", " << -x << " : "<< gm << "\t" 
	    << y << ", " << -x+1 << " : "<<g);
    result = (unsigned char)((double)ida[x][-y]*gm + 
			     (double)ida[x+1][-y]*g);
    //        Original[0] = Transformed[1];
    //        Original[1] = -Transformed[0];
    break;
  case W_SW:
    BIASDOUT(D_EL_BRESROUND,-x << ", " << y << " : "<< gm << "\t" 
	    << -x << ", " << y+1 << " : "<<g);
    result = (unsigned char)((double)ida[y][-x]*gm + 
			     (double)ida[y+1][-x]*g);
    //        Original[0] = -Transformed[0];
    //        Original[1] = Transformed[1];
    break;
  case NW_W:
    BIASDOUT(D_EL_BRESROUND,-x << ", " << -y << " : "<< gm << "  : "
	    <<(int)ida[-y][-x]<<"\t" << -x << ", " << -y-1 << " : "<<g
	    <<"  : "<<(int)ida[-y-1][-x]);
    result = (unsigned char)((double)ida[-y][-x]*gm + 
			     (double)ida[-y-1][-x]*g);
    BIASDOUT(D_EL_BRESROUND,"result: "<<(int)result);
    //        Original = Transformed * (-1);
    break;
  case N_NW:
    BIASDOUT(D_EL_BRESROUND,-y << ", " << -x << " : "<< gm << "\t" 
	    << -y << ", " << -x+1 << " : "<<g);
    result = (unsigned char)((double)ida[-x][-y]*gm + 
			     (double)ida[-x-1][-y]*g);
    //        Original[0] = -Transformed[1];
    //        Original[1] = -Transformed[0];
    break;
  case NE_N:
    BIASDOUT(D_EL_BRESROUND,-y << ", " << x << " : "<< gm << "\t" 
	    << -y << ", " << x+1 << " : "<<g);
    result = (unsigned char)((double)ida[-x][y]*gm + 
			     (double)ida[-x-1][y]*g);
    //        Original[0] = -Transformed[1];
    //        Original[1] = Transformed[0];
    break;
  case E_NE:
    BIASDOUT(D_EL_BRESROUND,x << ", " << -y << " : "<< gm << "\t" 
	    << x << ", " << -y+1 << " : "<<g);
    result = (unsigned char)((double)ida[-y][x]*gm + 
			     (double)ida[-y-1][x]*g);
    //cerr << " y " << -y << "  g " << gm << endl;
    //        Original[0] = Transformed[0];
    //        Original[1] = -Transformed[1];
    break;
  case LT_Invalid:
    BIASERR("Invalid LineType, maybe wrong slope in EpipolarLine");
    break;
  }
}

void EpipolarLine::CalcTransformed_()
{
  double dr=0.0;
  Vector2<double> Transfslope, Slope(data_[1], data_[0]);
//   dx_ = fabs(data_[0]);
//   dy_ = fabs(data_[1]);
  Sym_.Transform(Type_, Slope, Transfslope);
  dx_ = fabs(Transfslope[0]);  
  dy_ = fabs(Transfslope[1]);

  BIASDOUT(D_EL_TRANSF," dx_ "<<dx_<<"\tdy_ "<<dy_<<"\ttype: "<<Type_);

  Sym_.Transform(Type_, CloseIntersection_, TransformedStart_);
  Sym_.Transform(Type_, FarIntersection_, TransformedEnd_);

  dr = ceil(RMinLine_) - RMinLine_;
  BIASDOUT(D_EL_TRANSF," dr start: "<<dr<<"\tstart befor "
          <<(Vector3<double>)TransformedStart_);
  TransformedStart_[0] += dx_ * dr;
  TransformedStart_[1] += dy_ * dr;

  Sym_.InverseTransform(Type_, TransformedStart_, Start_);

  dr = RMaxLine_ - floor(RMaxLine_);
  BIASDOUT(D_EL_TRANSF," dr end: "<<dr<<"\tend befor "
          <<(Vector3<double>)TransformedEnd_);
  TransformedEnd_[0] -= dx_ * dr;
  TransformedEnd_[1] -= dy_ * dr;

  Sym_.InverseTransform(Type_, TransformedEnd_, End_);

  BIASDOUT(D_EL_TRANSF,"("<<CloseIntersection_[0]<<", "<<CloseIntersection_[1]
	  <<") -> ("<<FarIntersection_[0]<<", "<<FarIntersection_[1]
	  <<")  transf:  ("<<TransformedStart_[0]<<", "<<TransformedStart_[1]
	  <<") -> ("<<TransformedEnd_[0]<<", "<<TransformedEnd_[1]<<")");

  TransformedCurrent_ = TransformedStart_;
  Sym_.InverseTransform(Type_, TransformedCurrent_, Current_);

}

// the following is used in DetermineIntersectionsAndEpipoleDistances_
int CompIntersect(const void *left, const void *right)
{
  struct TIntersect *Left = (struct TIntersect *)left;
  struct TIntersect *Right = (struct TIntersect *)right;
  return (Left->EpipoleDistance < Right->EpipoleDistance);
}

bool
EpipolarLine::DetermineIntersectionsAndEpipoleDistances_(HomgPoint2D& point)
{
  bool result = true;
  int intersectioncount = 0;
  struct TIntersect edges[4]; // 0=left, 1=right, 2=top, 3=bottom
  HomgLine2D left(-1.0, 0.0, 0.0);
  HomgLine2D right(-1.0, 0.0, Width_-1);
  HomgLine2D top(0.0, -1.0, 0.0);
  HomgLine2D bottom(0.0, -1.0, Height_-1);

  edges[0].IntersectsInImage = edges[1].IntersectsInImage 
    = edges[2].IntersectsInImage = edges[3].IntersectsInImage = false;
  edges[0].EpipoleDistance = edges[1].EpipoleDistance 
    = edges[2].EpipoleDistance = edges[3].EpipoleDistance = -1.0;
  edges[0].Edge = LeftEdge;
  edges[1].Edge = RightEdge;
  edges[2].Edge = TopEdge;
  edges[3].Edge = BottomEdge;
  edges[0].Point = Intersection(left);
  edges[1].Point = Intersection(right);
  edges[2].Point = Intersection(top);
  edges[3].Point = Intersection(bottom);

  // deal with compiler precision
  for (unsigned int i=0; i<4; i++){
    if (edges[i].Point[0] < 0.0 &&  edges[i].Point[0] > -MY_DBL_PRECISION)
      edges[i].Point[0] = 0.0;
    if (edges[i].Point[1] < 0.0 &&  edges[i].Point[1] > -MY_DBL_PRECISION)
      edges[i].Point[1] = 0.0;
    if (edges[i].Point[0] > GetImageWidth() &&  
        edges[i].Point[0] < GetImageWidth()+MY_DBL_PRECISION)
      edges[i].Point[0] = GetImageWidth();
    if (edges[i].Point[1] > GetImageHeight() &&  
        edges[i].Point[1] < GetImageHeight()+MY_DBL_PRECISION)
      edges[i].Point[1] = GetImageHeight();
  }

  //  #ifdef BIAS_DEBUG
  //    cout << "untested intersections: " << endl;
  //    for (int i=0; i<4; i++){
  //      cout << i <<":  (" << edges[i].Point[0] << ", " 
  //  	 << edges[i].Point[1] << ",  "<< edges[i].Point[2] <<")\t" 
  //  	 << edges[i].EpipoleDistance << endl;
  //    }
  //  #endif

  if ((point[0]==0.0) || 
      ((edges[0].Point[2]!=0.0) &&
       (edges[0].Point[1]>=0 && edges[0].Point[1]<=Height_-1))){
    if(point[0]==0.0)
      edges[0].Point=point;
    edges[0].IntersectsInImage = true; intersectioncount++; 
    edges[0].EpipoleDistance = sqrt(pow(edges[0].Point[1]-Epipole_[1],2) + 
                                    pow(edges[0].Point[0]-Epipole_[0],2));
  }
  if ((point[0]==Width_-1) || 
      ((edges[1].Point[2]!=0.0) &&
       (edges[1].Point[1]>=0 && edges[1].Point[1]<=Height_-1))){
    if (point[0]==Width_-1)
      edges[1].Point=point;
    edges[1].IntersectsInImage = true; intersectioncount++; 
    edges[1].EpipoleDistance = sqrt(pow(edges[1].Point[1]-Epipole_[1],2) + 
                                    pow(edges[1].Point[0]-Epipole_[0],2));
  }
  if ((point[1]==0.0) || 
      ((edges[2].Point[2]!=0.0) &&
       (edges[2].Point[0]>=0 && edges[2].Point[0]<=Width_-1))){
    if (point[1]==0.0)
      edges[2].Point=point;
    edges[2].IntersectsInImage = true; intersectioncount++; 
    edges[2].EpipoleDistance = sqrt(pow(edges[2].Point[1]-Epipole_[1],2) + 
                                    pow(edges[2].Point[0]-Epipole_[0],2));
  }
  if ((point[1]==Height_-1) ||
      ((edges[3].Point[2]!=0.0) &&
       (edges[3].Point[0]>=0 && edges[3].Point[0]<=Width_-1))){
    if (point[1]==Height_-1)
      edges[3].Point=point;
    edges[3].IntersectsInImage = true;intersectioncount++; 
    edges[3].EpipoleDistance = sqrt(pow(edges[3].Point[1]-Epipole_[1],2) + 
                                    pow(edges[3].Point[0]-Epipole_[0],2));
  }

  //  #ifdef BIAS_DEBUG
  //    cout << "unsorted intersections: " << endl;
  //    for (int i=0; i<4; i++){
  //      cout << i <<":  (" << edges[i].Point[0] << ", " 
  //  	 << edges[i].Point[1] << ",  "<< edges[i].Point[2] <<")\t" 
  //  	 << edges[i].EpipoleDistance << endl;
  //    }
  //  #endif

  // sort Intersections according to epipole distance
  qsort(&edges, 4, sizeof(struct TIntersect), CompIntersect);

  //  #ifdef BIAS_DEBUG
  //    cout << "sorted intersections: " << endl;
  //    for (int i=0; i<4; i++){
  //      cout << i <<":  (" << edges[i].Point[0] << ", " 
  //  	 << edges[i].Point[1] << ")\t" << edges[i].EpipoleDistance << endl;
  //    }
  //  #endif

//   if (EpipolePosition_!= Center){

    FarIntersection_ = edges[0].Point;
    FarEdge_ = edges[0].Edge;
    RMaxLine_ = edges[0].EpipoleDistance;
    switch(intersectioncount){
    case 1:
      CloseIntersection_=edges[0].Point;
      CloseEdge_ = edges[0].Edge;
      RMinLine_ = edges[0].EpipoleDistance;
      break;
    case 2:
      CloseIntersection_=edges[1].Point;
      CloseEdge_ = edges[1].Edge;
      RMinLine_ = edges[1].EpipoleDistance;
      break;
    case 3:
      CloseIntersection_=edges[2].Point;
      CloseEdge_ = edges[2].Edge;
      RMinLine_ = edges[2].EpipoleDistance;
      break;
    case 4:
      CloseIntersection_=edges[3].Point;
      CloseEdge_ = edges[3].Edge;
      RMinLine_ = edges[3].EpipoleDistance;
      break;
    default:
      result = false;
#ifdef BIAS_DEBUG      
      BIASDOUT(D_EL_INT, "invalid intersection count "<<intersectioncount
               <<" of ep. line: "<<*this<<"  with epipole: "<<Epipole_
               <<"close intersect: "<<GetCloseIntersectionWithImageBorder()
               <<"far intersection: "<<GetFarIntersectionWithImageBorder()
               << " sorted intersections: ");
      for (int i=0; i<4; i++){
        BIASDOUT(D_EL_INT, i <<":  (" << edges[i].Point[0] << ", " 
                 << edges[i].Point[1] << ")\t" << edges[i].EpipoleDistance);
      }
      //BIASASSERT(intersectioncount>=2);
#endif
      break;
    }
  
//   } else {

//     CloseIntersection_ = Epipole_;
//     CloseEdge_ = BIASInvalidEdge;   
//     RMinLine_ = 0.0;
//     int i=0;
//     if (Angle_ >= 0){
//       // take the intersection not sitting on top edge
//       while (edges[i].Point[1]==0) i++;
//     } else { 
//       // take the intersection not sitting on bottom edge
//       while (edges[i].Point[1]==Height_) i++;
//     }
//     FarIntersection_=edges[i].Point;
//     FarEdge_ = edges[i].Edge;
//     RMaxLine_ = edges[i].EpipoleDistance;
//   }
  return result;
}

// enum TEpipolePosition
// EpipolarLine::DeterminePointPosition_(unsigned int width, unsigned int height,
//                                       HomgPoint2D& point)
// {
//   enum TEpipolePosition epos = HomgPoioInvalidPosition;
//   enum TLocation xpos = LOC_undef, ypos = LOC_undef;

//   if (point[0] < 0)
//     xpos = LOC_LEFT;
//   if (point[0] > width-1)
//     xpos = LOC_RIGHT;
//   if (point[0] >= 0 && point[0] <= width-1)
//     xpos = LOC_MIDDLE;

//   if (point[1] < 0)
//     ypos = LOC_UPPER;
//   if (point[1] > height-1)
//     ypos = LOC_LOWER;
//   if (point[1] >= 0 && point[1] <= height-1)
//     ypos = LOC_MIDDLE;

//   switch (xpos) {
//   case LOC_LEFT:
//     switch (ypos) {
//     case LOC_LOWER:
//       epos = MIPBottomLeft;
//       break;
//     case LOC_UPPER:
//       epos = MIPTopLeft;
//       break;
//     case LOC_MIDDLE:
//       epos = MIPLeft;
//       break;
//     default:
//       BIASERR("DeterminePointPosition_() no valid ypos");
//       break;
//     }
//     break;

//   case LOC_MIDDLE:
//     switch (ypos) {
//     case LOC_LOWER:
//       epos = MIPBottom;
//       break;
//     case LOC_UPPER:
//       epos = MIPTop;
//       break;
//     case LOC_MIDDLE:
//       epos = MIPCenter;
//       break;
//     default:
//       MIPERR("DeterminePointPosition_() no valid ypos");
//       break;
//     }
//     break;

//   case LOC_RIGHT:
//     switch (ypos) {
//     case LOC_LOWER:
//       epos = MIPBottomRight;
//       break;
//     case LOC_UPPER:
//       epos = MIPTopRight;
//       break;
//     case LOC_MIDDLE:
//       epos = MIPRight;
//       break;
//     default:
//       MIPERR("DeterminePointPosition_() no valid ypos");
//       break;
//     }
//     break;

//   default:
//       MIPERR("DeterminePointPosition_() no valid xpos");
//     break;
//   }
//   return epos;
// }


#include <limits> // jw
template <class PixelType>
void EpipolarLine::Draw(Image<PixelType>& im)
{
  unsigned int start[2], end[2];
  PixelType color[3]; 
  color[0]=PixelType(255);
  color[1]=PixelType(255);
  color[2]=PixelType(255);

//   cerr << "Epipole: "<<Epipole_<<"   PInfinity: "<<PInfinity_<<endl;
//   cerr << "EpipolePos: "<<EpipolePosition_<<"    pinfpos: "<<PInfinityPosition_<<endl;
//   cerr << GetCloseIntersectionWithImageBorder() << " \t\t" 
//        << GetFarIntersectionWithImageBorder() << endl;
  if (EpipolePosition_!=HomgPoint2D::Center){
    start[0]=(unsigned int)rint(GetCloseIntersectionWithImageBorder()[0]/
                                GetCloseIntersectionWithImageBorder()[2]);
    start[1]=(unsigned int)rint(GetCloseIntersectionWithImageBorder()[1]/
                                GetCloseIntersectionWithImageBorder()[2]);
  } else {
    start[0]=(unsigned int)rint(Epipole_[0]/Epipole_[2]);
    start[1]=(unsigned int)rint(Epipole_[1]/Epipole_[2]);
    ImageDraw<PixelType>::RectangleCenter(im, start, 5,color);
  }

  if (PInfinityPosition_!=HomgPoint2D::Center){
    end[0]=(unsigned int)rint(GetFarIntersectionWithImageBorder()[0]/
                     GetFarIntersectionWithImageBorder()[2]);
    end[1]=(unsigned int)rint(GetFarIntersectionWithImageBorder()[1]/
                     GetFarIntersectionWithImageBorder()[2]);
  } else {
    end[0]=(unsigned int)rint(PInfinity_[0]/PInfinity_[2]);
    end[1]=(unsigned int)rint(PInfinity_[1]/PInfinity_[2]);
    ImageDraw<PixelType>::RectangleCenter(im, end, 5,color);
  }
  //cerr<<"start"<<start[0]<<" "<<start[1]<<"  end: "<<end[0]<<" "<<end[1]<<endl;
  ImageDraw<PixelType>::Line(im, start, end,color);
}

template <class PixelType>
void EpipolarLine::DrawWhole(Image<PixelType>& im)
{
  unsigned int start[2], end[2];
  start[0]=(unsigned int)rint(GetCloseIntersectionWithImageBorder()[0]/
                     GetCloseIntersectionWithImageBorder()[2]);
  start[1]=(unsigned int)rint(GetCloseIntersectionWithImageBorder()[1]/
                     GetCloseIntersectionWithImageBorder()[2]);
  end[0]=(unsigned int)rint(GetFarIntersectionWithImageBorder()[0]/
                   GetFarIntersectionWithImageBorder()[2]);
  end[1]=(unsigned int)rint(GetFarIntersectionWithImageBorder()[1]/
                   GetFarIntersectionWithImageBorder()[2]);
  ImageDraw<PixelType>::Line(im, start, end);
}

#define INSTANCE(type) \
template void EpipolarLine::Draw<type>(Image<type>& im);\
template void EpipolarLine::DrawWhole<type>(Image<type>& im);

INSTANCE(unsigned char)
INSTANCE(float)

#ifdef BUILD_IMAGE_CHAR
INSTANCE(char)
#endif
#ifdef BUILD_IMAGE_DOUBLE
INSTANCE(double)
#endif
#ifdef BUILD_IMAGE_UINT
INSTANCE(unsigned int)
#endif
#ifdef BUILD_IMAGE_INT
INSTANCE(int)
#endif
#ifdef BUILD_IMAGE_SHORT
INSTANCE(short)
#endif
#ifdef BUILD_IMAGE_USHORT
INSTANCE(unsigned short)
#endif



