/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "HMatrix.hh"

#include <MathAlgo/SVD.hh>
#include <MathAlgo/Lapack.hh>




using namespace BIAS;
using namespace std;

HMatrix::HMatrix()
  : HMatrixBase()
{}

HMatrix::HMatrix(Matrix3x3<double> Mat)
  : HMatrixBase(Mat)
{}


HMatrix& HMatrix::operator=(const HMatrix& mat)
  
{
  HMatrixBase::operator=(mat);
  _Hinv=mat.GetHinv();
  return *this;
}

HMatrix::~HMatrix()
{}






int HMatrix::Compute(std::vector<BIAS::HomgPoint2D>& frompoint, 
                     std::vector<BIAS::HomgPoint2D>& topoint,
                     bool DeprecatedWarning){
  if (DeprecatedWarning) {
    BIASWARN("Call of deprecated function HMatrix::Compute which will be moved"
             " to HMatrixEstimation in the future. Last chance to update "
             <<" your code. ");
  }
  int result = 0;
  if ((int) frompoint.size()<4){
    BIASERR("at least 4 points correspondences are needed for HMatrixBase "
            <<"computation");
    return -1;
  }

  Matrix<double> A(2*frompoint.size(), 9);
  Vector<double> S;
  Matrix<double> U, VT;

  register unsigned int p=0, p2=0;
  for (p=0, p2=0; p<2*frompoint.size(); p+=2, p2++){
 
    A[p][0] = A[p][1] = A[p][2] = 0.0;
    A[p+1][3] = A[p+1][4] = A[p+1][5] = 0.0;

    A[p+1][0] = topoint[p2][2] * frompoint[p2][0];
    A[p+1][1] = topoint[p2][2] * frompoint[p2][1];
    A[p+1][2] = topoint[p2][2] * frompoint[p2][2];

    A[p][3] = - A[p+1][0];
    A[p][4] = - A[p+1][1];
    A[p][5] = - A[p+1][2];

    A[p][6] = topoint[p2][1] * frompoint[p2][0];
    A[p][7] = topoint[p2][1] * frompoint[p2][1];
    A[p][8] = topoint[p2][1] * frompoint[p2][2];
    
    A[p+1][6] = - topoint[p2][0] * frompoint[p2][0];
    A[p+1][7] = - topoint[p2][0] * frompoint[p2][1];
    A[p+1][8] = - topoint[p2][0] * frompoint[p2][2];
  }
  
  result = General_singular_value_decomposition(A, S, U, VT);
 
  if (result == 0){
    for (p=0; p<9; p++){
      GetData()[p] = VT[8][p];
    }
  }

  return result;
}

int 
HMatrix::Compute(unsigned int numpoints, Vector2<double> frompoint[], 
                 Vector2<double> topoint[], bool DeprecatedWarning)
{
  HomgPoint2D *from=new HomgPoint2D[numpoints], *to=new HomgPoint2D[numpoints];
  for (register unsigned int i=0; i< numpoints; i++){
    from[i][0]=frompoint[i][0];
    from[i][1]=frompoint[i][1];
    from[i][2]=1.0;
    to[i][0]=topoint[i][0];
    to[i][1]=topoint[i][1];
    to[i][2]=1.0;
  }
  int res = Compute(numpoints, from, to, DeprecatedWarning);
  delete[] from;
  delete[] to;
  return res;
}

int HMatrix::Compute(unsigned int numpoints, HomgPoint2D frompoint[], 
                     HomgPoint2D topoint[], bool DeprecatedWarning) {
  if (DeprecatedWarning) {
    BIASWARN("Call of deprecated function HMatrix::Compute which will be moved"
             " to HMatrixEstimation in the future. Last chance to update "
             <<" your code. ");
  }
  int result = 0;
  if (numpoints<4){
    BIASERR("at least 4 points correspondences are needed for HMatrixBase "
            <<"computation");
    return -1;
  }

  Matrix<double> A(2*numpoints, 9);
  Vector<double> S;
  Matrix<double> U, VT;

  register unsigned int p=0, p2=0;
  for (p=0, p2=0; p<2*numpoints; p+=2, p2++){ 
    A[p][0] = A[p][1] = A[p][2] = 0.0;
    A[p+1][3] = A[p+1][4] = A[p+1][5] = 0.0;

    A[p+1][0] = topoint[p2][2] * frompoint[p2][0];
    A[p+1][1] = topoint[p2][2] * frompoint[p2][1];
    A[p+1][2] = topoint[p2][2] * frompoint[p2][2];

    A[p][3] = - A[p+1][0];
    A[p][4] = - A[p+1][1];
    A[p][5] = - A[p+1][2];

    A[p][6] = topoint[p2][1] * frompoint[p2][0];
    A[p][7] = topoint[p2][1] * frompoint[p2][1];
    A[p][8] = topoint[p2][1] * frompoint[p2][2];
    
    A[p+1][6] = - topoint[p2][0] * frompoint[p2][0];
    A[p+1][7] = - topoint[p2][0] * frompoint[p2][1];
    A[p+1][8] = - topoint[p2][0] * frompoint[p2][2];
  }
  
  result = General_singular_value_decomposition(A, S, U, VT);
 
  if (result == 0){
    /* the solution is the unit singular vector corresponding to 
       the smallest singular value. Since S is sorted in descending order,
       h is in the last row of VT */
    //     double length = 0.0;
    //     for (p=0; p<9; p++){
    //       length += VT[8][p] * VT[8][p];
    //     }

    //     length = sqrt(length);

    // //     //BIASDOUT(D_HM_COMP," norm of h: " << length);
    // //     cout << " norm of h: " << length << endl;
    
    //     for (p=0; p<9; p++){
    //       GetData()[p] = VT[8][p] / length;
    //     }
    for (p=0; p<9; p++){
      GetData()[p] = VT[8][p];
    }
  }

  return result;
}

void HMatrix::GetMappedRange(const unsigned int OrigWidth, 
                             const unsigned int OrigHeight,
                             BIAS::Vector2<double> & corner,
                             double & sizeWidth, double & sizeHeight) const {
  double minX, maxX, minY, maxY; // extrema
  
  // rect. coordinates of the 4 src. corners.
  BIAS::HomgPoint2D mapped[4];
  
  // 0  3
  // 1  2
  mapped[0]= (*this) * BIAS::HomgPoint2D(0,         0,          1); 
  // H * '0' left,  upper
  
  mapped[1]= (*this) * BIAS::HomgPoint2D(0,         OrigHeight, 1); 
  // H * '1' left,  lower

  mapped[2]= (*this) * BIAS::HomgPoint2D(OrigWidth, OrigHeight, 1); 
  // H * '2' right, lower

  mapped[3]= (*this) * BIAS::HomgPoint2D(OrigWidth, 0,          1); 
  // H * '3' right, upper
      
  // calc. extrema(xmy) of mapped corners
  // init with first point:
  mapped[0].Homogenize(); // let w==1
  minX=maxX=mapped[0][0]; // x euclid
  minY=maxY=mapped[0][1]; // y suclid  
  // homogenize and calc extrema
  for (unsigned int i=1; i<4; i++) {
    mapped[i].Homogenize();
    // find new extrema
    minX = (mapped[i][0]<minX) ? mapped[i][0] : minX; // x
    maxX = (mapped[i][0]>maxX) ? mapped[i][0] : maxX;    
    minY = (mapped[i][1]<minY) ? mapped[i][1] : minY; // y
    maxY = (mapped[i][1]>maxY) ? mapped[i][1] : maxY;
  };
  
  corner=BIAS::Vector2<double>(minX, minY);
  sizeWidth=maxX-minX;
  sizeHeight=maxY-minY;  
}


double HMatrix::GetMappedError(const HomgPoint2D &PointPicOne,
                               const HomgPoint2D &PointPicTwo) const {
  
  HomgPoint2D x1 = PointPicOne;
  HomgPoint2D x2 = PointPicTwo;
  x1.Homogenize();
  x2.Homogenize();
  HomgPoint2D mapped((*this) * x1);
  mapped.Homogenize();
  return (mapped[0]-x2[0])*(mapped[0]-x2[0]) +
    (mapped[1]-x2[1])*(mapped[1]-x2[1]);
}

void HMatrix::GetJacobian(const HomgPoint2D& x, Matrix<double>& Jac) const {
#ifdef BIAS_DEBUG
  if (!x.IsHomogenized()) {
    BIASERR("supplied unhomogenized point !");
    abort();
  }
  if ((Jac.GetRows()!=2) ||(Jac.GetCols()!=2)) {
    BIASERR("Jacobian has wrong size !");
    abort();
  }
#endif

  // ( x' y' )^T = H( (x y)^T), assume w==w'==1
  const double denominv = 1.0 / ((*this)[2][0]*x[0]+(*this)[2][1]*x[1]+
    (*this)[2][2]);

  const double denomsquaredinv = denominv * denominv;
  const double enumx = ((*this)[0][0]*x[0]+(*this)[0][1]*x[1]+
    (*this)[0][2]) * denomsquaredinv;
  const double enumy = ((*this)[1][0]*x[0]+(*this)[1][1]*x[1]+
    (*this)[1][2]) * denomsquaredinv;
   
  // x'/dx
  Jac[0][0] = ((*this)[0][0]*denominv - (*this)[2][0]*enumx);
  // x'/dy
  Jac[1][0] = ((*this)[0][1]*denominv - (*this)[2][1]*enumx);
  // y'/dx                                        
  Jac[0][1] = ((*this)[1][0]*denominv - (*this)[2][0]*enumy);
  // y'/dy
  Jac[1][1] = ((*this)[1][1]*denominv - (*this)[2][1]*enumy);
}
