/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "HMatrixEstimation.hh"

#include <MathAlgo/SVD.hh>
#include <MathAlgo/Lapack.hh>
#include <Base/Geometry/KMatrix.hh>

using namespace BIAS;
using namespace std;

int HMatrixEstimation::Compute(const vector<BIAS::HomgPoint2D>& points1, 
                               const vector<BIAS::HomgPoint2D>& points2,
                               HMatrix& H) const {
 
  int result = 0;
  BIASASSERT(points1.size()==points2.size());
  if ((int) points1.size()<4){
    BIASERR("at least 4 points correspondences are needed for HMatrixBase "
            <<"computation");
    return -1;
  }

  const std::vector<BIAS::HomgPoint2D> *pp1 = &points1;
  const std::vector<BIAS::HomgPoint2D> *pp2 = &points2;
  std::vector<BIAS::HomgPoint2D> p1norm;
  std::vector<BIAS::HomgPoint2D> p2norm;
  KMatrix K1, K2;
  if (normalizeHartley_) {
    normalization_.Hartley(points1, p1norm, K1);
    normalization_.Hartley(points2, p2norm, K2);
    pp1 = &p1norm;
    pp2 = &p2norm;
  }

  Matrix<double> A(2*points1.size(), 9);
  Vector<double> S;
  Matrix<double> U, VT;

  register unsigned int p=0, p2=0;
  for (p=0, p2=0; p<2*(*pp1).size(); p+=2, p2++){
 
    A[p][0] = A[p][1] = A[p][2] = 0.0;
    A[p+1][3] = A[p+1][4] = A[p+1][5] = 0.0;

    A[p+1][0] = (*pp2)[p2][2] * (*pp1)[p2][0];
    A[p+1][1] = (*pp2)[p2][2] * (*pp1)[p2][1];
    A[p+1][2] = (*pp2)[p2][2] * (*pp1)[p2][2];

    A[p][3] = - A[p+1][0];
    A[p][4] = - A[p+1][1];
    A[p][5] = - A[p+1][2];

    A[p][6] = (*pp2)[p2][1] * (*pp1)[p2][0];
    A[p][7] = (*pp2)[p2][1] * (*pp1)[p2][1];
    A[p][8] = (*pp2)[p2][1] * (*pp1)[p2][2];
    
    A[p+1][6] = - (*pp2)[p2][0] * (*pp1)[p2][0];
    A[p+1][7] = - (*pp2)[p2][0] * (*pp1)[p2][1];
    A[p+1][8] = - (*pp2)[p2][0] * (*pp1)[p2][2];
  }
  
  result = General_singular_value_decomposition(A, S, U, VT);
 
  if (result == 0){
    for (p=0; p<9; p++){
      H.GetData()[p] = VT[8][p];
    }
  } else {
    H.SetZero();
  }

  // undo normalization
  if (normalizeHartley_) H = K2.Invert() * H * K1;

  return result;
}

int HMatrixEstimation::ComputeAffine(const vector<BIAS::HomgPoint2D>& points1,
                                     const vector<BIAS::HomgPoint2D>& points2,
                                     HMatrix& H) const {
  
  // use direct linear transformation on point pairs (p', p^T)
  // p^T crossproduct  HMatrix * p' = 0
  // leads to  A * hvector = 0
  // where A is a matrix compute from p and p'

  BIASASSERT(points1.size()==points2.size());
  
  // perform hartley normalization if desired
  const std::vector<BIAS::HomgPoint2D> *pp1 = &points1;
  const std::vector<BIAS::HomgPoint2D> *pp2 = &points2;
  std::vector<BIAS::HomgPoint2D> p1norm;
  std::vector<BIAS::HomgPoint2D> p2norm;
  KMatrix K1, K2;
  if (normalizeHartley_) {
    normalization_.Hartley(points1, p1norm, K1);
    normalization_.Hartley(points2, p2norm, K2);
    pp1 = &p1norm;
    pp2 = &p2norm;
  }

  // init model matrix with zero
  Matrix<double> A(2*points1.size(), 7, MatrixZero);
  H.SetIdentity();

  // fill rest of model matrix, same as for general H case, but leaving 
  // out last row of H (only 7 parameters of which one is redundant)  
  const unsigned int nummeas = 2*(*pp1).size();
  for (register unsigned int p=0, p2=0; p<nummeas; p+=2, p2++){
    A[p][3]   = -(A[p+1][0] = (*pp2)[p2][2] * (*pp1)[p2][0]);
    A[p][4]   = -(A[p+1][1] = (*pp2)[p2][2] * (*pp1)[p2][1]);
    A[p][5]   = -(A[p+1][2] = (*pp2)[p2][2] * (*pp1)[p2][2]);
    
    A[p][6]   =   (*pp2)[p2][1] * (*pp1)[p2][2];
    A[p+1][6] = - (*pp2)[p2][0] * (*pp1)[p2][2];
  }
  
  // solve using svd (at least 1d solution space)
  Vector<double> S;
  Matrix<double> U, VT; 
  if (General_singular_value_decomposition(A, S, U, VT) == 0){
    // normalize to H[2][2] = 1
    const HMATRIX_TYPE normalization = 1.0 / VT[6][6];
    HMATRIX_TYPE* pHData = H.GetData();
    for (int p=0; p<6; p++){
       *pHData++ = VT[6][p] * normalization;
    }    
  } else {
    // svd failed, no solution
    return -1;
  }

  // undo normalization
  if (normalizeHartley_) H = K2.Invert() * H * K1;

  return 0;
}
