/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "PMatrixEstimation.hh"
#include "MathAlgo/Minpack.hh"


BIAS::Matrix<double> globalM2;
BIAS::HomgPoint2D globalEpipole2;

#include <Base/Common/BIASpragma.hh>


using namespace BIAS;

/// error function, which punishes large differences of submatrix from 
/// a rotation matrix 
void DistanceFromRotation4Var(int *n, double *x, 
                              double *fvec, 
                              int *iflag) 
{
  if (*n!=4) {
    BIASERR("Caller supplied invalid n="<< (*n));
  }
      
  BIAS::Matrix<double> E2(3,1), BMat(1,3);
  for (int k=0; k< 3;k++){
    E2[k][0]   = globalEpipole2[k];
    BMat[0][k] = x[3]*x[k];
  }
  BIAS::Matrix<double> Outer = E2*BMat;
  BIAS::Matrix<double> SubMatrix = globalM2*x[3] + Outer;

  // enforce all rows to be orthogonal
  fvec[0] = SubMatrix.GetRow(0)*SubMatrix.GetRow(1);
  fvec[1] = SubMatrix.GetRow(1)*SubMatrix.GetRow(2);
  fvec[2] = SubMatrix.GetRow(2)*SubMatrix.GetRow(0);
  // enforce all rows to be of unit length
  fvec[3] =  fabs(1.0-(SubMatrix.GetRow(0) * SubMatrix.GetRow(0)));
  fvec[3] += fabs(1.0-(SubMatrix.GetRow(1) * SubMatrix.GetRow(1)));
  fvec[3] += fabs(1.0-(SubMatrix.GetRow(2) * SubMatrix.GetRow(2)));

}

int PMatrixEstimation::ComputeFromFDirect(BIAS::FMatrix &F, 
                                          const double& BaselineMagnitude, 
                                          BIAS::PMatrix &P1,
                                          BIAS::PMatrix &P2){

  HomgPoint2D Epipole1, Epipole2;
  if (F.GetEpipoles(Epipole1, Epipole2)<0) return -1;

  Epipole2.Normalize();
  Epipole2 *= BaselineMagnitude;

  // initially set p2 to be the identity in its left 3x3 i.e. no rotation
  InitP1P2Trans(P1, P2, Epipole2);
  
  // 
  RMatrix R;
  ComputeRotation(F, Epipole1, Epipole2, R);
  R = R.Transpose();
  
  // copy R into P2
  for (unsigned int i = 0; i < 3; i++)
    for (unsigned int j = 0; j < 3; j++)
      P2[i][j] = R[i][j];

  // sanity check: has R rank 3 ?
  SVD RSVD(R);
  Vector< double > S = RSVD.GetS();
  
  // catch division by zero problem in next check
  if (S[0]<1e-10) return -1;
  // sanity check on rank 3 of R
  if (S[2]/S[0]<1e-5) return -1;

  // if (P2_base.NormFrobenius()<1e-10) return -1;

  P1.InvalidateDecomposition();
  P2.InvalidateDecomposition();
  return 0;
} 

void PMatrixEstimation::ComputeRotation(BIAS::FMatrix &F,
                                        BIAS::HomgPoint2D &Epipole1,
                                        BIAS::HomgPoint2D &Epipole2,
                                        BIAS::RMatrix &R){


  // set rotation part as described in Foerstner
  Matrix3x3<double> ee = Epipole2.OuterProduct(Epipole1);
  Matrix3x3<double> e, eF; e.SetAsCrossProductMatrix(Epipole2);
  eF=e*F;
 
  eF *= ee.NormFrobenius();
  ee *= eF.NormFrobenius()/sqrt(2.0);

  R = eF+ee;
  
  // enforce right-handed rotation !
  R /= (R.det() < 0.0) 
    ? -sqrt(R[2][0]*R[2][0]+R[2][1]*R[2][1] + R[2][2]*R[2][2])
    :  sqrt(R[2][0]*R[2][0]+R[2][1]*R[2][1] + R[2][2]*R[2][2]);

  // check for positive optical axis of the second cemara because the first
  // cameras axis is (0 , 0 , 1) otherwise rotate around baseline (epipole) 
  if ( R[2][2] < 0.0){
    RMatrix rot(Epipole2,M_PI);
    R = rot*R;
  }
  R = R.Transpose();
}

void PMatrixEstimation::ComputeRotationCenter(BIAS::PMatrix P1, 
                                              BIAS::PMatrix P2,
                                              BIAS::RMatrix &R,
                                              BIAS::Vector3<double> &C){

  Vector<double> Epipole2 = P2*P1.GetC();
  Epipole2.MultiplyIP(Epipole2.NormL2());
  
  // compute homography for F
  Matrix3x3<double> H, H1,H2;
  for (unsigned int i = 0; i < 3; i++)
    for (unsigned int j = 0; j < 3; j++){
      H1[i][j] = P1[i][j];
      H2[i][j] = P2[i][j];
    }
  Matrix3x3<double> H1i;
  int ires = H1.Invert(H1i);
  if (ires!=0) BIASERR("error: Unable to invert matrix "<<H1<<std::endl);
  H = H2*H1i;
  H2.SetAsCrossProductMatrix(Epipole2);
  H1 = H2*H;
  FMatrix F(H1);

  HomgPoint2D e1, e2;
  F.GetEpipoles(e1,e2);
  ComputeRotation(F,e1,e2,R);
  
  // compute center
  for (unsigned int j = 0; j < 3; j++)
    Epipole2[j] = P2[j][3];

  C = R.Transpose()*Epipole2;
  C *= -1.0;
    

}
 
void PMatrixEstimation::ComputeFromFQuasiEuklid(BIAS::FMatrix &F, 
                                                double BaselineMagnitude, 
                                                BIAS::PMatrix &P1,
                                                BIAS::PMatrix &P2) {

  Matrix<double>    SubMatrix;
  Vector<double>    C(4);
  Matrix3x3<double> M2Matrix;
    
  BIASDOUT(BIAS_FMATRIXDEBUG, "F = " << F);

  HomgPoint2D Epipole1, Epipole2;
  F.GetEpipoles(Epipole1, Epipole2);

  Epipole2.Normalize();
  Epipole2 *= BaselineMagnitude;

  BIASDOUT(BIAS_FQuasiEuklid,"Epipole 1 = " << Epipole1 
           << " Epipole2 = " << Epipole2);


  // initially set p2 to be the identity in its left 3x3 i.e. no rotation
  InitP1P2Trans(P1, P2, Epipole2);
  BIASDOUT(BIAS_FQuasiEuklid,"After InitP1P2Trans P2= "<<P2);

  // find the decomposition of p2, and the closest p2 to the current setting.
  //FindClosestP2(F, P1, P2, M2Matrix, C, Epipole2);
  FindClosestP2(F, P2, M2Matrix, C, Epipole2);
  BIASDOUT(BIAS_FQuasiEuklid,"After FindClosestP2 P2 = "<< P2); 

  // now want to adjust the three DOF in b_trivec to make (m2+e2 b^T) as 
	// close to orthogonal as possible.

	// memcpy (pm_m2_matrix, m2_matrix, sizeof (pm_m2_matrix));
	// fm_compute_epipoles (f_matrix, &epipole1_trivec, 
	// &pm_epipole2_trivec);

 
  // set stop criterion for powell:
  double dMaxError = 1e-10;
  /// \todo dirty hack: set global variables, which are accessible in error
  /// function called from fortran-powell
  /// *** this is not thread-safe ! ***
  globalEpipole2 = Epipole2;
  globalM2 = M2Matrix;
  
  // set current value as start value
  Vector<double> CStart = C;

  int iError=0;
  // minimize error funtion DistanceFromRotation4Var on P using Powell
  if ((iError=Powell (&DistanceFromRotation4Var, 
                      CStart, C, dMaxError))<0) {
    BIASERR("Minimization did not succeed, error code is "<<iError);
  }
	BIASDOUT(BIAS_FQuasiEuklid,"C =" << C);

	//ho_trivec_outer_product (&pm_epipole2_trivec, &b_trivec, outer_matrix);
  Matrix<double> E2(3,1), BMat(1,3);
  for (int k=0; k< 3;k++){
    E2[k][0]   = Epipole2[k];
    BMat[0][k] = C[k];
  }
  Matrix<double> Outer = E2*BMat;
  BIASDOUT(BIAS_FQuasiEuklid,"M2Matrix = " << M2Matrix);
  BIASDOUT(BIAS_FQuasiEuklid,"Outer = " << Outer);
	
  SubMatrix = M2Matrix + Outer;
	BIASDOUT(BIAS_FQuasiEuklid,"SubMatrix = " << SubMatrix);
	
	// copy left part of P from submatrix
  for (int row = 0; row < 3; row++)
    for (int col = 0; col < 3; col++) {
      P2[row][col] = SubMatrix [row][col];
    }
  BIASDOUT(BIAS_FQuasiEuklid,"P2 = " << P2);

}


void PMatrixEstimation::InitP1P2Trans (BIAS::PMatrix &P1, 
                                       BIAS::PMatrix &P2,
                                       HomgPoint2D &Epipole2) { 
  //  BIASDOUT(BIAS_FMATRIXDEBUG,"In InitP1P2Trans E2 = " << Epipole2);
  P1.set_identity();
  P2.set_identity();
  
  P2[0][3] = Epipole2[0]; 
  P2[1][3] = Epipole2[1]; 
  P2[2][3] = Epipole2[2];
}


/*---------------------------------------------------------------------------
given approximate P2, compute the nearest value of P2 which is exactly
compatible with P1 and F, using the method in the IJCV paper.  variable names
are from the paper too.
----------------------------------------------------------------------------*/
void 
PMatrixEstimation::FindClosestP2 (BIAS::FMatrix &F, 
                                  //BIAS::PMatrix &P1,
                                  BIAS::PMatrix &P2,
                                  Matrix3x3<double> &M2Matrix,
                                  Vector<double> &C, HomgPoint2D& Epipole2 ) 
{

  BIASDOUT(BIAS_FMATRIXDEBUG,"F = " << F);
  Matrix3x3<double> SkewMatrix, SubMatrix;

  double DotProduct=0.0, Magnitude;

  Matrix<double> A(9,4);
  Vector<double> B(9);
  
  F.DecomposetoSR(SkewMatrix, M2Matrix);
  
  BIASDOUT(BIAS_PMATRIXDEBUG,"M2Matrix = " << M2Matrix);

  A[0][0] = M2Matrix [0][0];
  A[0][1] = Epipole2[0];
  A[0][2] = 0;
  A[0][3] = 0;
  B[0]    = P2[0][0];

  A[1][0] = M2Matrix[0][1];
  A[1][1] = 0;
  A[1][2] = Epipole2[0];
  A[1][3] = 0;
  B[1]    = P2[0][1];

  A[2][0] = M2Matrix[0][2];
  A[2][1] = 0;
  A[2][2] = 0;
  A[2][3] = Epipole2[0];
  B[2]    = P2[0][2];

  A[3][0] = M2Matrix[1][0];
  A[3][1] = Epipole2[1];
  A[3][2] = 0;
  A[3][3] = 0;
  B[3]    = P2[1][0];
  
  A[4][0] = M2Matrix[1][1];
  A[4][1] = 0;
  A[4][2] = Epipole2[1];
  A[4][3] = 0;
  B[4]    = P2[1][1];
  
  A[5][0] = M2Matrix[1][2];
  A[5][1] = 0;
  A[5][2] = 0;
  A[5][3] = Epipole2[1];
  B[5]    = P2[1][2];

  A[6][0] = M2Matrix[2][0];
  A[6][1] = Epipole2[2];
  A[6][2] = 0;
  A[6][3] = 0;
  B[6]    = P2[2][0];

  A[7][0] = M2Matrix[2][1];
  A[7][1] = 0;
  A[7][2] = Epipole2[2];
  A[7][3] = 0;
  B[7]    = P2[2][1];

  A[8][0] = M2Matrix[2][2];
  A[8][1] = 0;
  A[8][2] = 0;
  A[8][3] = Epipole2[2];
  B[8] = P2[2][2];

  BIASDOUT(BIAS_PMATRIXDEBUG,"A = " << A);
  BIASDOUT(BIAS_PMATRIXDEBUG,"B = " << B);

  //ma2_solve_equation_svd (a_matrix, x_vector, b_vector);
  Matrix<double> AT  = A.Transpose();
  BIASDOUT(BIAS_PMATRIXDEBUG,"AT = "  << AT);
  Vector<double> ATB = AT*B;
  BIASDOUT(BIAS_PMATRIXDEBUG,"ATB = " << ATB);
  Matrix<double> ATA = AT*A;
  BIASDOUT(BIAS_PMATRIXDEBUG,"ATA = " << ATA);
  
  Vector<double> X=Lapack_LU_linear_solve(ATA,ATB);
  BIASDOUT(BIAS_PMATRIXDEBUG,"Lapack_LU_linear_solve X = " << X);
  M2Matrix *= X[0];
  BIASDOUT(BIAS_PMATRIXDEBUG,"M2Matrix = " << M2Matrix);
  
  // ho_trivec_outer_product (&epipole2_trivec, b_trivec_ptr, outer_matrix);
  Matrix<double> E2(3,1), vMat(1,3);
  for (int k=0; k< 3;k++){
    E2[k][0]   = Epipole2[k];
    vMat[0][k] = X[k+1];
    C[k]       = X[k+1];
  }
  C[3]=1;

  Matrix<double> Outer=E2*vMat;
  SubMatrix = Outer + M2Matrix;
  
  Vector3<double> Col(P2[0][3], P2[1][3], P2[2][3]);
  
  Magnitude  = Col.norm_L2();
  DotProduct = Col * Epipole2;
    
  if (DotProduct > 0)
    Epipole2 *=  Magnitude;
  else
    Epipole2 *= -Magnitude;

  P2[0][0] = SubMatrix[0][0];
  P2[0][1] = SubMatrix[0][1];
  P2[0][2] = SubMatrix[0][2];
  P2[0][3] = Epipole2[0];
  P2[1][0] = SubMatrix[1][0];
  P2[1][1] = SubMatrix[1][1];
  P2[1][2] = SubMatrix[1][2];
  P2[1][3] = Epipole2[1];
  P2[2][0] = SubMatrix[2][0];
  P2[2][1] = SubMatrix[2][1];
  P2[2][2] = SubMatrix[2][2];
  P2[2][3] = Epipole2[2];

}
  




