/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BackwardMapping_hh__
#define __BackwardMapping_hh__

#include <Base/Image/Image.hh>
#include <Image/PyramidImage.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Math/Matrix2x2.hh>

/// if jacobian gets smaller than this, clip to this value to avoid singularity
/// when computing smoothing kernel
#define JACOBI_THRESH 1e-100

namespace BIAS {

  /** @enum InterpolationMethod
      @brief accuracy for resampling

      Trilinear averages different pyramid levels based on a local isotropic
      sampling frequency
      Anisotropic is the most expensive and uses the local jacobian for 
      anti-aliasing, while preserving texture direction which need 
      no smoothing */
  typedef enum InterpolationMethod { MapNearestNeighbor=1, MapBilinear, 
                                     MapBicubic, MapTrilinear, 
                                     MapAnisotropic };

  /** @class BackwardMapping
      @ingroup g_filter
      @brief  abstract base class to maps image src to image sink with 
      arbitrary continuous function

      This class implements a backward mapping and runs over all pixels
      of some output image. For each pixel it calls GetSourceCoordinates
      (which you have to implement in your derived class) to compute
      the coordinates in the source image. The local sampling frequency
      can be computed to avoid aliasing. Either you make an anlytic 
      implementation or simply use the good (but maybe slow) numerical 
      approximation from BackwardMapping.

      For instance if you scale an image by factor 2, you would implement
      GetSourceCoordinates as source_coord = 0.5 * sink_coord and the local
      sampling frequency would be 0.5. Sampling frequencies below 1 wont
      introduce aliasing, while those larger than 1 can.
      Therefore local low pass filtering is simulated by using a pyramid
      of the input image.

      The local sampling frequency of your mapping function can be approximated
      analyzing the jacobian of the function as a R^2->R^2 mapping, see
      HomographyMapping for an example or simply use the numerical solution in
      this class if you are not interested in speed. 
      However, using an image pyramid isotropic scaling is assumed, if you 
      resample by 0.3 in x-direction and 5.0 in y direction, both dirs are
      smoothed with the same sigma and you will lose image information. If
      you want to remove such effects, use anisotropic filtering. This
      is the slowest but best technique.

      To make a pyramid of correct size, Backwardmapping::UpdatePyramidSize
      maps the corners of your sink image and computes the local scale
      at these corners assuming maximum scale at one corner. If your
      derived mapping function has a maximum scale somewhere else, 
      you must override UpdatePyramidSize.

      If possible,bicubic interpolation is done in the image, at borders 
      (high pyramid levels) only bilinear. For efficiency, you can disable
      the usage of pyramids, speeding up the computation (bicubic, bilinear, 
      nearest neighbor). 

      @author koeser, 2005

  */ 
  template <class InputStorageType, class OutputStorageType> 
  class BIASImage_EXPORT BackwardMapping  {
  public:

    BackwardMapping()
    {
#ifdef BIAS_DEBUG
      rangeChecked_ = true;
#endif
      aliasing_ = false;
      autoPyramidSize_ = true;
      interpolationType_ = MapTrilinear;
    }

    virtual ~BackwardMapping() {};

    /** @name User interfaces
        @{
    */

    /** @brief backward mapping with various interpolations

    Sink must be initialized with correct size and background color for invalid
    pixels. This way you can call it several times to create a panorama from
    several source images. For each pixel in sink, GetSourceCoordinates is 
    called and the value in source interpolated.
    */
    int Map(const Image<InputStorageType>& src, 
            Image<OutputStorageType>& sink, InterpolationMethod=MapTrilinear); 

    /** @brief use this for subsequent calls to avoid pyramid recalculation 
        when using tri-filtering */
    int Map(Image<OutputStorageType>& sink);

    /** 
        @}
    */

    /** @brief sets pyramid size to pyramid and updates if neccessary */
    void SetPyramidSize(const int newsize);


    /** @brief calculates the bounding box in sink image, 
        where to do the backward mapping.

        The resulting coordinates must be valid in sink, and the br position
        should be outside the region.
        Implement this in a derived class for speeding up calculation.
        @todo fuse this function with ClipBoundingBoxToROICorners_
    */
    virtual inline int GetBoundingBox(unsigned int /*srcwidth*/, 
                                      unsigned int /*srcheight*/,
                                      unsigned int sinkwidth, 
                                      unsigned int sinkheight,
                                      unsigned int &tlx, unsigned int &tly, 
                                      unsigned int &brx, unsigned int &bry)
    { 
      tlx=0; tly=0;     
      brx=sinkwidth; bry=sinkheight;   
      return 0; 
    } 


    /** @brief use coordinates from sink to estimate maximum pyramid level 
        needed, update pyramid size */
    virtual void UpdatePyramidSize(const Image<OutputStorageType>& sink);

  protected:

    /** @name geometrical coordinate mapping functions
        @{
    */

    /** @brief actual mapping function which takes a point from sink
        and computes a point in source and a local sampling frequency */
    virtual int GetSourceCoordinates_(const HomgPoint2D& sink,
                                      HomgPoint2D& source) const;

    /** @brief numeric approximation of jacobian, reimplement in derived class
        to make an analytic jacobian */
    virtual int GetJacobian_(const HomgPoint2D& sink,
                             Matrix2x2<double>& Jacobian) const;

    /** 
        @}
    */

    /** @name pixel-value-mapping worker functions
        @{
    */

    /** @brief map image avoiding aliasing 

    This is almost correct regarding the sampling theorem, because it uses
    trilinear or anisotropic filtering over a pyramid. The local 
    sampling frequency
    is estimated by analyzing the jacobian of the homography as a R^2->R^2
    mapping. This is very slow. If possible, bicubic interpolation is done 
    in the image, at borders (high pyramid levels) only bilinear.<br>
    In auto mode pyramid size is checked 
    @author koeser 04/2005   */
    int MapTri_(Image<OutputStorageType>& sink);

    /** @brief bilinear, bicubc and nearest neighbor interpolation     */
    int MapBi_(const Image<InputStorageType>& source,
               Image<OutputStorageType>& sink,  
               InterpolationMethod interpolationQuality);




    /** @brief worker function, which handles bilinear/bicubic interpolation,
        pyramid levels, borders and weights */
    int GetImageValue_(const double& xsource, 
                       const double& ysource,
                       const Image<InputStorageType>& im,
                       unsigned int channel,
                       double& result);


    /** @brief worker function, which handles bilinear/bicubic interpolation,
        pyramid levels, borders and weights */
    int GetTrilinearImageValue_(const double& xsource, 
                                const double& ysource, 
                                const double& scale, 
                                unsigned int channel,
                                double& result);


    /** @brief use jacobian of mapping function to do anisotropic 
        antialiasing */
    int GetAnisotropicImageValue_(const double& xsource, 
                                  const double& ysource, 
                                  Matrix2x2<double>& Jacobian, 
                                  unsigned int channel,
                                  double& thereturnvalue);

    /** 
        @}
    */

    /** @brief clip interpolated value to outputstoragetype */
    void ClipValue_(const double& value, OutputStorageType& newvalue);

    /** @brief clip computed region in dest image to roi 
        @todo fuse this function with GetBoundingBox */
    void ClipBoundingBoxToROICorners_(BIAS::Image<OutputStorageType>& sink, 
                                      unsigned int& tlx, unsigned int& tly, 
                                      unsigned int& brx, unsigned int& bry);

    /** @brief compute maximum number of pyramid levels needed */
    virtual void BuildPyramid_(const Image<InputStorageType>& source, 
                               bool forceInit = true);

    /** @brief takes jacobian to compute gaussian kernel for 
        anisotropic smoothing */
    void JacobianToMask_(const Matrix2x2<double>& Jacobian, 
                         Matrix<double>& theMask) const;

    /// this contains a pyramid of the last source image for anti-aliasing
    PyramidImage<InputStorageType> pyramid_;
    /// the roi of the last source image
    ROI ROI_;

    /// pyramid size set by user or computed automatically ?
    bool autoPyramidSize_;

    /// which interpolation method are we using
    InterpolationMethod interpolationType_;

#ifdef BIAS_DEBUG
    /// the interpolated value is computed in double, 
    /// if clipvalue is not specialized for your outputstoragetype(e.g. ushort)
    /// this bool indicates that the generic conversion (e.g. double->ushort)
    /// could have introduced artefacts (65536.0 -> 1)
    bool rangeChecked_;
#endif

    /// set to true if possibility of aliasing is detected during mapping
    bool aliasing_;  
  };


#ifdef WIN32

  template <>
  void BackwardMapping<float, float>::ClipValue_(const double& value,
                                                 float& newvalue);

  template <>
  void BackwardMapping<unsigned char, 
                       unsigned char>::ClipValue_(const double& value,
                                                  unsigned char& newvalue);

#endif // WIN32

} // namespace 


#endif
