/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "ColorMultiHistogram.hh"
#include <Base/Image/ImageBase.hh>

using namespace BIAS;
using namespace std;


/**
   This class creates a ColorMultiHistogram in the HSL color space.
   @author ingo schiller
   @date march 2004 
 */
template<class StorageType>
ColorMultiHistogram<StorageType>::ColorMultiHistogram()
{
  minSat_=10;
  lambda_=20;
  binSize_=32;
  hist1_.SetMinSat(minSat_);
  hist1_.SetLambda(lambda_);
  hist1_.SetBinSize(binSize_);
  hist1_.SetExpDist(true); 
  hist2_.SetMinSat(minSat_);
  hist2_.SetLambda(lambda_);
  hist2_.SetBinSize(binSize_);
  hist2_.SetExpDist(true);
}

template<class StorageType>
ColorMultiHistogram<StorageType>::~ColorMultiHistogram() 
{
  
}


//set the region of interest in first histogram
template<class StorageType>
int ColorMultiHistogram<StorageType>::SetROI1(int ux, int uy, int lx, int ly){
  
  hist1_.SetROI(ux,uy,lx,ly);
 
  return 0;
}

//set the region of interest in second histogram
template<class StorageType>
int ColorMultiHistogram<StorageType>::SetROI2(int ux, int uy, int lx, int ly){
  
  hist2_.SetROI(ux,uy,lx,ly);
  return 0;
}


//Set the binsize, which is then set in the ColorHistograms
template<class StorageType>
int ColorMultiHistogram<StorageType>::SetBinSize(int binsize){
  hist1_.SetBinSize(binsize);
  hist2_.SetBinSize(binsize);
  return 0;
}

/**
   @returns the first ColorHistogram
*/
template<class StorageType>
ColorHistogram<StorageType> ColorMultiHistogram<StorageType>::GetHist1() const 
{
  return hist1_;
}


/**
   @returns the second ColorHistogram
*/
template<class StorageType>
ColorHistogram<StorageType> ColorMultiHistogram<StorageType>::GetHist2() const 
{
  return hist2_;
}


/**
   @returns the distance of the centers of the ROIs
   in pixels
*/
template<class StorageType>
void ColorMultiHistogram<StorageType>::GetDistROIs(std::vector<double>& dist) const 
{
  std::vector<int> first,second;
  dist.assign(2,0.0);
  
  hist1_.GetROI(first);
  hist2_.GetROI(second);
  
  dist[0]=(second[2]+second[0])/2-(first[2]+first[0])/2;
  dist[1]=(second[3]+second[1])/2-(first[3]+first[1])/2;
  
}


template<class StorageType>
int ColorMultiHistogram<StorageType>::Clear()
{
  hist1_.Clear();
  hist2_.Clear();
 
  return 0;
}

/**
   simply call the GenerateHist() from ColorHistogram twice here 
*/ 
template<class StorageType>
int ColorMultiHistogram<StorageType>::GenerateHist(BIAS::Image<StorageType>& img)
{
  hist1_.GenerateHist(img);
  hist2_.GenerateHist(img);
  return 0;
}

/**
   simply call the GenerateHist() from ColorHistogram twice here 
*/
template<>
int ColorMultiHistogram<unsigned char>::GenerateHist(BIAS::Image<unsigned char>& img)
{
  hist1_.GenerateHist(img);
  hist2_.GenerateHist(img);
 
  return 0;
}

/**
   simply call the CalcSimilarity from ColorHistograms here and 
   multiply the values. This is the distance of the MultiHistogram
*/ 
template<class StorageType>
double ColorMultiHistogram<StorageType>::CalcSimilarity(const BIAS::ColorMultiHistogram<StorageType> &chist) const
{
  //indicators for similarity
  double dist1=0,dist2=0; 
  dist1=hist1_.CalcSimilarity(chist.GetHist1());
  //cout<<"Dist1: "<<dist1<<endl;
  dist2=hist2_.CalcSimilarity(chist.GetHist2());  
  //cout<<"Dist2: "<<dist2<<endl;
  dist1+=dist2;
  dist1/=2;
  return dist1;
}

template<class StorageType> double 
ColorMultiHistogram<StorageType>::CalcSimilarityColorWeighted(const BIAS::ColorMultiHistogram<StorageType> &chist) const
{
  //indicators for similarity
  double dist1=0,dist2=0, w1, w2; 
  dist1=hist1_.CalcSimilarity(chist.GetHist1(), w1);
  //cout<<"Dist1: "<<dist1<<endl;
  dist2=hist2_.CalcSimilarity(chist.GetHist2(), w2);  
  //cout<<"Dist2: "<<dist2<<endl;
  dist1= (dist1*w1 + dist2*w2)  / (w1+w2);
  return dist1;
}

template<class StorageType> 
int ColorMultiHistogram<StorageType>::Update(const ColorMultiHistogram<StorageType> &hist,double factor) 
{
  hist1_.Update(hist.GetHist1(),factor);
  hist2_.Update(hist.GetHist2(),factor);
  return 0;
}


/**
   dumps information on MultiHistogram to ostreams
   format: first hist1, then hist2
   binsize tab ux tab uy tab lx tab ly
   hs histogram, as a matrix with binsize x binsize (tab separated)
   l values as a line with lenght binsize (tab separated)
*/
template<class StorageType>
void ColorMultiHistogram<StorageType>::Dump(std::ostream& os,std::ostream& os2)
{
  hist1_.Dump(os);
  hist2_.Dump(os2);
}

//reads the information of a histogram from disk
template<class StorageType>
int ColorMultiHistogram<StorageType>::Read(char *filename1, char *filename2)
{
  //cout<<"Reading the histograms"<<endl;
 
  hist1_.Read(filename1);
  hist2_.Read(filename2);

  return 0;
}



#define INST(type) template class ColorMultiHistogram<type>;

INST(float)
INST(unsigned char)

  // disabled untested instances, myriards of warnings
// #ifdef BUILD_IMAGE_INT
// INST(int);
// #endif
// #ifdef BUILD_IMAGE_CHAR
// INST(char);
// #endif
// #ifdef BUILD_IMAGE_SHORT
// INST(short);
// #endif
// #ifdef BUILD_IMAGE_USHORT
// INST(unsigned short);
// #endif
// #ifdef BUILD_IMAGE_USHORT
// INST(unsigned int);
// #endif
// #ifdef BUILD_IMAGE_DOUBLE
// INST(double);
// #endif
