/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Image/HomographyMapping.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Debug/TimeMeasure.hh>

using namespace BIAS;
using namespace std;

using namespace BIAS;
using namespace std;

int main(int argc, char *argv[])
{

  if (argc<2){
    BIASERR(argv[0]<<" <image-file> [hmatfile [width height]]");
    return -1;
  }
 
  Image<unsigned char> im;
  if (ImageIO::Load(argv[1], im)!=0){
    BIASERR("error loading "<<argv[1]);
    return -2;
  }
  int width  = im.GetWidth();
  int height = im.GetHeight();

  cout<<"ROI of source image is "<<*im.GetROI()
      <<" while image size is "<<width<<" x "<<height<<endl;

 
  
  HomographyMapping<unsigned char, unsigned char> Mapper;
  HMatrix H(MatrixIdentity);
  if (argc>2) {
    cout<<"Reading HMatrix "<<argv[2]<<endl;
    H.Load(argv[2]);
    cout<<"H is "<<H<<endl;
    H = H.Invert();
  } else {
 //    H[0][0]= 1.05;   H[0][1]= 1.1;      H[0][2] = -350.0;
//     H[1][0]=-1.1;    H[1][1]= 1.85;     H[1][2] = -100.0;
    double angle = -15.0 / 180.0 * M_PI;
    HMatrix HRot(MatrixIdentity);
    HRot[0][0]= cos(angle);   HRot[0][1] = sin(angle);   
    HRot[1][0]= -sin(angle);    HRot[1][1]= cos(angle); 
    HMatrix HScale(MatrixIdentity);
     HScale[0][0] = 1.0; HScale[1][1] = 5.0;
    //HScale[0][0] = 2.0; HScale[1][1] = 2.0; 
     HScale[1][2] = -400;
    H = HScale * HRot;
    cout<<"Using default homography "<<H<<endl;
  }

  if (argc>4) {
    width = atoi(argv[3]);
    height = atoi(argv[4]);
    cout<<"Using imagesize "<<width<<" x "<<height<<endl;
  } else {
    cout<<"Using input image size "<<width<<" x "<<height<<endl;
  }

  Image<unsigned char> 
    imresnearest(width, height, im.GetChannelCount()),
    imresbilinear(width, height, im.GetChannelCount()),
    imresbicubic(width, height, im.GetChannelCount()),
    imrestrilinear(width, height, im.GetChannelCount()),
    imresanisotropic(width, height, im.GetChannelCount());


  // trilinear mapping with automatic scale and pyramid size selection
  Mapper.SetHomography(H);
 
  cout<<"timing ..."<<endl<<flush;
  unsigned int numbermeasurements = 1;
  TimeMeasure t1, t2, t3, t4, t5;
  for (unsigned int i=0; i<numbermeasurements; i++) {
    t1.Start();
    // trilinear mapping with automatic scale and pyramid size selection
    Mapper.Map(im, imresnearest, MapNearestNeighbor);
    t1.Stop();
    
    t2.Start();
    // fast backward mapping (bilinaer) disregarding sampling theroem
    Mapper.Map(im, imresbilinear, MapBilinear);
    t2.Stop();

    t3.Start();
    // nn mapping 
    Mapper.Map(im, imresanisotropic, MapAnisotropic);
    t3.Stop();
    
    t4.Start();
    // trilinear mapping with automatic scale and pyramid size selection
    Mapper.Map(im, imrestrilinear, MapTrilinear);
    t4.Stop();

    t5.Start();
    // trilinear mapping with automatic scale and pyramid size selection
    Mapper.Map(im, imresbicubic, MapBicubic);
    t5.Stop();

  }
  cout <<"Anisotropic mapping took "
       <<t3.GetUserTime()/double(numbermeasurements)
       <<" ms, "<<endl
       <<"Trilinear mapping took "
       <<t4.GetUserTime()/double(numbermeasurements)
       <<" ms,"<<endl
       <<"Bicubic mapping took "
       <<t5.GetUserTime()/double(numbermeasurements)
       <<" ms,"<<endl
       <<"Bilinear mapping took "
       <<t2.GetUserTime()/double(numbermeasurements) 
       <<" ms,"<<endl
       <<"NN-mapping took "
       <<t1.GetUserTime()/double(numbermeasurements) <<" ms"<<endl;

  if (ImageIO::Write("image_homography_anisotropic.mip", imresanisotropic)!=0){
    BIASERR("error image");
    return -3;
  }
  if (ImageIO::Write("image_homography_trilinear.mip", imrestrilinear)!=0){
    BIASERR("error image");
    return -3;
  }
  if (ImageIO::Write("image_homography_bicubic.mip", imresbicubic)!=0){
    BIASERR("error image");
    return -3;
  }
  if (ImageIO::Write("image_homography_bilinear.mip", imresbilinear)!=0){
    BIASERR("error image");
    return -3;
  }
  if (ImageIO::Write("image_homography_nearest.mip", imresnearest)!=0){
    BIASERR("error image");
    return -3;
  }


  return 0;
}
