/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "HomographyMapping.hh"
#include <Geometry/RMatrix.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Geometry/HomgPoint2D.hh>

using namespace BIAS;

template <class InputStorageType, class OutputStorageType>
HomographyMapping<InputStorageType, OutputStorageType>::
~HomographyMapping() 
{}


template <class InputStorageType, class OutputStorageType>
HomographyMapping<InputStorageType, OutputStorageType>::
HomographyMapping() 
: H_(MatrixIdentity) 
{}


template <class InputStorageType, class OutputStorageType>
HomographyMapping<InputStorageType, OutputStorageType>::
HomographyMapping(const HomographyMapping & src) 
{
    (*this) = src;
}


template <class InputStorageType, class OutputStorageType>
HomographyMapping<InputStorageType, OutputStorageType> & 
HomographyMapping<InputStorageType, OutputStorageType>::
operator=(const HomographyMapping & src)
{
     H_ = src.H_;     
     return *this;
}



template <class InputStorageType, class OutputStorageType>
int HomographyMapping<InputStorageType, OutputStorageType>::
RotateToGrey(Image<InputStorageType>& src, 
             Image<OutputStorageType>& sink,
             float PhiX, float PhiY, float PhiZ, 
             const Matrix3x3<double> &K) 
{
  Matrix3x3<double> Kinv;
  Matrix3x3<double> KRKinv;
  Image<InputStorageType> source;

  double centerx,centery;

  if (!src.IsEmpty()){

    if(src.GetColorModel() != ImageBase::CM_Grey)
      ImageConvert::Convert(src, source, ImageBase::CM_Grey);
    else
      source=src;

    if(!sink.IsEmpty())
      sink.Release();

    sink.Init(source.GetWidth(),source.GetHeight(),source.GetChannelCount());

    // Prepare inverse camera mattix
    Kinv.set_identity();
    Kinv[0][0] = 1 / K[0][0];
    Kinv[1][1] = 1 / K[1][1];
    Kinv[0][2] = -K[0][2]/K[0][0];
    Kinv[1][2] = -K[1][2]/K[1][1];

    centerx = K[0][0];
    centery = K[1][1];

    RMatrix R;
    R.SetXYZ(PhiX,PhiY,PhiZ);
    KRKinv = K*R*Kinv;

    /*cout << "K=" << K << endl;
      cout << "R=" << R << endl;
      cout << "Kinv = " << Kinv << endl;
      cout << "KRKinv=" << KRKinv << endl;*/
    SetHomography(KRKinv);
    return Map(src, sink);
  } else{
    return -1;
  }
}



template <class InputStorageType, class OutputStorageType>
int HomographyMapping<InputStorageType, OutputStorageType>::
GetSourceCoordinates_(const HomgPoint2D& sink,
                      HomgPoint2D& source) const {
  // compute coordinates in source image with backward hoography
  H_.Mult(sink, source);
  source.Homogenize();
  return 0;
}

template <class InputStorageType, class OutputStorageType>
int HomographyMapping<InputStorageType, OutputStorageType>::
GetJacobian_(const HomgPoint2D& sink, Matrix2x2<double>& Jacobian) const {
  // analytic jacobian of homography
  H_.GetJacobian(sink, Jacobian);
  return 0;
}


using namespace std;

template <class InputStorageType, class OutputStorageType>
int HomographyMapping<InputStorageType, OutputStorageType>::
GetBoundingBox(unsigned int sourcewidth, unsigned int sourceheight,
               unsigned int /*sinkwidth*/, unsigned int /*sinkheight*/,
               unsigned int &tlx, unsigned int &tly, 
               unsigned int &brx, unsigned int &bry) {

  // do a forward mapping of the corners of the source image and compute
  // convex hull rectangle

  HomgPoint2D tl(0,0,1), br(sourcewidth-1, sourceheight-1,1), bbTL, bbBR;
  HomgPoint2D bl(0,sourceheight-1,1), tr(sourcewidth-1,0,1), bbBL, bbTR;
  
  HMatrix HI=H_.Invert();
  bbTL= HI*tl; bbTL.Homogenize();
  bbBR= HI*br; bbBR.Homogenize();
  bbBL= HI*bl; bbBL.Homogenize();
  bbTR= HI*tr; bbTR.Homogenize();
  
  double dtlx,dtly, dbrx,dbry;

  // min(x) and min(y)
  dtlx=(bbTL[0]+0.5); 
  if (bbBR[0]<dtlx) dtlx=(bbBR[0]);
  if (bbBL[0]<dtlx) dtlx=(bbBL[0]);
  if (bbTR[0]<dtlx) dtlx=(bbTR[0]);

  dtly=(bbTL[1]+0.5); 
  if (bbBR[1]<dtly) dtly=(bbBR[1]);
  if (bbBL[1]<dtly) dtly=(bbBL[1]);
  if (bbTR[1]<dtly) dtly=(bbTR[1]);

  // max(x) and max(y)
  dbrx=(bbTL[0]+0.5); 
  if (bbBR[0]>dbrx) dbrx=(bbBR[0]);
  if (bbBL[0]>dbrx) dbrx=(bbBL[0]);
  if (bbTR[0]>dbrx) dbrx=(bbTR[0]);

  dbry=(bbTL[1]+0.5); 
  if (bbBR[1]>dbry) dbry=(bbBR[1]);
  if (bbBL[1]>dbry) dbry=(bbBL[1]);
  if (bbTR[1]>dbry) dbry=(bbTR[1]);
  
  //// clip to image borders
  // tl is part of the bounding box
  if (dtlx<0) dtlx=0; 
  if (dtlx >= double(INT_MAX)) dtlx = INT_MAX;

  if (dtly<0) dtly=0; 
  if (dtly >= double(INT_MAX)) dtly = INT_MAX;
  
  // br is outside the boundinx box
  if (dbrx<=0) dbrx = 1; 
  if (dbrx>double(INT_MAX)) dbrx = INT_MAX;
  if (dbry<=0) dbry = 1; 
  if (dbry>double(INT_MAX)) dbry = INT_MAX;
  
  tlx = (unsigned int)dtlx; 
  tly = (unsigned int)dtly;
  brx = (unsigned int)ceil(dbrx); 
  bry = (unsigned int)ceil(dbry);
  
  return 0;
}

// TODO FIXME win32 template instantiation problem (JW)
template class HomographyMapping<unsigned char, unsigned char>;
template class HomographyMapping<float, float>;
template class HomographyMapping<unsigned char, float>;

#if defined(BUILD_IMAGE_CHAR)
// TODO FIXME win32 template instantiation problem (JW)
template class HomographyMapping<unsigned char, char>;
template class HomographyMapping<char, char>;
#endif

#if defined(BUILD_IMAGE_USHORT)
template class HomographyMapping<unsigned short,unsigned short>;
#endif

#if defined(BUILD_IMAGE_SHORT)
template class HomographyMapping<short, short>;
#endif

#if defined(BUILD_IMAGE_SHORT)&&defined(BUILD_IMAGE_USHORT)
template class HomographyMapping<unsigned short, short>;
#endif

#if defined(BUILD_IMAGE_INT)
template class HomographyMapping<int,int>;
#endif

#if defined(BUILD_IMAGE_USHORT)
template class HomographyMapping<unsigned short,float>;
#endif

#if defined(BUILD_IMAGE_USHORT) && defined(BUILD_IMAGE_INT)
template class HomographyMapping<unsigned short,int>;
#endif

#if defined(BUILD_IMAGE_DOUBLE)
template class HomographyMapping<double,double>;
#endif
