/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __PYRAMID_HH__
#define __PYRAMID_HH__

#include <Base/Common/BIASpragmaStart.hh>

#include <Base/Image/Image.hh>
#include <Utils/Param.hh>
#include <Filter/Rescale.hh>

#include <float.h>
#include <vector>

// for GetSingleImage
// border between pyramid level and fill color
#define PYRAMID_BORDER 2
#define PYRAMID_BACKGROUND 0

#define D_PYRAMID_CREATE      0x00000001
#define D_PYRAMID_SINGLE_IM  0x00000002

namespace BIAS 
{
  /** @class PyramidImage
      @ingroup g_image
      @brief class for holding downsampled images 

      pointer to the images are held in this class 
      @author woelk 09/2004 */
  template <class StorageType>
  class BIASImage_EXPORT PyramidImage : public std::vector<Image<StorageType>* >, public Debug
  {
  
  public:
    PyramidImage();

    PyramidImage(Param& para, const std::string& postfix="");

    PyramidImage(Param& para, const int grp, const std::string& postfix);

    PyramidImage(const PyramidImage<StorageType>& pim);

    ~PyramidImage();

    /** 
        @author woelk 09/2004 */
    void Init(const Image<StorageType>& image);

    /** 
        @author woelk 09/2004 */
    void Init(unsigned int width, unsigned int height, 
              unsigned int channelcount);

    /** also deletes images 
        @author woelk 09/2004 */
    void clear();

    inline void Release() { clear(); }

    /** 
        @author woelk 09/2004 */
    PyramidImage<StorageType>& 
    operator=(const PyramidImage<StorageType>& pim);

    /** downsamples from (*this)[0]
        assumes that all pointers e.g. (*this)[1] are valid 
        and point to an image of correct size
        returns 0 on success, anything else on failure 
        @author woelk */
    int Downsample();

    /** returns a single image containing all pyramid images
        @author woelk 09/2004 */
    void GetSingleImage(Image<StorageType>& im) const;

    void SetUID(BIAS::UUID uid);

    /** set the ROI for all images */
    int SetROI(unsigned minx, unsigned miny, unsigned maxx, unsigned maxy);

    inline bool IsEmpty() { return (this->size() == 0); }

    /** @brief sets all pixels in all images to zero
        @author woelk 10/2004 */
    void SetZero();

    /** @brief writes a file in mip format per pyramid level using prefix
        @author woelk 10/2004 */
    int WriteImages(const std::string& prefix);

    /////////////////////////////////////////////////////////////////////////
    //              get/set functions
    /////////////////////////////////////////////////////////////////////////

    inline void SetRescaleFactor(const double factor) 
    { *_RescaleFactor=factor; }

    inline double GetRescaleFactor() const { return *_RescaleFactor; }

    inline void SetPyramidSize(const int size) { *_PyramidSize=size; }
    
    inline int GetPyramidSize() const { return *_PyramidSize; }

    inline void SetLowPassType(const int lptype) 
    { _rescale.SetLowPassType(lptype); }

    inline int GetLowPassType() const { return _rescale.GetLowPassType(); }

    inline std::vector<double> GetFactors() const { return _vFactor; };

    inline std::vector<double>& GetFactors() { return _vFactor; };

    /// @author woelk, see _PostionOffset for details
    inline double GetPositionOffset() const { return _PositionOffset; };

  protected:
    // rescale filter for downsampling
    Rescale<StorageType, StorageType> _rescale;

    bool _ParamsAllocated;
    // param
    int *_PyramidSize;
    // the param of the rescale filter 
    double *_RescaleFactor;
    
    // _vFactor[i] is the dowsampling factor from (*this)[0] -> (*this)[i]
    std::vector<double> _vFactor; 
    
    /** Offset which needs to be used when transferring point locations 
        between pyramid levels:

        (2*xs+offset, 2*ys+offset) = (xb, yb) or
        (xs, ys) = ((xb-offset)/2.0, (yb-offset)/2.0)

        where (xs, ys) is the pixel position in pixel coo. the smaller  
        destination image and (xb, yb) is the position in pixel coo. in the 
        bigger source image.

        see Rescale::DowsampleBy2 for details
        @author woelk 05/2005 */
    double _PositionOffset;

    /** creates the levels by downsampling from (*this)[0] */
    void _CreateLevels();
    
    void _AddParam(Param& para, const int grp, const std::string& postfix);
    
  };

} // namespace BIAS


#include <Base/Common/BIASpragmaEnd.hh>

#endif
