
# include "RigParameter.hh"
# include "CameraParam.hh"
#ifdef BIAS_HAVE_XML2
#  include <Base/Common/XMLIO.hh>
#endif
#include <Image/SphericalUndistortion.hh>
#include <Image/MatlabTBUndistortion.hh>

using namespace std;
using namespace BIAS;


bool RigParameter::ReadFromCMP(const string &filename)
{
  char tmpline[256];
  string viewblock, token;
  unsigned int openbrace,linenum;
  int pos;
  bool BlockStarted=false;
     
  // at first clear internal data members
  Center_.clear();
  Rotation_.clear();
  Calibration_.clear();
  Camera_.clear();

  ifstream ifs(filename.c_str());
  if(!ifs){
    perror(filename.c_str());
    return false;
  }
  openbrace = 0;  linenum = 0;
    
  int NoRead = 0;

  while (ifs && NoRead < 6){
    ifs.getline(tmpline,sizeof(tmpline));
    linenum++;
    // remove any trailing CR or LF
    for (int i=strlen(tmpline)-1; i>0; i--)
      if (tmpline[i] == 13 || tmpline[i]==10) tmpline[i] =0;
    
    if (tmpline[0] == 0 || tmpline[0] == '#') {
      continue;
    }

    pos = 0;
    while (tmpline[pos] != 0){
      if (tmpline[pos] == '{') { 
        openbrace ++;
        if (!BlockStarted) {
          BlockStarted = true;
          //cout <<"started block at line "<<linenum<<" : "<<tmpline<<endl;
        }
      }
      if (tmpline[pos] == '}') openbrace --;
      if (openbrace == 0 && BlockStarted){
        //cout <<"finished block at line "<<linenum<<" : "<<tmpline<<endl;
        viewblock.append(tmpline);
        ParseView(viewblock);
        NoRead++;
        viewblock="";
        BlockStarted = false;
      }
      pos++;
    }
    if (BlockStarted)
      viewblock.append(tmpline);
  }
  ifs.close();
  return 0;
}


int  RigParameter::ParseView(const string &block)
{

  char delim;
  int startofblock,endoflastmatrix,startofmatrix,endoflastline;
  int startofline, endofline;
  
  std::vector<Matrix4x4<double> > M;
  M.resize(3);

  RMatrix R; 
  KMatrix K;
  
  startofblock = block.find('{', 0);
  endoflastmatrix = startofblock+1;
  for (int m=0; m<3; m++) {
    startofmatrix = block.find('{',endoflastmatrix);
    endoflastline = startofmatrix+1;
      
    for (int line=0; line<4; line++){
      startofline = block.find('{',endoflastline);  
      endofline = block.find('}',startofline);
      string fourfloats(block.substr(startofline+1,endofline-startofline-1));
      // 	cout <<"line: "<<line<<" : "<<fourfloats<<endl;
      stringstream iss(fourfloats);
      iss >>(M[m])[line][0]>>delim>>(M[m])[line][1]>>delim>>(M[m])[line][2]
          >>delim>>(M[m])[line][3];
      endoflastline = endofline;
    }
    endoflastmatrix = block.find('}',endoflastline);
  }
    
  // set data members for the first camera
  //cout << M[0] << M[1] << M[2] << endl;
  K.SetIdentity();
  for (unsigned int i = 0 ; i < 3; i++){
    for (unsigned int j = 0 ; j < 3; j++){
      R[j][i] = M[0][i][j];
    }
    
    if (i < 2){
      K[i][i] = 1/M[1][i][i];
      K[i][2] = -M[2][i][3];
    }
  }
  
  Rotation_.push_back(R);
  Calibration_.push_back(K);

  return 0;
    
}
  
void RigParameter::SetLadybugCenters(double value /*= 1.0 */){
  
  Vector3<double> CenterVec(0.0 , 0.0 , value);
  
  for (unsigned int i = 0; i < Calibration_.size(); i++){
    Center_.push_back(Rotation_[i]*CenterVec);
  }


}

void RigParameter::AdaptAxisAngleRepresentation(Vector3<double>& Axis,
                                                double& Angle,
                                                const Vector3<double>& RefAxis,
                                                const double& RefAngle) {
  // check if optical axes point into same half space
  if (RefAxis.ScalarProduct(Axis)<0.0) {
    Axis *= -1.0;
    Angle = 2.0 * M_PI - Angle;
  }
  // check 2pi wrap-around problem to avoid average of 0 and 2pi to be pi
  if (Angle>RefAngle) {
    // +/- 2pi does not change the angle: 
    while (Angle-RefAngle>M_PI) Angle -= 2.0*M_PI;
  } else {
    // +/- 2pi does not change the angle: 
    while (RefAngle-Angle>M_PI) Angle += 2.0*M_PI;
  }
}

bool RigParameter::Compute(std::vector<BIAS::PMatrix>& Pvec, 
                           const unsigned int& Rigsize,
                           const double& WeightOfCurrentParams){


  //TODO: Implement Camera Object into Rig functionality
  
#ifdef BIAS_DEBUG
  if (Pvec.size()%Rigsize) {
    BIASERR("Vector of cameras has not multiple size of Rigsize"); 
    abort();
  }

  if (WeightOfCurrentParams<0.0) {
    BIASERR("Useless negative weight !");
    abort();
  }

  if (WeightOfCurrentParams > 0.0) {
    if ((Center_.size()!=Rigsize)||(Rotation_.size()!=Rigsize)||
        (Calibration_.size()!=Rigsize)) {
      BIASERR("Cannot average differently sized vectors !");    
      abort();
    }
  }
#endif


  std::vector<Vector3<double> > dC(Pvec.size());
  std::vector<Vector3<double> > dRAx(Pvec.size()); // Rotation_ axes
  std::vector<double> dRAn(Pvec.size());           // Rotation_ angles
  RMatrix tmpR,firstCamR;

  std::vector<Vector3<double> > MeanDeltaRAx(Rigsize), RefAxes(Rigsize);
  std::vector<double> MeanDeltaRAn(Rigsize), RefAngle(Rigsize);

  if (WeightOfCurrentParams==0.0) {
    Center_.resize(Rigsize);
    Rotation_.resize(Rigsize);
    Calibration_.resize(Rigsize);

    firstCamR = Pvec[0].GetR();
    firstCamR.GetRotationAxisAngle(RefAxes[0], RefAngle[0]);
    for (unsigned int r = 1; r < Rigsize; r++) {
      tmpR = firstCamR.Transpose() * Pvec[r].GetR();
      tmpR.GetRotationAxisAngle(RefAxes[r], RefAngle[r]);
      MeanDeltaRAx[r] = BIAS::Vector3<double>(0,0,0);
      MeanDeltaRAn[r] = 0;
    }
  } else {
    for (unsigned int r = 0; r < Rigsize; r++) { 
      Rotation_[r].GetRotationAxisAngle(MeanDeltaRAx[r], MeanDeltaRAn[r]);
      MeanDeltaRAx[r] *= WeightOfCurrentParams*Pvec.size()/Rigsize;
      MeanDeltaRAn[r] *= WeightOfCurrentParams*Pvec.size()/Rigsize;
      RefAxes[r]  = MeanDeltaRAx[r];
      RefAngle[r] = MeanDeltaRAn[r];
    }
  } 

  for (unsigned int r = 0; r < Rigsize; r++){ 
    Center_[r]      *= WeightOfCurrentParams*Pvec.size()/Rigsize;
    Calibration_[r] *= WeightOfCurrentParams*Pvec.size()/Rigsize;
  }


   // loop over rigs
  unsigned int index=0;
  for (; index <= Pvec.size()-Rigsize; index+= Rigsize){
    // loop over cameras of the rig
    for (unsigned int r = 0; r < Rigsize; r++){
      // Pvec[index+r].InvalidateDecomposition();
      if (r == 0){
        // for first rig, get R and C
        firstCamR=Pvec[index+r].GetR();
        //cout <<"FirstCamR["<<index<<"]="<<firstCamR<<" with det="
        //    <<firstCamR.det()<<" and norm "<<firstCamR.NormFrobenius()<<endl;
        firstCamR.GetRotationAxisAngle(dRAx[index],dRAn[index]);
        //cout <<"dRAx ("<<index<<") is "<<dRAx[index]<<" with angle "<<dRAn[index]<<endl;
        AdaptAxisAngleRepresentation(dRAx[index],dRAn[index],
                                     RefAxes[0], RefAngle[0]);

        dC[index]=Pvec[index+r].GetC();
      } else {
        dC[index+r]= firstCamR * (Pvec[index+r].GetC()-dC[index]);
        // dC[index+r]= (Pvec[index+r].GetC()-dC[index]);
        //        tmpR=Pvec[index+r].GetR()*firstCamR.Transpose();

        tmpR = firstCamR.Transpose() * Pvec[index+r].GetR();
        tmpR.GetRotationAxisAngle(dRAx[index+r],dRAn[index+r]);
        AdaptAxisAngleRepresentation(dRAx[index+r],dRAn[index+r],
                                     RefAxes[r], RefAngle[r]);
        cout <<"["<<r<<"]"<<" Computed Rotation_ Axis "<<dRAx[index+r]<<" and "
             <<" Angle="<<dRAn[index+r]<<endl;
      }
    }

    dC[index]=Vector3<double>(0,0,0); 
    dRAx[index] =  Vector3<double>(1,0,0);
    dRAn[index] = 0;
  }
                                                                 
 

  for(index = 0 ; index <= dC.size()-Rigsize; index+=Rigsize){
    // loop over all cameras except the first one because its the reference
    // camera
    for (unsigned int r = 0; r < Rigsize; r++){
      cout <<"dc ("<<index+r<<") is "<<dC[index+r]<<endl;
      Center_[r] += dC[index+r];
      cout <<"dRAx ("<<index+r<<") is "<<dRAx[index+r]<<endl;
      MeanDeltaRAx[r] += dRAx[index+r];
      cout <<"MeandRAx is now "<<MeanDeltaRAx[r] <<endl;
      cout <<"dRAn ("<<index+r<<") is "<<dRAn[index+r]<<endl;
      MeanDeltaRAn[r] += dRAn[index+r];
      Calibration_[r] = KMatrix(Calibration_[r] + Pvec[index+r].GetK()); 
    }
  }
  double normalizationfactor =
    (double)Rigsize/((double)Pvec.size()*(1.0+WeightOfCurrentParams));

  cout <<"normalization is "<<normalizationfactor<<endl;
  for (unsigned int r = 0; r < Rigsize; r++){
    Center_[r] *= normalizationfactor;
    MeanDeltaRAn[r] *= normalizationfactor;
    while (MeanDeltaRAn[r]>2.0*M_PI) {
      cout <<"adapting -2pi ";
      MeanDeltaRAn[r]-=2.0*M_PI;
    }
    while (MeanDeltaRAn[r]<0.0) {
      cout <<"adapting +2pi ";
      MeanDeltaRAn[r] += 2.0*M_PI;
    }
    Calibration_[r] *= normalizationfactor;
    MeanDeltaRAx[r].Normalize();    
    Rotation_[r].Set(MeanDeltaRAx[r],MeanDeltaRAn[r]);
  }
  
  return true;  
}

#ifdef BIAS_HAVE_XML2
int RigParameter::WriteRigData(const std::string &filename) {

  XMLIO myXML;
  xmlNodePtr rootNode, childNode1, childNode2, childNode3;
  rootNode = myXML.create("rig");
  myXML.addAttribute(rootNode, "version", std::string("1.0") );
  for (unsigned int i=0; i<Camera_.size(); i++) {
    childNode1 = myXML.addChildNode(rootNode, "cameraDescription");
    Camera_[i].AddCameraDataToRig(myXML, childNode1);
    childNode2 = myXML.addChildNode(childNode1, "poseInRig");
    double yaw, pitch, roll;
    myXML.addComment(childNode2, "Yaw, Pitch and Roll are the Rotation_ angles of the camera around the rigs X, Y and Z Axes relative to the rig Rotation_. The Rotation_s are applied in the order XYZ or YPR.");
    Rotation_[i].GetRotationAnglesXYZ(pitch, yaw, roll);
    childNode3 = myXML.addChildNode(childNode2, "x");
    myXML.addContent(childNode3, Center_[i][0]);
    childNode3 = myXML.addChildNode(childNode2, "y");
    myXML.addContent(childNode3, Center_[i][1]);
    childNode3 = myXML.addChildNode(childNode2, "z");
    myXML.addContent(childNode3, Center_[i][2]);
    myXML.addComment(childNode2,"Rotation_ direction: from rig reference system to camera frame.");
    childNode3 = myXML.addChildNode(childNode2, "yaw");
    myXML.addContent(childNode3, yaw);
    childNode3 = myXML.addChildNode(childNode2, "pitch");
    myXML.addContent(childNode3, pitch);
    childNode3 = myXML.addChildNode(childNode2, "roll");
    myXML.addContent(childNode3, roll);
  }
  for (unsigned int i=0; i<SensorRotation_.size(); i++) {
    childNode1 = myXML.addChildNode(rootNode, "sensorDescription");
    childNode2 = myXML.addChildNode(childNode1, "poseInRig");
    double yaw, pitch, roll;
    SensorRotation_[i].GetRotationAnglesXYZ(pitch, yaw, roll);
    childNode3 = myXML.addChildNode(childNode2, "x");
    myXML.addContent(childNode3, SensorCenter_[i][0]);
    childNode3 = myXML.addChildNode(childNode2, "y");
    myXML.addContent(childNode3, SensorCenter_[i][1]);
    childNode3 = myXML.addChildNode(childNode2, "z");
    myXML.addContent(childNode3, SensorCenter_[i][2]);
    myXML.addComment(childNode2,"Rotation_ direction: from rig reference system to sensor frame.");

    childNode3 = myXML.addChildNode(childNode2, "yaw");
    myXML.addContent(childNode3, yaw);
    childNode3 = myXML.addChildNode(childNode2, "pitch");
    myXML.addContent(childNode3, pitch);
    childNode3 = myXML.addChildNode(childNode2, "roll");
    myXML.addContent(childNode3, roll);
  }
  myXML.write(filename); 
  return 0;
}
#endif //BIAS_HAVE_XML2


#ifdef BIAS_HAVE_XML2
int RigParameter::ReadRigData(const std::string &filename) {

  XMLIO myXML;
  xmlNodePtr rootNode, childNode1, childNode2;

  Camera_.clear();
  Calibration_.clear();
  Center_.clear();
  Rotation_.clear();

  rootNode = myXML.read(filename);
  if (rootNode==NULL) {
    BIASERR("Rig-Parameter-File not found or parse error: " << filename);
    return -1;
  }
 
  CameraParam cp;
  KMatrix km;
  Vector3<double> vp;
  RMatrix rm,srm;
  srm.SetIdentity();
  if (myXML.getNodeName(rootNode)!="rig" && 
      myXML.getNodeName(rootNode)!="Rig") {
    if (myXML.getNodeName(rootNode)!="camera" &&
        myXML.getNodeName(rootNode)!="Camera") {
      BIASERR("Root node has to be \"rig\" or \"camera\"");
      return -1;
    } else {
      //Root Node is "camera"
      int res = cp.Init(myXML, rootNode);
      if (res<0) return res;
      
      if (cp.IsSpherical()) {
        //cout <<"Spherical camera !"<<endl;
        SphericalUndistortion su(cp);
        km = su.GetKMatrix();
        //km.SetIdentity();
      } else {
        //cout <<"Perspective camera !"<<endl;
        km = cp.GetKMatrix();
      }

      vp[0]=0;
      vp[1]=0;
      vp[2]=0;
      rm.SetIdentity();
      srm.SetIdentity();
      Camera_.push_back(cp);
      Calibration_.push_back(km);
      Center_.push_back(vp);
      Rotation_.push_back(rm);
      // this is not correct here, we dont have a sensor for each camera !!!
      SensorRotation_.push_back(srm);
      SensorCenter_.push_back(vp);
    }
  } else {

    // Root node is "Rig"
    double vers=0;
    vers = myXML.getAttributeValueDouble(rootNode, "version");
    if (vers==0) {
      vers = myXML.getAttributeValueDouble(rootNode, "Version");
    }
    if (vers!=RIG_PARAMETER_CURRENT_XML_VERSION) {
      BIASWARN("Parameter File Version (" << vers << ")is Deprecated, "
              << "current version is " << RIG_PARAMETER_CURRENT_XML_VERSION);
      if (vers==0.8) {
        ReadRigDataV0_8(filename);
        return 0;
      } else {
        return -1;
      } 
    }

    childNode1 = myXML.getFirstChild(rootNode);
    if (childNode1 == NULL) {
      BIASERR("No CameraDescriptions in Rig, empty Rig???");
      return -1;        
    }
       
    do {  
      if (myXML.getNodeName(childNode1)=="cameraDescription") {
        childNode2 = myXML.getChild(childNode1, "camera");
        if (childNode2 == NULL) {
          BIASERR("\"CameraDescription\" without a \"Camera\" is not possible.");
          return -1;        
        }  else {
          int res = cp.Init(myXML, childNode2);
          if (res<0) return res;
          if (cp.IsSpherical()) {
            SphericalUndistortion su(cp);
            km = su.GetKMatrix();
            //km.SetIdentity();
          } else {
            km = cp.GetKMatrix();
          }
        }
        childNode2 = myXML.getChild(childNode1, "poseInRig");
        if (childNode2 == NULL) {
          BIASERR("\"cameraDescription\" without a \"poseInRig\" is not possible.");
          return -1;        
        }  else {
          double x, y, z, yaw, pitch, roll;
          x =   myXML.getNodeContentDouble(myXML.getChild(childNode2, "x"));
          y =   myXML.getNodeContentDouble(myXML.getChild(childNode2, "y"));
          z =   myXML.getNodeContentDouble(myXML.getChild(childNode2, "z"));
          yaw = myXML.getNodeContentDouble(myXML.getChild(childNode2,"yaw"));
          pitch=myXML.getNodeContentDouble(myXML.getChild(childNode2,"pitch"));
          roll =myXML.getNodeContentDouble(myXML.getChild(childNode2,"roll"));
          vp[0]=x;
          vp[1]=y;
          vp[2]=z;
          rm.SetXYZ(pitch, yaw, roll);
        }
        
        Camera_.push_back(cp);
        Calibration_.push_back(km);
        Center_.push_back(vp);
        Rotation_.push_back(rm);
      }
      if (myXML.getNodeName(childNode1)=="sensorDescription") {
        childNode2 = myXML.getChild(childNode1, "poseInRig");
        if (childNode2 == NULL) {
          BIASERR("\"sensorDescription\" without a \"poseInRig\" is not possible.");
          return -1;        
        }  else {
          double x, y, z, yaw, pitch, roll;
          x =   myXML.getNodeContentDouble(myXML.getChild(childNode2, "x"));
          y =   myXML.getNodeContentDouble(myXML.getChild(childNode2, "y"));
          z =   myXML.getNodeContentDouble(myXML.getChild(childNode2, "z"));
          yaw = myXML.getNodeContentDouble(myXML.getChild(childNode2,"yaw"));
          pitch=myXML.getNodeContentDouble(myXML.getChild(childNode2,"pitch"));
          roll =myXML.getNodeContentDouble(myXML.getChild(childNode2,"roll"));
          vp[0]=x;
          vp[1]=y;
          vp[2]=z;
          rm.SetXYZ(pitch, yaw, roll);

        }
        
        SensorCenter_.push_back(vp);
        SensorRotation_.push_back(rm);
      }
      childNode1 = myXML.getNextChild(childNode1); 
    } while (childNode1!=NULL);
  }      

  if (Camera_.size()==0) {
    return -1;
  }
  return 0;
}


int RigParameter::ReadRigDataV0_8(const std::string &filename) {

  XMLIO myXML;
  xmlNodePtr rootNode, childNode1, childNode2;

  Camera_.clear();
  Calibration_.clear();
  Center_.clear();
  Rotation_.clear();

  rootNode = myXML.read(filename);
  if (rootNode==NULL) {
    BIASERR("Rig-Parameter-File not found or parse error: " << filename);
    return -1;
  }
 
  CameraParam cp;
  KMatrix km;
  Vector3<double> vp;
  RMatrix rm,srm;
  srm.SetIdentity();

  // Root node is "Rig"
  childNode1 = myXML.getFirstChild(rootNode);
  if (childNode1 == NULL) {
    BIASERR("No CameraDescriptions in Rig, empty Rig???");
    return -1;        
  }
  
  do {  
    if (myXML.getNodeName(childNode1)=="CameraDescription") {
      childNode2 = myXML.getChild(childNode1, "Camera");
      if (childNode2 == NULL) {
        BIASERR("\"CameraDescription\" without a \"Camera\" is not possible.");
        return -1;        
      }  else {
        int res = cp.Init(myXML, childNode2);
	if (res<0) return res;
        if (cp.IsSpherical()) {
          SphericalUndistortion su(cp);
          km = su.GetKMatrix();
          //km.SetIdentity();
        } else {
          km = cp.GetKMatrix();
        }
      }
      childNode2 = myXML.getChild(childNode1, "PoseInRig");
      if (childNode2 == NULL) {
        BIASERR("\"CameraDescription\" without a \"PoseInRig\" is not possible.");
        return -1;        
      }  else {
        double x, y, z, yaw, pitch, roll;
        x =     myXML.getAttributeValueDouble(childNode2, "X");
        y =     myXML.getAttributeValueDouble(childNode2, "Y");
        z =     myXML.getAttributeValueDouble(childNode2, "Z");
        yaw =   myXML.getAttributeValueDouble(childNode2, "Yaw");
        pitch = myXML.getAttributeValueDouble(childNode2, "Pitch");
        roll =  myXML.getAttributeValueDouble(childNode2, "Roll");
        vp[0]=x;
        vp[1]=y;
        vp[2]=z;
        rm.SetXYZ(pitch, yaw, roll);
      }
      
      Camera_.push_back(cp);
      Calibration_.push_back(km);
      Center_.push_back(vp);
      Rotation_.push_back(rm);
    }
    childNode1 = myXML.getNextChild(childNode1); 
  } while (childNode1!=NULL);    

  if (Camera_.size()==0) {
    return -1;
  }
  return 0;
}
#endif //BIAS_HAVE_XML2

void RigParameter::SetData_DummyRig() {
  
  CameraParam cp;
  KMatrix km;
  Vector3<double> vp;
  RMatrix rm;
  
  cp.SetData_EVID31_147094();
  km = cp.GetKMatrix();
  vp.Set(-358.02,5.72,4.04);
  rm[0][0]=0.999918312;
  rm[0][1]=0.01265;
  rm[0][2]=0.0018;
  rm[1][0]=-0.01264;
  rm[1][1]=0.9998977;
  rm[1][2]=0.00569;
  rm[2][0]=0.00188;
  rm[2][1]=-0.00566;
  rm[2][2]=0.999976;
  Camera_.push_back(cp);
  Calibration_.push_back(km);
  Center_.push_back(vp);
  Rotation_.push_back(rm);
 
  cp.SetData_OMNITECH_ORIFL_190_3();
  SphericalUndistortion su(cp);
  km = su.GetKMatrix();
  vp.Set(9.87,8.76,7.65);
  rm.SetXYZ(0.22, 0.67, -0.005);
  Camera_.push_back(cp);
  Calibration_.push_back(km);
  Center_.push_back(vp);
  Rotation_.push_back(rm);
  
}

  
std::ostream& BIAS::operator<<(std::ostream &os, RigParameter &r) {
  std::vector<BIAS::RMatrix> rMat;
  r.GetRotations(rMat);
  std::vector<BIAS::Vector3<double> > cVec;
  r.GetCameraCenters(cVec);
  std::vector< BIAS::KMatrix > kMat;
  r.GetCalibrations(kMat);
  std::vector<CameraParam> vCParam;
  r.GetCameraParams(vCParam);
  
  for (unsigned int i=0; i<cVec.size(); i++) {
    os << "C["<<i<<"]="<<cVec[i]<<std::endl;
  }
  BIAS::Vector3<double> axis;
  double angle;
  for (unsigned int i=0; i<rMat.size(); i++) {
    rMat[i].GetRotationAxisAngle(axis, angle);
    os << "R["<<i<<"]: Axis="<<axis<<" angle="<<angle<<std::endl;
    // os << "R["<<i<<"]="<<r.Rotation__[i]<<std::endl;
  }
  for (unsigned int i=0; i<kMat.size(); i++) {
    os << "K["<<i<<"]="<<kMat[i]<<std::endl;
  }
  for (unsigned int i=0; i<kMat.size(); i++) {
    os << "Camera["<<i<<"]="<<std::endl;
    vCParam[i].DebugOut();
  }

  return os;
}


void RigParameter::GetKMatrices(std::vector<BIAS::KMatrix>& vKmatrix){

  vKmatrix.clear();
  BIAS::KMatrix km;
  for(unsigned int i=0;i<Camera_.size();i++){
	if(i<=Calibration_.size()){
	  km = Calibration_[i];
	
      if(DownSamplingFactor_<=0) DownSamplingFactor_=1.0;
      km[0][0]/=DownSamplingFactor_;
      km[1][1]/=DownSamplingFactor_;
  
      km[0][2]/=DownSamplingFactor_;
      km[1][2]/=DownSamplingFactor_;

      //now fill matrix vector with downsampled values
      vKmatrix.push_back(km);
	 }
  }
}

bool RigParameter::GetKMatrix(BIAS::KMatrix& matrix,int index) const
{
  matrix.set_zero();
  if((unsigned int)index>Calibration_.size())
    return false;
  else{    
    matrix = Calibration_[index];
    matrix[0][0]/=DownSamplingFactor_;
    matrix[1][1]/=DownSamplingFactor_;
    matrix[0][2]/=DownSamplingFactor_;
    matrix[1][2]/=DownSamplingFactor_;
    return true;
  } 
}

void RigParameter::SetDownsampling(const double factor)
{
  DownSamplingFactor_= factor;
  for(unsigned int i=0;i<Camera_.size();i++)
    Camera_[i].SetDownsampling(factor);
}
