/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/** @example ExampleTracker2
    This little example demonstrates the 


    @ingroup g_tracker
    @author woelk */

#include <bias_config.h>
#include <Base/Common/BIASpragma.hh>


#include <Matcher2D/Tracker.hh>
#include <Base/Image/ImageDraw.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Image/ImageIO.hh>
#include <FeatureDetector/CornerDetectorKLT.hh>
#include <FeatureDetector/CornerDetectorHarris.hh>
#include <FeatureDetector/CornerDetectorFoerstner.hh>
#include <Base/Debug/TimeMeasure.hh>

#include <deque>

using namespace BIAS;
using namespace std;

#define DEFAULT_PARA_FILE_NAME "Tracker.ini"

/** The preprocessor define type is used to choose between the image type
    StorageType is the internal type used to store the images data.
    CalculationType is the data type used for calculations.

    type 0: Images are of type unsigned char (i.e. byte)
            The internal calculation (e.g. gradinets) is done using int.
           
    type 1: Images are of type float (the can be of any type on disk, they are
            immediatly casted to float after loading them from disk).
            The internal calculation (e.g. gradinets) is done using float.
*/
#define type 1
#if type == 0
#define StorageType unsigned char
#define CalculationType int
#elif type == 1
#define StorageType float
#define CalculationType float
#endif

#define DEBUG false

/* The following variables should usually be class member variables.
   They are choosen to be global vriables in order to keep the example simple.
*/

// the corner detector class
CornerDetectorBase<StorageType> *cd = NULL;
// the tracker class
Tracker<StorageType, CalculationType> *tr = NULL;

// the pyramid images used by the tracker
PyramidImage<CalculationType> *pim[2]; // low pass filtered images
PyramidImage<CalculationType> *gx[2]; // horizontal gradient
PyramidImage<CalculationType> *gy[2]; // vertical gradient

int act=0; // actual index into the pyramid image arrays
int *cdt; // enum indicating which corner detector class to use
int *cdfreq; // corner detector is run at every cdfreq-th frame
int *pyind; //

///  the resulting data structs
deque<vector<HomgPoint2D> > p; //  the corners
deque<vector<KLT_TYPE> > residui; // mean absolute grey difference 
/** int representing success (=0) or failure reason (<0) see 
    TrackerBase::Track() for details */
deque<vector<int> > results; 

/// temporary vector for the 
vector<HomgPoint2D> mp;
vector<QUAL> qual; /// the displacement in the last iteration step
vector<HomgPoint2D> cp;
vector<KLT_TYPE> r; //
vector<int> res;

// how often has the tracker/corner detctro been run ?
int tnum=0, cdnum=0;

// time measure classes
TimeMeasure otimer; // overall timer
TimeMeasure ttimer; // timer used for tracking
TimeMeasure cdtimer; // timer for corner detection
// timer used for calculating the pyramid images and filtering
TimeMeasure pytimer; 

// should the tracks be drawn as trailing lines ?
bool draw_tracks=true;

/** draw the tracks with a circle around the actual corner position */
void Draw(Image<StorageType>& stim, Image<unsigned char>& im, 
          deque<vector<HomgPoint2D> >& p, int imnum)
{
#if type == 0
  ImageConvert::Convert(stim, im, ImageBase::CM_RGB);
#elif type == 1
  Image<unsigned char> ucim;
  ImageConvert::ConvertST(stim, ucim, ImageBase::ST_unsignedchar);
  ImageConvert::Convert(ucim, im, ImageBase::CM_RGB);
#endif

  unsigned char col[]={255, 0, 0};
  const int nump=p[0].size();
  unsigned start[2], end[2];
  const unsigned radius=1;
  int maxhist=0;

  if (imnum>0){
    for (int l=0; l<nump; l++){
      if (!p[imnum][l].IsAtInfinity() && !p[imnum-1][l].IsAtInfinity()){
        // circel around corner in actual image
        start[0]=(unsigned)rint(p[imnum][l][0]);
        start[1]=(unsigned)rint(p[imnum][l][1]);
        ImageDraw<unsigned char>::CircleCenterFilled(im, start[0], start[1], 
                                                     radius, col);
        if (draw_tracks){
          // lines for tracks
          for (int i=imnum; i>0; i--){
            if (p[i-1][l].IsAtInfinity()) {
              if ((imnum-i)>maxhist) maxhist=imnum-i;
              break;
            }
            start[0]=(unsigned)rint(p[i][l][0]);
            start[1]=(unsigned)rint(p[i][l][1]);
            end[0]=(unsigned)rint(p[i-1][l][0]);
            end[1]=(unsigned)rint(p[i-1][l][1]);
            if (i-1==0) maxhist=imnum;
            ImageDraw<unsigned char>::Line(im, start, end, col);
          }
        }
      } 
    }
  }
}


/** print tracks to stdout, just for debugging purposes */
void DumpTracks(deque<vector<HomgPoint2D> >& p)
{
  int numim=0;
  while (p[numim].size()>0 && numim<(int)p.size()) numim++;
  for (unsigned l=0; l<p[0].size(); l++){
    cout <<setw(3)<<l<<" : ";
    for (int i=0; i<numim; i++){
      cout.precision(4);
      fprintf(stdout, "%i:(%5.1f, %5.1f) ", i, p[i][l][0], p[i][l][1]);
    }
    cout << endl;
  }
}




/** the drawing algorithm needs one point at infinity between subsequent 
    tracks, therefore the last point of the last track is set to (0, 0, 0) */
void InsertZeroPoints(deque<vector<HomgPoint2D> >& p, int i)
{
  if (i>0){
    const int nump=p[i].size();
   
    for (int k=0; k<nump; k++){
      if (p[i][k].IsAtInfinity() && !p[i-1][k].IsAtInfinity()){
        p[i-1][k].Set(0.0, 0.0, 0.0);
      }
    }
  }
}

/** 
 */
void Track(const Image<CalculationType>& im, Image<unsigned char>& rgbim, int i)
{
  otimer.Start();
  TimeMeasure mcdtimer;

  /// the first thing that needs to be done is to calculate the pyramid images
  pytimer.Start();
  tr->PreparePyramide(im, *pim[act], *gx[act], *gy[act]);
  pytimer.Stop();

  /// secondly the tracking process is started
  ttimer.Start();
  tr->Track(*pim[act], *gx[act], *gy[act]);
  ttimer.Stop();
  tnum++;

  // now initially tell the tracker where to track
  if (i==0){
    cdtimer.Start();
    // the gradient based corner detector differ slightly to the Susan
    // corner detector regarding the interface
    CornerDetectorGradient<StorageType, CalculationType> *cdgrad
      = dynamic_cast<CornerDetectorGradient<StorageType, CalculationType> *>(cd);
    if (cdgrad){
      if (DEBUG) cout << "using gradients for corner detector\n\n";
      cdgrad->Detect(*(*gx[act])[*pyind], *(*gy[act])[*pyind], mp, qual);
    } else {
      if (DEBUG) cout << "using image for corner detector\n";
      cd->Detect(*(*pim[act])[*pyind], mp, qual);
    } 
    if (DEBUG) cout << "initially found "<<mp.size()<<" points\n";

    // let the tracker know of the point positions
    tr->AddNumPoints(mp, qual);
    cdtimer.Stop();
    cdnum++;
  }

  // remember the corners in the global data struct
  tr->GetPoints(mp);
  p.push_back(mp);
  if (DEBUG) cout << "added "<<p[0].size()<<" points\n";
    
  // now draw the tracks into rgbim
  Draw(*(*pim[act])[*pyind], rgbim, p, i);
    

  // at every cdfreq-th frame the lost tracks indicated by points at infinity
  // (i.e. p[2]==0) are filled up again
  if (i!=0 && (i % *cdfreq)==0){
    // run the corner detector
    cdtimer.Start();
    // the gradient based corner detector differ slightly to the Susan
    // corner detector regarding the interface
    CornerDetectorGradient<StorageType, CalculationType> *cdgrad
      =dynamic_cast<CornerDetectorGradient<StorageType,CalculationType> *>(cd);
    if (cdgrad){
      if (DEBUG) cout << "using gradients for corner detector\n\n";
      cdgrad->Detect(*(*gx[act])[*pyind], *(*gy[act])[*pyind], cp, qual);
    } else {
      if (DEBUG) cout << "using image for corner detector\n";
      cd->Detect(*(*pim[act])[*pyind], cp, qual);
    } 

    // the track length in the globals data structs are determiend
    // by 'zero points'
    InsertZeroPoints(p, i);

    if (false) DumpTracks(p);
    
    // now only the points where tracking failed are filled up
    int num=tr->FillUpPoints(cp, qual);
    if (DEBUG) cout << "filled up "<<num<<" points\n";
    // get the newly filled up points and remeber them in the global deque
    tr->GetPoints(mp);
    // restrict the maximum track length in the global data struct (optional)
    p.pop_back();
    p.push_back(mp);
    
    cdtimer.Stop();
    if (false) DumpTracks(p);
    cdnum++;
  }
  // now remember the residui (mean grey value difference)
  tr->GetResiduiMAD(r);
  residui.push_back(r);
  // and the success indicator
  tr->GetResults(res);
  results.push_back(res);
  otimer.Stop();
}

/// mark unnecessairy parameter as enhanced
void MarkEnhancedAndSetDefaults(Param& para)
{
  para.SetEnhancedFlag("MaxError-TrackerBase-Tracker", true);
  para.SetEnhancedFlag("HalfWinSize-TrackerBase-Tracker", true);
  para.SetEnhancedFlag("MaxIterations-TrackerBase-Tracker", true);
  para.SetEnhancedFlag("MaxResiduum-TrackerBase-Tracker", true);
  para.SetEnhancedFlag("PyIndex-Tracker", true);

  para.SetEnhancedFlag("MaxNumCorners-CornerDetectorBase", true);
  para.SetEnhancedFlag("SubpixelAccuracy-CornerDetectorGradient", true);
  para.SetEnhancedFlag("TrackerType-Tracker", true);
//   para.SetEnhancedFlag("", true);
//   para.SetEnhancedFlag("", true);
//   para.SetEnhancedFlag("", true);
//   para.SetEnhancedFlag("", true);
  //para.SetEnhancedFlag("", true);

  para.SetParamDouble("Factor-Rescale-PyramidImage", 4.0);
  para.SetParamInt("PyramidSize-PyramidImage", 2);
  para.SetParamInt("HalfWinSize-StructureTensor-CornerDetectorGradient", 2);
  para.SetParamInt("MinDistance-CornerDetectorGradient", 5);
//   para.SetParamEnum("LowPassType-Rescale-PyramidImage", LowPassBase<float, float>::LPT_Mean);
//   para.SetParamEnum("GradientType-StructureTensor-CornerDetectorGradient", GradientBase<float, float>::GT_Sobel3x3);
}

/// the main function
int main(int argc, char *argv[])
{
  TimeMeasure timer;
  timer.Start();
  Param para;

  Image<unsigned char> ucim, rgbim;
  Image<StorageType> im, im2;
  int grp=para.GetFreeGroupID();
  tr = new Tracker<StorageType, CalculationType>(para, grp, "");
  para.SetGroupName(grp, "Tracker");

  grp=para.GetFreeGroupID();
  for (int i=0; i<2; i++){
    pim[i] = new PyramidImage<CalculationType>(para, grp, "");
    gx[i] = new PyramidImage<CalculationType>(para, grp, "");
    gy[i] = new PyramidImage<CalculationType>(para, grp, "");
  }
  para.SetGroupName(grp, "PyramidImage");


  // allocate the differnet corner detectors
  grp=para.GetFreeGroupID();
  CornerDetectorKLT<StorageType, CalculationType> cdklt(para, grp, "");
  CornerDetectorHarris<StorageType, CalculationType> cdharris(para, grp, "");
  CornerDetectorFoerstner<StorageType, CalculationType> cdfoerstner(para, grp, "");
  cdt=CornerDetectorBase<StorageType>::
    AddCornerDetectorType(para, "CornerDetectorType", grp);
  // parameter for the frequency of the corner detector
  cdfreq=para.AddParamInt("CornerDetectorFreq", "corner detector fills up points every n-th frame", 5, 0, INT_MAX, 0, grp);
  para.SetGroupName(grp, "CornerDetector");
                   
  //
  MarkEnhancedAndSetDefaults(para);
  /** read parameter from command line and DEFAULT_PARA_FILE_NAME */
  int argind = para.UpdateParameter(argc, argv, DEFAULT_PARA_FILE_NAME);
  pyind = para.GetParamInt("PyIndex-Tracker");

  // use the corner detector choosen by the parameter CornerDetectorType
  switch (*cdt){
  case CornerDetectorBase<StorageType>::CDT_Susan:
#if type == 0
    cd = &cds;
#else
    BIASERR("susan only implemented for unsigned char");
    cd = NULL;
#endif
    if (DEBUG) cout << "using susan\n";
    break;
  case CornerDetectorBase<StorageType>::CDT_KLT:
    cd = &cdklt;
    if (DEBUG) cout << "using klt\n";
    break;
  case CornerDetectorBase<StorageType>::CDT_Harris:
    cd = &cdharris;
    if (DEBUG) cout << "using harris\n";
    break;
  case CornerDetectorBase<StorageType>::CDT_Foerstner:
    cd = &cdfoerstner;
    if (DEBUG) cout << "using foerstner\n";
    break;
  default:
    BIASERR("unknown corner detector type");
    cd = NULL;
    break;
  }
  cd->AddDebugLevel(D_CD_WRITE_DEBUG_IM);
  //cd->AddDebugLevel(D_CD_FEATURES);

  //tr->AddDebugLevel(D_TRACKER_INIT);
  //tr->AddDebugLevel(D_TRACKER_PARAM);
  //tr->AddDebugLevel(D_TRACKER_TRACK);
  tr->AddDebugLevel(D_TRACKER_WRITE_IM);
  //tr->AddDebugLevel(D_TRACKER_FILL_UP);

  // check if enough images were given on command line
  if (argc-argind<2 || argind<1){
    cerr << argv[0] << " [parameter] <im1> <im2> [ <im3> ... ] \n";
    para.Usage(cerr);
    para.DisableDestructorWarning();
    return -2;
  } 

  // tell the tracker to adopt to new parameter
  tr->ParameterChanged();

  // 
  Rescale<float, float> resc;

  // loop for every image
  for (int i=argind; i<argc; i++){
    timer.Stop();
    cout << "running at " << 1e6/(double)timer.GetRealTime() << " fps\r";
    timer.Reset();
    timer.Start();

    // act is the index into the actual 
    // toggle act between 0 and 1
    act=!act;
#if type == 0
    if (ImageIO::Load(argv[i], im)!=0){
#else
    if (ImageIO::Load(argv[i], ucim)!=0){
#endif
      BIASERR("error loading image "<<argv[i]);
      para.DisableDestructorWarning();
      return -1;
    } else {
      if (DEBUG) cerr << "read "<<argv[i]<<endl;
    }

#if type != 0
    if (ImageConvert::ConvertST(ucim, im, ImageBase::ST_float)!=0){
      BIASERR("error converting image "<<argv[i]);
    }
#endif

    // downsample the images read from disk by factor 2 (optional)
    if (i==argind) im2.Init(im.GetWidth()>>1, im.GetHeight()>>1, 1);
    resc.DownsampleBy2Grey(im, im2);
    //im2=im;

    // if a region of interest needs to be used, it should be set here
    //im.SetROI(0, 40, 320, 190);

    // now call the function that does the tracking
    Track(im2, rgbim, i-argind);

    // write result image to disk
    ostringstream name;
    name << "track-"<<setw(4)<<setfill('0')<<i-argind<<".mip";
    ImageIO::Write(name.str(), rgbim);
    
  }
    
  // write a Birchfeld style tracker file
  if (DEBUG) tr->WriteTrackFile(p, results, "features.txt");
    
  if (DEBUG) cout << "\nTracks: "<<endl;
  if (DEBUG) DumpTracks(p);
  if (DEBUG) cout << "\n";

  cout << "preparing pyramid images took :"
       <<pytimer.GetRealTime()/(float)tnum
       <<" us / image\n";
  
  cout << "tracking took :"<<ttimer.GetRealTime()/(float)tnum
       <<" us / image   user "<<ttimer.GetUserTime()/(float)tnum
       <<" ms /image \n";
  
  cout << "detecting corners took: "<<cdtimer.GetRealTime()/(float)cdnum
       <<" us / detection\n";

  cout << "overall tracking corners took: "<<otimer.GetRealTime()/(float)tnum
       <<" us / detection\n";

  para.DisableDestructorWarning();
  return 0;
}

