/* 
This file is part of the BIAS library (Basic ImageAlgorithmS),
however, it contains other contributions.
For these marked parts a different license may apply, please read
the documentation of the marked parts for further details. 
For the rest of this file read the following:

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "TrackerBase.hh"
#include <algorithm>
#include "MathAlgo/SVD.hh"

#include <Base/Common/BIASpragma.hh>

using namespace BIAS;
using namespace std;

//////////////////////////////////////////////////////////////////////////
//                 operators
//////////////////////////////////////////////////////////////////////////

ostream& BIAS::operator<<(std::ostream& os, const enum ETrackerType& tt)
{
  switch(tt){
  case TT_Invalid:
    os << "invalid";
    break;
   case TT_Simple:
    os << "TT_Simple";
    break; 
   case TT_Weighted:
    os << "TT_Weighted";
    break; 
   case TT_Affine4:
    os << "TT_Affine4";
    break; 
   case TT_Affine6:
    os << "TT_Affine6";
    break; 
   case TT_SimpleAffine4:
    os << "TT_SimpleAffine4";
    break; 
   case TT_SimpleAffine6:
    os << "TT_SimpleAffine6";
    break; 
  default:
    os << "unkown tracker type "<<(int)tt;
    break;
  }
  return os;
}


ostream& BIAS::operator<<(ostream& os, const enum ERejectionType& rt)
{
  switch (rt){
  case RT_Invalid:
    os << "invalid";
    break;
   case RT_MAD:
    os << "RT_MAD";
    break; 
   case RT_X86:
    os << "RT_X86";
    break; 
  default:
    os << "unkown rejection type "<<(int)rt;
    break;
  }
  return os;
}

//////////////////////////////////////////////////////////////////////////
//                 implementation
//////////////////////////////////////////////////////////////////////////

#define SMALL_DET 1e-8
#define DEFAULT_MAX_AFFINECHANGE 0.001


template <class StorageType>
TrackerBase<StorageType>::
TrackerBase() 
  : Debug()
{
  InitMembers_();
  _ParamsAllocated = true;

  _HalfWinSize = new int;
  *_HalfWinSize = 3;
  _MaxIterations = new int;
  *_MaxIterations = 10;
  _MaxError = new double;
  *_MaxError = 0.01;
  _MaxAffineChange = new double;
  *_MaxAffineChange = DEFAULT_MAX_AFFINECHANGE;

  _MaxResiduumMAD = new double;
  *_MaxResiduumMAD = 10.0;
  _TrackerType = new int;
  *_TrackerType = TT_Simple;
  _RejectionType = new int;
  *_RejectionType = RT_MAD;

  _LastHalfWinSize=_WinSize=0;
}

template <class StorageType>
TrackerBase<StorageType>::
TrackerBase(Param& para, const std::string& postfix) 
  : Debug()
{
  InitMembers_();
  int grp = para.GetFreeGroupID();
  AddParam_(para, grp, "-TrackerBase"+postfix);
  if (para.IsUsedGroupID(grp)) para.SetGroupName(grp, "TrackerBase");
  _ParamsAllocated = false;
  
  _LastHalfWinSize = _WinSize = 0;
}

template <class StorageType>
TrackerBase<StorageType>::
TrackerBase(Param& para, const int grp, const std::string& postfix) 
  : Debug()
{
  InitMembers_();
  AddParam_(para, grp, "-TrackerBase"+postfix);
  _ParamsAllocated = false;
  _LastHalfWinSize = _WinSize = 0;
}

template <class StorageType>
TrackerBase<StorageType>::
~TrackerBase()
{
  if (_ParamsAllocated) {
    delete _HalfWinSize;
    delete _MaxIterations;
    delete _MaxError;
    delete _MaxResiduumMAD;
    delete _TrackerType;
    delete _RejectionType;
  }
}

template <class StorageType>
void TrackerBase<StorageType>::
InitMembers_() {
  _im1 = NULL;
  _im2 = NULL;
  _gradim1x = NULL;
  _gradim2x = NULL; 
  _gradim1y = NULL;
  _gradim2y = NULL;
  _HalfWinSize = _MaxIterations  = NULL; 
  _MaxError    = _MaxResiduumMAD = NULL;
  _TrackerType = _RejectionType  = NULL;
  _maxx1 = _maxy1 = _minx1 = _miny1 = 0;
  _maxx2 = _maxy2 = _minx2 = _miny2 = 0;
  _AffineResults.resize(1);
  _AffineResults[0].SetIdentity();
  _MaxAffineChange = NULL;
}

template <class StorageType>
void TrackerBase<StorageType>::
Init_(StorageType **ida1, StorageType **ida2, StorageType **gradx1, 
      StorageType **grady1, StorageType **gradx2, StorageType **grady2, 
      unsigned minx1, unsigned miny1, unsigned maxx1, unsigned maxy1,
      unsigned minx2, unsigned miny2, unsigned maxx2, unsigned maxy2)
{
  bool resize=(*this->_HalfWinSize != this->_LastHalfWinSize);
  if (resize){
    Resize_(*this->_HalfWinSize);
  }

  _minx1=minx1+*_HalfWinSize+1;
  _miny1=miny1+*_HalfWinSize+1;
  _maxx1=maxx1-*_HalfWinSize-1;
  _maxy1=maxy1-*_HalfWinSize-1;
  _minx2=minx2+*_HalfWinSize+1;
  _miny2=miny2+*_HalfWinSize+1;
  _maxx2=maxx2-*_HalfWinSize-1;
  _maxy2=maxy2-*_HalfWinSize-1;

  BIASCDOUT(D_TRACKERB_INIT, "tracking roi im1 : ( "<<_minx1<<", "<<_miny1
            <<")  <-->  ( "<<_maxx1<<", "<<_maxy1<<")\n");
  BIASCDOUT(D_TRACKERB_INIT, "tracking roi im2 : ( "<<_minx2<<", "<<_miny2
            <<")  <-->  ( "<<_maxx2<<", "<<_maxy2<<")\n");

  _ida1=ida1;
  _ida2=ida2;
  _gradx1=gradx1;
  _gradx2=gradx2;
  _grady1=grady1;
  _grady2=grady2;
  BIASASSERT(*_HalfWinSize>0);
  BIASASSERT(_maxy1>_miny1 && _maxx1>_minx1);
  BIASASSERT(_maxy2>_miny2 && _maxx2>_minx2);

  if (resize && *_TrackerType==TT_Weighted){
    CalcGaussianWeightMatrix_();
  }
}

template <class StorageType>
void TrackerBase<StorageType>::
Init(Image<StorageType>& im1, Image<StorageType>& im2, 
     Image<StorageType>& gradx1, Image<StorageType>& grady1, 
     Image<StorageType>& gradx2, Image<StorageType>& grady2)
{
  unsigned ytlx, ytly, ybrx, ybry;
  grady1.GetROI()->GetCorners(ytlx, ytly, ybrx, ybry);
  unsigned xtlx, xtly, xbrx, xbry;
  gradx1.GetROI()->GetCorners(xtlx, xtly, xbrx, xbry);
  unsigned minx1, miny1, maxx1, maxy1;
  minx1=(ytlx>xtlx)?(ytlx):(xtlx);
  miny1=(ytly>xtly)?(ytly):(xtly);
  maxx1=(ybrx<xbrx)?(ybrx):(xbrx); 
  maxy1=(ybry<xbry)?(ybry):(xbry);

  grady2.GetROI()->GetCorners(ytlx, ytly, ybrx, ybry);
  gradx2.GetROI()->GetCorners(xtlx, xtly, xbrx, xbry);
  unsigned minx2, miny2, maxx2, maxy2;
  minx2=(ytlx>xtlx)?(ytlx):(xtlx);
  miny2=(ytly>xtly)?(ytly):(xtly);
  maxx2=(ybrx<xbrx)?(ybrx):(xbrx); 
  maxy2=(ybry<xbry)?(ybry):(xbry);

  BIASASSERT(im2.SamePixelAndChannelCount(im1));
  BIASASSERT(gradx1.SamePixelAndChannelCount(im1));
  BIASASSERT(grady1.SamePixelAndChannelCount(im1));
  BIASASSERT(gradx2.SamePixelAndChannelCount(im1));
  BIASASSERT(grady2.SamePixelAndChannelCount(im1));

  _im1 = &im1;
  _im2 = &im2;

  _gradim1x = &gradx1;
  _gradim2x = &gradx2; 
  _gradim1y = &grady1;
  _gradim2y = &grady2;

  Init_(im1.GetImageDataArray(), im2.GetImageDataArray(),
        gradx1.GetImageDataArray(), grady1.GetImageDataArray(),
        gradx2.GetImageDataArray(), grady2.GetImageDataArray(),
        minx1, miny1, maxx1, maxy1, minx2, miny2, maxx2, maxy2);
  
}


template <class StorageType>
int TrackerBase<StorageType>::
Track(HomgPoint2D& p1, HomgPoint2D& p2, HomgPoint2D& p2tracked, 
      KLT_TYPE& error, int &iter, KLT_TYPE& residuumMAD, KLT_TYPE& residuumMSD,
      Matrix<KLT_TYPE>& cov,
      const Matrix2x2<KLT_TYPE>& AffinePred) {
  int res=-20;
  if (p1[2]==1.0){
    bool FirstPassDone = false;
    switch(*_TrackerType){
    case TT_Weighted:
      res=TrackWeighted_(p1.GetData(), p2.GetData(), p2tracked.GetData(), 
                         error, iter);
      cov.newsize(2,2);
      EvaluateResult_(res, residuumMAD, residuumMSD, cov.GetData());
      break;
    case TT_Simple:
    case TT_SimpleAffine4:
    case TT_SimpleAffine6:

      res=TrackSimple_(p1.GetData(), p2.GetData(), p2tracked.GetData(), error, 
                       iter);
      cov.newsize(2,2);
      EvaluateResult_(res, residuumMAD, residuumMSD, cov.GetData());
      FirstPassDone = true;
      break;
    }
    bool Only4Parameters = false;
    // second pass of possible multipass, affine estimation
    switch(*_TrackerType){
    case TT_SimpleAffine4:
    case TT_Affine4: Only4Parameters = true;
    case TT_SimpleAffine6:
    case TT_Affine6: {
      if (_AffineResults.size()==0) _AffineResults.resize(1);
#ifdef BIAS_DEBUG
      if (AffinePred.NormFrobenius()<1e-10) {
        BIASERR("affine guess seems unitialized, set at least identity");
      }
#endif    
      if (FirstPassDone && res==TRACKER_SUCCESS)  p2 = p2tracked;  
      res =
        TrackAffine_(p1.GetData(), p2.GetData(), AffinePred,
                     p2tracked.GetData(), _AffineResults[0],
                     error, iter, cov, residuumMAD, residuumMSD, 
                     Only4Parameters);
      break;
    }

    default:
      if (!FirstPassDone) BIASERR("unknown tracker type "<<*_TrackerType);
      break;
    }
    
  } else {
    res = TRACKER_IGNOREPOINT;
  }

  return res;
}


template <class StorageType>
void TrackerBase<StorageType>::
Track(std::vector<HomgPoint2D>& p1, std::vector<HomgPoint2D>& p2, 
      std::vector<HomgPoint2D>& p2tracked, std::vector<KLT_TYPE>& error, 
      std::vector<int>& numiter, std::vector<KLT_TYPE>& residuiMAD, 
      std::vector<KLT_TYPE>& residuiMSD, std::vector<int>& res, 
      std::vector<Matrix<KLT_TYPE> >& cov,
      const std::vector<Matrix2x2<KLT_TYPE> >& AffinePred)
{
#ifdef TIMING
  TimeMeasure timer;
  timer.Start();
#endif
  const unsigned num=p1.size();
  BIASASSERT(num==p2.size() && num==p2tracked.size() && num==error.size());
  BIASASSERT(num==numiter.size() && num==res.size() && num==residuiMSD.size()&&
             num==residuiMAD.size() && num==cov.size());
 
  // firt pass of possible multi pass tracking
  bool FirstPassDone = false;
  switch(*_TrackerType){
  case TT_Weighted:
    for (unsigned i=0; i<num; i++){
      if (p1[i][2]==1.0){
#ifdef BIAS_DEBUG
        BIASASSERT(p2[i].IsHomogenized());
#endif    
        res[i]=
          TrackWeighted_(p1[i].GetData(), p2[i].GetData(), 
                         p2tracked[i].GetData(), error[i], numiter[i]);
        cov[i].newsize(2,2);
        EvaluateResult_(res[i], residuiMAD[i], residuiMSD[i],cov[i].GetData());
      } else {
        res[i] = TRACKER_IGNOREPOINT;
      }
    }
    break;
  case TT_Simple:
  case TT_SimpleAffine4:
  case TT_SimpleAffine6:
  
    for (unsigned i=0; i<num; i++){
      if (p1[i][2]==1.0){
#ifdef BIAS_DEBUG
        BIASASSERT(p2[i].IsHomogenized());
        BIASCDOUT(D_TRACKERB_INPUT, "attemting to track "<<setw(4)<<i<<"  "
                  <<p1[i]<<" -> "<<p2[i]);
#endif    
        res[i]=
          TrackSimple_(p1[i].GetData(), p2[i].GetData(),p2tracked[i].GetData(),
                       error[i], numiter[i]);
        cov[i].newsize(2,2);
        EvaluateResult_(res[i], residuiMAD[i], residuiMSD[i],cov[i].GetData());

      } else {
        res[i] = TRACKER_IGNOREPOINT;
      }
    }
    FirstPassDone = true;
    break;
  }
  bool Only4Parameters = false;
  // second pass of possible multipass, affine estimation
  switch(*_TrackerType){
  case TT_SimpleAffine4:
  case TT_Affine4: Only4Parameters = true;
  case TT_SimpleAffine6:
  case TT_Affine6:
                      
    BIASASSERT(AffinePred.size()==num);
    if (_AffineResults.size()!=num) {
      _AffineResults.resize(num);
    }
        
    for (unsigned i=0; i<num; i++){
      if (p1[i][2]==1.0) {
        // exploit result of first part
        if (FirstPassDone && res[i]==TRACKER_SUCCESS)  p2[i] = p2tracked[i];
#ifdef BIAS_DEBUG
        BIASASSERT(p2[i].IsHomogenized());
        BIASCDOUT(D_TRACKERB_INPUT, "attemting to aff-track "<<setw(4)<<i<<"  "
                  <<p1[i]<<" -> "<<p2[i]);
#endif    
        res[i]=
          TrackAffine_(p1[i].GetData(), p2[i].GetData(), AffinePred[i],
                       p2tracked[i].GetData(), _AffineResults[i],
                       error[i], numiter[i], cov[i],
                       residuiMAD[i], residuiMSD[i], Only4Parameters);
        //cout<<_AffineResults[i]<<endl;
      } else {
        res[i] = TRACKER_IGNOREPOINT;
      }
    }
    break;
  
  default:
    if (!FirstPassDone) {
      BIASERR("unknown tracker type "<<*_TrackerType);
    }
    break;
  }


  switch(*_RejectionType){
  case RT_MAD:
    RejectMAD_(residuiMAD, res);
    break;
  case RT_X86:
    RejectX86_(residuiMSD,  res);
    break;
  default:
    BIASERR("unknown rejection method "<<*_RejectionType);
    break;
  }
    
#ifdef TIMING
  timer.Stop();
  cerr << "TrackerBase tracking took "<<timer.GetRealTime()<<" us\n";
#endif
}


///////////////////////////////////////////////////////////////////
//  tracking
///////////////////////////////////////////////////////////////////

template <class StorageType>
int TrackerBase<StorageType>::
TrackSimple_(KLT_TYPE p1[2], KLT_TYPE p2[2], KLT_TYPE result[2], 
             KLT_TYPE& error, int &iter)
{
#ifdef TIMING
  TimeMeasure timer;
  timer.Start();
#endif
  if (*this->_HalfWinSize != _LastHalfWinSize){
    Resize_(*this->_HalfWinSize);
  }
  BIASCDOUT(D_TRACKERB_KLT,"trying to track p1: (" << p1[0] << ", " << p1[1] 
            << ") <--> p2: (" <<p2[0] << ", " << p2[1]<<")\n");

  int res=0;
  int rh, ch; 
  register KLT_TYPE dI, ex, ey, dx, dy, det;
  KLT_TYPE lsterror=FLT_MAX;
  const int hws=*this->_HalfWinSize;
  const int winsize=_WinSize;
  const int maxiter=*this->_MaxIterations;
  const KLT_TYPE maxerr=*this->_MaxError;

  BIASCDOUT(D_TRACKERB_INIT, "using hws: "<<hws<<"\tmaxiter: "<<maxiter
            <<"\tmaxerr: "<<maxerr<<endl);

  if (p1[0]<this->_minx1 || p1[0]>=this->_maxx1 || 
      p1[1]<this->_miny1 || p1[1]>=this->_maxy1 ||
      p2[0]<this->_minx2 || p2[0]>=this->_maxx2 || 
      p2[1]<this->_miny2 || p2[1]>=this->_maxy2 ){
    BIASCDOUT(D_TRACKERB_KLT, "p1 ("<<p1[0]<<", "<<p1[1]<<")   p2 ("<<p2[0]
              <<", "<<p2[1]<<")\n");
    BIASCDOUT(D_TRACKERB_KLT, "tracking roi im1 : ( "<<_minx1<<", "<<_miny1
              <<")  <-->  ( "<<_maxx1<<", "<<_maxy1<<")\n");
    BIASCDOUT(D_TRACKERB_KLT, "tracking roi im2 : ( "<<_minx2<<", "<<_miny2
              <<")  <-->  ( "<<_maxx2<<", "<<_maxy2<<")\n");
    BIASCDOUT(D_TRACKERB_KLT, "points do not lie in ROI, returning -4\n");
    return TRACKER_OUTOFROI;
  } else {
    BIASCDOUT(D_TRACKERB_KLT, "tracking roi im1 : ( "<<_minx1<<", "<<_miny1
              <<")  <-->  ( "<<_maxx1<<", "<<_maxy1<<")\n");
    BIASCDOUT(D_TRACKERB_KLT, "tracking roi im2 : ( "<<_minx2<<", "<<_miny2
              <<")  <-->  ( "<<_maxx2<<", "<<_maxy2<<")\n");
  }

  iter=0;

  result[0]=p2[0];
  result[1]=p2[1];
  this->MyBilinearRegion1_(p1[0],p1[1],hws);
  BIASCDOUT(D_TRACKERB_BILINEAR, "_bl1: "<<_bl1<<"\n_gx1: "<<_gx1
            <<"\n_gy1: "<<_gy1<<endl);
#ifdef TIMING
  timer.Stop();
  cerr << "TrackerBase MyBilinearRegion1_ took "<<timer.GetRealTime()
       <<" us = "<<timer.GetCycleCount()<<" cycles \n";
  timer.Reset();
  timer.Start();
#endif

  do {
    _gxx=_gxy=_gyy=ex=ey=0.0;
    MyBilinearRegion2_(result[0],result[1],hws);
    BIASCDOUT(D_TRACKERB_BILINEAR, "_bl2: "<<_bl2<<"\n_gx2: "<<_gx2
              <<"\n_gy2: "<<_gy2<<endl);
   
    // initialize G and e
    for (rh=0; rh<winsize; rh++){
      for (ch=0; ch<winsize; ch++){
        _gsx[rh][ch] = _gx1[rh][ch] + _gx2[rh][ch];
        _gsy[rh][ch] = _gy1[rh][ch] + _gy2[rh][ch];
        _gxx += _gsx[rh][ch]*_gsx[rh][ch];
        _gyy += _gsy[rh][ch]*_gsy[rh][ch];
        _gxy += _gsx[rh][ch]*_gsy[rh][ch];
        dI = _bl1[rh][ch] - _bl2[rh][ch];
        ex += _gsx[rh][ch]*dI;
        ey += _gsy[rh][ch]*dI;
#ifdef BIAS_DEBUG
        if (isnan(_gxx) || isinf(_gxx) ||
            isnan(_gxy) || isinf(_gxy) ||
            isnan(_gyy) || isinf(_gyy) ) {
          cout << "_gx1: "<<_gx1<<"\n_gx2: "<<_gx2
               << "\n_gy1: "<<_gy1<<"\n_gy2: "<<_gy2
               << "\n_bl1: "<<_bl1<<"\n_bl2: "<<_bl2
               << "\np1: "<<p1[0]<<" "<<p1[1]<<endl
               << "\np2: "<<p2[0]<<" "<<p2[1]<<endl
               << "\nroi1 ("<<_minx1<<", "<<_miny1<<") <--> ("
               <<_maxx1<<", "<<_maxy1<<")"
               << "\nroi2 ("<<_minx2<<", "<<_miny2<<") <--> ("
               <<_maxx2<<", "<<_maxy2<<")\n";
        }
        BIASASSERT(!isnan(_gxx) && !isinf(_gxx));
        BIASASSERT(!isnan(_gxy) && !isinf(_gxy));
        BIASASSERT(!isnan(_gyy) && !isinf(_gyy));
#endif
      } 
    }
    BIASCDOUT(D_TRACKERB_KLT, "iter: "<<iter<<"  _gxx: "<<_gxx<<" _gxy: "<<_gxy
              <<" _gyy: "<<_gyy<<" ex: "<<ex<<" ey: "<<ey<<endl);
    det = _gxx*_gyy-_gxy*_gxy;
    if (fabs(det)<SMALL_DET) {
      BIASDOUT(D_TRACKERB_KLT, 
               "unable to invert G, det: "<<det);
      res = TRACKER_NOSTRUCTURE;
      break;
    }
    _idet=1.0/det;
    dx = (_gyy*ex - _gxy*ey)*_idet; 
    dy = (_gxx*ey - _gxy*ex)*_idet; 

    BIASCDOUT(D_TRACKERB_KLT, "     det: "<<det<<" _idet: "<<_idet
              <<" dx: "<<dx<<" dy: "<<dy<<endl);
    
    //error = sqrt(dx*dx + dy*dy);
    error = (fabs(dx)>fabs(dy))?fabs(dx):fabs(dy);

    /*
    if (error>lsterror){
      BIASCDOUT(D_TRACKERB_KLT, 
                "error increased ("
                <<lsterror<<" -> "<<error
                <<"), stopping iteration and returning 1\n");
      res = TRACKER_ERRORINCREASED;
      error=lsterror;
      break;
      } */
    lsterror=error;

    result[0] += dx;
    result[1] += dy;

    BIASCDOUT(D_TRACKERB_KLT,"     disp after "<<iter<<" iteration: "
              <<result[0]-p2[0]<<", "<<result[1]-p2[1]<<" error: "<< error
              << "  ("<<result[0]<<", "<<result[1]<<")\n");

    //#define D_TRACK
#ifdef D_TRACK
    fprintf(stdout, "iter: %i\t_gxx: %f\t_gxy: %f\t_gyy: %f\tex: %f\tey: %f\n",
            iter, _gxx, _gxy, _gyy, ex, ey);
    fprintf(stdout, "      \tdx: %f\tdy: %f\tcoo: (%f, %f)\n", 
            dx, dy, result[0], result[1]);
#endif

    iter++;
    
    if (result[0]<this->_minx2 || result[0]>=this->_maxx2 ||
        result[1]<this->_miny2 || result[1]>=this->_maxy2){
      BIASCDOUT(D_TRACKERB_KLT, 
                "feature slid out of image, returning -2");
      res = TRACKER_OUTOFIMAGE;
      break;
    }
  } while (error>maxerr && iter<maxiter);
    
  if (res==0 && error>maxerr && iter==maxiter) {
    res = TRACKER_TOOMANYITER;
    BIASCDOUT(D_TRACKERB_KLT,"error > "<<maxerr
              <<" and max iteration count reached: "<<iter<<" returning -3");
  } else if (res==0) {
    MyBilinearRegion2_(result[0],result[1],hws);
  }

#ifdef TIMING
  timer.Stop();
  cerr << "i "<<iter<<" res "<<res<<" ";
  cerr << "TrackSimple_ took "<<timer.GetRealTime()<<" us = "
       <<timer.GetCycleCount()<<" cycles\n";
#endif
  return res;
}

template <class StorageType>
int TrackerBase<StorageType>::
TrackWeighted_(KLT_TYPE p1[2], KLT_TYPE p2[2], KLT_TYPE result[2], 
               KLT_TYPE& error, int &iter)
{
  if (*this->_HalfWinSize != this->_LastHalfWinSize){
    Resize_(*this->_HalfWinSize);
    CalcGaussianWeightMatrix_();
  }
  BIASCDOUT(D_TRACKERB_KLT,"p1: " << p1[0] << ", " << p1[1] 
            << "\tp2: " <<p2[0] << ", " << p2[1]<<endl);

  int res=TRACKER_SUCCESS;
  int rh, ch;
  
  register KLT_TYPE dI, ex, ey, dx, dy, det;
  KLT_TYPE lsterror=FLT_MAX;
  const int hws=*this->_HalfWinSize;
  const int winsize=this->_WinSize;
  const int maxiter=*this->_MaxIterations;
  const KLT_TYPE maxerr=*this->_MaxError;

  if (p1[0]<this->_minx1 || p1[0]>=this->_maxx1 || 
      p1[1]<this->_miny1 || p1[1]>=this->_maxy1 ||
      p2[0]<this->_minx2 || p2[0]>=this->_maxx2 || 
      p2[1]<this->_miny2 || p2[1]>=this->_maxy2 ){
    BIASCDOUT(D_TRACKERB_KLT, "p1 ("<<p1[0]<<", "<<p1[1]<<")   p2 ("<<p2[0]
              <<", "<<p2[1]<<")\n");
    BIASCDOUT(D_TRACKERB_KLT, "tracking roi im1 : ( "<<_minx1<<", "<<_miny1
              <<")  <-->  ( "<<_maxx1<<", "<<_maxy1<<")\n");
    BIASCDOUT(D_TRACKERB_KLT, "tracking roi im2 : ( "<<_minx2<<", "<<_miny2
              <<")  <-->  ( "<<_maxx2<<", "<<_maxy2<<")\n");
    BIASCDOUT(D_TRACKERB_KLT, "points do not lie in ROI, returning -4\n");
    return TRACKER_OUTOFROI;
  } else {
    BIASCDOUT(D_TRACKERB_KLT, "tracking roi im1 : ( "<<_minx1<<", "<<_miny1
              <<")  <-->  ( "<<_maxx1<<", "<<_maxy1<<")\n");
    BIASCDOUT(D_TRACKERB_KLT, "tracking roi im2 : ( "<<_minx2<<", "<<_miny2
              <<")  <-->  ( "<<_maxx2<<", "<<_maxy2<<")\n");
  }

  iter=0;

  result[0]=p2[0];
  result[1]=p2[1];
  //  MyBilinearRegion(_ida1,_gradx1,_grady1,p1[0],p1[1],hws,_bl1,_gx1,_gy1);
  this->MyBilinearRegion1_(p1[0],p1[1],hws);
  do {
    _gxx=_gxy=_gyy=ex=ey=0.0;
    //     MyBilinearRegion(_ida2,_gradx2,_grady2,result[0],result[1],hws,
    //                      _bl2,_gx2,_gy2);
    this->MyBilinearRegion2_(result[0],result[1],hws);
    // initialize G and e
    for (rh=0; rh<winsize; rh++){
      for (ch=0; ch<winsize; ch++){
        this->_gsx[rh][ch] = 
          _weight[rh][ch] * (this->_gx1[rh][ch] + this->_gx2[rh][ch]);
        this->_gsy[rh][ch] = 
          _weight[rh][ch] * (this->_gy1[rh][ch] + this->_gy2[rh][ch]);
        _gxx += this->_gsx[rh][ch]*this->_gsx[rh][ch];
        _gyy += this->_gsy[rh][ch]*this->_gsy[rh][ch];
        _gxy += this->_gsx[rh][ch]*this->_gsy[rh][ch];
        dI = _weight[rh][ch] * (this->_bl1[rh][ch] - this->_bl2[rh][ch]);
        ex += this->_gsx[rh][ch]*dI;
        ey += this->_gsy[rh][ch]*dI;
      } 
    }
    BIASCDOUT(D_TRACKERB_KLT, "_gxx: "<<_gxx<<" _gxy: "<<_gxy
              <<" _gyy: "<<_gyy<<" ex: "<<ex<<" ey: "<<ey<<endl);
    det = _gxx*_gyy-_gxy*_gxy;
    if (fabs(det)<SMALL_DET) {
      BIASCDOUT(D_TRACKERB_KLT, 
                "unable to invert G, det: "<<det<<endl);

      res = TRACKER_NOSTRUCTURE;
      break;
    }
    _idet=1.0/det;
    dx = (_gyy*ex - _gxy*ey)*_idet; 
    dy = (_gxx*ey - _gxy*ex)*_idet; 

    BIASCDOUT(D_TRACKERB_KLT, "det: "<<det<<" _idet: "<<_idet
              <<" dx: "<<dx<<" dy: "<<dy<<endl);
    
    //error = sqrt(dx*dx + dy*dy);
    error = (fabs(dx)>fabs(dy))?fabs(dx):fabs(dy);
    /*
      if (error>lsterror){
      BIASCDOUT(D_TRACKERB_KLT, "error increased ("
      <<lsterror<<" -> "<<error
      <<"), stopping iteration and returning 1\n");
      res = TRACKER_ERRORINCREASED;
      error=lsterror;
      break;
      }
    */
    lsterror=error;

    result[0] += dx;
    result[1] += dy;

    BIASCDOUT(D_TRACKERB_KLT,"disp after "<<iter<<" iteration: "
              <<result[0]-p2[0]<<", "<<result[1]-p2[1]<<" error: "<< error
              << "  ("<<result[0]<<", "<<result[1]<<")\n");

    iter++;
    
    if (result[0]<this->_minx2 || result[0]>=this->_maxx2 ||
        result[1]<this->_miny2 || result[1]>=this->_maxy2){
      BIASCDOUT(D_TRACKERB_KLT, 
                "feature slid out of image, returning -2\n");
      res = TRACKER_OUTOFIMAGE;
      break;
    }
  } while (error>maxerr && iter<maxiter);
    
  if (res==0 && error>maxerr) {
    res = TRACKER_TOOMANYITER;
    BIASCDOUT(D_TRACKERB_KLT,"error > maxerror : returning -3\n");
  } else if (res==0) {
    MyBilinearRegion2_(result[0],result[1],hws);
  }

#ifdef TIMING
  timer.Stop();
  cerr << "i "<<iter<<" res "<<res<<" ";
  cerr << "TrackWeighted_ took "<<timer.GetRealTime()<<" us = "
       <<timer.GetCycleCount()<<" cycles\n";
#endif

  return res;
}





template <class StorageType>
int TrackerBase<StorageType>::
EvaluateResult_(int& res, KLT_TYPE& mad, KLT_TYPE& msd, KLT_TYPE cov[4])
{
#ifdef TIMING
  TimeMeasure timer;
  timer.Start();
#endif
  
  if (res==0){
    const int wsize = _WinSize*_WinSize;
    KLT_TYPE *f1=_bl1.GetData(), *f2=_bl2.GetData(), *end=f1+wsize;
    mad = 0.0;
    msd = 0.0;
    KLT_TYPE tmp_res=0.0;
    while (f1<end){
      tmp_res = *f1++ - *f2++;
      msd += tmp_res * tmp_res;
      mad += fabs(tmp_res);
    }
     
    // now calculate the covariance matrix
    // see Reinhards dissertation page 67, equation 5.9
    // this is the inverse of the matrix 
    // the reference variance is sigma
    BIASASSERT(wsize>3);
    double sigmasquared = msd / (double)(wsize - 2);

    // compute mean squared difference
    msd /= (KLT_TYPE)wsize;
    mad /= (KLT_TYPE)wsize;

    // the inverse of the structure tensor matrix is given by
    // 1/det * | _gyy -_gxy| 
    //         |-_gxy  _gxx|
    cov[0] = _gyy * _idet * sigmasquared;
    cov[1] = cov[2]= -_gxy * _idet * sigmasquared;
    cov[3] = _gxx * _idet * sigmasquared;
// #ifdef BIAS_DEBUG
//     cout << "mad "<<mad<<" sigmasq "<<sigmasquared<<" cov[0][0] "<<cov[0]
//          <<" cov[1][1] "<<cov[3];
//     double meanx=0.0, meany=0.0;
//     for (int x=0; x<7; x++){
//       for (int y=0; y<7; y++){
//         meanx+=fabs(_gx1[y][x]);
//         meany+=fabs(_gy1[y][x]);
//       }
//     }
//     meanx/=49.0;
//     meany/=49.0;
//     cout <<"  gx: "<<meanx <<" gy: "<<meany<<" idet "<<_idet<<" _gxx "
//          <<_gxx<<" _gyy "<<_gyy<<endl;
// #endif
  }

#ifdef TIMING
  timer.Stop();
  cerr << "TrackerBase::EvaluateResult_ took "<<timer.GetRealTime()<<" us = "
       <<timer.GetCycleCount()<<" cycles\n";
#endif
  return res;
}



template <class StorageType>
int TrackerBase<StorageType>::
RejectMAD_(std::vector<KLT_TYPE>& sad, std::vector<int>& res)
{
  KLT_TYPE thresh=*_MaxResiduumMAD;
  unsigned num = res.size();

  for (unsigned i=0; i<num; i++){
    if (res[i]>=TRACKER_SUCCESS && sad[i]>=thresh){
      res[i]=TRACKER_MEANABSDIF;
    }
  }
  return 0;
}

template <class StorageType>
int TrackerBase<StorageType>::
RejectX86_(std::vector<KLT_TYPE>& ssd, std::vector<int>& res)
{
#ifdef TIMING
  TimeMeasure timer;
  timer.Start();
#endif
  int num=ssd.size();

  _ssd_residui.clear();
  _diff_from_median.clear();
  _ssd_residui.reserve(num);
  _diff_from_median.reserve(num);
  _all_ssd_residui.resize(num);
  _all_diff_from_median.resize(num);

  // remove all points where tracking failed
  for (int i=0; i<num; i++){
    if (res[i]>=TRACKER_SUCCESS){
      _ssd_residui.push_back(_all_ssd_residui[i]);
    }
  }
  unsigned num2=_ssd_residui.size();
  BIASCDOUT(D_TRACKERB_REJECT, "using "<<num2
            <<" valid residuis for median calculation"<<endl);

  _Median.Compute(_ssd_residui);
  _median_ssd_residui = _Median.GetMedian();
  BIASCDOUT(D_TRACKERB_REJECT,"median residuum is "<<_median_ssd_residui
            <<endl);

  for (unsigned i=0; i<num2; i++){
    _diff_from_median.push_back(fabs(_ssd_residui[i] - _median_ssd_residui));
    //  BIASCDOUT(D_TRACKERB_REJECT, setw(4)<<i<<" residuum "<<setprecision(2)
    //               <<setw(10)<<_ssd_residui[i]<<"\tdiff_from_median "
    //               <<setprecision(2)<<setw(10)<<_diff_from_median[i]<<endl);
  }

  _Median.Compute(_diff_from_median);
  _threshold = 5.2 * _Median.GetMedian();
  BIASCDOUT(D_TRACKERB_REJECT, "median of diff from median is "
            <<_Median.GetMedian()<<endl);
  BIASCDOUT(D_TRACKERB_REJECT, "x84 threshold is "<<_threshold<<endl);

  unsigned rejected=0;
  for (int i=0; i<num; i++){
    _all_diff_from_median[i] = fabs(_all_ssd_residui[i] - _median_ssd_residui);
    if (res[i]>=0 && (_all_diff_from_median[i] > _threshold)){
      res[i]=TRACKER_X84;
      BIASCDOUT(D_TRACKERB_REJECT, "rejecting point "<<setw(4)<<i
                <<": diff from median "<<fixed<<right<<setprecision(2)
                <<setw(10)<<_all_diff_from_median[i]<<endl);
      rejected++;
    } else if (res[i]==TRACKER_MEANABSDIF && 
               (_all_diff_from_median[i] < _threshold)){
      res[i]=0;
      BIASCDOUT(D_TRACKERB_REJECT, "accepting point "<<setw(4)<<i
                <<": diff from median "<<fixed<<right<<setprecision(2)
                <<setw(10)<<_all_diff_from_median[i]<<endl);
    }
  }
  BIASCDOUT(D_TRACKERB_REJECT, "rejected "<<rejected<<" points "<<endl);

#ifdef BIAS_DEBUG
  if (rejected>num2/2){
    cout << "rejected more than half of tracks ("<<rejected<<" of "<<num2
         <<")aborting\n";
    for (unsigned i=0; i<num2; i++){
      cout << setw(4)<<i<<" residuum "<<setprecision(4)<<setw(10)<<fixed
           <<right<<_ssd_residui[i]<<"\tdiff_from_median "<<setprecision(4)
           <<setw(10)<<fixed<<right<<_diff_from_median[i]<<endl;
    }
    
    sort(_ssd_residui.begin(), _ssd_residui.end());
    sort(_diff_from_median.begin(), _diff_from_median.end());
    cout << "sorted\n";
    for (unsigned i=0; i<num2; i++){
      cout << setw(4)<<i<<" residuum "<<setprecision(4)<<setw(10)<<fixed
           <<right<<_ssd_residui[i]<<"\tdiff_from_median "<<setprecision(4)
           <<setw(10)<<fixed<<right<<_diff_from_median[i]<<endl;
    }

    abort();
  }
#endif

#ifdef TIMING
  timer.Stop();
  cerr << "TrackerBase::Rejectx86 took "<<timer.GetRealTime()
       <<" us = "<<timer.GetCycleCount()<<" cycles \n";
#endif
  return 0;
}


template <class StorageType>
void TrackerBase<StorageType>::
CalcGaussianWeightMatrix_()
{
  _weight.newsize(this->_WinSize, this->_WinSize);
  // now calculate the gaussian weight matrix
  double sigma = ((double)this->_WinSize)/6.0;
  double fac = 1.0 / (2.0 * sigma * sigma);
  double sum=0.0, tmp;
  register int i, j, x, y, xp, xm, yp, ym;
  const int hws=*this->_HalfWinSize;
  for (i= hws; i >= 0; i--){
    for (j= hws; j >= 0; j--){
      tmp=exp(-(double)(i*i+j*j)*fac);
      yp=hws+i;
      ym=hws-i;
      xp=hws+j;
      xm=hws-j;
      _weight[ym][xm] =_weight[yp][xm] = _weight[ym][xp] = _weight[yp][xp] = 
        (KLT_TYPE)(tmp);
    }
  }
    
  for (y=0; y<this->_WinSize; y++){
    for (x=0; x<this->_WinSize; x++){
      sum+=_weight[y][x];
    } 
  }
  BIASCDOUT(D_TRACKERB_INIT, "_weight: "<<_weight<<endl);
  BIASCDOUT(D_TRACKERB_INIT, "_weight sum before normalization: "<<sum<<endl);

  //normalization
  sum=1.0/sum; // division is slow
  for (i= hws; i >= 0; i--){
    for (j= hws; j >= 0; j--){
      yp=hws+i;
      ym=hws-i;
      xp=hws+j;
      xm=hws-j;
      tmp=_weight[ym][xm];
      _weight[ym][xm] = _weight[yp][xm] = _weight[ym][xp] =
        _weight[yp][xp] = tmp * sum;
    }
  }
  BIASCDOUT(D_TRACKERB_INIT, "_weight: "<<_weight<<endl);
#ifdef BIAS_DEBUG
  sum=0.0;
  for (y=0; y<this->_WinSize; y++){
    for (x=0; x<this->_WinSize; x++){
      sum+=_weight[y][x];
    } 
  }
  BIASCDOUT(D_TRACKERB_INIT,"_weight sum after normalization: "<<sum<<endl);
#endif  
}


template <class StorageType>
void TrackerBase<StorageType>::
AddParam_(Param& para, const int grp, const std::string& postfix)
{
  if (para.CheckParam("MaxIterations"+postfix)){ 
    _MaxIterations=para.GetParamInt("MaxIterations"+postfix);
  } else { 
    _MaxIterations=
      para.AddParamInt("MaxIterations"+postfix, 
                       "maximum number of iterations", 
                       10, 1, INT_MAX, 0, grp);
  }

  if (para.CheckParam("HalfWinSize"+postfix)){ 
    _HalfWinSize=para.GetParamInt("HalfWinSize"+postfix);
  } else { 
    _HalfWinSize=
      para.AddParamInt("HalfWinSize"+postfix, 
                       "half size of support window used in gradient"
                       " descent tracking", 
                       3, 1, INT_MAX, 0, grp);
  }

  if (para.CheckParam("MaxError"+postfix)){ 
    _MaxError=para.GetParamDouble("MaxError"+postfix);
  } else { 
    _MaxError=
      para.AddParamDouble("MaxError"+postfix, 
                          "iterations are aborted when displacement"
                          " drops below MaxError", 
                          0.01, 0.0, DBL_MAX, 0, grp);
  }
  if (para.CheckParam("MaxAffineChange"+postfix)){ 
    _MaxAffineChange=para.GetParamDouble("MaxAffineChange"+postfix);
  } else { 
    _MaxAffineChange=
      para.AddParamDouble("MaxAffineChange"+postfix, 
                          "dont stop unless affine entries fall below",
                          DEFAULT_MAX_AFFINECHANGE, 0.0, DBL_MAX, 0, grp);
  }

  if (para.CheckParam("MaxResiduumMAD"+postfix)){ 
    _MaxResiduumMAD=para.GetParamDouble("MaxResiduumMAD"+postfix);
  } else { 
    _MaxResiduumMAD=
      para.AddParamDouble("MaxResiduumMAD"+postfix, 
                          "tracking failed when mean image difference is above"
                          " MaxResiduum after otherwise successfull tracking", 
                          10.0, 0.0, DBL_MAX, 0, grp);
  }

  vector<string> enums;
  vector<int> values;
  enums.push_back("TT_Simple");values.push_back(TT_Simple);
  enums.push_back("TT_Weighted");values.push_back(TT_Weighted);
  enums.push_back("TT_Affine4");values.push_back(TT_Affine4);
  enums.push_back("TT_Affine6");values.push_back(TT_Affine6);
  enums.push_back("TT_SimpleAffine4");values.push_back(TT_SimpleAffine4);
  enums.push_back("TT_SimpleAffine6");values.push_back(TT_SimpleAffine6);

  if (para.CheckParam("TrackerType"+postfix)){ 
    _TrackerType=para.GetParamEnum("TrackerType"+postfix);
  } else { 
    _TrackerType=
      para.AddParamEnum("TrackerType"+postfix, "type of tracker to use", 
                        enums, TT_Simple, &values, 0, grp); 
  }

  enums.clear();
  values.clear();
  enums.push_back("RT_MAD");values.push_back(RT_MAD);
  enums.push_back("RT_X86");values.push_back(RT_X86);
  //enums.push_back(""); values.push_back();
  if (para.CheckParam("RejectionType"+postfix)){ 
    _RejectionType=para.GetParamEnum("RejectionType"+postfix);
  } else { 
    _RejectionType=
      para.AddParamEnum("RejectionType"+postfix,"type of rejection alg to use",
                        enums, RT_MAD, &values, 0, grp); 
  }

}


// ***************** parts, which are not under LGPL *************************
// The function TrackAffine_ is based on the affine tracker by Thorsten
// Thormaehlen. Some of the worker functions have been rewritten, some
// have been adapted, see below for the conditions on this software.


template <class StorageType>
int TrackerBase<StorageType>::
TrackAffine_(KLT_TYPE p1[2], KLT_TYPE p2[2], 
             const Matrix2x2<KLT_TYPE>& p2A,
             KLT_TYPE result[2],  Matrix2x2<KLT_TYPE>& resultA,
             KLT_TYPE& error, int &iteration, Matrix<KLT_TYPE>& Cov,
             KLT_TYPE& mad, KLT_TYPE& msd,
             bool Only4Parameters, bool ComputeCovariance) {

  SetDebugLevel(D_TRACKERB_AFFINE | D_TRACKERB_KLT);
  BIASDOUT(D_TRACKERB_AFFINE, "Called Affine for "
           <<p1[0]<<";"<<p1[1]<<", guess is "<<p2[0]<<";"<<p2[1]
           <<" affine guess is "<< p2A);

  const int width = _WinSize;
  const int height = _WinSize;
  const int hw = width/2;
  const int hh = height/2;
  const KLT_TYPE fhw = hw;
  const KLT_TYPE fhh = hh;
 
  const int max_iterations = *this->_MaxIterations;
  const KLT_TYPE maxerr    = *this->_MaxError;
  const KLT_TYPE th_aff    = *this->_MaxAffineChange;
  
  KLT_TYPE dx=0, dy=0;
  iteration = 0;
  int status = 0;
  
  Matrix<KLT_TYPE> T(Only4Parameters?4:6,Only4Parameters?4:6);
  Vector<KLT_TYPE> a(Only4Parameters?4:6);

  Cov.newsize((Only4Parameters)?4:6,(Only4Parameters)?4:6);


  resultA = p2A;
  KLT_TYPE *Axx = &(resultA[0][0]);
  KLT_TYPE *Ayx = &(resultA[0][1]); 
  KLT_TYPE *Axy = &(resultA[1][0]);
  KLT_TYPE *Ayy = &(resultA[1][1]);

  result[0] = p2[0];
  result[1] = p2[1];

  bool convergence = false;

 
  // Allocate memory for windows
  KLT_TYPE *imgdiff = new KLT_TYPE[width * height];
  KLT_TYPE *gradx   = new KLT_TYPE[width * height];
  KLT_TYPE *grady   = new KLT_TYPE[width * height];

  do  {
    KLT_TYPE ul_x =  result[0] + *Axy *  fhh  - *Axx * fhw;
    KLT_TYPE ul_y =  result[1] + *Ayy *  fhh  - *Ayx * fhw; 
    KLT_TYPE ll_x =  result[0] - *Axx *  fhw  - *Axy * fhh; 
    KLT_TYPE ll_y =  result[1] - *Ayx *  fhw  - *Ayy * fhh;
    KLT_TYPE ur_x =  result[0] + *Axx *  fhw  + *Axy * fhh; 
    KLT_TYPE ur_y =  result[1] + *Ayx *  fhw  + *Ayy * fhh;
    KLT_TYPE lr_x =  result[0] + *Axx *  fhw  - *Axy * fhh; 
    KLT_TYPE lr_y =  result[1] + *Ayx *  fhw  - *Ayy * fhh;

    if ( ul_x  < this->_minx2 ||  ul_x  > this->_maxx2 ||
         ll_x  < this->_minx2 ||  ll_x  > this->_maxx2 ||
         ur_x  < this->_minx2 ||  ur_x  > this->_maxx2 ||
         lr_x  < this->_minx2 ||  lr_x  > this->_maxx2 ||
         ul_y  < this->_miny2 ||  ul_y  > this->_maxy2 ||
         ll_y  < this->_miny2 ||  ll_y  > this->_maxy2 ||
         ur_y  < this->_miny2 ||  ur_y  > this->_maxy2 ||
         lr_y  < this->_miny2 ||  lr_y  > this->_maxy2) {
      status = TRACKER_OUTOFIMAGE;
      BIASDOUT( D_TRACKERB_KLT, "out of image");
      break;
    }
    
    _am_computeIntensityDifferenceAffine(*_im1, *_im2, 
                                         p1[0], p1[1], result[0], result[1],  
                                         *Axx, *Ayx , *Axy, *Ayy,
                                         hw, hh, imgdiff);
    
    _am_getGradientWinAffine(result[0], result[1], 
                             *Axx, *Ayx , *Axy, *Ayy,
                             hw, hh, gradx, grady);

    if (Only4Parameters){
    
      _am_compute4by1ErrorVector(imgdiff, gradx, grady, hw, hh, a);
      _am_compute4by4GradientMatrix(gradx, grady, hw, hh, T);
	
      status = _am_gauss_jordan_elimination(T,a);
      	
      *Axx += a[0];
      *Ayx += a[1];
      *Ayy = *Axx;
      *Axy = -(*Ayx);
	
      result[0] += dx = a[2];
      result[1] += dy = a[3];
    } else {
     
      _am_compute6by1ErrorVector(imgdiff, gradx, grady, hw, hh, a);
      _am_compute6by6GradientMatrix(gradx, grady, hw, hh, T);
      
      status = _am_gauss_jordan_elimination(T,a);
	
      *Axx += a[0];
      *Ayx += a[1];
      *Axy += a[2];
      *Ayy += a[3];

      result[0] += dx = a[4];
      result[1] += dy = a[5];
          
    }
         
    // old upper left corner - new upper left corner 
    ul_x -=  result[0] - *Axx * fhw + *Axy * fhh;  
    ul_y -=  result[1] - *Ayx * fhw + *Ayy * fhh; 
    // old lower left corner - new lower left corner 
    ll_x -=  result[0] - *Axx * fhw - *Axy * fhh;  
    ll_y -=  result[1] - *Ayx * fhw - *Ayy * fhh;
    // old upper right corner - new upper right corner 
    ur_x -=  result[0] + *Axx * fhw + *Axy * fhh;  
    ur_y -=  result[1] + *Ayx * fhw + *Ayy * fhh;
    // old lower right corner - new lower right corner 
    lr_x -=  result[0] + *Axx * fhw - *Axy * fhh;  
    lr_y -=  result[1] + *Ayx * fhw - *Ayy * fhh;
    
    convergence = (fabs(dx) < maxerr && fabs(dy) < maxerr  &&
                   fabs(ul_x) < th_aff && fabs(ul_y) < th_aff &&
                   fabs(ll_x) < th_aff && fabs(ll_y) < th_aff &&
                   fabs(ur_x) < th_aff && fabs(ur_y) < th_aff &&
                   fabs(lr_x) < th_aff && fabs(lr_y) < th_aff);
    
    if (status == TRACKER_NOSTRUCTURE) {
      BIASDOUT( D_TRACKERB_KLT, "no structure for tracking");
      break;
    }
    iteration++;
 
  }  while ( !convergence  && iteration < max_iterations); 
  if (iteration >= max_iterations) {
    status = TRACKER_TOOMANYITER;
    BIASDOUT( D_TRACKERB_KLT, "too many iter: "<<iteration
              <<" maxiter="<<max_iterations);
  }
  mad = 0;
  msd = 0;
  if ((status==TRACKER_TOOMANYITER) || (status==TRACKER_SUCCESS)) {
    _am_computeIntensityDifferenceAffine(*_im1, *_im2, p1[0], p1[1], 
                                         result[0], result[1], 
                                         *Axx, *Ayx , *Axy, *Ayy,
                                         hw, hh, imgdiff);

    const unsigned int wh = width*height;
    register KLT_TYPE* fw = imgdiff;
    for (unsigned int h=0; h<wh ; h++) {
      const KLT_TYPE cur = fabsf(*fw++);
      mad += cur;
      msd += cur*cur;
    }
    
    double sigma = msd / (double)(wh - (Only4Parameters)?4:6);
    mad /= double(wh);
    msd /= double(wh);
 
    if (ComputeCovariance) {
      if (Only4Parameters)
        _am_compute4by4GradientMatrix(gradx, grady, hw, hh, T);
      else 
        _am_compute6by6GradientMatrix(gradx, grady, hw, hh, T);
      SVD covsvd(T);
      Cov = covsvd.Invert() * sigma;
    } else {
      // pseudo covariance
      Cov.SetZero();
      if (Only4Parameters) {
        Cov[2][2] = dx * dx;
        Cov[3][3] = dy * dy;
      } else {
        Cov[4][4] = dx * dx;
        Cov[5][5] = dy * dy;
      }
    }
  }
  
  error = (fabs(dx)>fabs(dy))?fabs(dx):fabs(dy);

  // Check whether window is out of bounds
  if (result[0]<this->_minx2 || result[0] > this->_maxx2  || 
      result[1]<this->_miny2 || result[1] > this->_maxy2)
    status = TRACKER_OUTOFIMAGE;


  // Free memory 
  delete imgdiff;
  delete gradx;
  delete grady;

  BIASDOUT( D_TRACKERB_KLT, "status at end is "<<status);
 
  // Return appropriate value 
  return status;
}



/********************************************************************** 
* CONSISTENCY CHECK OF FEATURES BY AFFINE MAPPING (BEGIN)
* 
* Created by: Thorsten Thormaehlen (University of Hannover) June 2004    
* thormae@tnt.uni-hannover.de
* 
* Permission is granted to any individual or institution to use, copy, modify,
* and distribute this part of the software, provided that this complete 
* authorship and permission notice is maintained, intact, in all copies. 
*
* This software is provided  "as is" without express or implied warranty.
*
* 
*/

int _am_gauss_jordan_elimination(Matrix<KLT_TYPE>& a,
                                 Vector<KLT_TYPE>& b) {
  const int n = a.GetRows();
  Vector<int> indxc(n),indxr(n),ipiv(n);
  int i,j,k,l,ll;
  KLT_TYPE big,dum,pivinv,tmp;
  int col = 0;
  int row = 0;

  for (j=0;j<n;j++) ipiv[j]=0;
  for (i=0;i<n;i++) {
    big=0.0;
    for (j=0;j<n;j++)
      if (ipiv[j] != 1)
        for (k=0;k<n;k++) {
          if (ipiv[k] == 0) {
            if (fabs(a[j][k]) >= big) {
              big=fabs(a[j][k]);
              row=j;
              col=k;
            }
          } else if (ipiv[k] > 1) return TRACKER_NOSTRUCTURE;
        }
    ++(ipiv[col]);
    if (row != col) {
      for (l=0;l<n;l++) {
        tmp = a[row][l];
        a[row][l] = a[col][l];
        a[col][l] = tmp;
      }
      tmp = b[row];
      b[row] = b[col];
      b[col] = tmp;
    }
    indxr[i]=row;
    indxc[i]=col;
    if (a[col][col] == 0.0) return TRACKER_NOSTRUCTURE;
    pivinv=1.0/a[col][col];
    a[col][col]=1.0;
    for (l=0;l<n;l++) a[col][l] *= pivinv;
    b[col] *= pivinv;
    for (ll=0;ll<n;ll++)
      if (ll != col) {
        dum=a[ll][col];
        a[ll][col]=0.0;
        for (l=0;l<n;l++) a[ll][l] -= a[col][l]*dum;
        b[ll] -= b[col]*dum;
      }
  }
  for (l=n-1;l>=0;l--) {
    if (indxr[l] != indxc[l])
      for (k=0;k<n;k++) {
        tmp = a[k][indxr[l]];
        a[k][indxr[l]] = a[k][indxc[l]];
        a[k][indxc[l]] = tmp;
      }
  }
 
  return TRACKER_SUCCESS;
}

template <class StorageType>
void TrackerBase<StorageType>::
_am_getGradientWinAffine(const KLT_TYPE& x, const KLT_TYPE& y,     
                         const KLT_TYPE& Axx, const KLT_TYPE& Ayx,
                         const KLT_TYPE& Axy, const KLT_TYPE& Ayy,  
                         const int hw, const int hh,   
                         KLT_TYPE* out_gradx,    
                         KLT_TYPE* out_grady) {
  register int i, j;
  KLT_TYPE mi, mj;
  for (j = -hh ; j <= hh ; j++) {
    for (i = -hw ; i <= hw ; i++)  {
      mi = Axx * i + Axy * j;
      mj = Ayx * i + Ayy * j;
      *out_gradx++ = _gradim2x->BilinearInterpolation(x+mi, y+mj);
      *out_grady++ = _gradim2y->BilinearInterpolation(x+mi, y+mj);
    }
  }
}

template <class StorageType>
void TrackerBase<StorageType>::
_am_computeIntensityDifferenceAffine(Image<StorageType>& img1,
                                     Image<StorageType>& img2,
                                     const KLT_TYPE& x1, const KLT_TYPE& y1,
                                     const KLT_TYPE& x2, const KLT_TYPE& y2,   
                                     const KLT_TYPE& Axx, const KLT_TYPE& Ayx, 
                                     const KLT_TYPE& Axy, const KLT_TYPE& Ayy, 
                                     const int hw, const int hh, 
                                     KLT_TYPE* imgdiff) {
  KLT_TYPE g1, g2;
  register int i, j;

  // int mi, mj;
  // should this be float ???
  KLT_TYPE mi, mj;
  for (j = -hh ; j <= hh ; j++)
    for (i = -hw ; i <= hw ; i++)  {
      g1 = img1.BilinearInterpolation(x1+i, y1+j);
      mi = Axx * i + Axy * j;
      mj = Ayx * i + Ayy * j;
      g2 = img2.BilinearInterpolation(x2+mi, y2+mj);
      *imgdiff++ = g1 - g2;
    }
}


void _am_compute6by6GradientMatrix(KLT_TYPE* gradx,
                                   KLT_TYPE* grady,
                                   const int hw, 
                                   const int hh,
                                   Matrix<KLT_TYPE>& T) {
  register int i, j;
  KLT_TYPE gx, gy, gxx, gxy, gyy,  x, y, xx, xy, yy;
 
  
  // init with zero
  for (j = 0 ; j < 6 ; j++)  {
    for (i = j ; i < 6 ; i++)  {
      T[j][i] = 0.0;
    }
  }
  
  for (j = -hh ; j <= hh ; j++) {
    for (i = -hw ; i <= hw ; i++)  {
      gx = *gradx++;
      gy = *grady++;
      gxx = gx * gx;
      gxy = gx * gy;
      gyy = gy * gy;
      x = i; 
      y = j; 
      xx = x * x;
      xy = x * y;
      yy = y * y;
      
      T[0][0] += xx * gxx; 
      T[0][1] += xx * gxy;
      T[0][2] += xy * gxx;
      T[0][3] += xy * gxy;
      T[0][4] += x  * gxx;
      T[0][5] += x  * gxy;
	
      T[1][1] += xx * gyy;
      T[1][2] += xy * gxy;
      T[1][3] += xy * gyy;
      T[1][4] += x  * gxy;
      T[1][5] += x  * gyy;
			 
      T[2][2] += yy * gxx;
      T[2][3] += yy * gxy;
      T[2][4] += y  * gxx;
      T[2][5] += y  * gxy;
	 
      T[3][3] += yy * gyy;
      T[3][4] += y  * gxy;
      T[3][5] += y  * gyy; 

      T[4][4] += gxx; 
      T[4][5] += gxy;
      
      T[5][5] += gyy; 
    }
  }
  
  for (j = 0 ; j < 5 ; j++)  {
    for (i = j+1 ; i < 6 ; i++)  {
      T[i][j] = T[j][i];
    }
  }

}



void _am_compute6by1ErrorVector(KLT_TYPE* imgdiff,
                                KLT_TYPE* gradx,
                                KLT_TYPE* grady,
                                const int hw,
                                const int hh,
                                Vector<KLT_TYPE>& e) {
  register int i, j;
  register KLT_TYPE diff,  diffgradx,  diffgrady;

  /* Set values to zero */  
  for(i = 0; i < 6; i++) e[i] = 0.0; 
  
  /* Compute values */
  for (j = -hh ; j <= hh ; j++) {
    for (i = -hw ; i <= hw ; i++)  {
      diff = *imgdiff++;
      diffgradx = diff * (*gradx++);
      diffgrady = diff * (*grady++);
      e[0] += diffgradx * i;
      e[1] += diffgrady * i;
      e[2] += diffgradx * j; 
      e[3] += diffgrady * j; 
      e[4] += diffgradx;
      e[5] += diffgrady; 
    }
  }
  
  for(i = 0; i < 6; i++) e[i] *= 0.5;
  
}


void _am_compute4by4GradientMatrix(KLT_TYPE* gradx,
                                   KLT_TYPE* grady,
                                   const int hw,
                                   const int hh,
                                   Matrix<KLT_TYPE>& T)
{
  register int i, j;
  KLT_TYPE gx, gy, x, y;
 
  
  // Set values to zero 
  for (j = 0 ; j < 4 ; j++)  {
    for (i = 0 ; i < 4 ; i++)  {
      T[j][i] = 0.0;
    }
  }
  
  for (j = -hh ; j <= hh ; j++) {
    for (i = -hw ; i <= hw ; i++)  {
      gx = *gradx++;
      gy = *grady++;
      x = i; 
      y = j; 
      T[0][0] += (x*gx+y*gy) * (x*gx+y*gy);
      T[0][1] += (x*gx+y*gy)*(x*gy-y*gx);
      T[0][2] += (x*gx+y*gy)*gx;
      T[0][3] += (x*gx+y*gy)*gy;
   
      T[1][1] += (x*gy-y*gx) * (x*gy-y*gx);
      T[1][2] += (x*gy-y*gx)*gx;
      T[1][3] += (x*gy-y*gx)*gy;
     
      T[2][2] += gx*gx;
      T[2][3] += gx*gy;
      
      T[3][3] += gy*gy;
    }
  }
  
  for (j = 0 ; j < 3 ; j++)  {
    for (i = j+1 ; i < 4 ; i++)  {
      T[i][j] = T[j][i];
    }
  }

}

void _am_compute4by1ErrorVector(KLT_TYPE* imgdiff,
                                KLT_TYPE* gradx,
                                KLT_TYPE* grady,
                                const int hw, 
                                const int hh,
                                Vector<KLT_TYPE>& e){
  register int i, j;
  register KLT_TYPE diff,  diffgradx,  diffgrady;

  /* Set values to zero */  
  for(i = 0; i < 4; i++) e[i] = 0.0; 
  
  /* Compute values */
  for (j = -hh ; j <= hh ; j++) {
    for (i = -hw ; i <= hw ; i++)  {
      diff = *imgdiff++;
      diffgradx = diff * (*gradx++);
      diffgrady = diff * (*grady++);
      e[0] += diffgradx * i + diffgrady * j;
      e[1] += diffgrady * i - diffgradx * j;
      e[2] += diffgradx;
      e[3] += diffgrady;
    }
  }
  
  for(i = 0; i < 4; i++) e[i] *= 0.5;
  
}

/**************** end affine part  ********************************/ 

//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////

template class TrackerBase<unsigned char>; 
template class TrackerBase<float>; 

// fill in instances as required
#ifdef BUILD_IMAGE_INT
template class TrackerBase<int>; 
#endif
#ifdef BUILD_IMAGE_CHAR
template class TrackerBase<char>; 
#endif
#ifdef BUILD_IMAGE_SHORT
template class TrackerBase<short>; 
#endif
#ifdef BUILD_IMAGE_USHORT
template class TrackerBase<unsigned short>; 
#endif
#ifdef BUILD_IMAGE_UINT
template class TrackerBase<unsigned int>; 
#endif
#ifdef BUILD_IMAGE_DOUBLE
#endif
