/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __TrackerBase__hh__
#define __TrackerBase__hh__

#include <Base/Common/BIASpragmaStart.hh>

#include <Base/Debug/Debug.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Math/Matrix2x2.hh>
#include <Base/Image/Image.hh>
#include <Utils/Param.hh>
#include <MathAlgo/Median1D.hh>

#include <vector>

#define KLT_TYPE double

#define D_TRACKERB_INIT        0x00000001
#define D_TRACKERB_KLT         0x00000002
#define D_TRACKERB_BILINEAR    0x00000004
#define D_TRACKERB_INPUT       0x00000008
#define D_TRACKERB_REJECT      0x00000010
#define D_TRACKERB_AFFINE         0x00000020

#define TRACKER_SUCCESS      0 
#define TRACKER_NOSTRUCTURE -1 
#define TRACKER_OUTOFIMAGE  -2
#define TRACKER_TOOMANYITER -3
#define TRACKER_OUTOFROI    -4 
#define TRACKER_MEANABSDIF  -5
#define TRACKER_IGNOREPOINT -6 
#define TRACKER_X84         -7 
#define TRACKER_ERRORINCREASED  1



namespace BIAS {

  /** defines type or combination of tracker(s) used: simple/weighted estimate
      a dx,dy, Affine4/Affine6 estimate an affine transformation with 4 resp. 6
      parameters and the SimpleAffine start estimating dx,dy and afterwards
      refine with affine */
  enum ETrackerType {TT_Invalid=-1, TT_Simple, TT_Weighted, TT_Affine4,
                     TT_Affine6, TT_SimpleAffine4, TT_SimpleAffine6};

  enum ERejectionType {RT_Invalid=-1, RT_MAD, RT_X86};

  std::ostream& operator<<(std::ostream& os, const enum ETrackerType& tt);

  std::ostream& operator<<(std::ostream& os, const enum ERejectionType& rt);

  /** @class TrackerBase
      @ingroup g_tracker
      @brief  

      Base class for the different tracking algorithms, defining the interfaces
      for the tracking functions. Also stores some basic options and datas, 
      i.e. the pointer to the image datas. See ExampleTrackerBase for brief 
      exemplary usage.

      \todo Implement DLL - add lost of missing explicit instantiations (JW)

      @author woelk 09/2004 */
  template <class StorageType>
  class BIASMatcher2D_EXPORT TrackerBase : public Debug
  {
  public :
    TrackerBase();

    TrackerBase(Param& para, const std::string& postfix="");

    TrackerBase(Param& para, const int grp, const std::string& postfix);

    ~TrackerBase();

    /** Track from im1 to im2, 
        gradx1/2: horizontal gradient imgs
        grady1/2: vertical gradient imgs
        The ROI must be set correctly in gradient images (representing the 
        gradient filter mask size). */
    void Init(Image<StorageType>& im1, Image<StorageType>& im2, 
              Image<StorageType>& gradx1, Image<StorageType>& grady1, 
              Image<StorageType>& gradx2, Image<StorageType>& grady2);
    
    /** Calculates correspondence from image1 to image2. The two images and 
        corresponding gradients must specified using the function Init()
        before calling this function.
        
        The algorithm searches for best correspondence to point p1 from image1 
        in image2. The search starts at p2 in image 2.
        Upon completition 
        - result      : contains the best match, 
        - error       : contains the change in displacement calculated in 
                        the last iteration, 
        - iter        : contains then number of iterations used,
        - residuumMAD : contains the mean absolute grey value difference 
                        between p1 in image 1 and p2tracked in image 2
                        over the tracking support window,
        - residuumMSD : contains the squared grey value difference between p1 
                        in image 1 and p2tracked in image 2 summed over the 
                        tracking support window,
        - cov         : contains an approximation to the covarinace matrix
        image1 and result in image2 over the tracking window ( = sum of
        absolute difference divided by number of pixels in tracking value ).
            
        All homogenous points must be homogenized (i.e. w=1.0).
        
        The algorithms either terminates if the displacement resulting from
        one iteration drops below *_Maxerror or the maximum number of 
        iterations *_MaxIterations is reached.

        @return   see defines
        - 0 on success (error< maxerror)
        - -1 when no spatial structure is present
        - -2 when point slid out of image
        - -3 when maxiter is reached and error is above maxerror
        - -4 when a point lies outside of valid region in images
        - -5 when point was rejected by MAD criterion
        - -6 when input point was at infinity
        - -7 when point is rejected by X86
        - 1 when error increased from last iteration 
        @author woelk */
    int Track(HomgPoint2D& p1, HomgPoint2D& p2, HomgPoint2D& p2tracked, 
              KLT_TYPE& error, int &iter, KLT_TYPE& residuumMAD,
              KLT_TYPE& residuumMSD, Matrix<KLT_TYPE>& cov,
              const Matrix2x2<KLT_TYPE> & AffinePred =
              Matrix2x2<KLT_TYPE>());

    /** Same as Track(), but a correspondence is searched for every point 
        in the vector p1. */
    void Track(std::vector<HomgPoint2D>& p1, 
               std::vector<HomgPoint2D>& p2, 
               std::vector<HomgPoint2D>& p2tracked, 
               std::vector<KLT_TYPE>& error, 
               std::vector<int>& numiter,
               std::vector<KLT_TYPE>& residuiMAD, 
               std::vector<KLT_TYPE>& residuiMSD, 
               std::vector<int>& res,
               std::vector<Matrix<KLT_TYPE> >& cov,
               const std::vector<Matrix2x2<KLT_TYPE> >& AffinePred =
               std::vector<Matrix2x2<KLT_TYPE> >());


    /** @brief get results of affine tracking */
    inline void GetAffineResults(std::vector<Matrix2x2<KLT_TYPE> > & Results) {
      Results = _AffineResults;
    }

    /** @brief get result of affine tracking */
    inline void GetAffineResult(Matrix2x2<KLT_TYPE> & Result) {
      BIASASSERT(_AffineResults.size()>0);
      Result = _AffineResults[0];
    }

    ////////////////////////////////////////////////////////////////////////

    inline void SetMaxError(const KLT_TYPE maxerr) { *_MaxError=maxerr; }

    inline void SetMaxResiduumMAD(const KLT_TYPE maxres) 
    { *_MaxResiduumMAD=maxres; }

    inline void SetMaxIterations(const int maxiter){ *_MaxIterations=maxiter; }
    
    inline void SetHalfWinSize(const int hws) { *_HalfWinSize=hws; }
    
    /// these are needed in the highlevel Tracker class
    inline int *GetMaxIterationsPointer() {return _MaxIterations;};
    inline int *GetHalfWinSizePointer() {return _HalfWinSize;};
    inline KLT_TYPE *GetMaxErrorPointer() {return _MaxError;};
    inline int *GetTrackerType() { return _TrackerType;};
    inline int *GetRejectionType() { return _RejectionType; };

  protected:
    bool _ParamsAllocated;
    int *_MaxIterations; /// iteration stops after _MaxIterations
    int *_HalfWinSize; /// use support window of size 2*hws+1
    /// iteration stops if position refinement is less than *_Maxerror
    KLT_TYPE *_MaxError; 
    /// iteration stops if position refinement is less than *_Maxerror
    KLT_TYPE *_MaxAffineChange; 

    /// outlier rejection via MAD 
    KLT_TYPE *_MaxResiduumMAD; 
    /// the tracker type TT_Simple and TT_Weighted
    int *_TrackerType;
    /// The rejection type: valid RT_MAD and RT_X86
    int *_RejectionType;

    ///////////////////////////////////////////////////////////////////////////
    /// the image data arrays
    StorageType **_ida1, **_ida2;
    /// pointers to images
    Image<StorageType> *_im1, *_im2, 
      *_gradim1x, *_gradim2x, *_gradim1y, *_gradim2y;

    /// gradients image data arrays
    StorageType **_gradx1, **_grady1, **_gradx2, **_grady2;
    /// the image borders modulo filter and support windows or ROI
    /// for image 1 and image 2
    int _maxx1, _maxy1, _minx1, _miny1;
    int _maxx2, _maxy2, _minx2, _miny2;
    ///remember last support window size (for recalculation of gaussian weight)
    int _LastHalfWinSize, _WinSize;
    /// used in iteration
    Matrix2x2<KLT_TYPE> _G, _Ginv;
    Vector2<KLT_TYPE> _e, _d, _disp;
    Matrix<KLT_TYPE> _gx, _gy, _bl1, _bl2, _gx1, _gy1, _gx2, _gy2, _gsx, _gsy;
    KLT_TYPE _gxx, _gxy, _gyy, _idet;
    Matrix<KLT_TYPE> _weight; // the weight matrix used in TrackWeighted_

    /// used in RejectX86_
    Median1D<KLT_TYPE> _Median; // median calculation class
    std::vector<KLT_TYPE> _all_ssd_residui; // sum of sq. grey value diff
    /// sum of sq. grey value diff only for points where tracking succeeded
    std::vector<KLT_TYPE> _ssd_residui; 
    KLT_TYPE _median_ssd_residui; // median of _ssd_residui
    /// _diff_from_median = fabs(_ssd_residui - _median_ssd_residui)
    std::vector<KLT_TYPE> _diff_from_median;
    /// same as _diff_from_median but also for points where tracking was not
    /// sucessfull
    std::vector<KLT_TYPE> _all_diff_from_median;
    /// the rejection threshold calculated by X86 
    /// _threshold = 5.2 * median(_diff_from_median)
    KLT_TYPE _threshold; 

    /// results after affine tracking
    std::vector<Matrix2x2<KLT_TYPE> > _AffineResults;

    ///////////////////////////////////////////////////////////////////
    //  initialization
    ///////////////////////////////////////////////////////////////////

    /** @brief set member pointers to NULL */
    void InitMembers_();

    /// minx/y maxx/y defines a ROI
    void Init_(StorageType **ida1, StorageType **ida2, 
               StorageType **gradx1, StorageType **grady1, 
               StorageType **gradx2, StorageType **grady2, 
               unsigned minx1, unsigned miny1, unsigned maxx1, unsigned maxy1,
               unsigned minx2, unsigned miny2, unsigned maxx2, unsigned maxy2);

    ///////////////////////////////////////////////////////////////////
    //  tracking
    ///////////////////////////////////////////////////////////////////

    /** The correspondence calculation is computed using an iterative gradient
        descent algorithm. See Lukas and Kanade: "An iterative image 
        registration technique with an application to stereo vision", 1986 
        for details. The symmetric approach as described by Stan Birchfield 
        is implemented.
        @author woelk */
    int TrackSimple_(KLT_TYPE p1[2], KLT_TYPE p2[2], KLT_TYPE result[2], 
                     KLT_TYPE& error, int &iter);

    /** Same as TrackSimple_ but also uses the weighting matrix _weight.
        @author woelk */
    int TrackWeighted_(KLT_TYPE p1[2], KLT_TYPE p2[2], KLT_TYPE result[2], 
                       KLT_TYPE& error, int &iter);

    /** @brief track using affine warp
        @param p1 point in image 1
        @param p2 prediction for point in image 2
        @param p2A prediction for affine warp
        @param result result after tracking
        @param resultA affine warp after tracking
        @param error maxnorm of last iteration step
        @param iter number of iterations done
        @param Cov resulting covariance
        @param sad sum of absolute differences
        @param ssd sum of squared differences
        @param Only4 only estimate A, not dx,dy
        @param ComputeCovariance, if true Cov is computed using svd

        This implementation is based on Thorsten Thormaelen's version,
        see cpp file.

        @todo really compute cov with svd ???, conceptual problem: where to 
        compute residui, where covariance ???

        Not evaluated, tested or optimized by now !
        @author koeser 06/2005
    */
    int TrackAffine_(KLT_TYPE p1[2], KLT_TYPE p2[2], 
                     const Matrix2x2<KLT_TYPE>& p2A,
                     KLT_TYPE result[2],  Matrix2x2<KLT_TYPE>& resultA,
                     KLT_TYPE& error, int &iter, Matrix<KLT_TYPE>& Cov,
                     KLT_TYPE& mad, KLT_TYPE& msd,
                     bool Only4 = false, bool ComputeCovariance = true);
				



    // ------------ affine worker functions ----------------------------
    void _am_getGradientWinAffine(const KLT_TYPE& x, const KLT_TYPE& y,
                                  const KLT_TYPE& Axx, const KLT_TYPE& Ayx,
                                  const KLT_TYPE& Axy, const KLT_TYPE& Ayy,   
                                  const int halfwidth, const int halfheight, 
                                  KLT_TYPE* out_gradx,     
                                  KLT_TYPE* out_grady);   

    void _am_computeIntensityDifferenceAffine(Image<StorageType>& img1,
                                              Image<StorageType>& img2,
                                              const KLT_TYPE& x1, 
                                              const KLT_TYPE& y1,    
                                              const KLT_TYPE& x2, 
                                              const KLT_TYPE& y2,     
                                              const KLT_TYPE& Axx, 
                                              const KLT_TYPE& Ayx, 
                                              const KLT_TYPE& Axy, 
                                              const KLT_TYPE& Ayy,   
                                              const int halfwidth,
                                              const int halfheight, 
                                              KLT_TYPE* imgdiff);  

    // -------------------------------------------------------------------


    /* TODO inverse compositional approach */
    //int TrackBaker_(KLT_TYPE p1[2], KLT_TYPE p2[2], KLT_TYPE result[2], 
    //                  KLT_TYPE& error, int &iter);

    /* TODO compensation for illumination changes */
    //int RefineIllum_();


    ///////////////////////////////////////////////////////////////////
    //  calculation of cov, sad and ssd
    ///////////////////////////////////////////////////////////////////


    /** Uses _bl1, _bl2, _gxx, _gxy and _gyy to calculate the residui
        and covarainec matrix and must hence be called immediatly after the 
        Track*_() function call!!
        @author woelk */
    int EvaluateResult_(int& res,KLT_TYPE& mad,KLT_TYPE& msd, KLT_TYPE cov[4]);

    ///////////////////////////////////////////////////////////////////
    //  outlier rejection
    ///////////////////////////////////////////////////////////////////

    /** 
        @author woelk */
    int RejectMAD_(std::vector<KLT_TYPE>& sad, std::vector<int>& res);

    /** 
        @author woelk */
    int RejectX86_(std::vector<KLT_TYPE>& ssd, std::vector<int>& res);

    ///////////////////////////////////////////////////////////////////
    //  helper functions
    ///////////////////////////////////////////////////////////////////
    
    /** resizes the internal matrices _gx1, _gy1, _bl1, _gx2, _gy2 and _bl2
        @author woelk */
    inline void Resize_(const int halfwinsize);

    /** 
        @author woelk */
    void AddParam_(Param& para, const int grp, const std::string& postfix);
    
    /** Constructs the weight matrix _weight with size _WinSize, _WinSize 
        and fills it with gaussian weights
        @author woelk */
    void CalcGaussianWeightMatrix_();

    ////////////////////////////////////////////////////////////////////
    //  bilinear interpolation functions 
    ////////////////////////////////////////////////////////////////////
    
    /// variables needed in MyBilienarRegion?_
    unsigned rf, rc, cf, cc, r, c, x_floor, y_floor, x_floor_end, y_floor_end;
    /// variables needed in MyBilienarRegion?_
    KLT_TYPE ul, ur, ll, lr, brdx, brdy, brdxy;
    
    /** Fills the matrices _gx1, _gy1 and _bl1 by interpolating 
        _gradx1, _grady1 and _ida1 at positions between (x-hws, y-hws) and
        (x+hws, y+hws).
        @author woelk */
    inline void MyBilinearRegion1_(KLT_TYPE x, KLT_TYPE y, int hws);

    /** Fills the matrices _gx2, _gy2 and _bl2 by interpolating 
        _gradx2, _grady2 and _ida2 at positions between (x-hws, y-hws) and
        (x+hws, y+hws).
        @author woelk */
    inline void MyBilinearRegion2_(KLT_TYPE x, KLT_TYPE y, int hws);


    ///////////////////////////////////////////////////////////////////////
    // debugging functions
    ///////////////////////////////////////////////////////////////////////


  }; // class

  ////////////////////////////////////////////////////////////////////
  //  inline functions
  ////////////////////////////////////////////////////////////////////

  template <class StorageType>
  inline void TrackerBase<StorageType>::
  MyBilinearRegion1_(KLT_TYPE x, KLT_TYPE y, int hws)
  {
    x_floor = (unsigned int)floor(x);
    y_floor = (unsigned int)floor(y);
    x_floor_end = x_floor+hws+1;
    y_floor_end = y_floor+hws+1;
    brdy = y - (KLT_TYPE)(y_floor);
    brdx = x - (KLT_TYPE)(x_floor);
    brdxy = brdx*brdy;
    for (rf=y_floor-hws, rc=rf+1, r=0; rf<y_floor_end; rf++, rc++, r++){
      for (cf=x_floor-hws, cc=cf+1, c=0; cf<x_floor_end; cf++, cc++, c++){
        ul = (KLT_TYPE)this->_ida1[rf][cf];
        ur = (KLT_TYPE)this->_ida1[rf][cc];
        ll = (KLT_TYPE)this->_ida1[rc][cf];
        lr = (KLT_TYPE)this->_ida1[rc][cc];
        _bl1[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(this->_gradx1[rf][cf]);
        ur = (KLT_TYPE)(this->_gradx1[rf][cc]);
        ll = (KLT_TYPE)(this->_gradx1[rc][cf]);
        lr = (KLT_TYPE)(this->_gradx1[rc][cc]);
        _gx1[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(this->_grady1[rf][cf]);
        ur = (KLT_TYPE)(this->_grady1[rf][cc]);
        ll = (KLT_TYPE)(this->_grady1[rc][cf]);
        lr = (KLT_TYPE)(this->_grady1[rc][cc]);
        _gy1[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
      }
    }
  }

  template <class StorageType>
  inline void TrackerBase<StorageType>::
  MyBilinearRegion2_(KLT_TYPE x, KLT_TYPE y, int hws)
  {
    x_floor = (unsigned int)floor(x);
    y_floor = (unsigned int)floor(y);
    x_floor_end = x_floor+hws+1;
    y_floor_end = y_floor+hws+1;
    brdy = y - (KLT_TYPE)(y_floor);
    brdx = x - (KLT_TYPE)(x_floor);
    brdxy = brdx*brdy;
    for (rf=y_floor-hws, rc=rf+1, r=0; rf<y_floor_end; rf++, rc++, r++){
      for (cf=x_floor-hws, cc=cf+1, c=0; cf<x_floor_end; cf++, cc++, c++){
        ul = (KLT_TYPE)this->_ida2[rf][cf];
        ur = (KLT_TYPE)this->_ida2[rf][cc];
        ll = (KLT_TYPE)this->_ida2[rc][cf];
        lr = (KLT_TYPE)this->_ida2[rc][cc];
        _bl2[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(this->_gradx2[rf][cf]);
        ur = (KLT_TYPE)(this->_gradx2[rf][cc]);
        ll = (KLT_TYPE)(this->_gradx2[rc][cf]);
        lr = (KLT_TYPE)(this->_gradx2[rc][cc]);
        _gx2[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(this->_grady2[rf][cf]);
        ur = (KLT_TYPE)(this->_grady2[rf][cc]);
        ll = (KLT_TYPE)(this->_grady2[rc][cf]);
        lr = (KLT_TYPE)(this->_grady2[rc][cc]);
        _gy2[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
      }
    }
  }

  template <class StorageType>
  inline void TrackerBase<StorageType>::
  Resize_(const int halfwinsize)
  {
    _LastHalfWinSize=halfwinsize;
    _WinSize=(halfwinsize<<1)+1;
    _gx1.newsize(_WinSize, _WinSize);
    _gy1.newsize(_WinSize, _WinSize);
    _gx2.newsize(_WinSize, _WinSize);
    _gy2.newsize(_WinSize, _WinSize);
    _gsx.newsize(_WinSize, _WinSize);
    _gsy.newsize(_WinSize, _WinSize);
    _bl1.newsize(_WinSize, _WinSize);
    _bl2.newsize(_WinSize, _WinSize);
  
  }


} // namespace

void _am_compute4by1ErrorVector(KLT_TYPE* imgdiff,
                                KLT_TYPE* gradx,
                                KLT_TYPE* grady,
                                const int hw, 
                                const int hh,
                                BIAS::Vector<KLT_TYPE>& e);

void _am_compute4by4GradientMatrix(KLT_TYPE* gradx,
                                   KLT_TYPE* grady,
                                   const int hw,
                                   const int hh,
                                   BIAS::Matrix<KLT_TYPE>& T);

int  _am_gauss_jordan_elimination(BIAS::Matrix<KLT_TYPE>& a,
			     BIAS::Vector<KLT_TYPE>& b);
void _am_compute6by1ErrorVector(KLT_TYPE* imgdiff,
                                KLT_TYPE* gradx,
                                KLT_TYPE* grady,
                                const int hw,
                                const int hh,
                                BIAS::Vector<KLT_TYPE>& e);

void _am_compute6by6GradientMatrix(KLT_TYPE* gradx,
                                   KLT_TYPE* grady,
                                   const int hw, 
                                   const int hh,
                                   BIAS::Matrix<KLT_TYPE>& T);




#include <Base/Common/BIASpragmaEnd.hh>

#endif // __TrackerBase__hh__
