/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __ExampleRANSAC_Plane_hh__
#define __ExampleRANSAC_Plane_hh__

#include <Base/Common/BIASpragmaStart.hh>


#include "MathAlgo/RANSAC.hh"
#include "MathAlgo/SVD.hh"
#include "Base/Geometry/HomgPoint3D.hh"
#include "Base/Geometry/HomgPlane3D.hh"

#include <string>

namespace BIAS {

  class PlaneRANSAC : public RANSAC<HomgPlane3D>
  {
    /** this is a class evaluating the mean of the 
        double data assuming some outlier 
        using the RANSAC algorithm */
  public:
    inline PlaneRANSAC();

    inline ~PlaneRANSAC();

    inline void Init(std::vector<HomgPoint3D>& data, unsigned int data_length, 
                     double inlier_distance_threshold, bool refine_solution, 
                     unsigned int sample_size, double expected_inlier_fraction,
                     double solution_error_threshold);

    inline int GetSampleSolutions(std::vector<unsigned int> &which_samples,
                                  std::vector<HomgPlane3D> &solutions);

    inline bool EvaluateSolution(const HomgPlane3D& solution,  
                                 bool existing_solution_flag,
                                 std::vector<bool>& inlier,    
                                 int& accept_count,
                                 double& evaluate_score,
                                 bool& ok_to_terminate_flag);

    inline bool RefineSolution(std::vector<unsigned int>& which_samples,
                               HomgPlane3D& solution,
                               std::vector<bool>& new_inliers);

  private:
    std::vector<HomgPoint3D> _vData; 
    // data  closer to solution than _dInlierDistanceThreshold will be 
    // regarded as inlying
    double _dInlierDistanceThreshold;
    // if a sample yields more than _dExpectedInlierFraction inliers and 
    // the evaluate_score is lower than _dSolutionErrorThreshold, 
    // RANSACing will stop
    double _dExpectedInlierFraction;
    double _dSolutionErrorThreshold;
  };
  
  PlaneRANSAC::PlaneRANSAC() : RANSAC<HomgPlane3D>()
  {}

  PlaneRANSAC::~PlaneRANSAC() 
  {}

  void PlaneRANSAC::Init(std::vector<HomgPoint3D>& data, 
                          unsigned int data_length, 
                          double inlier_distance_threshold, 
                          bool refine_solution, unsigned int sample_size,
                          double expected_inlier_fraction, 
                          double solution_error_threshold)
  {
    RANSAC<HomgPlane3D>::Init(sample_size, 3, refine_solution);
    _vData=data;
    _uiDataSize=data_length;
    _dInlierDistanceThreshold=inlier_distance_threshold;
    _dExpectedInlierFraction=expected_inlier_fraction;
    _dSolutionErrorThreshold=solution_error_threshold;
  }
  
  int 
  PlaneRANSAC::GetSampleSolutions(std::vector<unsigned int> &which_samples,
                                   std::vector<HomgPlane3D> &solutions)
  {
    solutions.clear();
    solutions.push_back(HomgPlane3D(_vData[which_samples[0]],
                                    _vData[which_samples[1]],
                                    _vData[which_samples[2]]));

    return solutions.size();
  }

  bool PlaneRANSAC::EvaluateSolution(const HomgPlane3D& solution,  
                                      bool existing_solution_flag,
                                      std::vector<bool>& inlier,    
                                      int& accept_count,
                                      double& evaluate_score,
                                      bool& ok_to_terminate_flag)
  {
    bool good_flag=true;
    inlier.clear();
    inlier.resize(_uiDataSize);

    int my_accept_count = 0;
    evaluate_score = 0.0;

    for (unsigned int i=0; i<_uiDataSize; i++){
      inlier[i] = (solution.DistanceSquared(_vData[i]) < 
                   _dInlierDistanceThreshold);
      if (inlier[i]) {
        my_accept_count++;
        evaluate_score += solution.DistanceSquared(_vData[i]);
        if (existing_solution_flag && my_accept_count>accept_count) break;
      }
    }
    if (my_accept_count!=0) {
      evaluate_score/=my_accept_count;
    } else {
      evaluate_score=DBL_MAX;
      good_flag=false;
    }
    accept_count=my_accept_count;
    
    ok_to_terminate_flag=
      ((((double)accept_count/(double)_uiDataSize) > _dExpectedInlierFraction)
      && (evaluate_score < _dSolutionErrorThreshold));
    std::string bo=(ok_to_terminate_flag)?("true"):("false");
   
    
    return good_flag;
  }

  bool PlaneRANSAC::RefineSolution(std::vector<unsigned int>& which_samples,
                                    HomgPlane3D& solution,
                                    std::vector<bool>& new_inliers
                                    )
  {
    BIAS::Matrix<double>  A;

    // fill the 3d  points rowwise into A using the scheme
    const unsigned int nr_of_points = which_samples.size();
    std::cout<<"RefineSolution has "<<nr_of_points<<" inliers"<<std::endl;
    A.newsize( nr_of_points, 4 );
  
    for (unsigned int i=0; i<nr_of_points; i++) {
        //        -1            X_i             Y_i             Z_i 
      A[i][0] = _vData[which_samples[i]][0];
      A[i][1] = _vData[which_samples[i]][1];
      A[i][2] = _vData[which_samples[i]][2];    
      A[i][3] = _vData[which_samples[i]][3];
    }
    A = A.Transpose()*A;
    // get a solution (linear fit) of this equation system by its SVD
    SVD svd(A , 1e-5, false);
  
  // cout<<"U = " << svd.GetU() <<endl;
  // cout<<"S = " << svd.GetS() <<endl;
  // cout<<"V = " << svd.GetV() <<endl;
  // cout<<"Nullspace dim = " << svd.NullspaceDim() <<endl;;
  
    Vector4<double> regplane;
    if (svd.Rank()<3) return false;
    regplane = svd.GetNullvector(); // get the last Nullvector  
    solution = HomgPlane3D(regplane);
    // cout <<"solution: " << solution << endl;
  
    //  Vector3<double> n_vec;
    //double d;
  // decompose the solution into HNF of a plane
  //d = solution[0];
  //n_vec[0] = solution[1];
  //n_vec[1] = solution[2];
  //n_vec[2] = solution[3]; 
  // now (d, n_vec ) is a plane in normalform
  
  // convert to HNF representation: 
  //d = d / n_vec.norm_L2(); // divide by actual length for HNF
  //n_vec.normalize_length();// normalize n_vec to length 1 (for HNF)
  
  // cout << " plane in HNF:  d = "<<d<<endl;
  // cout << " n_vce'0 : " << n_vec << endl;  
 
    return true;
  }

} // namespace 

#include <Base/Common/BIASpragmaEnd.hh>


#endif // __ExampleRANSAC_double_hh__
