/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __Minpack_hh__
#define __Minpack_hh__

#include "bias_config.h" 

#include <Base/Math/Matrix.hh>
#include <Base/Math/Vector.hh>

#include "Lapack.hh"

#define MINPACK_DEF_TOLERANCE 1e-7
#define LM_DEF_MAX_ITER -1


#define LM_IMPROPER_INPUT -1
#define LM_OK 0 
#define LM_TOO_MANY_IT 5
#define LM_SMALL_FTOL  6.
#define LM_SMALL_XTOL  7
#define LM_SMALL_GTOL  8

namespace BIAS{
/** @addtogroup mathalgo
    \brief C++ Wrappers for important Lapack fortran77 functions
    @{
*/

  /** Solve a system of n non-linear equations with n unkowns using the
      powell hybrid method. Jacobian is calculated numerical by 
      forward-difference approximation. The error is assumed to be in the 
      range of machine precision.
      See minpack/hybrd1.f
      @author woelk
      @param fcn is the name of the user-supplied subroutine which
             calculates the functions at x and returns the values in fvec.
             fcn must be declared e.g. in the user calling program

             @param iflag should not be changed by fcn unless
             the user wants to terminate execution, in this case set iflag 
             to a negative integer.

             @param n is a positive integer input variable set to the number
             of functions and variables.

             @param x is an array of length n. on input x must contain
             an initial estimate of the solution vector. on output x
             contains the final estimate of the solution vector.

             @param fvec is an output array of length n which contains
             the functions evaluated at the output x.

       @param Tolerance is a nonnegative input variable. termination occurs
       when the algorithm estimates that the relative error
       between x and the solution is at most Tolerance.

       @return 0 if everything worked fine

       if the user has terminated execution, return is the value of iflag,
       see description of fcn. Otherwise, return means
         - return=-1   improper input parameters.
         - return=0   algorithm estimates that the relative error
                      between x and the solution is at most tol.
         - return=2   number of calls to fcn has reached or exceeded 200*(n+1)
         - return=3   Tolerance is too small. no further improvement in
                      the approximate solution x is possible.
         - return=4   iteration is not making good progress.
  */
	BIASMathAlgo_EXPORT
  int Powell(void (*fcn)(int *n, double *x, double *fvec, int *iflag), 
             Vector<double>& InitialGuess, Vector<double>& Result, 
             double Tolerance=MINPACK_DEF_TOLERANCE);

  /** solve a system of n non-linear equations with n unkowns using the
      powell hybrid method. Jacobian is aproximated numericaly by 
      forward-difference approximation.
      See Powell for further parameter documentation
      See minpack/hybrd.f
      @author woelk */
	BIASMathAlgo_EXPORT
  int PowellExtended(void (*fcn)(int *n, double *x, double *fvec, int *iflag), 
                     Vector<double>& InitialGuess, Vector<double>& Result, 
                     double EpsFcn, double Tolerance=MINPACK_DEF_TOLERANCE);
  
  /** Solves an overdetermined system of n non-linear equations with m 
      parameters using the Levenberg-Marquardt algorithm. A least squares 
      solution is used. Jacobian is given analytically.
      
      @param fcn minimization function, where m(=sizeof fvec) and n(=sizeof x)
             are the numbers of residuals/vars passed via fvec/x
      @param m number of residuals, m>=n
      @param n number of variables, m>=n
      @param InitialGuess start point, size must be n
      @param Result solution after optimization (output), size must be n

      @return 0 means success in optimization  
      
      return values in detail:
        - return = -1: Improper input parameters.
        - return =  0: a termination condition became true
        - return =  5: Number of calls to fcn has reached or exceeded maxfev.
        - return =  6: ftol is too small. No further reduction in sum of 
                       squares possible.
        - return =  7: xtol is too small. No further improvement in
                       approximate solution x is possible.
        - return =  8: gtol is too small. fvec is orthogonal to the columns of 
                       the jacobian to machine precision.

      See minpack/lmder1.f 
      @author woelk */
	BIASMathAlgo_EXPORT
  int LevenbergMarquardt(void (*fcn)(int *m, int *n, double *x, 
				     double *fvec, double *fjac, 
				     int *ldfjac, int *iflag), 
                         int m, int n, Vector<double>& InitialGuess, 
                         Vector<double>& Result,
                         double Tolerance=MINPACK_DEF_TOLERANCE);
  
  /** Solves an overdetermined system of n non-linear equations with m 
      parameters using the Levenberg-Marquardt algorithm. A least squares 
      solution is used. Jacobian is given analytically.
      See minpack/lmder.f 
      @author woelk */
	BIASMathAlgo_EXPORT
  int LevenbergMarquardtExtended(void (*fcn)(int *m, int *n, double *x, 
                                             double *fvec, double *fjac, 
                                             int *ldfjac, int *iflag), 
                                 int m, int n, Vector<double>& InitialGuess, 
                                 Vector<double>& Result, 
                                 double Tolerance=MINPACK_DEF_TOLERANCE);

  /** Solves an overdetermined system of n non-linear equations with m 
      parameters using the Levenberg-Marquardt algorithm. A least squares 
      solution is used. Jacobian is calculated numericaly by 
      forward-difference approximation.
      See minpack/lmdif1.f 
      @param MaxIter max number of error-function evaluations. Note that the 
      complete numerical approximation of the jacobian is always done as a 
      whole before this value is checked. Set to -1 if you want to use the 
      default number of the fortran code
      @author woelk */
 BIASMathAlgo_EXPORT
 int LevenbergMarquardt(void (*fcn)(int *m, int *n, double *x, 
                                     double *fvec, int *iflag), 
                         int M, int N, Vector<double>& InitialGuess, 
                         Vector<double>& Result, 
                         double Tolerance=MINPACK_DEF_TOLERANCE,
                         int MaxIter=LM_DEF_MAX_ITER);

  /** Solves an overdetermined system of n non-linear equations with m 
      parameters using the Levenberg-Marquardt algorithm. A least squares 
      solution is used. Jacobian is calculated numericaly by 
      forward-difference approximation.
      See minpack/lmdif.f 
      @param ResultingJacobian pass NULL if you are not interested in the 
      jacobian at the minimum, otherwise pass a pointer to double [n*m]
      @param ipvt pass NULL if you are not interested in the 
      jacobian at the minimum otherwise pass a pointer to int [n]
      @author woelk */
  BIASMathAlgo_EXPORT
int LevenbergMarquardtExtended(void (*fcn)(int *m, int *n, double *x, 
                                             double *fvec, int *iflag), 
                                 int M, int N, Vector<double>& InitialGuess, 
                                 Vector<double>& Result, double EpsFcn,
                                 double Tolerance=MINPACK_DEF_TOLERANCE,
                                 int MaxIter=LM_DEF_MAX_ITER,
                                 double* ResultingJacobian=NULL,
                                 int* ipvt=NULL, 
                                 double *SumOfSquaredErrors=NULL);
  // @}
} // namespace


#endif // __Minpack_hh__
