/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __RANSAC_hh__
#define __RANSAC_hh__

#include <Base/Common/BIASpragmaStart.hh>

#include <Base/Debug/Debug.hh>
#include <Base/Math/Random.hh>	
#include <Base/Geometry/HomgPoint2D.hh>	
#include <Base/Math/Matrix.hh>	

#include <algorithm>
#include <float.h>
#include <iterator>
#include <vector>

#define D_RANSAC_SAMPLES      0x0001
#define D_RANSAC_SOLVE_MASTER 0x0002

#define RANSAC_DEFAULT_PROBOFGOODSOLUTION 0.98
#define RANSAC_DEFAULT_SOSMALLSOLUTIONCHANGE 0.02

namespace BIAS {
  /** @class RANSAC
      @ingroup g_mathalgo
      @brief generic abstract RANSAC class,  RANSAC - RANdom SAmple Consensus
      
      To use this class derive a class from it in which
      - GetSampleSolutions(...)
      - EvaluateSolution(...)
      - RefineSolution(...)
      are overloaded.
      GenerateSamples(...) can be overloaded if for instance an evenspaced 
      sampling is desired.
         
      For an example of how to use this class see 
      ./Examples/ExampleRANSAC_double.hh and ./Examples/ExampleRANSAC.cpp
       
      the function SolveMaster was ported from TargetJR -> PMatrixRANSAC by 
      frahm and from there on further ported to this class from woelk

      Since this is a generic templated class the implementation has to go 
      into the header file.

      @author frahm, woelk */  

  template <class SolutionType>
  class BIASMathAlgo_EXPORT RANSAC : public Debug 
  {
    
  public:
    RANSAC();
  
    /** Construct a RANSAC object that takes samples of size sample_size,
        returning at most max_solutions_per_sample solutions each time.
        If the refine_solution flag is true, a routine has been supplied to
        refine the initial fits before evaluation. */
    RANSAC(unsigned int sample_size, unsigned int max_solutions_per_sample = 1,
           bool refine_solution = false);
  
    virtual ~RANSAC();

    void Init(unsigned int sample_size,unsigned int max_solutions_per_sample=1,
              bool refine_solution = false);

    /** @brief Explicitly set max number of samples
        
        Ransacing will stop after max_samples samples, regardless
        of solution quality.  Normally, more than a few thousand samples 
        signifies a problem. */
    void SetMaxSamples(unsigned int max_samples) 
    { _uiMaxSamples = max_samples; }

    /** @brief Explicitly set number of samples

        Explicitly set number of samples to make in SolveMaster, i.e.
        Value of ComputeSampleCount() is ignored if this function is called
        with a strictly positive value uiSamples.
        Note:
        SolveMaster performs at most min(_uiMaxSamples, uiSamples) iterations
        @author koeser
     */
    void SetSampleCount(unsigned int uiSamples) {
      _uiExplicitSampleNo = uiSamples;
    }

    /** @brief MAIN function called by user

        This function is called to run the computation.
        On output, solution is set to the best solution found and
        inliers indicate which data were inliers to that solution.
        The inlying_data_fraction is responsible for the maximal number of 
        sample sets taken from the data. The smaller the 
        inlying_data_fraction, the more sample sets are taken 

        @attention Do NOT override this function in derived classes, unless
        you want to do modifications in the RANSAC _algorithm_ itself (and 
        usually you dont want to do this ...)
        input : inlying_data_fraction   expected inlier fraction
        output: solution                solution
                inliers                 inlier flags 
        @return <0 indicates an error, otherwise number of chosen sample
        @author zisserman, frahm, woelk */
    virtual int SolveMaster(double inlying_data_fraction,
                            SolutionType &solution,
                            std::vector<bool> &inliers);
  
    /** @brief Compute solution(s) for the given set of samples
        @return the number of solutions found.

        The which_samples array will be of length sample_size, selecting the
        indices to be fit to from the subclass's data array. */
    virtual int GetSampleSolutions(std::vector<unsigned int> &which_samples,
                                   std::vector<SolutionType> &solutions);

    /** @brief evaluate the goodness of a solution
     *
     *  For the specified solution, set
     *  - an accept count indicating the number of data which are inliers 
     *    to the given solution,
     *  - the inlier flag for each datum to true if the datum is inlying
     *  - an evaluate score (typically the average residual) indicating 
     *    the goodness of fit of the inliers - the solver code requires a 
     *    positive number where zero is a perfect solution, larger positive
     *    scores mean a worse solution.
     *  - an ok_to_terminate flag which is set TRUE if the solution
     *    is good enough that the robust solver can discontinue search for
     *    a better solution.
     *  
     *  If existing_solution_flag is true, then the best accept count 
     *  so far is in accept_count.  The function can terminate early if
     *  it is not going to beat this.
     *
     *  A skeleton implementation of this function would be written as follows:
     *  \verbatim
     *  int my_accept_count = 0;
     *  for (unsigned int i=0; i<_uiDataSize; i++) {
     *    inlier[i] = Consistent(corner_matches[i], solution);
     *    if (inlier[i]) {
     *      my_accept_count++;
     *      evaluate_score += Residual(corner_matches[i], solution);
     *      if (existing_solution_flag && my_accept_count>accept_count) 
     *        break;
     *    }
     *  }
     *  accept_count=my_accept_count;
     *  evaluate_score=(accept_count!=0)?(evaluate_score/accept_count):
     *    (DBL_MAX);
     *  if (evaluate_score < solution_error_threshold && 
     *      (double)accept_count/(double)_uiDataSize > min_inlier_fraction) 
     *    ok_to_terminate_flag=true; 
     *  else 
     *    ok_to_terminate_flag=false;
     *  return (accept_count!=0) // maybe even (accept_count>inl_frac)
     *  \endverbatim
     *
     *  @param solution (in) solution to be evaluated
     *  @param existing_solution_flag (in) if true the best accept_cout so far 
     *                                is given in accept_count
     *  @param accept_count (in/out) input: best accept_cout so far if 
     *                      existing_solution_flag==true, output: number of 
     *                      inliers to the current solution
     *  @param inlier (out) inlier[i] ==> datum i is inlying 
     *  @param evaluate_score (out) a score for the solution, 0 means ideal 
     *                        solution, larger score means worse solution
     *  @param ok_to_termiate_flag (out) if true, RANSACing will stop */
    virtual 
    bool EvaluateSolution(const SolutionType& solution,
                          bool existing_solution_flag,
                          std::vector<bool>& inlier, int& accept_count,
                          double& evaluate_score, bool& ok_to_terminate_flag);
    
    /** @brief refine previously computed solution based on its inliers

        If implemented, this function should compute a more accurate solution,
        e.g. a least squares on all inliers, possibly using as an initial 
        estimate the solution in solution.
        The output array new_inliers should be set to reflect the inliers to 
        the new solution. 
        @return false if no better solution than the initial could be found
    */
    virtual 
    bool RefineSolution(std::vector<unsigned int>& which_samples,
                        SolutionType& solution,
                        std::vector<bool>& new_inliers);

    /** @brief pick a set of samples

        Override this to specialize the technique used to generate samples.
        By default, random sampling is assumed but implementations may wish to 
        use the latin-square sampling provided by generate_sample_evenspace 
        (below), or some other problem-specific sampling strategy. */
    virtual 
    bool GenerateSamples(int sample_index, 
                         std::vector<unsigned int>& which_samples);
  
    bool GenerateSamplesEvenspace(int sample_index, 
                                  std::vector<unsigned int>& sample_table);

    bool GenerateSamplesRandom(int sample_index,
                               std::vector<unsigned int>& sampletable);

    // Return the best EvaluateScore obtained by a previous call of SolveMaster
    double GetFinalScore() const { return _dBestSampleEvaluateScore; }

  protected:
    /// number of samples available
    unsigned _uiSampleSize;

    /// max number of solutions that are to be computed for one sample set
    unsigned _uiMaxSolutionsPerSample;
    
    /// number of data elements needed to compute a solution
    unsigned _uiDataSize;
    
    /// do we want to refine a good sample
    bool _bRefineSolution;

    /// sampled best values
    int    _iBestSampleAcceptCount;
    double _dBestSampleEvaluateScore;

    /// refined best values
    int    _iBestInlierAcceptCount;
    double _dBestInlierEvaluateScore;

    /// absolute max number of samples
    unsigned int _uiMaxSamples;

    /// use this number of samples
    unsigned int _uiExplicitSampleNo;

    /// prob. for computing number of samples, set close to 1
    double _dProbabilityOfGoodSolution;
    
    /// If all other conditions are right, and the fractional change in the 
    /// solution statistics is less than this, stop sampling.
    double _dSoSmallSolutionChange;

    /// if we already have a solution for which EvaluateSolution returned true
    bool _got_solution_flag;

    /// if true, the first acceptable solution is returned
    bool _greedy;

  private:
    static int ComputeSampleCount(double inlying_data_fraction, 
                                  double desired_prob_good, 
                                  unsigned int sample_size);
  };



  /************************************************************/
  /* now the implementation                                   */
  /************************************************************/


  template <class SolutionType>
  RANSAC<SolutionType>::RANSAC()
  {
    _dProbabilityOfGoodSolution = RANSAC_DEFAULT_PROBOFGOODSOLUTION;
    _dSoSmallSolutionChange = RANSAC_DEFAULT_SOSMALLSOLUTIONCHANGE;
    _uiSampleSize=_uiMaxSolutionsPerSample=_uiDataSize=_uiMaxSamples=0;
    _uiExplicitSampleNo = 0;
    _bRefineSolution = false;
    _iBestSampleAcceptCount   = _iBestInlierAcceptCount   = -1;
    _dBestSampleEvaluateScore = _dBestInlierEvaluateScore = -1.0;
    _got_solution_flag = false;
    _greedy = false;
  }

  template <class SolutionType>
  RANSAC<SolutionType>::RANSAC(unsigned int sample_size, 
                               unsigned int max_solutions_per_sample,
                               bool refine_solution)
  {
    _dProbabilityOfGoodSolution = RANSAC_DEFAULT_PROBOFGOODSOLUTION;
    _dSoSmallSolutionChange = RANSAC_DEFAULT_SOSMALLSOLUTIONCHANGE;
    Init(sample_size, max_solutions_per_sample, refine_solution);
    _uiDataSize = 0;
    _iBestSampleAcceptCount   = _iBestInlierAcceptCount   = -1;
    _dBestSampleEvaluateScore = _dBestInlierEvaluateScore = -1.0;
    _got_solution_flag = false;
    _greedy = false;
  }

  template <class SolutionType>
  void RANSAC<SolutionType>::Init(unsigned int sample_size, 
                                  unsigned int max_solutions_per_sample,
                                  bool refine_solution)
  {
    _uiSampleSize=sample_size;
    _uiMaxSolutionsPerSample=max_solutions_per_sample;
    _bRefineSolution=refine_solution;
    _uiMaxSamples = 0;
    _uiExplicitSampleNo = 0;
  }

  template <class SolutionType>
  RANSAC<SolutionType>::~RANSAC()
  {}

  /* @author frahm, woelk */
  template <class SolutionType>
  int RANSAC<SolutionType>::SolveMaster(double inlying_data_fraction,
                                        SolutionType& solution,
                                        std::vector<bool>& inliers)
  {
#ifdef BIAS_DEBUG
    if (_uiDataSize==0){
      BIASERR("_uiDataSize should be set in derived class first");
      return -1;
    }
    if (inlying_data_fraction>1.0 || inlying_data_fraction<0.0){
      BIASERR("invalid inlying_data_fraction: "<<inlying_data_fraction);
      return -1;
    }
#endif

    if (_uiDataSize < _uiSampleSize) {
      BIASERR("SolveMaster: _uiDataSize < _uiSampleSize !");
      return -1;
    }

    std::vector<bool> inlier_set(_uiDataSize, false);

    std::vector<SolutionType> multisoln_table(_uiMaxSolutionsPerSample);
  
    // check whether _uiExplicitSampleNo has been set (to a positive value)
    int sample_count;
    if (_uiExplicitSampleNo>0) {
      // yes the user specified to make exactly _uiExplicitSampleNo samples
      sample_count = _uiExplicitSampleNo; 
      BIASDOUT(D_RANSAC_SOLVE_MASTER, "RANSAC: Using "<< sample_count
             <<" samples (explicitly specified).");
    }
    else {
      // compute the number of samples needed using Torr's thesis
      sample_count = ComputeSampleCount(inlying_data_fraction, 
                                        _dProbabilityOfGoodSolution,
                                        _uiSampleSize);

      BIASDOUT(D_RANSAC_SOLVE_MASTER, "RANSAC: Need at most "<< sample_count
               <<" samples for "<< inlying_data_fraction <<" inlier fraction");
                
    }

    // Override computed bailout if specified
    if (_uiMaxSamples != 0 && (unsigned)sample_count > _uiMaxSamples) {
      sample_count = _uiMaxSamples;
      BIASDOUT(D_RANSAC_SOLVE_MASTER,"RANSAC: But will bail out after " 
               << sample_count );
    }

    // offsets of samples for current solution
    std::vector<unsigned int> which_samples;

    // number of solutions per sample
    int nsolutions;

    // number of inliers for current solution
    int sample_accept_count; 
    
    // score for sample set(resp. current solution), is set by EvaluateSolution
    double sample_evaluate_score; 

    // found a solution that good that we can finish, set by EvaluateSolution
    bool ok_to_terminate_flag;

    // current solution is accepted by EvaluateSolution
    bool good_flag;

    // number of inliers for current solution after refinement
    int inlier_accept_count;

    // score for current solution after refinement
    double inlier_evaluate_score;

    // indicates that the current is the best solution so far
    bool best_so_far_flag;

    // indicate whether we have ever reached the checkpoint before
    // fast termination can only happen at the second time we reach the 
    // checkpoint (or if _greedy is set)
    bool checkpoint_set_flag = false;
    
    // backup of solution evaluation at last time we were at checkpoint
    int sample_checkpoint_count = 0;
    int inlier_checkpoint_count = 0;

    // if ok_to_terminate is set true, fast termination is enabled if
    // termination_check_active_flag is true.
    bool termination_check_active_flag = false;

    int sample_index;
    for (sample_index = 0; sample_index < sample_count; sample_index++){

      // jmf for skipping if we haven't enough matches
      if (GenerateSamples(sample_index, which_samples)){

        nsolutions = GetSampleSolutions(which_samples, multisoln_table);
        for (int solution_index=0; solution_index<nsolutions;solution_index++){
          SolutionType* active_solution_ptr = &multisoln_table[solution_index];

          if (_got_solution_flag) // ?
            sample_accept_count = _iBestSampleAcceptCount;
        
          ok_to_terminate_flag = false;
          sample_evaluate_score = DBL_MAX;
          good_flag = 
            EvaluateSolution(*active_solution_ptr, _got_solution_flag, 
                             inlier_set, sample_accept_count, 
                             sample_evaluate_score, ok_to_terminate_flag); 

          BIASDOUT(D_RANSAC_SOLVE_MASTER,"sample: "<<sample_index<<"("
                   <<solution_index<<")"
                   <<" solution: "<<(*active_solution_ptr)
                   <<" inliers: " <<sample_accept_count<<"/"<<_uiDataSize
                   <<" score: "<< sample_evaluate_score);        
          inlier_accept_count=-1;
          inlier_evaluate_score = 1e20;
          if (good_flag && _bRefineSolution) { 
            which_samples.clear();
            for (unsigned int i=0; i<_uiDataSize; i++)
              if (inlier_set[i]) which_samples.push_back(i);
            if (RefineSolution(which_samples, *active_solution_ptr, 
                               inlier_set)){
              if (_got_solution_flag)
                inlier_accept_count = _iBestInlierAcceptCount;
              good_flag = 
                EvaluateSolution (*active_solution_ptr, _got_solution_flag,
                                  inlier_set, inlier_accept_count,
                                  inlier_evaluate_score, ok_to_terminate_flag);
              if (good_flag) {
                BIASDOUT(D_RANSAC_SOLVE_MASTER, "RANSAC: minimal/refined "
                         <<sample_accept_count<<"/"
                         <<inlier_accept_count<<" "<<sample_evaluate_score<<"/"
                         <<inlier_evaluate_score);
              }
            } else 
              BIASDOUT(D_RANSAC_SOLVE_MASTER,
                       "RANSAC: (note) refine_solution returned false");
          } else { // if (good_flag && _bRefineSolution) { 
            BIASDOUT(D_RANSAC_SOLVE_MASTER,
                     "RANSAC: (note) either good_flag=false or "
                     <<"_bRefineSolution=false");
          }
	
          // the criteria for deciding the best solution is the following -
          // (a) if the accept_count is higher than the previous best, take it,
          // (b) if the accept_count is equal to the previous best, and 
          //     the evaluate score is smaller than the previous best, take it.
          if (good_flag) {
            best_so_far_flag = false;
	  
            if (!_got_solution_flag) {
              _got_solution_flag = true;
              best_so_far_flag = true;
            } else {
              if (!_bRefineSolution) {
                if (sample_accept_count > _iBestSampleAcceptCount ||
                    (sample_accept_count == _iBestSampleAcceptCount &&
                     sample_evaluate_score < _dBestSampleEvaluateScore))
                  best_so_far_flag = true;
              } else {
                if (inlier_accept_count > _iBestInlierAcceptCount ||
                    (inlier_accept_count == _iBestInlierAcceptCount &&
                     inlier_evaluate_score < _dBestInlierEvaluateScore))
                  best_so_far_flag = true;
              }
            }

            // save all the scores associated with this result.
            // future evaluations will be tested against these accept scores.
            if (best_so_far_flag) {
              _iBestSampleAcceptCount = sample_accept_count;
              _dBestSampleEvaluateScore = 
                (sample_evaluate_score<inlier_evaluate_score) ?
                sample_evaluate_score : inlier_evaluate_score;
              _iBestInlierAcceptCount = inlier_accept_count;
              _dBestInlierEvaluateScore = inlier_evaluate_score;
              
              // save inliers and solution
              inliers = inlier_set;
              solution = *active_solution_ptr;
        
              if (_bRefineSolution){
                BIASDOUT(D_RANSAC_SOLVE_MASTER,"RANSAC: Sample "<<sample_index
                         <<" is best so far.  " << "Inliers/score "
                         <<_iBestInlierAcceptCount << "/" 
                         <<_dBestInlierEvaluateScore<<": "<<solution);
              } else {
                BIASDOUT(D_RANSAC_SOLVE_MASTER, "RANSAC: Sample "<<sample_index
                         <<" is best so far.  " << "Inliers/score "
                         << _iBestSampleAcceptCount << "/"
                         <<_dBestSampleEvaluateScore<<": "<<solution);
              }

              
              if (ok_to_terminate_flag){
                termination_check_active_flag = true;
                BIASDOUT(D_RANSAC_SOLVE_MASTER, 
                         "RANSAC: termination_check_active_flag = true");
              }
            } // if (best_so_far_flag) {
          } // if (good_flag) {
        } // for (int solution_index=0; solution_index<nsolutio
              
        // this is all fast termination stuff
        if (_got_solution_flag) {
          // in greedy mode, we dont want to check the improvement
          if (_greedy && termination_check_active_flag) break;
          if (!checkpoint_set_flag)
            checkpoint_set_flag = true;
          else {
            // check if we have already found such a good solution that we can 
            // terminate (termination_check_active_flag)
            // and if the best-so-far-solution has not increased 
            // significantly compared to the previous solution
            if (!_bRefineSolution) {
              if (termination_check_active_flag &&
                  ((double) (_iBestSampleAcceptCount - sample_checkpoint_count)
                   /(double)sample_checkpoint_count < _dSoSmallSolutionChange))
                break;
            } else {
              if (termination_check_active_flag &&
                  ((double) (_iBestInlierAcceptCount - inlier_checkpoint_count)
                   /(double)inlier_checkpoint_count < _dSoSmallSolutionChange))
                break;
            }
          }
	
          // save the current best values to measure the improvement next time
          inlier_checkpoint_count = _iBestInlierAcceptCount;
          sample_checkpoint_count = _iBestSampleAcceptCount;
        }
      } else { // if (GenerateSamples
        BIASERR("GenerateSamples failed");
      }
    } // for (int sample_index = 0 ...
  
    if (!_got_solution_flag) {
      BIASERR("RANSAC: no solution found\n");
      // prepare return value "failed"
      sample_index = -2;
    } else {
      if (_bRefineSolution){
        BIASDOUT(D_RANSAC_SOLVE_MASTER,"\nRANSAC: using sample "
                 <<" with inliers/score "<<_iBestInlierAcceptCount << "/" 
                 <<_dBestInlierEvaluateScore<<" as solution : "
                 <<solution<<"\n\n");
      } else {
        BIASDOUT(D_RANSAC_SOLVE_MASTER, "\nRANSAC: using sample "
                 <<" with inliers/score "<< _iBestSampleAcceptCount << "/"
                 <<_dBestSampleEvaluateScore<<" as solution : "
                 <<solution<<"\n\n");
      }
    }
    // reset best so far values
    _iBestSampleAcceptCount   = -1;
    _dBestSampleEvaluateScore = -1;
    _iBestInlierAcceptCount   = -1;
    _dBestInlierEvaluateScore = -1;
    _got_solution_flag = false;
    return sample_index;
  }
 

  template <class SolutionType>
  int RANSAC<SolutionType>::
  GetSampleSolutions(std::vector<unsigned int> &which_samples, 
                     std::vector<SolutionType> &solutions)
  {
    BIASERR("this function should be implemented in derived class");
    return -1;
  }


  template <class SolutionType>
  bool RANSAC<SolutionType>::
  EvaluateSolution(const SolutionType& solution, bool existing_solution_flag,
                   std::vector<bool>& out_inlier_flags, int& accept_count_ptr,
                   double& evaluate_score_ptr, bool& ok_to_terminate_flag_ptr)
  {
    BIASERR("this function should be implemented in derived class");
    return false;
  }

  template <class SolutionType>
  bool RANSAC<SolutionType>::
  RefineSolution(std::vector<unsigned int>& which_samples, 
                 SolutionType& solution,
                 std::vector<bool>& new_inliers)
  {
    BIASERR("this function should be implemented in derived class");
    return false;
  }

  template <class SolutionType>
  bool RANSAC<SolutionType>::
  GenerateSamples(int sample_index, std::vector<unsigned int>& which_samples)
  {
    return GenerateSamplesRandom(sample_index, which_samples);
  }

  template <class SolutionType>
  bool RANSAC<SolutionType>::
  GenerateSamplesEvenspace(int sample_index, 
                           std::vector<unsigned int>& which_samples)
  {
    bool res=false;
    BIASERR("unfinished code");
    /*
    // from sa_generate_sample_evenspace
  
    assert((int)_sample_size <= sample_table.length());
    assert((int)_sample_size < data_size);
    
    // basic_offset starts sample_size samples evenly spaced from the data_size set. 
    unsigned basic_offset = data_size / _sample_size; 
    unsigned offset = basic_offset + sample_index / basic_offset;
    unsigned data_index = sample_index % data_size;

    if (offset < basic_offset + data_size) {
    for(unsigned count = 0; count < _sample_size; ++count) {
    sample_table[count] = data_index;
      
    data_index += offset;

    if ((int)data_index >= data_size)
    data_index = data_index % data_size;
    }
    }
    else {
    if (offset == basic_offset + data_size && (sample_index % basic_offset == 0))
    cerr << "RANSAC: Evenspace sampling has reached repetition stage, changing to random sampling\n";
    
    generate_sample_random (sample_index, sample_table);
    }
    */
    return res;
  }

  template <class SolutionType>
  bool RANSAC<SolutionType>::
  GenerateSamplesRandom(int sample_index, 
                        std::vector<unsigned int>& which_samples)
  {
    bool res=true;
#ifdef BIAS_DEBUG
    if (_uiSampleSize==0){
      BIASERR("call Init or appropriate constructor befor using RANSAC");
      res=false;
    }
    if (_uiSampleSize>=_uiDataSize){
      BIASERR("_uiDataSize should be set in derived class befor using RANSAC");
      res=false;
    }
#endif
    // initialzing random with 0 and 1 does not seem to work well
    Random ran(sample_index+rand());
    unsigned int count=0;
    unsigned int index;
    which_samples.clear();
    do {
      index = ran.GetUniformDistributedInt((unsigned int)0, _uiDataSize-1);
      if (find(which_samples.begin(), which_samples.end(), index)==
          which_samples.end()){
        which_samples.push_back(index);
        count++;
      }
    } while (count<_uiSampleSize);

#ifdef BIAS_DEBUG
    if ((GetDebugLevel() & D_RANSAC_SAMPLES)!=0){
      BIASDOUT(D_RANSAC_SAMPLES, "---------- sample "<<sample_index
               <<" ---------------------------------"<<"\nnow using data: ");
      std::ostream_iterator<int> oi(GetDebugStream(), " ");
      std::copy(which_samples.begin(), which_samples.end(), oi);
      GetDebugStream() << std::endl;
    }
#endif

    return res;
  }


  template <class SolutionType>
  int RANSAC<SolutionType>::
  ComputeSampleCount(double inlying_data_fraction, double desired_prob_good, 
                     unsigned int sample_size)
  {
    if (inlying_data_fraction == 1.0)
      // the analytic expression below cant cope with this so do it explicitly.
      return 1;

    // from Torr's thesis.
    return int(log (1.0 - desired_prob_good) /
               log (1.0 - pow (inlying_data_fraction, (double) sample_size)));
  }


} // namespace

#include <Base/Common/BIASpragmaEnd.hh>

#endif   //__RANSAC_hh__
