/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Image/Camera.hh>
#include <Image/CameraParam.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageAttributes.hh>
#include <iostream>
#include <fstream>

#ifdef   WIN32
#include "Base/Common/getopt_W32.h"
#else  //not WIN32
#include <unistd.h>
#endif //WIN32

using namespace std;
using namespace BIAS;


void usage()
{
  cout <<"Usage: biasextract [options] Image ..."<<endl
       <<" Options: "<<endl
       <<"-a      : extract ascii string"<<endl
       <<"-t      : extract timestamp as [sec usec]"<<endl
       <<"-T      : extract timestamp as [msec]"<<endl
       <<"-s      : suppress current filename in output"<<endl
       <<"-P      : extract P-Matrix"<<endl
       <<"-p      : Decompose P-Matrix and extract axis-angle orientation"
       <<endl
       <<"-x      : extract pmatrix as xml data"<<endl
       <<"-K      : extract internal parameter matrix"<<endl
       <<"-R      : extract rotation matrix"<<endl
       <<"-r      : extract axis-angle orientation"<<endl
       <<"-C      : extract camera center position"<<endl
       <<"-A      : extract viewing direction"<<endl	  
       <<"-V      : extract 'down' vector"<<endl	  
       <<"-H      : extract 'right' vector"<<endl	  
       <<"-e      : compute K from EXIF data"<<endl
       <<"-c      : compute CameraParam from KMatrix in header"<<endl
       <<"-E      : compute CameraParam from EXIF data"<<endl
       <<"-u      : extract UID"<<endl
       <<"-O      : extract orientation"<<endl       
       <<"-o file : write output to file"<<endl;
}


/** @class biasextract
@ingroup g_tools
@brief extract header data from image files: uid, time stamp orientation,
 KMatrix from EXIF data, ..
@author evers */
int main(int argc,char *argv[]) 
{
  bool GetASCIIString = false;
  bool GetTimeStamp = false;
  bool GetTimeMsec = false;
  bool GetPMatrix = false;
  bool GetR = false;  
  bool GetC = false;  
  bool GetK = false;  
  bool GetAxisAngle = false;
  bool GetKFromEXIF = false;
  bool GetCameraParamFromEXIF = false;
  bool GetCameraParamFromK = false;
  bool GetPXML = false;
  bool GetUID = false;
  bool GetOri = false;
  bool GetA = false;
  bool GetH = false;
  bool GetV = false;
  string Outfile;
  string PXMLName;
  ofstream ofs;

  bool SuppressFilename = false;
  //int c;
  int c=1;
  while(c >0) {
    c = getopt(argc,argv,"aceo:ptrsux:ETPKRCAHVO");
    if (c>0){
      switch(c) {
      case 'a': GetASCIIString =true;
	break;
      case 't': GetTimeStamp = true;
        break;
      case 'e': GetKFromEXIF = true;
        break;
      case 'E': GetCameraParamFromEXIF = true;
        break;
      case 'c': GetCameraParamFromK = true;
        break;
      case 'T': 
        GetTimeStamp = true;
        GetTimeMsec = true;	
        break;
      case 'P': GetPMatrix = true;
        break;
      case 'p': 
        GetPMatrix = true; 
        GetK = true; 
        GetR = true; 
        GetAxisAngle = true;
        GetC = true;  
        break;
      case 'R': GetR = true;  
        break;
      case 'r': GetAxisAngle = true;
        break;
      case 'C': GetC = true;  
        break;
      case 'A': GetA = true;  
        break;
      case 'H': GetH = true;  
        break;
      case 'V': GetV = true;  
        break;
      case 'K': GetK = true;  
        break;
      case 'u': GetUID = true;
        break;
      case 'O': GetOri = true;
        break;
      case 'o': Outfile = optarg;
        break;
      case 's': SuppressFilename = true;
        break;
      case 'x':  GetPXML = true;
        PXMLName = optarg;
        break; 
      default:
        usage();
        //exit(1);
        return 0; // for Dart testing 
      }
    }
  }


  if (argc-optind < 1) {
    cout <<"==> To few arguments."<<endl;
    usage();
    exit(1);
  }
  if (Outfile != "") {
    ofs.open(Outfile.c_str());
    if (!ofs) {
      perror(Outfile.c_str());
      exit(1);
    }
    cout <<"Writing to file: "<<Outfile<<endl;
    cout.rdbuf(ofs.rdbuf());
  } else {
    //     ofs = cout;
    cout <<"Writing to stdout "<<endl;
  }

  //  Image<unsigned char> tmp;
  ImageBase tmp;
  
  for ( int i=optind; i<argc;i++){
    //     ImageIO::Read(argv[i],pic);
    //    ImageIO::Load(argv[i],pic);
    ImageIO::Load(argv[i],tmp);
    Camera<unsigned char> pic(tmp);    
    pic.ParseMetaData();
    if(GetASCIIString) {
      cerr<<"ASCII: "<<pic.GetASCIIString()<<endl;
    }

    if (GetKFromEXIF) {
      ImageAttributes Att;
      Att.InitExifTags();
      if (Att.Read(argv[i])!=BIAS_IMAGEATTRIBUTES_SUCCESS) {
        cerr<<"error reading exif tags ..."<<endl;
      } else {
        KMatrix KExif;
        KExif.SetIdentity();
        bool success = true;
        success = success && (Att.GetFocalLengthXPixel(KExif[0][0])>=0);
        success = success && (Att.GetFocalLengthYPixel(KExif[1][1])>=0);
        success = success && (Att.GetPrincipalPoint(KExif[0][2],
                                                    KExif[1][2] )>=0);
        if (success) {
          cerr<<"K: ";
          cout<<KExif<<endl;
        } else {
          cerr<<"Error, not enough EXIF data to get K:"<<KExif<<endl;
        }
      }
    }
    if (GetCameraParamFromEXIF) {
#ifdef BIAS_HAVE_XML2
      CameraParam cp;
      if (cp.ReadFromEXIF(argv[i])==0) {
        stringstream cpfilename;
        cpfilename <<argv[i]<<".cameraparam.xml";
        cp.WriteCameraData(cpfilename.str());
      } else {
        cerr<<"Error, not enough EXIF data to construct cameraparam"<<endl;
      }
#else
      cerr<<"Missing XML2 support led to missing class BIAS::CameraParam, "
          <<"cant extract EXIF."<<endl;
#endif
    }
    pic.ParseMetaData();
    if (!SuppressFilename) cout <<argv[i]<<"\t"<<endl;
    if (GetUID)   cout<<pic.GetUID()<<"  ";  
    if (GetTimeStamp) {
      long sec,usec;
      pic.GetTime(sec,usec);
     // cout <<"true msec: "<<pic.GetTime()<<endl;
      if (GetTimeMsec) cout<< (sec % 1000) * 1000 + (usec / 1000);
      else cout<<sec<<"  "<<usec;
    }
    if (GetOri) {
      float y,p,r;
      pic.GetOrientation(y,p,r);
      cout<<y<<"   "<<p<<"   "<<r;
    }
    if (GetPMatrix || GetPXML) {
      PMatrix P;
      // check if valid Pose in meta data
      if (pic.IsRValid() && pic.IsCValid()) {
        // ok at least we have a pose, so compose P
        KMatrix K(MatrixIdentity);
        // check if we also have a calibration
        if (pic.IsKValid()) K = pic.GetK();
        else cerr<<"No K for composing P, assuming identity"<<endl;
        P.Compose(K, pic.GetR(), pic.GetC());
      } else if (pic.IsPValid()) {
        P = pic.GetP();
      } else {
        // no P so return zero matrix
        P.SetZero();
      }
      if (GetPMatrix) 
        cout<< P<<endl;
      if (GetPXML) {
#ifdef BIAS_HAVE_XML2
        P.XMLWrite(PXMLName);
#else
        cerr<<"No XML support found, so no writing of P as xml !"<<endl;
#endif
      }
    }
    if (GetK) {
      KMatrix K;
      if (pic.IsKValid()) {
        K = pic.GetK();
      } else {
        pic.GetP().GetK();
      }
      cout<<"K: "<<K<<endl;
    }
    if (GetR){
      PMatrix P = pic.GetP();
      cout<<"R: "<<P.GetR()<<endl;
    }
    if (GetAxisAngle){
      PMatrix P = pic.GetP();
      Vector3<double> axis;
      double angle;
      RMatrix R;
      R = P.GetR();
      R.GetRotationAxisAngle(axis, angle);
      cout<<"r[axis, angle]: "<<axis[0]<<" "<<axis[1]
          <<" "<<axis[2]<<" "<<angle<<endl;
      cout<<"|axis|= "<<axis.NormL2()<<endl;
      cout<<"R: "<<R<<endl;
      RMatrix temp;
      temp.Set(axis, angle);
      cout<<"R-axisAngle:\n "<<R-temp<<endl;

    }
    if (GetC){
      PMatrix P = pic.GetP();
      cout<<"C: "<<P.GetC()<<endl;
    }
    if (GetA){
      PMatrix P = pic.GetP();
      cout<<"A: "<<P.GetA()<<endl;
    }
    if (GetH){
      PMatrix P = pic.GetP();
      cout<<"H: "<<P.GetH()<<endl;
    }
    if (GetV){
      PMatrix P = pic.GetP();
      cout<<"V: "<<P.GetV()<<endl;
    }

    if (GetCameraParamFromK) {
#ifdef BIAS_HAVE_XML2
      KMatrix K;
      bool HaveK = true;
      if (pic.IsKValid()) {
        K = pic.GetK();
      } else if (pic.IsPValid()) {
        K = pic.GetP().GetK();
      } else {
        HaveK = false;
      }
      if (HaveK) {
        CameraParam cp(K, pic.GetWidth(), pic.GetHeight());
        stringstream cpfilename;
        cpfilename <<argv[i]<<".cameraparam.xml";
        cp.WriteCameraData(cpfilename.str());
      } else {
        cerr<<"Error, no K to construct cameraparam"<<endl;
      }
#else
      cerr<<"Missing XML2 support led to missing class BIAS::CameraParam, "
          <<"cant extract EXIF."<<endl;
#endif
    }
    cout<<endl;
  }

  if (Outfile != "")  
    ofs.close();

  return 0;
}
