/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Base/Image/ImageBase.hh>
#include <Base/Image/ImageIO.hh>
#include <stdio.h>
#include <string>
#include "Geometry/PMatrix.hh"
#include <Geometry/FMatrix.hh>
#include <Geometry/EpipolarLine.hh>
#include <Image/Camera.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Image/ImageDraw.hh>
#include <FeatureDetector/CornerDetectorHarris.hh>
#include <Utils/ThreeDOut.hh>
#include <Gui/GuiEnhanced.hh>

#define BIASOUT(theArg) cout<<theArg<<endl<<flush

using namespace BIAS;
using namespace std;

void usage() {
  cout << "Usage: biasshowepi [--fmatrix f.mat]"
    " [--vrml] [--onlyonewindow]  image1 image2"
       <<endl;
  cout <<"This program opens the image files given by the command line"
       <<" and extracts PMatrices from the headers, computes FMatrices "
       <<"(Alternatively pass FMatrix file on the command line)."<<endl
       <<"Then runs corner detector in image1 and shows epipolar lines"<<endl
       <<" in image2 "<<endl<<endl
       <<"You can also left-click in the left image to define a point which is"
       <<" enlarged and right-click in the right image to enlarge the region"
       <<" there. If you pass  --onlyonewindow  the large regions are"
       <<" displayed in the same window, this avoids async replies."<<endl;
  
}

/** @class biasshowepi
    @ingroup g_tools
    @brief show epipolar lines for image pair
    @author koeser */
int main (int argc, char *argv[])
{
 
  if (argc<3) { 
    usage();
    return -1;
  }
  FMatrix F;
  
  Param theParams(true);
  theParams.AddParamString("fmatrix","const fmatrix file","");
  theParams.AddParamBool("vrml","want vrml model",false);
  bool* OnlyOneWindow = 
    theParams.AddParamBool("onlyonewindow","draw everything in one window, "
                           "avoiding async replies",false);
  BIAS::CornerDetectorHarris<unsigned char, float> Harris(theParams);
  ThreeDOutParameters param3D;
  param3D.CameraStyle = PyramidMesh;
  ThreeDOut Scene(param3D);

  vector<HomgPoint2D> Features;
  unsigned int numfeatures = 0;
    
  // image filenames to be filled in from somewhere
  vector<string> filenames;
  int iFirstImgArg = theParams.ParseCommandLine(argc,argv);
  for (int argnr=iFirstImgArg; argnr<argc; argnr++) {
    cout <<"using file "<<argv[argnr]<<endl;
    filenames.push_back(argv[argnr]);
  }


  bool  FFromCommandLine = false;
  string* FFileName = NULL;
  if ((FFileName = theParams.GetParamString("fmatrix"))!=NULL) {
    if (*FFileName!="") {
      if (!F.Load(*FFileName)) {
        cout <<"error reading fmatrix from file "<< (*FFileName)<<endl;
        exit(-1);
      }
      FFromCommandLine = true;
      cout<<"Using const FMAtrix for all image pairs !"<<endl;
    }
  } 

  const unsigned int numimages = filenames.size();    
  
  // the depth map for the currrent image
  Image<float> DepthMap;
  
  // the image under inspection and the previous image
  Camera<unsigned char> CurImage[2];
  PMatrix CurP[2];
  int CurIndex = 1;


  Image<unsigned char> gim1, gim2, gim;
  GuiEnhanced gui;
  GuiEnhanced *zoom1, *zoom2; 
  if (! *OnlyOneWindow) {
    zoom1 = new GuiEnhanced;
    zoom2 = new GuiEnhanced;
  }
  
  for (unsigned int i=0; i<numimages; i++) {
    bool ShowEpiLines = true;
    // swap the current and last index
    if (CurIndex==0) CurIndex = 1; else CurIndex = 0;
    if (ImageIO::Load(filenames[i], CurImage[CurIndex])!=0) {
      BIASERR("Error loading image "<<filenames[i]<<" SKIP this image.");
      continue;
    } 
    if (CurImage[CurIndex].GetColorModel()!=ImageBase::CM_Grey) {
      BIASERR("No grey image, converting ...");
      ImageConvert::IP_ToGrey(CurImage[CurIndex]);
    }
    CurImage[CurIndex].ParseMetaData();
    BIASOUT("Cur image is "<<i<<" with name "<<filenames[i]);
    
    CurP[CurIndex] = CurImage[CurIndex].GetP();
    // #### debugging code for evaluating bbc principal point (koeser) #####
    bool SwapYCoordinateOfPrincipalPoint = false;
    if (SwapYCoordinateOfPrincipalPoint) {
      BIASWARN("Not using principal point as in image header, but inverting"
               " y coordinate ! !!!");
      KMatrix K = CurP[CurIndex].GetK();
      RMatrix R = CurP[CurIndex].GetR();
      Vector3<double> Center =  CurP[CurIndex].GetC();
      K[1][2] = CurImage[CurIndex].GetHeight()-1-K[1][2];
      CurP[CurIndex].Compose(K,R,Center);
    }
    // ###############################################################
    if (CurP[CurIndex].NormFrobenius()<0.1) {
      BIASERR("No PMatrix in header, trying to compose.");
      KMatrix K;
      RMatrix R;
      Vector3<double> C;
      R = CurImage[CurIndex].GetR();
      if (R.NormFrobenius()<0.1){
        BIASERR("No R found. Giving up.");
        continue;
      }
      K = CurImage[CurIndex].GetK();
      if (K.NormFrobenius()<0.1){
        BIASERR("No K found. Giving up.");
        continue;
      }
      C = CurImage[CurIndex].GetC();
      CurP[CurIndex].Compose(K,R,C);
      cout<<"Composed P is "<<CurP[CurIndex]<<", composed from K="<<K
          <<" R="<<R<<" C="<<C<<endl;
    }

    
   
    BIASOUT("P is "<<CurP[CurIndex]);
    RGBAuc CamColor((i==0)?255:0,0,(i!=0)?255:0,0);
    Scene.AddPMatrix(CurP[CurIndex], CurImage[CurIndex].GetWidth(),
                     CurImage[CurIndex].GetHeight(), CamColor);

    if (i==0) continue;

    Image<unsigned char> GreyImage;
    ImageConvert::ToGrey(CurImage[CurIndex], GreyImage);
    Image<float> FloatImage;
    ImageConvert::ConvertST(GreyImage, FloatImage, ImageBase::ST_float);
    vector<float> Qualities;
    Features.clear();
    Harris.Detect(GreyImage, Features, Qualities);

    numfeatures = Features.size();
    BIASOUT("Have "<<numfeatures<<" 2D features.");
    Vector3<double> C(CurP[CurIndex].GetC());
    
    stringstream ssname;
    ssname<<"cams"<<i<<".wrl";
    if (theParams.GetParamBool("vrml")!=NULL) {
      if (*theParams.GetParamBool("vrml")) Scene.VRMLOut(ssname.str());    
    } else cout <<"no vrml"<<endl;
    Image<unsigned char> &im1=CurImage[CurIndex];
    Image<unsigned char> &im2=CurImage[1-CurIndex];
    bool  KeyEvent = true, Pressed = false;
    char key;
    int Button, Xpos,  Ypos;
    unsigned int f = 0;
    Image<unsigned char> z1,z2;
    HomgPoint2D x(im1.GetWidth()/2,im1.GetHeight()/2);
    bool UpdateImages = true;
    while (ShowEpiLines) {
      if (UpdateImages) {
        gim2 = im2;
        gim1 = im1;
        UpdateImages = false;
      }
      if (KeyEvent && key == ' ') {
        x = Features[f++];
        if (f==numfeatures) f = 0;
        cout <<"Feature "<<f<<" of "<<numfeatures<<endl;
      }
      if (KeyEvent && key == 8) {
        if (f==0) f = numfeatures-1;
        x = Features[f--];
        cout <<"Feature "<<f<<" of "<<numfeatures<<endl;
      }
      else if (Pressed) {
        if (Button==1 || Button==3) 
          gim2 = im2;
        
        if (Button==1) {
          // update main window for cutout
          gim1 = im1;
         
        }
        gim.SetROI(0,0, im1.GetWidth(), im1.GetHeight());
        gim.Paste2ROI(gim1);
        gim.SetROI(im1.GetWidth(),0, im1.GetWidth()*2, im1.GetHeight());
        gim.Paste2ROI(gim2);
        gui.ShowImage(gim);
        const int scale = 5;
        const int cutoutsize = 21;
        const int imagesize = (2*cutoutsize)*scale;
        switch (Button) {
        case 1:
          x[0] = Xpos;
          x[1] = Ypos;
          x[2] = 1.0;
          gui.GetZoomImage(Xpos,Ypos,scale,cutoutsize,z1);
          ImageDraw<unsigned char>::RectangleCenter(z1,104 ,104, 25);
          
          if (! *OnlyOneWindow) zoom1->ShowImage(z1);
          else {
            for (unsigned int xx=0; xx<z1.GetWidth(); xx++) {
              for (unsigned int yy=0; yy<z1.GetHeight(); yy++) {
                gim1.GetImageDataArray()[yy][xx] = 
                  z1.GetImageDataArray()[yy][z2.GetChannelCount()*xx];            
              }
            }
          }
          break;
        case 3: {
          gui.GetZoomImage(Xpos,Ypos,scale,cutoutsize,z2);
          HomgPoint2D epipole1, epipole2;
          Matrix3x3<double> CutoutTransform;
          CutoutTransform.SetIdentity();
          CutoutTransform[0][0] = CutoutTransform[1][1] = 1.0 / scale;
          CutoutTransform[0][2] = double(Xpos-cutoutsize-im1.GetWidth());
          CutoutTransform[1][2] = double(Ypos-cutoutsize);
          
          FMatrix FCutout(CurP[CurIndex], CurP[1-CurIndex]);
          FCutout = CutoutTransform.Transpose()* FCutout;
          FCutout.GetEpipolesHomogenized(epipole1, epipole2);
          HomgPoint2D::EPosition epos2 = 
            epipole2.DeterminePosition(imagesize, imagesize);
          EpipolarLine eline;
        
          // draw epipolar line
          eline = FCutout * x;
          eline.Recalc(imagesize, imagesize, epipole2, epos2);
          eline.DrawWhole(z2);
          if (! *OnlyOneWindow) zoom2->ShowImage(z2);
          else {
            for (unsigned int xx=0; xx<z2.GetWidth(); xx++) {
              for (unsigned int yy=0; yy<z2.GetHeight(); yy++) {
                gim2.GetImageDataArray()[yy][xx] = 
                  z2.GetImageDataArray()[yy][z2.GetChannelCount()*xx]; 
              }
            }
          }
          break;
        }
        }
      }
      FMatrix F12(CurP[CurIndex], CurP[1-CurIndex]);
      // compute epipoles in big image
      HomgPoint2D epipole1, epipole2;
      F12.GetEpipolesHomogenized(epipole1, epipole2);
      
      
      
      // draw the source point in gim1 
      ImageDraw<unsigned char>::
        RectangleCenter(gim1, (int)x[0], (int)x[1], 5);
      ImageDraw<unsigned char>::
        RectangleCenter(gim1, (int)x[0], (int)x[1], 7);
      ImageDraw<unsigned char>::
        RectangleCenter(gim1, (int)x[0], (int)x[1], 9);
      ImageDraw<unsigned char>::
        RectangleCenter(gim1, (int)x[0], (int)x[1], 11);
      
      // compute the projection of the epipolar plane for that point
      EpipolarLine eline1;
      eline1 = epipole1.CrossProduct(x);
      HomgPoint2D::EPosition epos1 = 
        epipole1.DeterminePosition(im1.GetWidth(), im1.GetHeight());
      eline1.Recalc(im1.GetWidth(), im1.GetHeight(), epipole1, epos1);
      eline1.DrawWhole(gim1);

      
      HomgPoint2D::EPosition epos2 = 
        epipole2.DeterminePosition(im2.GetWidth(), im2.GetHeight());
 
      EpipolarLine eline;
        
      // draw epipolar line
      eline = F12 * x;
      eline.Recalc(im2.GetWidth(), im2.GetHeight(), epipole2, epos2);
      eline.DrawWhole(gim2);

      
      if (gim.IsEmpty())
        gim.Init(im1.GetWidth()*2, im1.GetHeight(), im1.GetChannelCount());
      gim.SetROI(0,0, im1.GetWidth(), im1.GetHeight());
      gim.Paste2ROI(gim1);
      gim.SetROI(im1.GetWidth(),0, im1.GetWidth()*2, im1.GetHeight());
      gim.Paste2ROI(gim2);
      gui.ShowImage(gim);

      gui.WaitForEvent(KeyEvent,key,Pressed,Button,Xpos,Ypos);
      if (KeyEvent && key=='q') ShowEpiLines = false;
    }
  }

  if (! *OnlyOneWindow) {
    delete zoom1;
    delete zoom2;
  }
  return 0;
    
}












