/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageDraw.hh>
#include <Gui/ConsoleRedirectIO.hh>
#include <Gui/ImageCanvas.hh>
#include <Image/Camera.hh>
#include <Geometry/PMatrix.hh>
#include <Geometry/FMatrix.hh>
#include <Geometry/EpipolarLine.hh>

#include <wx/wx.h>
#include <wx/splitter.h>

#include <iostream>

using namespace std;
using namespace BIAS;

/*
===============================================================================
Constants
===============================================================================
*/

const int HALF_CUTOUT_SIZE = 30;
const int CUTOUT_SIZE =  2 * HALF_CUTOUT_SIZE;
const int SCALE = 4;

/*
===============================================================================
ComputeCutoutImage
===============================================================================
*/

void ComputeCutoutImage(Image<unsigned char>& src, Image<unsigned char>& dst,
                        int srcX, int srcY, int srcWidth, int srcHeight, int scale)
{
  int dstWidth  = srcWidth  * scale;
  int dstHeight = srcHeight * scale;

  if ((int)dst.GetWidth() != dstWidth || (int)dst.GetHeight() != dstHeight)
    dst.Release();

  if (dst.IsEmpty())
    dst.Init(dstWidth, dstHeight, 3);

  unsigned char** srcIda = src.GetImageDataArray();
  unsigned char** dstIda = dst.GetImageDataArray();

  for (int x = srcX; x < srcX + srcWidth; x++)
  {
    for (int y = srcY; y < srcY + srcHeight; y++)
    {
      unsigned char pixel[3];
      if (x < 0 || y < 0 || x >= (int)src.GetWidth() || y >= (int)src.GetHeight())
      {
        pixel[0] = 0;
        pixel[1] = 0;
        pixel[2] = 0;
      }
      else
      {
        pixel[0] = srcIda[y][3*x+0];
        pixel[1] = srcIda[y][3*x+1];
        pixel[2] = srcIda[y][3*x+2];
      }

      int startX = (x - srcX) * scale;
      int startY = (y - srcY) * scale;

      for (int i = startX; i < startX + scale; i++)
      {
        for (int j = startY; j < startY + scale; j++)
        {
          if (i >= 0 && j >= 0 && i < (int)dst.GetWidth() && j < (int)dst.GetHeight())
          {
            dstIda[j][3*i+0] = pixel[0];
            dstIda[j][3*i+1] = pixel[1];
            dstIda[j][3*i+2] = pixel[2];
          }
        }
      }
    }
  }
}

/*
===============================================================================
PlainImageCanvas - Image canvas that does not show zoom frames and redirects
mouse click events to its parent.
===============================================================================
*/

class PlainImageCanvas : public ImageCanvas
{
public:
  PlainImageCanvas(wxWindow* parent, wxStatusBar* sbar = NULL, int sbarid = 0);

private:
  void OnLeftDown(wxMouseEvent& event);
  void OnRightDown(wxMouseEvent& event);
  void OnMotion(wxMouseEvent& event);

  DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(PlainImageCanvas, ImageCanvas)
EVT_LEFT_DOWN(PlainImageCanvas::OnLeftDown)
EVT_RIGHT_DOWN(PlainImageCanvas::OnRightDown)
EVT_MOTION(PlainImageCanvas::OnMotion)
END_EVENT_TABLE()

PlainImageCanvas::PlainImageCanvas(wxWindow* parent, wxStatusBar* sbar, int sbarid)
  : ImageCanvas(parent, sbar, sbarid)
{
}

void PlainImageCanvas::OnLeftDown(wxMouseEvent& event)
{
  if (GetParent() != NULL)
  {
    GetParent()->ProcessEvent(event);
  }
}

void PlainImageCanvas::OnRightDown(wxMouseEvent& event)
{
  if (GetParent() != NULL)
  {
    GetParent()->ProcessEvent(event);
  }
}

void PlainImageCanvas::OnMotion(wxMouseEvent& event)
{
  if (GetParent() != NULL)
  {
    GetParent()->ProcessEvent(event);
  }
  event.Skip();
}

/*
===============================================================================
CutoutFrame - Frame showing a cutout of a larger image
===============================================================================
*/

class CutoutFrame : public wxFrame
{
public:
  CutoutFrame(wxWindow* parent, const wxString& title);

  void ShowImage(Image<unsigned char>& image, const string& name);
  void SetTransform(int offsetX, int offsetY, int scale);

private:
  int               _OffsetX, _OffsetY, _Scale;
  PlainImageCanvas* _ImageCanvas;

  void OnMotion(wxMouseEvent& event);
  void OnLeaveWindow(wxMouseEvent& event);
  void OnClose(wxCloseEvent& event);

  DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(CutoutFrame, wxFrame)
EVT_MOTION(CutoutFrame::OnMotion)
EVT_LEAVE_WINDOW(CutoutFrame::OnLeaveWindow)
EVT_CLOSE(CutoutFrame::OnClose)
END_EVENT_TABLE()

CutoutFrame::CutoutFrame(wxWindow* parent, const wxString& title)
  : wxFrame(parent, -1, title,
            wxDefaultPosition, wxDefaultSize,
            wxDEFAULT_FRAME_STYLE | wxFRAME_FLOAT_ON_PARENT)
{
  _OffsetX = 0;
  _OffsetY = 0;
  _Scale   = 1;

  _ImageCanvas = new PlainImageCanvas(this);

  CreateStatusBar(1);

  wxBoxSizer* sizer = new wxBoxSizer(wxHORIZONTAL);
  sizer->Add(_ImageCanvas);
  SetSizer(sizer);
}

void CutoutFrame::ShowImage(Image<unsigned char>& image, const string& name)
{
  _ImageCanvas->Show(image, name);
  _ImageCanvas->SetScrollbars(0, 0, 0, 0);
  GetSizer()->SetItemMinSize(_ImageCanvas, image.GetWidth(), image.GetHeight());
  Fit();
}

void CutoutFrame::SetTransform(int offsetX, int offsetY, int scale)
{
  _OffsetX = offsetX;
  _OffsetY = offsetY;
  _Scale   = scale;
}

void CutoutFrame::OnMotion(wxMouseEvent& event)
{
  wxClientDC dc(this);
  PrepareDC(dc);
  wxPoint pos = event.GetPosition();
  int x = dc.DeviceToLogicalX(pos.x) / _Scale + _OffsetX;
  int y = dc.DeviceToLogicalY(pos.y) / _Scale + _OffsetY;
  stringstream text;
  text << "(x, y) = (" << x << ", " << y << ")";
  GetStatusBar()->SetStatusText(text.str().c_str());
}

void CutoutFrame::OnLeaveWindow(wxMouseEvent&  /* event */)
{
  GetStatusBar()->SetStatusText("");
}

void CutoutFrame::OnClose(wxCloseEvent& event)
{
  if (event.CanVeto())
  {
    Hide();
    event.Veto();
  }
  else
  {
    Destroy();
  }
}

/*
===============================================================================
MainFrame
===============================================================================
*/

class MainFrame : public wxFrame
{
public:
  MainFrame();

  void SetImages(Camera<unsigned char>& image1, const string& name1,
                 Camera<unsigned char>& image2, const string& name2);

private:
  PlainImageCanvas*     _ImageCanvas1;
  PlainImageCanvas*     _ImageCanvas2;
  Camera<unsigned char> _Image1, _Image2;
  std::string           _ImageName1, _ImageName2;
  CutoutFrame*          _CutoutFrame1;
  CutoutFrame*          _CutoutFrame2;
  wxPoint               _SelectedPoint1; // selected point in left image
  wxPoint               _SelectedPoint2; // selected point in right image

  void OnLeftDown(wxMouseEvent& event);

  DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(MainFrame, wxFrame)
EVT_LEFT_DOWN(MainFrame::OnLeftDown)
END_EVENT_TABLE()

MainFrame::MainFrame()
  : wxFrame(NULL, -1, "BiasShowEpiWX")
{
  _CutoutFrame1 = new CutoutFrame(this, "Cutout 1");
  _CutoutFrame2 = new CutoutFrame(this, "Cutout 2");

  CreateStatusBar();

  _ImageCanvas1 = new PlainImageCanvas(this, GetStatusBar(), 0);
  _ImageCanvas2 = new PlainImageCanvas(this, GetStatusBar(), 0);
}

void MainFrame::SetImages(Camera<unsigned char>& image1, const string& name1,
                          Camera<unsigned char>& image2, const string& name2)
{
  _Image1 = image1;
  _Image2 = image2;
  _ImageName1 = name1;
  _ImageName2 = name2;

  _ImageCanvas1->Show(_Image1, _ImageName1);
  _ImageCanvas2->Show(_Image2, _ImageName2);
  _ImageCanvas1->SetClearDrawingContext(false);
  _ImageCanvas2->SetClearDrawingContext(false);

  wxBoxSizer* sizer = new wxBoxSizer(wxHORIZONTAL);
  sizer->Add(_ImageCanvas1, 1, wxEXPAND);
  sizer->AddSpacer(3);
  sizer->Add(_ImageCanvas2, 1, wxEXPAND);

  sizer->SetItemMinSize(_ImageCanvas1, 100, 100);
  sizer->SetItemMinSize(_ImageCanvas2, 100, 100);

  SetSizer(sizer);
  sizer->SetSizeHints(this);

  SetSize(900, 600); // default main frame size
}

void MainFrame::OnLeftDown(wxMouseEvent& event)
{
  if (event.GetEventObject() == _ImageCanvas1)
  {
    wxClientDC dc(this);
    _ImageCanvas1->PrepareDC(dc);
    wxPoint pos = event.GetPosition();
    _SelectedPoint1.x = dc.DeviceToLogicalX(event.GetX());
    _SelectedPoint1.y = dc.DeviceToLogicalY(event.GetY());

    HomgPoint2D point(_SelectedPoint1.x, _SelectedPoint1.y, 1.0);

    // images including selected point and epipolar lines
    Image<unsigned char> PaintedImage1 = _Image1;
    Image<unsigned char> PaintedImage2 = _Image2;

    // --- LEFT IMAGE ---

    // draw the selected point in PaintedImage1
    ImageDraw<unsigned char>::RectangleCenter(
      PaintedImage1, _SelectedPoint1.x, _SelectedPoint1.y, CUTOUT_SIZE);

    // compute epipoles in big image
    FMatrix F12(_Image1.GetP(), _Image2.GetP());
    HomgPoint2D epipole1, epipole2;
    F12.GetEpipolesHomogenized(epipole1, epipole2);

    // compute the projection of the epipolar plane for that point
    EpipolarLine eline1 = epipole1.CrossProduct(point);
    HomgPoint2D::EPosition epos1 =
      epipole1.DeterminePosition(_Image1.GetWidth(), _Image1.GetHeight());
    eline1.Recalc(_Image1.GetWidth(), _Image1.GetHeight(), epipole1, epos1);
    eline1.DrawWhole(PaintedImage1);

    // --- RIGHT IMAGE ---

    HomgPoint2D::EPosition epos2 =
      epipole2.DeterminePosition(_Image2.GetWidth(), _Image2.GetHeight());

    // draw epipolar line
    EpipolarLine eline = F12 * point;
    eline.Recalc(_Image2.GetWidth(), _Image2.GetHeight(), epipole2, epos2);
    eline.DrawWhole(PaintedImage2);

    // show images (and restore scrollbar positions after that)
    int x1, y1, x2, y2;
    _ImageCanvas1->GetViewStart(&x1, &y1);
    _ImageCanvas1->GetViewStart(&x2, &y2);
    _ImageCanvas1->Show(PaintedImage1, _ImageName1);
    _ImageCanvas2->Show(PaintedImage2, _ImageName2);
    _ImageCanvas1->Scroll(x1, y1);
    _ImageCanvas1->Scroll(x2, y2);

    // --- CUTOUT FOR LEFT IMAGE ---
    Image<unsigned char> CutoutImage;
    ComputeCutoutImage(_Image1, CutoutImage,
                       _SelectedPoint1.x - HALF_CUTOUT_SIZE,
                       _SelectedPoint1.y - HALF_CUTOUT_SIZE,
                       CUTOUT_SIZE, CUTOUT_SIZE, SCALE);
    ImageDraw<unsigned char>::RectangleCenter(CutoutImage,
      2 + (_SelectedPoint1.x - (_SelectedPoint1.x - HALF_CUTOUT_SIZE)) * SCALE,
      2 + (_SelectedPoint1.y - (_SelectedPoint1.y - HALF_CUTOUT_SIZE)) * SCALE, SCALE);
    ImageDraw<unsigned char>::RectangleCenter(CutoutImage,
      2 + (_SelectedPoint1.x - (_SelectedPoint1.x - HALF_CUTOUT_SIZE)) * SCALE,
      2 + (_SelectedPoint1.y - (_SelectedPoint1.y - HALF_CUTOUT_SIZE)) * SCALE, SCALE * 4);

    _CutoutFrame1->ShowImage(CutoutImage, _ImageName1);
    _CutoutFrame1->SetTransform(_SelectedPoint1.x - HALF_CUTOUT_SIZE,
                                _SelectedPoint1.y - HALF_CUTOUT_SIZE, SCALE);
    _CutoutFrame1->Show(TRUE);
    _CutoutFrame1->Fit();
  }

  // store selected point when clicked in image 2
  if (event.GetEventObject() == _ImageCanvas2)
  {
    wxClientDC dc(this);
    _ImageCanvas2->PrepareDC(dc);
    wxPoint pos = event.GetPosition();
    _SelectedPoint2.x = dc.DeviceToLogicalX(event.GetX());
    _SelectedPoint2.y = dc.DeviceToLogicalY(event.GetY());
  }

  // update cutout of second image on click or when epipolar line was changed
  if ((event.GetEventObject() == _ImageCanvas1 && _CutoutFrame2->IsShown()) ||
      event.GetEventObject() == _ImageCanvas2)
  {
    Image<unsigned char> CutoutImage;
    ComputeCutoutImage(_Image2, CutoutImage,
                       _SelectedPoint2.x - HALF_CUTOUT_SIZE,
                       _SelectedPoint2.y - HALF_CUTOUT_SIZE,
                       CUTOUT_SIZE, CUTOUT_SIZE, SCALE);

    HomgPoint2D point(_SelectedPoint1.x, _SelectedPoint1.y, 1.0);

    HomgPoint2D epipole1, epipole2;
    Matrix3x3<double> CutoutTransform;
    CutoutTransform.SetIdentity();
    CutoutTransform[0][0] = CutoutTransform[1][1] = 1.0 / (double)SCALE;
    CutoutTransform[0][2] = (double)(_SelectedPoint2.x - HALF_CUTOUT_SIZE);
    CutoutTransform[1][2] = (double)(_SelectedPoint2.y - HALF_CUTOUT_SIZE);

    FMatrix FCutout(_Image1.GetP(), _Image2.GetP());
    FCutout = CutoutTransform.Transpose() * FCutout;
    FCutout.GetEpipolesHomogenized(epipole1, epipole2);
    HomgPoint2D::EPosition epos2 = epipole2.DeterminePosition(CUTOUT_SIZE*SCALE, CUTOUT_SIZE*SCALE);

    // draw epipolar line
    EpipolarLine eline;
    eline = FCutout * point;
    eline.Recalc(CUTOUT_SIZE*SCALE, CUTOUT_SIZE*SCALE, epipole2, epos2);
    eline.DrawWhole(CutoutImage);

    _CutoutFrame2->ShowImage(CutoutImage, _ImageName2);
    _CutoutFrame2->SetTransform(_SelectedPoint2.x - HALF_CUTOUT_SIZE,
                                _SelectedPoint2.y - HALF_CUTOUT_SIZE, SCALE);
    _CutoutFrame2->Show(TRUE);
    _CutoutFrame2->Fit();
  }
}

/*
===============================================================================
BiasShowEpiApp
===============================================================================
*/

/** @class biasshowepiwx
    @ingroup g_tools
    @brief show epipolar lines for image pair using wx-gui
    @author jkollmann/koeser */

class BiasShowEpiApp : public wxApp
{
public:
  virtual bool OnInit();
};

bool BiasShowEpiApp::OnInit()
{
  // utility: create a Redirection console for cout, cerr etc.
  ConsoleRedirectIO();

  // parse command line
  if (argc < 3)
  {
    cout << "No images given." << endl;
    cout << "Usage: biasshowepiwx image1.ppm image2.ppm." << endl;
    cout << "images must contain pmatrix in metadata." << endl;
    return false;
  }
  string fileName1 = argv[1];
  string fileName2 = argv[2];

  // load images
  Camera<unsigned char> image1, image2;
  if (ImageIO::Load(fileName1, image1) != 0)
  {
    cout << "Failed to load " << fileName1 << endl;
    return false;
  }
  if (ImageIO::Load(fileName2, image2) != 0)
  {
    cout << "Failed to load " << fileName2 << endl;
    return false;
  }
  image1.ParseMetaData();
  image2.ParseMetaData();

  // show frame
  MainFrame* frame = new MainFrame;
  frame->Show(TRUE);
  frame->SetImages(image1, fileName1, image2, fileName2);
  SetTopWindow(frame);

  return TRUE;
}

IMPLEMENT_APP(BiasShowEpiApp)
