 /* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ThreeDOut.hh"
#include <Geometry/RMatrix.hh>

#ifdef BIAS_HAVE_OPENGL
#  include <GL/gl.h>
#  include <GL/glu.h>
#  include <Utils/StatekeeperGL.hh>
#endif


#define ASSERT_MAP_ELEMENT(themap, key) if (themap.find(key)==themap.end()) { \
  BIASERR("Tried to access map with key " \
  <<key<<". Undefined Element." \
  <<" Aborting."); abort(); }

#define DEGREE_PER_RAD (180.0/M_PI)


using namespace BIAS;
using namespace std;


ThreeDOutParameters::ThreeDOutParameters()
{
  PointStyle                = Point;
  CameraStyle               = PyramidMesh;
  LineStyle                 = Solid;
  DrawUncertaintyEllipsoids = false;
  PointSize                 = 0.1;
  LineWidth                 = 1;
}

ThreeDOut:: ThreeDOut(const ThreeDOutParameters& params){
  NextPointIndex_     = 0;
  NextLineIndex_      = 0;
  NextEllipsoidIndex_ = 0;
  NextPIndex_         = 0;
  NextImageIndex_     = 0;
  NextIndexedFaceSetIndex_ = 0;
  SetParams(params);
}

ThreeDOut::~ThreeDOut(){}

void ThreeDOut::SetParams(const ThreeDOutParameters& params){
  Params_ = params;
}

// ------------------------ 3d add data functions ------------------------

unsigned int ThreeDOut::AddPoint(const BIAS::Vector3<double> &v,
                                 const BIAS::RGBAuc &Color)
{
  Point3DData D;
  D.Color = Color;
  D.Point = v;
  mapPoints_[NextPointIndex_] = D;
  return NextPointIndex_++;
}

unsigned int ThreeDOut::AddPoint(const BIAS::HomgPoint3D &v, 
                                 const BIAS::RGBAuc &Color)
{
#ifdef BIAS_DEBUG
  if (!v.IsHomogenized()) BIASERR("Supplied unhomogenized point !");
#endif
  Vector3<double> Center(v[0], v[1], v[2]);
  return AddPoint(Center, Color);
}

unsigned int ThreeDOut::AddLine(const BIAS::Vector3<double> &Start,
                                const BIAS::Vector3<double> &End,
                                const BIAS::RGBAuc &Color)
{
  Line3DData D;
  D.Start = Start;
  D.End   = End;
  D.Color = Color;

  mapLines_[NextLineIndex_] = D;
  return NextLineIndex_++;
}

unsigned int ThreeDOut::AddLine(const BIAS::HomgPoint3D &Start,
                                const BIAS::HomgPoint3D &End, 
                                const BIAS::RGBAuc &Color)
{
#ifdef BIAS_DEBUG
  if (!Start.IsHomogenized()) BIASERR("Supplied unhomogenized start point !");
  if (!End.IsHomogenized()) BIASERR("Supplied unhomogenized end point !");
#endif
  Vector3<double> s(Start[0], Start[1], Start[2]);
  Vector3<double> e(End[0], End[1], End[2]);
  return AddLine(s, e, Color);
}

unsigned int ThreeDOut::AddEllipsoid(BIAS::CovMatrix3x3 &cov, 
                                     BIAS::Vector3<double> &C,
                                     const BIAS::RGBAuc &Color,
                                     const double& dConfidenceRegionScale) 
{
  Ellipsoid3DData Data;
  Matrix3x3<double> VT;
  RMatrix VInv;
  Vector3<double> S;
  cov.GetPCA(S, VT);
  VInv = VT.Transpose();
  VInv.GetRotationAxisAngle(Data.RotAxis, Data.dAngle);
  Data.s0 = sqrt(S[0]) * dConfidenceRegionScale;
  Data.s1 = sqrt(S[1]) * dConfidenceRegionScale;
  Data.s2 = sqrt(S[2]) * dConfidenceRegionScale;
  Data.C = C;
  Data.Color = Color;
  // can y and z be exchanged by svd, so that we dont capture it ?

  mapEllipsoids_[NextEllipsoidIndex_] = Data;
  return NextEllipsoidIndex_++;
}

unsigned int ThreeDOut::AddEllipsoid(BIAS::CovMatrix3x3 &cov, 
                                     BIAS::HomgPoint3D &C,
                                     const BIAS::RGBAuc &Color,
                                     const double &dConfidenceRegionScale)
{
#ifdef BIAS_DEBUG
  if (!C.IsHomogenized()) BIASERR("Supplied unhomogenized point !");
#endif
  Vector3<double> Center(C[0], C[1], C[2]);
  return AddEllipsoid(cov, Center, Color, dConfidenceRegionScale);
}

unsigned int ThreeDOut::AddImage(const Vector3<double>& UL,
                                 const Vector3<double>& UR,
                                 const Vector3<double>& LL,
                                 const Vector3<double>& LR,
                                 unsigned int Width, unsigned int Height,
                                 unsigned char* pData, unsigned int channels) 
{
#ifdef BIAS_DEBUG
  if (Width>UINT_MAX/2) {
    BIASERR("Width "<<Width<<" is out of bounds !");
    abort();
  }
  if (Height>UINT_MAX/2) {
    BIASERR("Height "<<Height<<" is out of bounds !");
    abort();
  }
  if ((channels!=1)&&(channels!=3)) {
    BIASERR("Unknown channel count:"<<channels);
    abort();
  }
#endif
  // find min w with w>=width and w=2^n, needed by opengl
  unsigned int w(1), h(1);
  while (w<Width)  w<<=1;
  while (h<Height) h<<=1;

  Image3DData &ImagePatch(mapImages_[NextImageIndex_]); 
  if (!ImagePatch.ImageData.IsEmpty()) ImagePatch.ImageData.Release();
  ImagePatch.ImageData.Init(w, h, channels);
  ImagePatch.ImageData.SetROI(0, 0, Width, Height);
  unsigned char** pColRow = ImagePatch.ImageData.GetImageDataArray();
  unsigned char* pOrig = pData;
  for (unsigned int i=0; i<Height; i++) {
    memcpy(pColRow[i], pOrig,  channels*Width);
    pOrig += channels*Width;
  }
  ImagePatch.UL = UL;
  ImagePatch.UR = UR;
  ImagePatch.LL = LL;
  ImagePatch.LR = LR;

  return NextImageIndex_++;
}


// wrapper
unsigned int ThreeDOut::AddPMatrix(BIAS::Matrix3x4<double> &M, 
                                   const unsigned int& width,
                                   const unsigned int& height,
                                   const BIAS::RGBAuc &Color,
                                   const double &dScale,
                                   const string& name) 
{
  PMatrix P( M );
  return AddPMatrix(P, 
    width,
    height,
    Color,
    dScale,
    name);
}


unsigned int ThreeDOut::AddPMatrix(BIAS::PMatrix &P, 
                                   const unsigned int& width,
                                   const unsigned int& height,
                                   const BIAS::RGBAuc &Color,
                                   const double &dScale,
                                   const string& name) 
{
  Vector3<double> d,A;
  HomgPoint2D ThisCorner;
  Matrix3x3<double> R, iKRt;
  KMatrix K;
  PMatrix3DData pd;
  pd.name = name;
  pd.Color=Color;
  P.GetK(K);
  P.GetR(R);
  P.GetC(pd.C);
  P.GetA(A); 
  int ires = (K * R.Transpose()).Invert(iKRt);
  if (ires!=0) BIASERR("error computing iKRt.");
  BIASDOUT(D_3DOUT_PDECOMP, "Decomposing P="<<P<<" into K="<<K<<" R="
    <<R<<" C="<<pd.C<<" A="<<A);
  for (unsigned int i=0; i<5; i++){
    switch (i) {
    case 0:
      ThisCorner[0] = 0.0;
      ThisCorner[1] = 0.0;
      break;
    case 1:
      ThisCorner[0] = width-1;
      ThisCorner[1] = 0.0;
      break;
    case 2:
      ThisCorner[0] = width-1;
      ThisCorner[1] = height-1;
      break;
    case 3:
      ThisCorner[0] = 0.0;
      ThisCorner[1] = height-1;
      break;
    case 4:
      ThisCorner[0] = K[0][2];
      ThisCorner[1] = K[1][2];
      break;
    }
    ThisCorner[2] = 1.0;
    // get backprojection ray d from camera center to image point ThisCorner
    iKRt.Mult(ThisCorner,d);

    // normalize projection of d on A to length 1
    d *= dScale / (d.ScalarProduct(A));

    // attach ray d to camera center to construct a 3d point
    d += pd.C;
    switch (i) {
    case 0: pd.UL=d; break;
    case 1: pd.UR=d; break;
    case 2: pd.LR=d; break;
    case 3: pd.LL=d; break;
    case 4: pd.PP=d; break;
    }
  } // for (unsigned int i=0; i<4; i++){
  BIASDOUT(D_3DOUT_PDECOMP, "Created points ul="<<pd.UL<<" ur="<<pd.UR<<" lr="
    <<pd.LR<<" ll="<<pd.LL<<" and principle point at "<<pd.PP);
  pd.Up[0] = R[0][1];
  pd.Up[1] = R[1][1];
  pd.Up[2] = R[2][1];
  pd.Up *= -dScale /2.0;
  mapPMatrices_[NextPIndex_] = pd;


  return NextPIndex_++;
}

unsigned int ThreeDOut::
AddTriangleMesh(const BIAS::TriangleMesh& mesh, 
		const std::string& name)
{
  IndexedFaceSet Data;
  Data.name = name;
  mesh.GetTexture(Data.texture);
  Data.texture.Flip();
  mesh.GetVertices(Data.vertexSet);
  mesh.GetTexCoords(Data.texCoordSet);
  mesh.GetTriangleIndices(Data.faceIndices);

  mapIndexedFaceSet_[NextIndexedFaceSetIndex_] = Data;

//   cout<<"got Texture of size "<<Data.texture.GetWidth()<<"x"
//       <<Data.texture.GetHeight()<<"\n";
//   cout<<"got vertices "<<Data.vertexSet.size()<<endl;
//   cout<<"got tex coords "<<Data.texCoordSet.size()<<endl;
//   cout<<"got triangles "<<Data.faceIndices.size()<<endl;
  
  
  return NextIndexedFaceSetIndex_++;
}



// -------------------------- 3D data update functions -----------------------


void ThreeDOut::UpdatePoint(const unsigned int& PointIndex,
                            const BIAS::Vector3<double> &v,
                            const BIAS::RGBAuc &Color)
{
#ifdef BIAS_DEBUG
  ASSERT_MAP_ELEMENT(mapPoints_, PointIndex);
#endif
  Point3DData& D = mapPoints_[PointIndex];
  D.Point = v;
  D.Color = Color;
}

void ThreeDOut::UpdatePoint(const unsigned int& PointIndex,
                            const BIAS::HomgPoint3D &v,
                            const BIAS::RGBAuc &Color)
{
#ifdef BIAS_DEBUG
  if (!v.IsHomogenized()) BIASERR("Supplied unhomogenized point !");
#endif
  Vector3<double> V(v[0], v[1], v[2]);
  UpdatePoint(PointIndex, V, Color);
}

void ThreeDOut::UpdatePMatrix(const unsigned int& PIndex, 
                              BIAS::PMatrix &P, 
                              const unsigned int& width,
                              const unsigned int& height,
                              const BIAS::RGBAuc &Color,
                              const double& dScale)
{
  unsigned int backupnextpindex = NextPIndex_;
  NextPIndex_ = PIndex;
  AddPMatrix(P, width, height, Color, dScale);
  NextPIndex_ = backupnextpindex;
}

void ThreeDOut::UpdatePMatrixColor(const unsigned int& PIndex, 
                                   const BIAS::RGBAuc &Color)
{
  mapPMatrices_[PIndex].Color = Color;
}

void ThreeDOut::UpdateLine(const unsigned int& LineIndex,
                           const BIAS::Vector3<double> &Start,
                           const BIAS::Vector3<double> &End,
                           const BIAS::RGBAuc &Color)
{
#ifdef BIAS_DEBUG
  ASSERT_MAP_ELEMENT(mapLines_, LineIndex);
#endif
  Line3DData& D = mapLines_[LineIndex];
  D.Start = Start;
  D.End = End;
  D.Color = Color;
}

void ThreeDOut::UpdateLine(const unsigned int& LineIndex,
                           const BIAS::HomgPoint3D &Start,
                           const BIAS::HomgPoint3D &End, 
                           const BIAS::RGBAuc &Color) 
{
#ifdef BIAS_DEBUG
  if (!Start.IsHomogenized()) BIASERR("Supplied unhomogenized start point !");
  if (!End.IsHomogenized()) BIASERR("Supplied unhomogenized end point !");
#endif
  Vector3<double> s(Start[0], Start[1], Start[2]);
  Vector3<double> e(End[0], End[1], End[2]);
  UpdateLine(LineIndex, s, e, Color);
}


void ThreeDOut::UpdateEllipsoid(const unsigned int& EIndex,
                                BIAS::CovMatrix3x3 &cov, 
                                BIAS::Vector3<double> &C,
                                const BIAS::RGBAuc &Color,
                                const double &dConfidenceRegionScale) 
{
#ifdef BIAS_DEBUG
  ASSERT_MAP_ELEMENT(mapEllipsoids_, EIndex);
#endif
  Ellipsoid3DData& Data = mapEllipsoids_[EIndex];
  Matrix3x3<double> VT;
  RMatrix VInv;
  Vector3<double> S;
  cov.GetPCA(S, VT);
  VT = VT.Transpose();
  VInv = VT; // .Invert();
  VInv.GetRotationAxisAngle(Data.RotAxis, Data.dAngle);
  Data.s0 = sqrt(S[0]) * dConfidenceRegionScale;
  Data.s1 = sqrt(S[1]) * dConfidenceRegionScale;
  Data.s2 = sqrt(S[2]) * dConfidenceRegionScale;
  Data.C = C;
  Data.Color = Color;
}

void ThreeDOut::UpdateEllipsoid(const unsigned int& EIndex,
                                BIAS::CovMatrix3x3 &cov, 
                                BIAS::HomgPoint3D &C,
                                const BIAS::RGBAuc &Color, 
                                const double& dConfidenceRegionScale)
{
#ifdef BIAS_DEBUG
  if (!C.IsHomogenized()) BIASERR("Supplied unhomogenized point !");
#endif
  Vector3<double> Center(C[0], C[1], C[2]);
  UpdateEllipsoid(EIndex, cov, Center, Color, dConfidenceRegionScale);
}

void ThreeDOut::RemovePoint(const unsigned int& PointIndex)
{
  ASSERT_MAP_ELEMENT(mapPoints_, PointIndex);
  // uncontrolled erase leads to crash in case of non existing PointIndex:
  mapPoints_.erase(mapPoints_.find(PointIndex));
}

void ThreeDOut::RemoveAllPoints() 
{
  mapPoints_.clear();
}

void ThreeDOut::RemoveLine(const unsigned int& LineIndex)
{
  ASSERT_MAP_ELEMENT(mapLines_, LineIndex);
  // uncontrolled erase leads to crash in case of non existing PointIndex:
  mapLines_.erase(mapLines_.find(LineIndex));
}

void ThreeDOut::RemoveAllLines()
{
  mapLines_.clear();
}

void ThreeDOut::RemoveEllipsoid(const unsigned int& EIndex)
{
  ASSERT_MAP_ELEMENT(mapEllipsoids_, EIndex);
  // uncontrolled erase leads to crash in case of non existing PointIndex:
  mapEllipsoids_.erase(mapEllipsoids_.find(EIndex));
}

void ThreeDOut::RemoveAllEllipsoids() {
  mapEllipsoids_.clear();
}

void ThreeDOut::RemoveImage(const unsigned int& EIndex)
{
  ASSERT_MAP_ELEMENT(mapImages_, EIndex);
  // uncontrolled erase leads to crash in case of non existing PointIndex:
  mapImages_.erase(mapImages_.find(EIndex));
}

void ThreeDOut::RemoveAllImages() 
{
  mapImages_.clear();
}

void ThreeDOut::RemoveAllPMatrices()
{
  mapPMatrices_.clear();
}

void ThreeDOut::RemovePMatrix(int index)
{
  map<unsigned int, PMatrix3DData>::iterator it = mapPMatrices_.find(index);
  if (it!=mapPMatrices_.end()) {
    mapPMatrices_.erase(it);
  } else {
    BIASERR("Tried to erase non-existing pmatrix:"<<index);
  }
}

// ---------------------------- OPENGL functions ----------------------------
#ifdef BIAS_HAVE_OPENGL

void ThreeDOut::OpenGLOutPoints(const BIAS::RGBAuc &Color)
{
  glColor4ubv(Color.GetData());
  glPointSize((float)Params_.PointSize);
  map<unsigned int, Point3DData>::iterator it; 
  switch (Params_.PointStyle) {
  case Point:
    glBegin(GL_POINTS);
    for (it=mapPoints_.begin(); it!=mapPoints_.end(); it++){
      glVertex3dv(it->second.Point.GetData());
    }
    glEnd();
    break;
  case PointSprite:
    BIASERR("PointSprites not yet implemented");
    break;
  case Sphere:
    BIASERR("Sphere not yet implemented");
    break;
  case Box:
    BIASERR("Points as Box not yet implemented");
    break;
  }  
}

void ThreeDOut::OpenGLOutPoints(){  
  map<unsigned int, Point3DData>::iterator it; 
  glPointSize((float)Params_.PointSize);
  switch (Params_.PointStyle) {
  case Point:
    glBegin(GL_POINTS);
    for (it=mapPoints_.begin(); it!=mapPoints_.end(); it++){
      glColor4ubv(it->second.Color.GetData());
      glVertex3dv(it->second.Point.GetData());
    }
    glEnd();
    break;
  case PointSprite:
    BIASERR("PointSprites not yet implemented");
    break;
  case Sphere:
    BIASERR("Sphere not yet implemented");
    break;
  case Box:
    BIASERR("Points as Box not yet implemented");
    break;
  }  
}

void ThreeDOut::OpenGLOutPMatrices_(const BIAS::RGBAuc &Color, bool UseOneColor)
{
  StatekeeperGL state;  

  unsigned int index=0;

  BIAS::RGBAuc col;
  switch (Params_.CameraStyle){
  case PyramidMesh: 
    {
      GLenum cap = GL_LINE_LOOP;

      glDisable(GL_TEXTURE_2D);
      if (UseOneColor)
        glColor4ubv(Color.GetData());

      map<unsigned int, PMatrix3DData>::iterator itP;
      for (itP=mapPMatrices_.begin(); itP!=mapPMatrices_.end(); itP++){
        const PMatrix3DData *it = &(itP->second);
        glPushName(BIAS_SELECTION_OFFSET_PMATRIX +index);
        {
          if (!UseOneColor)
            glColor4ubv(it->Color.GetData());
          glBegin( cap );
          glVertex3dv(it->C.GetData());
          glVertex3dv(it->UR.GetData());
          glVertex3dv(it->UL.GetData());
          glEnd();

          glBegin( cap );
          glVertex3dv(it->C.GetData());
          glVertex3dv(it->LR.GetData());
          glVertex3dv(it->UR.GetData());
          glEnd();

          glBegin( cap );
          glVertex3dv(it->C.GetData());
          glVertex3dv(it->LL.GetData());
          glVertex3dv(it->LR.GetData());
          glEnd();

          glBegin( cap );
          glVertex3dv(it->C.GetData());
          glVertex3dv(it->UL.GetData());
          glVertex3dv(it->LL.GetData());
          glEnd();
        }
        glPopName();
        index++;
      }
    }
    break;
  case PyramidSolid:
    {
      float mat_diffuse[] =  {Color[0]*0.7f, Color[1]*0.7f, Color[2]*0.7f, 1.0f};
      float mat_ambient[] =  {Color[0]*0.1f, Color[1]*0.1f, Color[2]*0.1f, 1.0f};
      float mat_shininess[] =  {8.0 };
      Vector3<double> normal;

      const GLenum cap = GL_POLYGON;
      glShadeModel(GL_FLAT);
      glFrontFace(GL_CW);
      glCullFace(GL_BACK);
      glEnable(GL_CULL_FACE);
      glEnable(GL_NORMALIZE);
      glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS,mat_shininess);

      map<unsigned int, PMatrix3DData>::iterator itP;
      for (itP=mapPMatrices_.begin(); itP!=mapPMatrices_.end(); itP++){
        const PMatrix3DData *it = &(itP->second);

        if (UseOneColor) col = Color;
        else col = it->Color;

        glPushName(BIAS_SELECTION_OFFSET_PMATRIX +index);
        {
          glColor4ubv(&col[0]);
          mat_diffuse[0] = col[0]/255.0f*0.9f;
          mat_diffuse[1] = col[1]/255.0f*0.9f;
          mat_diffuse[2] = col[2]/255.0f*0.9f;
          mat_diffuse[3] = 1.0f;

          mat_ambient[0] = col[0]/255.0f*0.2f;
          mat_ambient[1] = col[1]/255.0f*0.2f;
          mat_ambient[2] = col[2]/255.0f*0.2f; 
          mat_ambient[3] = 1.0f;

          glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
          glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);

          glBegin( cap ); 
          (it->C - it->UL).CrossProduct(it->C - it->UR, normal);
          glNormal3f((float)normal[0], (float)normal[1], (float)normal[2]);
          glVertex3dv(it->C.GetData());
          glVertex3dv(it->UL.GetData());
          glVertex3dv(it->UR.GetData());
          glEnd();

          glBegin( cap ); 
          (it->C - it->UR).CrossProduct(it->C - it->LR, normal);
          glNormal3f(-(float)normal[0], -(float)normal[1], -(float)normal[2]);
          glVertex3dv(it->C.GetData());
          glVertex3dv(it->UR.GetData());
          glVertex3dv(it->LR.GetData());
          glEnd();

          glBegin( cap ); 
          (it->C - it->LR).CrossProduct(it->C - it->LL, normal);
          glNormal3f(-(float)normal[0], -(float)normal[1], -(float)normal[2]);
          glVertex3dv(it->C.GetData());
          glVertex3dv(it->LR.GetData());
          glVertex3dv(it->LL.GetData());
          glEnd();

          glBegin( cap ); 
          (it->C - it->LL).CrossProduct(it->C - it->UL, normal);
          glNormal3f((float)normal[0], (float)normal[1], (float)normal[2]);
          glVertex3dv(it->C.GetData());
          glVertex3dv(it->LL.GetData());
          glVertex3dv(it->UL.GetData());
          glEnd();

          // camera plane 
          glBegin( cap ); 
          (it->C - it->LR).CrossProduct(it->C - it->UR, normal);
          glNormal3f((float)normal[0], (float)normal[1], (float)normal[2]);
          glVertex3dv(it->UL.GetData());
          glVertex3dv(it->LR.GetData());
          glVertex3dv(it->UR.GetData());
          glVertex3dv(it->UL.GetData());
          glVertex3dv(it->LL.GetData());
          glVertex3dv(it->LR.GetData());
          glEnd();
        }
        glPopName();
        index++;
      }
    }
    break;
  case Arrow:
    {  
      map<unsigned int, PMatrix3DData>::iterator itP;
      for (itP=mapPMatrices_.begin(); itP!=mapPMatrices_.end(); itP++){
        const PMatrix3DData *it = &(itP->second);
        glPushName(BIAS_SELECTION_OFFSET_PMATRIX +index);
        {
          glColor4ubv(it->Color.GetData());
          glBegin(GL_POINTS);
          glVertex3dv(it->C.GetData());
          glEnd();
          glColor3f(1.0,1.0,0.0);
          glBegin(GL_LINES);
          glVertex3dv(it->C.GetData() );
          glVertex3dv(it->PP.GetData());
          glVertex3dv(it->C.GetData() );
          glVertex3dv( ((it->Up) +(it->C)).GetData() );
          glEnd();
        }
        glPopName();
        index++;
      }
    }
    break;
  default:
    BIASERR("unknown camera drawing style "<<Params_.CameraStyle);
    break;
  }
}

void ThreeDOut::OpenGLOutPMatrices(const BIAS::RGBAuc &Color){
  OpenGLOutPMatrices_(Color, true);
}

void ThreeDOut::OpenGLOutPMatrices()
{  
  BIAS::RGBAuc Color;
  OpenGLOutPMatrices_(Color, false);
}

void ThreeDOut::OpenGLOutImages(const bool & disableBlending)
{
  map<unsigned int, Image3DData>::iterator it=mapImages_.begin();
  // nothing to do ?
  if (it==mapImages_.end()) return;

  StatekeeperGL state(GL_ENABLE_BIT|GL_COLOR_BUFFER_BIT|GL_CURRENT_BIT );

  const GLenum  cap = GL_QUADS;

  GLuint texid;
  glGenTextures( 1, &texid ) ;
  glBindTexture( GL_TEXTURE_2D, texid) ;
  glEnable(GL_TEXTURE_2D);

  /// \todo image interface (rgb<=>grey)
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);    
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);


  GLenum GLColorModel = 
    (it->second.ImageData.GetChannelCount()==3)? GL_RGB : GL_LUMINANCE;

  // assuming all images of same size
  // use roi to determine factor
  double factorx(1.0), factory(1.0);
  if (it!=mapImages_.end()) {
    unsigned int x1,x2,y1,y2;
    it->second.ImageData.GetROI()->GetCorners(x1,y1,x2,y2);
    factorx = (double)x2/(double)it->second.ImageData.GetWidth();
    factory = (double)y2/(double)it->second.ImageData.GetHeight();
  }
  for (it=mapImages_.begin(); it!=mapImages_.end(); it++){
    /// \todo It is very inefficient to recreate image on each call (jw)
    glTexImage2D(GL_TEXTURE_2D, 0,GLColorModel,
      it->second.ImageData.GetWidth(),
      it->second.ImageData.GetHeight(), 0, GLColorModel,
      GL_UNSIGNED_BYTE,
      (GLvoid *) it->second.ImageData.GetImageData()); 

    // no transparency?
    if (disableBlending) 
      glDisable(GL_BLEND);

    glEnable(GL_TEXTURE_2D);
    glBindTexture( GL_TEXTURE_2D, texid ) ;
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);  
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);    

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    {

      glBegin(cap); 
      glVertex3dv(it->second.UL.GetData()); 
      glTexCoord2d(0.0,0.0);
      glVertex3dv(it->second.LL.GetData());
      glTexCoord2d(factorx,0.0);   
      glVertex3dv(it->second.LR.GetData());
      glTexCoord2d(factorx,factory);
      glVertex3dv(it->second.UR.GetData());
      glTexCoord2d(0.0,factory);
      glEnd();  

    }
    glPopMatrix();
  }
  glDeleteTextures(1, &texid);
}


void ThreeDOut::OpenGLOutLines(const BIAS::RGBAuc &Color){
  StatekeeperGL state(GL_ENABLE_BIT); 

  const GLenum cap = GL_LINE_LOOP;
  glDisable(GL_TEXTURE_2D);
  glColor4ubv(Color.GetData()); // same color for alls
  map<unsigned int, Line3DData>::iterator it;
  for (it=mapLines_.begin(); it!=mapLines_.end(); it++){
    glBegin( cap );
    glVertex3dv(it->second.Start.GetData());
    glVertex3dv(it->second.End.GetData());
    glEnd();
  }
}

void ThreeDOut::OpenGLOutLines(){
  StatekeeperGL state(GL_ENABLE_BIT); 

  const GLenum cap = GL_LINE_LOOP;
  glDisable(GL_TEXTURE_2D);
  map<unsigned int, Line3DData>::iterator it;
  for (it=mapLines_.begin(); it!=mapLines_.end(); it++){
    // a different colro for each 
    glColor4ubv(it->second.Color.GetData());

    glBegin( cap );
    glVertex3dv(it->second.Start.GetData());
    glVertex3dv(it->second.End.GetData());
    glEnd();
  }
}


void ThreeDOut::OpenGLOutEllipsoids(const BIAS::RGBAuc & /*Color*/){
  BIASERR("Not implemented with const color");
}


void ThreeDOut::OpenGLOutEllipsoids(){
  GLUquadric *myQuad;
  myQuad = gluNewQuadric();
  map<unsigned int, Ellipsoid3DData>::iterator it;
  for (it=mapEllipsoids_.begin(); it!=mapEllipsoids_.end(); it++){
    glPushMatrix();
    // move to right location
    glTranslated(it->second.C[0], it->second.C[1], it->second.C[2]);
    glRotated(it->second.dAngle * DEGREE_PER_RAD, // gl wants degree
      it->second.RotAxis[0], it->second.RotAxis[1], 
      it->second.RotAxis[2]);
    glScaled(it->second.s0, it->second.s1, it->second.s2);
    // gl wants opacity not transparency
    glColor4ubv(it->second.Color.GetData());
    gluSphere(myQuad, 1, 20, 20);
    glPopMatrix();
  }
  gluDeleteQuadric(myQuad);
}


void ThreeDOut::OpenGLOut(const BIAS::RGBAuc &Color){
  // no color arg: 
  OpenGLOutImages();
  // override color:
  OpenGLOutPMatrices(Color);
  OpenGLOutPoints(Color);
  OpenGLOutLines(Color);
  OpenGLOutEllipsoids(Color);
}


void ThreeDOut::OpenGLOut(){
  OpenGLOutImages();
  OpenGLOutPMatrices();
  OpenGLOutPoints();
  OpenGLOutLines();
  OpenGLOutEllipsoids();
}
#endif // BIAS_HAVE_OPENGL


// --------------------------- VRML functions ---------------------------

int ThreeDOut::VRMLOutPoints(ostream& VRMLFile, const BIAS::RGBAuc &Color){
  if (mapPoints_.size()==0) return 0;
  map<unsigned int, Point3DData>::iterator it = mapPoints_.begin();
  while (it!=mapPoints_.end()) {
    // this not really correct because it changes the color of the
    // points and does not only change their visualization
    it->second.Color = Color;
    it++;
  }
  return VRMLOutPoints(VRMLFile);
}

int ThreeDOut::VRMLOutPointsAsPointSet_(ostream &VRMLFile)
{

  VRMLFile<<
    "Shape { "<<endl<< 
    "  geometry PointSet {"<<endl<<
    "    color Color {"<<endl<<
    "      color ["<<endl;

  double r,g,b;
  map<unsigned int, Point3DData>::iterator it = mapPoints_.begin();
  while (it!=mapPoints_.end()) {
    Point3DData &Da = it->second;
    r = (double)Da.Color[0] / 255.0; 
    g = (double)Da.Color[1] / 255.0;
    b = (double)Da.Color[2] / 255.0;
    if (it!=mapPoints_.begin()) // add a "," behind last line 
      VRMLFile<<","<<endl;
    VRMLFile <<
      "      "<< r << " " << g  << " " << b;
    it++;
  }

  VRMLFile<<endl<<
    "      ]"<<endl<<
    "    }"<<endl<<
    "    coord Coordinate {"<<endl<<
    "      point ["<<endl;

  streamsize w=8;  
  it=mapPoints_.begin();
  while (it!=mapPoints_.end()) {
    Point3DData &Db = it->second;
    if (it!=mapPoints_.begin()) // add a "," behind last line 
      VRMLFile<<","<<endl;
    VRMLFile<< 
      "      "
      << setw(w) << Db.Point[0] << " "
      << setw(w) << Db.Point[1] << " "
      << setw(w) << Db.Point[2];
    it++;
  }

  VRMLFile << endl<<
    "      ]" << endl<<
    "    }" << endl<<
    "  }" << endl<<
    "}" << endl;

  return 0;
}



int ThreeDOut::VRMLOutPointsAsBoxes_(ostream &VRMLFile)
{
  VRMLFile<<
    "Shape { "<<endl<< 
    "  geometry IndexedFaceSet {"<<endl<<
    "    color Color {"<<endl<<
    "      color ["<<endl;

  double r,g,b;
  map<unsigned int, Point3DData>::iterator it = mapPoints_.begin();
  while (it!=mapPoints_.end()) {
    Point3DData &Da = it->second;
    r = (double)Da.Color[0] / 255.0; 
    g = (double)Da.Color[1] / 255.0;
    b = (double)Da.Color[2] / 255.0;
    if (it!=mapPoints_.begin()) // add a "," behind last line 
      VRMLFile<<","<<endl;    
    for (unsigned int i=0;i<7;i++)
      VRMLFile <<
      "      "<< r << " " << g  << " " << b<<","<<endl;
    VRMLFile <<
      "      "<< r << " " << g  << " " << b;
    it++;   
  }

  VRMLFile<<endl<<
    "      ]"<<endl<<
    "    }"<<endl<<
    "    coord Coordinate {"<<endl<<
    "      point ["<<endl;


  float size = float(Params_.PointSize / 2.0);
  streamsize w=8;  
  it=mapPoints_.begin();
  while (it!=mapPoints_.end()) {
    Point3DData &Db = it->second;    
    if (it!=mapPoints_.begin()) // add a "," behind last line 
      VRMLFile<<","<<endl<<endl;
    // top four points
    VRMLFile<< 
      "      "
      << setw(w) << Db.Point[0]-size << " "
      << setw(w) << Db.Point[1]+size << " "
      << setw(w) << Db.Point[2]-size;
    VRMLFile<< ","<<endl;
    VRMLFile<< 
      "      "
      << setw(w) << Db.Point[0]+size << " "
      << setw(w) << Db.Point[1]+size << " "
      << setw(w) << Db.Point[2]-size;
    VRMLFile<< ","<<endl;   
    VRMLFile<< 
      "      "
      << setw(w) << Db.Point[0]+size << " "
      << setw(w) << Db.Point[1]+size << " "
      << setw(w) << Db.Point[2]+size;
    VRMLFile<< ","<<endl;    
    VRMLFile<< 
      "      "
      << setw(w) << Db.Point[0]-size << " "
      << setw(w) << Db.Point[1]+size << " "
      << setw(w) << Db.Point[2]+size;
    VRMLFile<< ","<<endl;    
    // bottom four points
    VRMLFile<< 
      "      "
      << setw(w) << Db.Point[0]-size << " "
      << setw(w) << Db.Point[1]-size << " "
      << setw(w) << Db.Point[2]-size;
    VRMLFile<< ","<<endl;
    VRMLFile<< 
      "      "
      << setw(w) << Db.Point[0]+size << " "
      << setw(w) << Db.Point[1]-size << " "
      << setw(w) << Db.Point[2]-size;
    VRMLFile<< ","<<endl;    
    VRMLFile<< 
      "      "
      << setw(w) << Db.Point[0]+size << " "
      << setw(w) << Db.Point[1]-size << " "
      << setw(w) << Db.Point[2]+size;
    VRMLFile<< ","<<endl;    
    VRMLFile<< 
      "      "
      << setw(w) << Db.Point[0]-size << " "
      << setw(w) << Db.Point[1]-size << " "
      << setw(w) << Db.Point[2]+size;
    it++;
  }

  VRMLFile << endl<<
    "      ]" << endl<<
    "    }" << endl<< // end of Coordinate
    "    coordIndex ["<<endl;
  for (unsigned int i=0;i<mapPoints_.size();i++) {
    if (i!=0)
      VRMLFile<<","<<endl<<endl;
    VRMLFile<<"          "<<
      i*8+0<<", "<<i*8+1<<", "<<i*8+2<<", "<<i*8+3<<", -1,"<<
      " # top side"<<endl<<"          "<<
      i*8+7<<", "<<i*8+6<<", "<<i*8+5<<", "<<i*8+4<<", -1,"<<
      " # bottom side"<<endl<<"          "<<
      i*8+4<<", "<<i*8+5<<", "<<i*8+1<<", "<<i*8+0<<", -1,"<<
      " # back side"<<endl<<"          "<<
      i*8+3<<", "<<i*8+2<<", "<<i*8+6<<", "<<i*8+7<<", -1,"<<
      " # front side"<<endl<<"          "<<
      i*8+0<<", "<<i*8+3<<", "<<i*8+7<<", "<<i*8+4<<", -1,"<<
      " # left side"<<endl<<"          "<<
      i*8+2<<", "<<i*8+1<<", "<<i*8+5<<", "<<i*8+6<<", -1";
  }

  VRMLFile<<endl<<
    "     ]"<<endl<<
    "     ccw FALSE"<<endl<<
    "     colorPerVertex TRUE"<<endl<<
    "  }" << endl<<
    "}" << endl;

  return 0;
}

int ThreeDOut::VRMLOutPoints(ostream& VRMLFile){
  if (mapPoints_.size()==0) return 0;
  int res=0;
  switch (Params_.PointStyle) {
  case Point:
    res=VRMLOutPointsAsPointSet_(VRMLFile);
    break;
  case PointSprite:
    BIASERR("PointSprites not yet implemented");
    break;
  case Sphere:
    res = VRMLOutPointsAsSpheres_(VRMLFile);
    break;
  case Box:
    res = VRMLOutPointsAsBoxes_(VRMLFile);
    break;
  }    
  return res;
}

int ThreeDOut::VRMLOutPMatrices(ostream& VRMLFile, const BIAS::RGBAuc &Color)
{
  switch (Params_.CameraStyle) {
  case Arrow: return VRMLOutPMatricesAsArrows_(VRMLFile, Color);
  case PyramidMesh: return VRMLOutPMatricesAsMesh_(VRMLFile, Color);
  default:BIASERR("Not yet implmented");
  }
  return -1;

}

int ThreeDOut::VRMLOutPMatricesAsMesh_(ostream& VRMLFile, 
                                       const BIAS::RGBAuc &Color)
{
  double r = 0;
  double g = 0;
  double b = 0;
  // VRML transparency=0 is opaque
  double transparency = 0; 

  map<unsigned int, PMatrix3DData>::iterator itP;
  r =  (double)Color[0] / 255.0;
  g =  (double)Color[1] / 255.0;
  b =  (double)Color[2] / 255.0;
  transparency = VRMLtransparency(Color);
  unsigned int num=0;
  // create vrml model of camera
  VRMLFile << "Shape {" << endl;
  // write camera center and 4 points in the image plane
  VRMLFile <<"  geometry IndexedLineSet {" << endl;
  VRMLFile <<"    coord Coordinate {"      << endl;
  VRMLFile <<"      point [ "              << endl;
  for (itP=mapPMatrices_.begin(); itP!=mapPMatrices_.end(); itP++){
    const PMatrix3DData *it = &(itP->second); 
    VRMLFile 
      <<"        "<<it->C[0]<<" "<<it->C[1]<<" "<<it->C[2]<< " , "<< endl
      <<"        "<<it->UL[0]<<" "<<it->UL[1]<<" "<<it->UL[2] <<" , " <<endl
      <<"        "<<it->UR[0]<<" "<<it->UR[1]<<" " <<it->UR[2]<<" , " << endl
      <<"        "<<it->LR[0]<<" "<<it->LR[1]<<" " <<it->LR[2]<<" , " << endl
      <<"        "<<it->LL[0]<<" "<<it->LL[1]<<" " <<it->LL[2]<<" , " << endl;
    num++;
  }
  VRMLFile<<" ] " << endl;
  VRMLFile <<"    }" << std::endl;
  VRMLFile << "    coordIndex [" << endl;
  for (unsigned int i=0;i<num; i++) {
    unsigned n=i*5;
    VRMLFile <<"      "<<n<<", "<<n+2<<", "<<n+1<<", -1" <<endl;
    VRMLFile <<"      "<<n<<", "<<n+1<<", "<<n+4<<", -1" <<endl;
    VRMLFile <<"      "<<n<<", "<<n+4<<", "<<n+3<<", -1" <<endl;
    VRMLFile <<"      "<<n<<", "<<n+3<<", "<<n+2<<", -1" <<endl;
  }
  VRMLFile << "       ] " << endl;
  VRMLFile << "  }" << endl;    // terminates IndexedLineSet
  // set color of lines
  VRMLFile << "  appearance Appearance { " << endl;
  VRMLFile << "    material Material {"<<endl
    << "      diffuseColor  " << " "<<r<<" "<<g<<" "<<b<<endl
    << "      emissiveColor " << " "<<r<<" "<<g<<" "<<b<<endl
    <<"     }"<< endl;   // terminates material
  VRMLFile << "  }"   << endl;  // terminates appearance 
  VRMLFile << "}" << endl;      // terminates shape
  return 0;
}

int ThreeDOut::VRMLOutPMatricesAsArrows_(ostream& VRMLFile, const BIAS::RGBAuc &Color){
  double r = 0;
  double g = 0;
  double b = 0;
  double transparency = 0;
  map<unsigned int, PMatrix3DData>::iterator itP;
  r =  (double)Color[0] / 255.0;
  g =  (double)Color[1] / 255.0;
  b =  (double)Color[2] / 255.0;
  transparency = VRMLtransparency(Color);
  unsigned int num=0;
  // first: one switch-case to write entry
  VRMLFile << "Shape {" << endl; 
  VRMLFile << "   geometry IndexedLineSet {" << endl;
  VRMLFile << "      coord Coordinate {" << endl;
  VRMLFile << "         point [" << endl;
  Vector3<double> up;
  // then: for-loop with switch-case to write cams efficient
  for (itP=mapPMatrices_.begin(); itP!=mapPMatrices_.end(); itP++){
    const PMatrix3DData *it = &(itP->second);
    // write the line
    up = it->C + it->Up;
    VRMLFile 
      << "      " <<up[0]<<" "<<up[1]<<" "<<up[2] <<", "<<endl
      << "   " <<it->C[0]<<" "<<it->C[1]<<" "<<it->C[2]<<","<<endl
      << "      " <<it->PP[0]<<" "<<it->PP[1]<<" "<<it->PP[2] <<", "<<endl;   
    num++;
  }
  VRMLFile<< "         	     ]" << endl;
  VRMLFile << "      }"<<endl;
  VRMLFile << "   coordIndex [ "<<endl;
  for (unsigned int i=0;i<num; i++)
    VRMLFile <<"      "<<i*3<<"  "<<i*3+1<<"  "<<i*3+2<<"  -1"<<endl;
  VRMLFile <<" ]"<< endl;
  VRMLFile << "   color Color {"<<endl;
  VRMLFile << " 	   color [ "<<endl;
  for (unsigned int i=0;i<num; i++)
    VRMLFile <<r<<" "<<g<<" "<<b<<", "  <<r<<" "<<g<<" "<<b<<" ,"<<endl;
  VRMLFile << " ] "<<endl;
  VRMLFile << "   }\n}\n}"<< endl;
  return 0;
}





int ThreeDOut::VRMLOutPMatrices(ostream& VRMLFile){
  double r = 0;
  double g = 0;
  double b = 0;
  double transparency = 0;
  bool CenterAsSphere = true;
  map<unsigned int, PMatrix3DData>::iterator itP;
  for (itP=mapPMatrices_.begin(); itP!=mapPMatrices_.end(); itP++){
    const PMatrix3DData *it = &(itP->second);
    switch (Params_.CameraStyle) {
    case PyramidSolid:
      r =  (double)it->Color[0] / 255.0;
      g =  (double)it->Color[1] / 255.0;
      b =  (double)it->Color[2] / 255.0;
      transparency =  VRMLtransparency(it->Color);

      // create vrml model of camera
      VRMLFile << std::endl;
      //       if (it->name.length()>0) 
      // 	VRMLFile <<"  DEF "<< it->name;
      VRMLFile << "  Shape {" << std::endl;

      // write camera center and 4 points in the image plane
      if (it->name.length()>0) 
        VRMLFile <<"         geometry DEF "<<it->name;	  
      else
        VRMLFile <<"         geometry";
      VRMLFile <<" IndexedFaceSet {" << std::endl;
      VRMLFile <<"            coord Coordinate {" << std::endl;
      VRMLFile <<"              point [ " << std::endl;
      VRMLFile <<"                       "<<it->C[0]<<" "<<it->C[1]<<" "
        <<it->C[2]<< " , " << std::endl;
      VRMLFile <<"                       "<<it->UL[0]<<" "<<it->UL[1]
      <<" "<<it->UL[2] <<" , " <<std:: endl;
      VRMLFile <<"                       "<<it->UR[0]<<" "<<it->UR[1]<<" "
        <<it->UR[2]<<" , " <<  std::endl;
      VRMLFile <<"                       "<<it->LR[0]<<" "<<it->LR[1]<<" "
        <<it->LR[2]<<" , " <<  std::endl;
      VRMLFile <<"                       "<<it->LL[0]<<" "<<it->LL[1]<<" "
        <<it->LL[2]  <<" ] " <<  std::endl;

      VRMLFile <<"            }" << std::endl;

      //  write triangulation for topology
      VRMLFile << "              coordIndex [" << endl;
      VRMLFile << "                           0, 2, 1, -1, " << std::endl;
      VRMLFile << "                           0, 1, 4, -1, " << std::endl;
      VRMLFile << "                           0, 4, 3, -1, " << std::endl;
      VRMLFile << "                           0, 3, 2, -1, " << std::endl;
      VRMLFile << "                           1, 2, 4, -1, " << std::endl;
      VRMLFile << "                           4, 2, 3, -1 ]" << std::endl;

      VRMLFile << "  }" << std::endl << std::endl; // terminates IndexedFaceSet

      VRMLFile << "         appearance Appearance { " << endl;

      VRMLFile << "              material Material { diffuseColor  "
        << " "<<r<<" "<<g<<" "<<b<< std::endl //terminates material
        << " transparency  "<< transparency<<"}"<<std::endl;
      VRMLFile << "} "   << std::endl;  // terminates appearance 
      VRMLFile << "   }" << std::endl;  // terminates shape

      break;
    case PyramidMeshWithArrow:
    case PyramidMesh:
      r =  (double)it->Color[0] / 255.0;
      g =  (double)it->Color[1] / 255.0;
      b =  (double)it->Color[2] / 255.0;

      // create vrml model of camera
      VRMLFile << endl;
      VRMLFile << "Shape {" << endl;

      // write camera center and 4 points in the image plane
      VRMLFile <<"  geometry IndexedLineSet {" << endl;
      VRMLFile <<"    coord Coordinate {"      << endl;
      VRMLFile <<"      point [ "              << endl;
      VRMLFile <<"        "<<it->C[0]<<" "<<it->C[1]<<" "
        <<it->C[2]<< " , " << endl;
      VRMLFile <<"        "<<it->UL[0]<<" "<<it->UL[1]
      <<" "<<it->UL[2] <<" , " <<std:: endl;
      VRMLFile <<"        "<<it->UR[0]<<" "<<it->UR[1]<<" "
        <<it->UR[2]<<" , " << endl;
      VRMLFile <<"        "<<it->LR[0]<<" "<<it->LR[1]<<" "
        <<it->LR[2]<<" , " << endl;
      VRMLFile <<"        "<<it->LL[0]<<" "<<it->LL[1]<<" "
        <<it->LL[2]  <<" ] " << endl;
      VRMLFile <<"    }" << std::endl;

      // conect the points with lines
      VRMLFile << "    coordIndex [" << endl;
      VRMLFile << "      0, 2, 1, -1, " << endl;
      VRMLFile << "      0, 1, 4, -1, " << endl;
      VRMLFile << "      0, 4, 3, -1, " << endl;
      VRMLFile << "      0, 3, 2, -1] " << endl;
      VRMLFile << "  }" << endl;    // terminates IndexedLineSet

      // set color of lines
      VRMLFile << "  appearance Appearance { " << endl;
      VRMLFile << "    material Material {"<<endl
        << "      diffuseColor  " << " "<<r<<" "<<g<<" "<<b<<endl
        << "      emissiveColor " << " "<<r<<" "<<g<<" "<<b<<endl
        <<"     }"<< endl;   // terminates material
      VRMLFile << "  }"   << endl;  // terminates appearance 
      VRMLFile << "}" << endl;      // terminates shape
      if (Params_.CameraStyle!=PyramidMeshWithArrow) break;
    case Arrow: 
      r =  (double)it->Color[0] / 255.0;
      g =  (double)it->Color[1] / 255.0;
      b =  (double)it->Color[2] / 255.0;

      if (CenterAsSphere) {
        VRMLFile << "Transform {"<<endl;
        VRMLFile << "  translation "
          << it->C[0] << " " <<it->C[1] << " " <<it->C[2]
          << endl;
          VRMLFile << "  children ";
          VRMLFile <<"   Shape {"<<endl;
          VRMLFile << "    appearance Appearance { " << endl;
          VRMLFile << "	    material Material { " << endl;
          VRMLFile << "	      transparency      "
            << VRMLtransparency(it->Color) <<endl;
          VRMLFile << "	      diffuseColor " << r << " " << g << " "<<b << endl;
          VRMLFile << "	    }" << endl; // close Material
          VRMLFile << "    }" << endl; // close Appearance
          VRMLFile << "    geometry Sphere  { radius "
            << 0.1*(it->C-it->PP).NormL2()<<" }" 
            << endl;
          VRMLFile << "  }" << endl; // close Shape
          VRMLFile << "}" << endl; // close transform
      } else {
        // write the point
        VRMLFile << "Shape {" << endl; 
        VRMLFile << "   geometry PointSet {"<<endl;
        VRMLFile << "      coord Coordinate {"<<endl;
        VRMLFile << "         point ["<<it->C[0]<<" "<<it->C[1]<<" "<<it->C[2]
        <<"]" << endl << "      }"<<endl;
        VRMLFile << "      color Color { color [ 0 1 0, 0 1 0 ] }"<<endl;
        VRMLFile << "   }\n}"<<endl;
      }

      // write the line
      VRMLFile << "Shape {" << endl; 
      VRMLFile << "   geometry IndexedLineSet {" << endl;
      VRMLFile << "      coord Coordinate {" << endl;
      VRMLFile << "         point [" << endl;
      VRMLFile << "    	 	         " <<it->C[0]<<" "<<it->C[1]<<" "
        <<it->C[2]<<","<<endl;
      VRMLFile << "    	 	         " <<it->PP[0]<<" "<<it->PP[1]<<" "
        <<it->PP[2] << "         	     ]" << endl;
      VRMLFile << "      }"<<endl;
      VRMLFile << "   coordIndex [ 0 1 -1 ]"<< endl;
      VRMLFile << "   color Color {"<<endl;
      VRMLFile << " 	   color [ "<<r<<" "<<g<<" "<<b<<", "
        <<r<<" "<<g<<" "<<b<<" ] "<<endl;
      VRMLFile << "   }\n}\n}"<< endl;

      break;
    default: BIASERR("Desired Camera drawing style is not implemented."
               <<" Skipping Cameras.");
      return -1;
    }
  }
  return 0;
}


int ThreeDOut::VRMLOutLines(ostream& VRMLFile, const BIAS::RGBAuc &Color){
  double r = (double)Color[0] / 255.0;
  double g = (double)Color[1] / 255.0;
  double b = (double)Color[2] / 255.0;
  // double a = (double)it->second.Color[3] / 255.0;
  switch (Params_.LineStyle) {
  case Solid:
    for (map<unsigned int, Line3DData>::iterator it=mapLines_.begin();
      it!=mapLines_.end(); it++) {
        VRMLFile << std::endl;
        VRMLFile << "  Shape {" << std::endl;
        VRMLFile <<"         geometry IndexedLineSet {" << std::endl;
        VRMLFile <<"            coord Coordinate {" << std::endl;
        VRMLFile <<"              point [ " << std::endl;
        VRMLFile <<"                       "<<it->second.Start[0]<<" "
          << it->second.Start[1]<<" " <<it->second.Start[2]<< " , "
          << std::endl;
        VRMLFile <<"                       "<<it->second.End[0]<<" "
          <<it->second.End[1] <<" "<<it->second.End[2] <<" ] "<<std::endl;

        VRMLFile <<"            }" << std::endl;
        VRMLFile << "              coordIndex [  0, 1, -1 ]" << std::endl;
        VRMLFile << "           color Color {" << std::endl;
        VRMLFile <<"              color [  "<<r<<" "<<g<<" "<<b<<" ]"<<std::endl;
        VRMLFile << "           } " <<std:: endl;   // terminates Color
        VRMLFile << "           colorPerVertex FALSE " << std::endl;
        VRMLFile << "  }" << std::endl << std::endl; // terminates IndexedLineSet
        VRMLFile << "   }" << std::endl;  // terminates shape
      } 
      break;
  default: BIASERR("Unsupported LineStyle, line drawing disabled."); return -1;
  }
  return 0;
}


int ThreeDOut::VRMLOutIndexedFaceSets(ostream& VRMLFile) 
{
  unsigned int i = 0;
  map<unsigned int, IndexedFaceSet>::iterator it;
  for (it = mapIndexedFaceSet_.begin(); it!=mapIndexedFaceSet_.end(); it++) {
    VRMLFile << std::endl;
    VRMLFile << "Shape {" << std::endl;
    //define apperance
    VRMLFile << "appearance Appearance { " << std::endl;
    VRMLFile << "texture  PixelTexture { " << std::endl;
    //fill in PixelTexture Size and colors in hex?
    VRMLFile << "\t image "
             <<it->second.texture.GetWidth()<<" "
             <<it->second.texture.GetHeight()<<" "
             <<it->second.texture.GetChannelCount()<<std::endl;    
    //    VRMLFile << "\t       ";
    unsigned int channelColor;
    unsigned int writeCounter = 0;
    for(unsigned int h=0; h<it->second.texture.GetHeight(); h++) {
      VRMLFile << " ";
      for(unsigned int w=0; w<it->second.texture.GetWidth(); w++) {
        VRMLFile <<"0x";
        for(unsigned int c=0; c<it->second.texture.GetChannelCount(); c++) {
          channelColor= it->second.texture.PixelValue(w,h,(unsigned short)c);
          VRMLFile.flags(ios_base::hex);
          VRMLFile.fill('0');
          VRMLFile.width(2);
          VRMLFile<<channelColor;
        }
        writeCounter++; 
        if(writeCounter>=8) {
          VRMLFile<<std::endl;
          writeCounter=0;
        }
        VRMLFile.flags(ios_base::dec);
        VRMLFile <<" ";
      }
      VRMLFile<<std::endl;
    }
    VRMLFile.flags(ios_base::dec);
    VRMLFile << "  } \n }\n";//end appearance

    //define geometry
    if (it->second.name.length()>0) 
      VRMLFile <<" geometry DEF "<<it->second.name;	  
    else
      VRMLFile <<" geometry";
  VRMLFile <<" IndexedFaceSet {" << std::endl;
  //write all vertex coordinates
  VRMLFile <<"    coord Coordinate {" << std::endl;
  VRMLFile <<"      point [ " << std::endl;
  for(i = 0; i<it->second.vertexSet.size()-1; i++) {
    VRMLFile <<it->second.vertexSet[i][0]<<" "
	     <<it->second.vertexSet[i][1]<<" "
	     <<it->second.vertexSet[i][2]<<", "
	     << std::endl;
  }
  VRMLFile <<it->second.vertexSet[it->second.vertexSet.size()-1][0]<<" "
	   <<it->second.vertexSet[it->second.vertexSet.size()-1][1]<<" "
	   <<it->second.vertexSet[it->second.vertexSet.size()-1][2]<<" ]"
	   <<std::endl;
  VRMLFile <<" }" << std::endl; //closing coord
    
    //write all texture coordinates
    VRMLFile <<"    texCoord TextureCoordinate {" << std::endl;
    VRMLFile <<"      point [ ";
    for(i = 0; i<it->second.texCoordSet.size()-1; i++) {
      VRMLFile <<it->second.texCoordSet[i][0]<<" "
               <<it->second.texCoordSet[i][1]<<", "
               << std::endl;
    }
    VRMLFile <<it->second.texCoordSet[it->second.texCoordSet.size()-1][0]<<" "
             <<it->second.texCoordSet[it->second.texCoordSet.size()-1][1]<<" ]"
             <<std::endl;
    VRMLFile <<" }" << std::endl; //closing texCoord
    
    //define faces per indices
    VRMLFile <<"      coordIndex [ ";
    for(i = 0; i<it->second.faceIndices.size()-1; i++) {
      for(int j=0; j<(it->second.faceIndices[i]).size(); j++)
        VRMLFile <<it->second.faceIndices[i][j]<<", ";
      VRMLFile<<" -1,"<< std::endl;
    }
    for(int j=0; 
        j<(it->second.faceIndices[it->second.faceIndices.size()-1]).size(); j++)
      VRMLFile<<it->second.faceIndices[it->second.faceIndices.size()-1][j]<<", ";
    VRMLFile<<" -1 ]"<< std::endl;
    VRMLFile <<"  }" << std::endl;//closing IndexedFaceSet
    VRMLFile <<"}" << std::endl;//closing SHAPE
 
  }
  return 0;
}

int ThreeDOut::VRMLOutImages(ostream& VRMLFile){

  map<unsigned int, Image3DData>::iterator it;
  for (it = mapImages_.begin(); it!=mapImages_.end(); it++) {

    VRMLFile << std::endl;
    VRMLFile << "Shape {" << std::endl;

    // write a indexedfaceset as the pixel grid, first all grid points
    VRMLFile <<"  geometry IndexedFaceSet {" << std::endl;
    VRMLFile <<"    coord Coordinate {" << std::endl;
    VRMLFile <<"      point [ " << std::endl;

    // get roi which is the original image
    // ImageData may be larger, because it is padded to power of two
    unsigned int x1,y1,x2,y2;
    it->second.ImageData.GetROI()->GetCorners(x1,y1,x2,y2);
    const unsigned int width  = x2;
    const unsigned int height = y2;
    for (unsigned int y=0; y<=height; y++) {
      for (unsigned int x=0; x<=width; x++) {
        const double xfrac = (double(x)-0.5) / double(width-1);
        const double yfrac = (double(y)-0.5) / double(height-1);
        Vector3<double> CurPoint(it->second.UL + 
          xfrac*(it->second.UR-it->second.UL)+
          yfrac*( (1.0-xfrac) * 
          (it->second.LL-it->second.UL)+
          xfrac*(it->second.LR-it->second.UR)));
        VRMLFile <<"        "
          << CurPoint[0]<<" "<<CurPoint[1]<<" "<<CurPoint[2]<<","
          << std::endl;
      }
    } 
    VRMLFile <<"      ]" << std::endl;
    VRMLFile <<"    }" << std::endl;
    //  now all the faces, the 3d pixels:
    VRMLFile <<"    coordIndex [" << endl;
    for (unsigned int x=0; x<width;x++) {
      for (unsigned int y=0; y<height;y++) {
        VRMLFile <<"      "<< (x+y*(width+1)) <<",";
        VRMLFile <<"      "<< (x+1+y*(width+1)) <<",";
        VRMLFile <<"      "<< (x+1+(y+1)*(width+1)) <<",";
        VRMLFile <<"      "<< (x+(y+1)*(width+1)) <<", -1"<<endl;
      }
    }
    VRMLFile << "    ]" << std::endl;

    // set color of the points
    VRMLFile << "    color Color {" << std::endl;
    VRMLFile << "      color [ " << std::endl;
    unsigned char** ppData = it->second.ImageData.GetImageDataArray();
    double r = 1.0;
    double g = 1.0;
    double b = 1.0;
    for (unsigned int x=0; x<width; x++) {
      for (unsigned int y=0; y<height; y++) {
        if (it->second.ImageData.GetChannelCount()==1) {
          r = g = b = ((double)ppData[y][x])/ 255.0;;
        } else if (it->second.ImageData.GetChannelCount()==3) {
          r = (double)ppData[y][3*x] / 255.0;
          g = (double)*(&(ppData[y][3*x])+1) / 255.0; 
          b = (double)*(&(ppData[y][3*x])+2) / 255.0; 
        } else {
          BIASERR("Not grey or rgb, cannot display !");
        }
        VRMLFile << " "<<r<<" "<<g<<" "<<b<<" ,"<<std::endl;
      }
    }
    VRMLFile << "      ]" <<std::endl;
    VRMLFile << "    }" <<std:: endl;   // terminates Color
    VRMLFile << "    colorPerVertex    FALSE " <<std:: endl; 
    VRMLFile << "  }" << std::endl << std::endl; // terminates IndexedFaceSet
    VRMLFile << "}" << std::endl;  // terminates shape
  }
  return 0;
}


int ThreeDOut::VRMLOutLines(ostream& VRMLFile){
  switch (Params_.LineStyle) {
  case Solid:
    for (map<unsigned int, Line3DData>::iterator it=mapLines_.begin();
      it!=mapLines_.end(); it++) {
        double r = (double)it->second.Color[0] / 255.0;
        double g = (double)it->second.Color[1] / 255.0;
        double b = (double)it->second.Color[2] / 255.0;
        // double a = (double)it->second.Color[3] / 255.0;
        VRMLFile << std::endl;
        VRMLFile << "  Shape {" << std::endl;
        VRMLFile <<"         geometry IndexedLineSet {" << std::endl;
        VRMLFile <<"            coord Coordinate {" << std::endl;
        VRMLFile <<"              point [ " << std::endl;
        VRMLFile <<"                       "<<it->second.Start[0]<<" "
          << it->second.Start[1]<<" " <<it->second.Start[2]<< " , "
          << std::endl;
        VRMLFile <<"                       "<<it->second.End[0]<<" "
          <<it->second.End[1] <<" "<<it->second.End[2] <<" ] "<<std::endl;

        VRMLFile <<"            }" << std::endl;
        VRMLFile << "              coordIndex [  0, 1, -1 ]" << std::endl;
        VRMLFile << "           color Color {" << std::endl;
        VRMLFile <<"              color [  "<<r<<" "<<g<<" "<<b<<" ]"<<std::endl;
        VRMLFile << "           } " <<std:: endl;   // terminates Color
        VRMLFile << "           colorPerVertex FALSE " << std::endl;
        VRMLFile << "  }" << std::endl << std::endl; // terminates IndexedLineSet
        VRMLFile << "   }" << std::endl;  // terminates shape
      } 
      break;
  default: 
    BIASERR("Unsupported LineStyle, line drawing disabled."); 
    return -1;
  }
  return 0;
}

int ThreeDOut::VRMLOutEllipsoids(ostream& VRMLFile, const BIAS::RGBAuc &Color){
  for (map<unsigned int, Ellipsoid3DData>::iterator it=mapEllipsoids_.begin();
    it!=mapEllipsoids_.end(); it++) {
      Ellipsoid3DData& E = it->second;
      VRMLFile << "Transform {"<<endl;
      VRMLFile << "  scaleOrientation "<< E.RotAxis[0] <<" " 
        << E.RotAxis[1]<<" " <<E.RotAxis[2] <<" "<< E.dAngle <<" " 
        << endl;
      VRMLFile << "  scale "<< E.s0 <<" "<< E.s1<< " "<< E.s2 << endl;
      VRMLFile << "  translation "
        << E.C[0] << " " << E.C[1] << " " << E.C[2] << endl;
      VRMLFile << "  children Shape {"<<endl;
      VRMLFile << "    appearance Appearance { " << endl;
      VRMLFile << "	    material Material { " << endl;
      VRMLFile << "	      transparency      "<<VRMLtransparency(Color)<< endl;
      VRMLFile << "	      diffuseColor " << (double)Color[0]/255.0 << " "
        << (double)Color[1]/255.0<< " " 
        << (double)Color[2]/255.0 << endl;
      VRMLFile << "	    }" << endl; // close Material
      VRMLFile << "    }" << endl; // close Appearance
      VRMLFile << "    geometry Sphere  { radius 1.0 }" << endl;
      VRMLFile << "  }" << endl; // close Shape
      VRMLFile << "}" << endl; // close Transform
    }
    return 0;
}

int ThreeDOut::VRMLOutPointsAsSpheres_(std::ostream &VRMLFile) {
  double thesize = Params_.PointSize;
  for (map<unsigned int, Point3DData>::iterator it=mapPoints_.begin();
    it!=mapPoints_.end(); it++) {
      Point3DData& E = it->second;
      VRMLFile << "Transform {"<<endl;
      VRMLFile << "  translation "
        << E.Point[0] << " " << E.Point[1] << " " << E.Point[2]
        << endl;
        VRMLFile << "  children ";
        VRMLFile <<"   Shape {"<<endl;
        VRMLFile << "    appearance Appearance { " << endl;
        VRMLFile << "	    material Material { " << endl;
        VRMLFile << "	      transparency      "<<VRMLtransparency(E.Color)<<endl;
        VRMLFile << "	      diffuseColor " 
          << (double)E.Color[0]/255.0 << " "
          << (double)E.Color[1]/255.0 << " " 
          << (double)E.Color[2]/255.0 << endl;
        VRMLFile << "	    }" << endl; // close Material
        VRMLFile << "    }" << endl; // close Appearance
        VRMLFile << "    geometry Sphere  { radius "<<thesize<<" }" << endl;
        VRMLFile << "  }" << endl; // close Shape
        VRMLFile << "}" << endl; // close transform
    }
    return 0;
}


int ThreeDOut::VRMLOutEllipsoids(ostream& VRMLFile){
  for (map<unsigned int, Ellipsoid3DData>::iterator it=mapEllipsoids_.begin();
    it!=mapEllipsoids_.end(); it++) {
      Ellipsoid3DData& E = it->second;
      VRMLFile << "Transform {"<<endl;
      VRMLFile << "  scaleOrientation "<< E.RotAxis[0] <<" " 
        << E.RotAxis[1]<<" " <<E.RotAxis[2] <<" "<< E.dAngle <<" " 
        << endl;
      VRMLFile << "  scale "<< E.s0 <<" "<< E.s1<< " "<< E.s2 << endl;
      VRMLFile << "  translation "
        << E.C[0] << " " << E.C[1] << " " << E.C[2] << endl;
      VRMLFile << "  children Shape {"<<endl;
      VRMLFile << "    appearance Appearance { " << endl;
      VRMLFile << "	    material Material { " << endl;
      VRMLFile << "	      transparency      "
        <<VRMLtransparency(E.Color)<<endl;
      VRMLFile << "	      diffuseColor " << (double)E.Color[0]/255.0 << " "
        << (double)E.Color[1]/255.0<< " " 
        << (double)E.Color[2]/255.0 << endl;
      VRMLFile << "	    }" << endl; // close Material
      VRMLFile << "    }" << endl; // close Appearance
      VRMLFile << "    geometry Sphere  { radius 1.0 }" << endl;
      VRMLFile << "  }" << endl; // close Shape
      VRMLFile << "}" << endl; // close Transform
    }
    return 0;
}

int ThreeDOut::VRMLOut(const string &sFilename, const BIAS::RGBAuc &Color){
  ofstream VRMLFile;
  VRMLFile.open(sFilename.c_str());
  if (!VRMLFile) { 
    BIASERR("Error opening VRML file for writing: "<<sFilename);
    return -1;
  }
  VRMLOutWriteHeader(VRMLFile);
  VRMLOutIndexedFaceSets(VRMLFile);
  VRMLOutEllipsoids(VRMLFile, Color); 
  VRMLOutLines(VRMLFile, Color);
  VRMLOutPoints(VRMLFile, Color);
  VRMLOutImages(VRMLFile);
  VRMLOutPMatrices(VRMLFile, Color);
  VRMLFile.close();
  return 0 ;
}

int ThreeDOut::VRMLOut(const string &sFilename){
  ofstream VRMLFile;
  VRMLFile.open(sFilename.c_str());
  if (!VRMLFile) { 
    BIASERR("Error opening VRML file for writing: "<<sFilename);
    return -1;
  }
  VRMLOutWriteHeader(VRMLFile);
  VRMLOutIndexedFaceSets(VRMLFile);
  VRMLOutPoints(VRMLFile);
  VRMLOutEllipsoids(VRMLFile); 
  VRMLOutLines(VRMLFile);
  VRMLOutImages(VRMLFile);
  VRMLOutPMatrices(VRMLFile);
  VRMLFile.close();
  return 0;
}

int ThreeDOut::VRMLOutWriteHeader(std::ostream& vrml){
  vrml << "#VRML V2.0 utf8" << endl;
  vrml << "# File written by Basic Image AlgorithmS (BIAS) library" << endl;
  vrml << "# http://www.mip.informatik.uni-kiel.de"
    <<endl;
  return 0;
}

// ------------------------------ other functions ----------------------------

void ThreeDOut::Dump(){
  cout <<endl<< "Dumping ThreeDOut object:"<<endl;
  cout <<"Points: -----------------------------------------"<<endl;
  for(map<unsigned int, Point3DData>::iterator it = mapPoints_.begin();
    it != mapPoints_.end(); it++) {
      cout <<"["<<it->first<<"] "<<it->second.Point<<endl;
    }
    cout<<endl<< "Lines: ------------------------------------------"<<endl;
    for(map<unsigned int, Line3DData>::iterator it = mapLines_.begin();
      it != mapLines_.end(); it++) {
        cout <<"["<<it->first<<"] "<<it->second.Start<<" -> "
          <<it->second.End<<endl;
      }
      cout<<endl<< "PMatrices: -------------------------------------- "<<endl;
      unsigned int i=0;
      for(map<unsigned int, PMatrix3DData>::iterator 
        it = mapPMatrices_.begin(); it != mapPMatrices_.end(); it++) {
          cout <<"["<<(i++)<<"] UL="<<it->second.UL<<" UR="<<it->second.UR
            <<" LL="<<it->second.LL<<" LR="
            <<it->second.LR<<" C="<<it->second.C<<endl;
        }
        cout<<endl<< "Ellipsoids: -------------------------------------"<<endl;
        for(map<unsigned int, Ellipsoid3DData>::iterator it = mapEllipsoids_.begin();
          it != mapEllipsoids_.end(); it++) {
            cout <<"["<<it->first<<"] Center at "<<it->second.C<<endl;
          }
          cout<<endl<<flush;

}


Vector3<double> ThreeDOut::GetCenterOfPoints(){
  Vector3<double> center(0.0);
  unsigned int count=0;
  map<unsigned int, Point3DData>::iterator it; 
  for (it=mapPoints_.begin(); it!=mapPoints_.end(); it++) {
    count ++;
    center += it->second.Point;
  }
  center /= double(count);
  return center;
}


void ThreeDOut::GetBoundingBox(Vector3<double> &min, Vector3<double> &max)
{
  for (unsigned int i=0;i<3; i++) {
    min[i] = DBL_MAX;
    max[i] = DBL_MIN;
  }

  map<unsigned int, Point3DData>::iterator it;
  for (it=mapPoints_.begin(); it!=mapPoints_.end(); it++) {
    for (unsigned int i=0;i<3; i++) {
      if (it->second.Point[i] < min[i]) min[i] = it->second.Point[i];
      if (it->second.Point[i] > max[i]) max[i] = it->second.Point[i];
    }
  }
}


void ThreeDOut::RemoveAll(){
  RemoveAllPoints();
  RemoveAllLines();
  RemoveAllEllipsoids();
  RemoveAllImages();
  RemoveAllPMatrices();
}

// OpenGL interpretes alpha as opacity 
// while VRML uses its inverse, transaparency JW
double ThreeDOut::VRMLtransparency(const float & alpha){
  // our alpha stores opacity (as in GL)
  // but VRML uses transparency and normalized scaling
  return double( (1.0-alpha)/255.0 );
}

double ThreeDOut::VRMLtransparency(const BIAS::Vector4<unsigned char> & vRgba){
  return VRMLtransparency( float(vRgba[3]) );
}

